/*
 * arch/arm/mach-cxd4115/gic.c
 *
 * MPCore Distributed Interrupt Controller
 *
 * Copyright 2007,2008,2009 Sony Corporation
 *
 * This code is based on arch/arm/common/gic.c.
 */
/*
 *  linux/arch/arm/common/gic.c
 *
 *  Copyright (C) 2002 ARM Limited, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Interrupt architecture for the GIC:
 *
 * o There is one Interrupt Distributor, which receives interrupts
 *   from system devices and sends them to the Interrupt Controllers.
 *
 * o There is one CPU Interface per CPU, which sends interrupts sent
 *   by the Distributor, and interrupts generated locally, to the
 *   associated CPU. The base address of the CPU interface is usually
 *   aliased so that the same address points to different chips depending
 *   on the CPU it is accessed from.
 *
 * Note that IRQs 0-31 are special - they are local to each CPU.
 * As such, the enable set/clear, pending set/clear and active bit
 * registers are banked per-cpu for these sources.
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/smp.h>
#include <linux/cpumask.h>

#include <asm/irq.h>
#include <asm/io.h>
#include <asm/mach/irq.h>
#include <asm/arch/gic.h>

#ifndef CONFIG_SMP
#undef hard_smp_processor_id
#include <asm/arch/smp.h>
#endif

#ifdef CONFIG_MPCORE_GIC_INIT
/* List of edge trigger IRQs */
static u8 edge_irqs[] __initdata = {
	IRQ_GPIO_RISE(0), IRQ_GPIO_RISE(1), IRQ_GPIO_RISE(2), IRQ_GPIO_RISE(3),
	IRQ_GPIO_RISE(4), IRQ_GPIO_RISE(5), IRQ_GPIO_RISE(6), IRQ_GPIO_RISE(7),
	IRQ_GPIO_RISE(8), IRQ_GPIO_RISE(9), IRQ_GPIO_RISE(10),IRQ_GPIO_RISE(11),
	IRQ_GPIO_RISE(12),IRQ_GPIO_RISE(13),IRQ_GPIO_RISE(14),IRQ_GPIO_RISE(15),
	IRQ_GPIO_RISE(16),IRQ_GPIO_RISE(17),IRQ_GPIO_RISE(18),IRQ_GPIO_RISE(19),
	IRQ_GPIO_RISE(20),IRQ_GPIO_RISE(21),IRQ_GPIO_RISE(22),
	IRQ_GPIO_FALL(0), IRQ_GPIO_FALL(1), IRQ_GPIO_FALL(2), IRQ_GPIO_FALL(3),
	IRQ_GPIO_FALL(4), IRQ_GPIO_FALL(5), IRQ_GPIO_FALL(6), IRQ_GPIO_FALL(7),
	IRQ_GPIO_FALL(8), IRQ_GPIO_FALL(9), IRQ_GPIO_FALL(10),IRQ_GPIO_FALL(11),
	IRQ_GPIO_FALL(12),IRQ_GPIO_FALL(13),IRQ_GPIO_FALL(14),IRQ_GPIO_FALL(15),
	IRQ_GPIO_FALL(16),IRQ_GPIO_FALL(17),IRQ_GPIO_FALL(18),IRQ_GPIO_FALL(19),
	IRQ_GPIO_FALL(20),IRQ_GPIO_FALL(21),IRQ_GPIO_FALL(22),
	88, 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 
	100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111,
	IRQ_VBUSVALID_RISE,IRQ_VBUSVALID_FALL,IRQ_AVALID_RISE,IRQ_AVALID_FALL,
	IRQ_SESSEND_RISE,IRQ_SESSEND_FALL,
#ifdef CONFIG_ARCH_CXD4120
	IRQ_GPIO_RISE(23), IRQ_GPIO_FALL(23), 148, 149, 203, 204,
#endif
};
#endif /* CONFIG_MPCORE_GIC_INIT */

static DEFINE_SPINLOCK(irq_controller_lock);

/*
 * Routines to acknowledge, disable and enable interrupts
 *
 * Linux assumes that when we're done with an interrupt we need to
 * unmask it, in the same way we need to unmask an interrupt when
 * we first enable it.
 *
 * The GIC has a separate notion of "end of interrupt" to re-enable
 * an interrupt after handling, in order to support hardware
 * prioritisation.
 *
 * We can make the GIC behave in the way that Linux expects by making
 * our "acknowledge" routine disable the interrupt, then mark it as
 * complete.
 */
static void gic_ack_irq(unsigned int irq)
{
	spin_lock(&irq_controller_lock);
	writel(GIC_IRQ_BITMASK(irq), VA_GIC_DIST + GIC_DIST_ENABLE_CLEAR + GIC_IRQ_OFFSET(irq));
	writel(irq, VA_GIC_CPU + GIC_CPU_EOI);
	mmiowb();
	spin_unlock(&irq_controller_lock);
}

static void gic_mask_irq(unsigned int irq)
{
	spin_lock(&irq_controller_lock);
	writel(GIC_IRQ_BITMASK(irq), VA_GIC_DIST + GIC_DIST_ENABLE_CLEAR + GIC_IRQ_OFFSET(irq));
	mmiowb();
	spin_unlock(&irq_controller_lock);
}

static void gic_unmask_irq(unsigned int irq)
{
	spin_lock(&irq_controller_lock);
	writel(GIC_IRQ_BITMASK(irq), VA_GIC_DIST + GIC_DIST_ENABLE_SET + GIC_IRQ_OFFSET(irq));
	spin_unlock(&irq_controller_lock);
}

#ifdef CONFIG_SMP
static void gic_set_cpu(unsigned int irq, cpumask_t mask_val)
{
	const void __iomem *reg = VA_GIC_DIST + GIC_DIST_TARGET + (irq & ~GIC_TARGET_BITS_MASK);
	unsigned int shift = (irq & GIC_TARGET_BITS_MASK) << 3;
	unsigned int cpu = first_cpu(mask_val);
	u32 val;

	spin_lock(&irq_controller_lock);
	irq_desc[irq].cpu = cpu;
	val = readl(reg) & ~(0xff << shift);
	val |= 1 << (phys_cpuid(cpu) + shift);
	writel(val, reg);
	spin_unlock(&irq_controller_lock);
}
#endif

static struct irq_chip gic_chip = {
	.name		= "GIC",
	.ack		= gic_ack_irq,
	.mask		= gic_mask_irq,
	.unmask		= gic_unmask_irq,
#ifdef CONFIG_SMP
	.set_affinity	= gic_set_cpu,
#endif
};

void __init gic_dist_init(unsigned int gic_nr, const void __iomem *base, unsigned int irq_start)
{
	unsigned int max_irq, i;
	u32 cpumask = 1 << hard_smp_processor_id();
#ifdef CONFIG_MPCORE_GIC_INIT
	u8 *p;
	unsigned int offset, shift, tmp;
#endif

	cpumask |= cpumask << 8;
	cpumask |= cpumask << 16;

#ifdef CONFIG_MPCORE_GIC_INIT
	writel(0, base + GIC_DIST_CTRL);
#endif

	/*
	 * Find out how many interrupts are supported.
	 */
	max_irq = readl(base + GIC_DIST_CTR) & 0x1f;
	max_irq = (max_irq + 1) * 32;

	/*
	 * The GIC only supports up to 1020 interrupt sources.
	 * Limit this to either the architected maximum, or the
	 * platform maximum.
	 */
	if (max_irq > max(1020, NR_IRQS))
		max_irq = max(1020, NR_IRQS);

#ifdef CONFIG_MPCORE_GIC_INIT
	/*
	 * Set all global interrupts to be level triggered, N:N model
	 */
	offset = (IRQ_GIC_START >> GIC_CONFIG_BITS_SHIFT) << 2;
	for (i = 32; i < max_irq; i += 16, offset += 4)
		writel(0, base + GIC_DIST_CONFIG + offset);

	/*
         * Set edge trigger
         *  SLOW: set one by one. (read-modify-write)
	 */
	for (i = 0, p = edge_irqs; i < sizeof edge_irqs; i++, p++) {
		offset = (*p >> GIC_CONFIG_BITS_SHIFT) << 2;
		shift = (*p & GIC_CONFIG_BITS_MASK) * GIC_CONFIG_BITS_WIDTH;
		tmp = readl(base + GIC_DIST_CONFIG + offset);
		tmp |= GIC_DIST_CONFIG_EDGE << shift;
		writel(tmp, base + GIC_DIST_CONFIG + offset);
	}

	/*
	 * Set all global interrupts to this CPU only.
	 */
	offset = (IRQ_GIC_START >> GIC_TARGET_BITS_SHIFT) << 2;
	for (i = 32; i < max_irq; i += 4, offset += 4)
		writel(cpumask, base + GIC_DIST_TARGET + offset);

	/*
	 * Set priority on all interrupts.
	 *   INT0..31 are alias.
	 */
	for (i = 0, offset = 0; i < max_irq; i += 4, offset += 4)
		writel(0xa0a0a0a0, base + GIC_DIST_PRI + offset);

	/*
	 * Disable all interrupts.
	 *   INT29..31 are alias.
	 */
	for (i = 0, offset = 0; i < max_irq; i += 32, offset += 4)
		writel(0xffffffff, base + GIC_DIST_ENABLE_CLEAR + offset);
#endif /* CONFIG_MPCORE_GIC_INIT */
	mmiowb();

	/*
	 * Setup the Linux IRQ subsystem.
	 */
	for (i = irq_start; i < max_irq; i++) {
		set_irq_chip(i, &gic_chip);
		set_irq_handler(i, handle_level_irq);
		set_irq_flags(i, IRQF_VALID | IRQF_PROBE);
	}

#ifdef CONFIG_MPCORE_GIC_INIT
	writel(1, base + GIC_DIST_CTRL);
#endif
}

void __cpuinit gic_cpu_init(unsigned int gic_nr, const void __iomem *base)
{
#ifdef CONFIG_SMP
	int i;

	/*
	 * Set priority on INT0..31 for this CPU.
	 */
	for (i = 0; i < 32; i += 4)
		writel(0xa0a0a0a0, VA_GIC_DIST + GIC_DIST_PRI + i);
#endif

	writel(0xf0, base + GIC_CPU_PRIMASK);
	writel(1, base + GIC_CPU_CTRL);
}

#ifdef CONFIG_SMP
void gic_raise_softirq(cpumask_t cpumask, unsigned int irq)
{
	unsigned long map = *cpus_addr(cpumask);

	writel(map << 16 | irq, VA_GIC_DIST + GIC_DIST_SOFTINT);
}
#endif

void raise_irq(unsigned int cpu, unsigned int irq)
{
	writel((1 << (16+cpu)) | irq, VA_GIC_DIST + GIC_DIST_SOFTINT);
}
EXPORT_SYMBOL(raise_irq);

struct gic_info {
	u32 prio[NR_IRQS/4];
	u32 target[NR_IRQS/4];
	u32 config[NR_IRQS/16];
	/* CPU I/F */
	u32 primask;
	u32 binary;
};
static struct gic_info gic_save;

void gic_suspend(void)
{
#ifdef CONFIG_MPCORE_GIC_INIT
	u32 *p;
	int i;

	/* save Priority Regs. */
	p = gic_save.prio;
	for (i = 0; i < NR_IRQS; i+=4) {
		*p++ = readl(VA_GIC_DIST + GIC_DIST_PRI + i);
	}
	/* save Target Regs. */
	p = gic_save.target;
	for (i = 0; i < NR_IRQS; i+=4) {
		*p++ = readl(VA_GIC_DIST + GIC_DIST_TARGET + i);
	}
	/* save Config Regs. */
	p = gic_save.config;
	for (i = 0; i < NR_IRQS; i+=16) {
		*p++ = readl(VA_GIC_DIST + GIC_DIST_CONFIG + i);
	}
#endif /* CONFIG_MPCORE_GIC_INIT */
	gic_save.primask = readl(VA_GIC_CPU + GIC_CPU_PRIMASK);
}

void gic_resume(void)
{
#ifdef CONFIG_MPCORE_GIC_INIT
	u32 *p;
	int i;

	/* restore Priority Regs. */
	p = gic_save.prio;
	for (i = 0; i < NR_IRQS; i+=4, p++) {
		writel(*p, VA_GIC_DIST + GIC_DIST_PRI + i);
	}
	/* save Target Regs. */
	p = gic_save.target;
	for (i = 0; i < NR_IRQS; i+=4, p++) {
		writel(*p, VA_GIC_DIST + GIC_DIST_TARGET + i);
	}
	/* save Config Regs. */
	p = gic_save.config;
	for (i = 0; i < NR_IRQS; i+=16, p++) {
		writel(*p, VA_GIC_DIST + GIC_DIST_CONFIG + i);
	}
#endif /* CONFIG_MPCORE_GIC_INIT */
	writel(gic_save.primask, VA_GIC_CPU + GIC_CPU_PRIMASK);
	writel(1, VA_GIC_CPU + GIC_CPU_CTRL);
}
