/*--------------------------------------------------------------------------*
 * drivers/usb/mcd/usb_mcd_bgadget.c
 * 
 * USB controller driver
 * 
 * Copyright 2005,2006,2008,2009 Sony Corporation
 * 
 * This code is based on musb_bgadget.c,
 * Copyright 2005 Mentor Graphics Corporation
 *
 * This file is part of the Inventra Controller Driver for Linux.
 *
 * The Inventra Controller Driver for Linux is free software; you
 * can redistribute it and/or modify it under the terms of the GNU
 * General Public License version 2 as published by the Free Software
 * Foundation.
 *
 * The Inventra Controller Driver for Linux is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with The Inventra Controller Driver for Linux ; if not,
 * write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA  02111-1307  USA
 *
 * ANY DOWNLOAD, USE, REPRODUCTION, MODIFICATION OR DISTRIBUTION
 * OF THIS DRIVER INDICATES YOUR COMPLETE AND UNCONDITIONAL ACCEPTANCE
 * OF THOSE TERMS.THIS DRIVER IS PROVIDED "AS IS" AND MENTOR GRAPHICS
 * MAKES NO WARRANTIES, EXPRESS OR IMPLIED, RELATED TO THIS DRIVER.
 * MENTOR GRAPHICS SPECIFICALLY DISCLAIMS ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY; FITNESS FOR A PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT.  MENTOR GRAPHICS DOES NOT PROVIDE SUPPORT
 * SERVICES OR UPDATES FOR THIS DRIVER, EVEN IF YOU ARE A MENTOR
 * GRAPHICS SUPPORT CUSTOMER.
 *---------------------------------------------------------------------------*/

#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/timer.h>
#include <linux/module.h>
#include <linux/smp.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/stat.h>
#include <linux/moduleparam.h>
#include <linux/dmapool.h>
#include <linux/dma-mapping.h>

#include <linux/usb.h>
#include <linux/usb/ch9.h>
#include <linux/usb_gadget.h>
#include <linux/usb/mcd/usb_otg_control.h>
#include <linux/usb/mcd/usb_mcd_ioctl.h>
#include "../core/hub.h"
#include "../core/hcd.h"

#include "usb_mcd_cnf.h"
#include "usb_mcd_debug.h"
#include "usb_mcd_dma.h"
#include "usb_mcd_otg.h"
#include "usb_mcd_gadget.h"
#include "usb_mcd_host.h"
#include "usb_mcd_device.h"
#include "usb_mcd_access.h"
#include "usb_mcd_musbhdrc.h"
#include "usb_mcd_fifo.h"

#ifdef MCD_DMA
#include "usb_mcd_dma.h"
#define USB_REQ_DMA_MODE1 1
#define USB_REQ_DMA_MODE0 2
#define USB_REQ_CPU       3
#define USB_INIT_REQ_PHASE1       1
#define USB_INIT_REQ_PHASE2       2

#define DUMMY_USB_IOCTL_TESTMODE_NORMAL	/* ƥȥ⡼        */
						/* TESTMODE_NORMALϥߡ */
						/* TESTMODE_NORMALꤵ */
						/* Ƥʤˤ⤷ʤ */
#define MCD_DMA_FEWBYTE_USING_CPU_TRANSFER	/* 4byteʲΥ硼ȥѥ */
						/* ȤCPUž */
#define MCD_DMA_FEWBYTE  4
#define USB_DMA_SHORTPKT_USING_CPU_TRANSFER	/* 4byte饤ȤΤ   */
						/* 硼ȥѥåȤCPUž */
//#define MCD_LOCK_DEBUG
#endif

/**************************************************************************
Gadget Functions
**************************************************************************/

/* endpoint operations */
STATIC int mcd_gadget_enable_end(struct usb_ep *ep, const struct usb_endpoint_descriptor *desc);
STATIC int mcd_gadget_disable_end(struct usb_ep *ep);

STATIC struct usb_request *mcd_gadget_alloc_request(struct usb_ep *ep, gfp_t gfp_flags);
STATIC void mcd_gadget_free_request(struct usb_ep *ep, struct usb_request *req);
STATIC void *mcd_gadget_alloc_buffer(struct usb_ep *ep, unsigned bytes, dma_addr_t * dma, gfp_t gfp_flags);
STATIC void mcd_gadget_free_buffer(struct usb_ep *ep, void *buf, dma_addr_t dma, unsigned bytes);
STATIC int mcd_gadget_queue(struct usb_ep *ep, struct usb_request *req, gfp_t gfp_flags);
STATIC int mcd_gadget_exec_request(struct usb_ep *ep, struct usb_request *pRequest, gfp_t gfp_flags);
STATIC int mcd_gadget_dequeue(struct usb_ep *ep, struct usb_request *req);
int mcd_gadget_set_halt(struct usb_ep *ep, int value);
STATIC int mcd_gadget_fifo_status(struct usb_ep *ep);
STATIC void mcd_gadget_fifo_flush(struct usb_ep *ep);
STATIC void mcd_gadget_fifo_flush_flush_only(struct usb_ep *ep);

/* general operations */
STATIC int mcd_gadget_get_frame(struct usb_gadget *gadget);
STATIC int mcd_gadget_wakeup(struct usb_gadget *gadget);
STATIC int mcd_gadget_set_self_powered(struct usb_gadget *gadget, int is_selfpowered);
STATIC int mcd_gadget_vbus_session(struct usb_gadget *gadget, int is_active);
STATIC int mcd_gadget_vbus_draw(struct usb_gadget *gadget, unsigned mA);
STATIC int mcd_gadget_pullup(struct usb_gadget *gadget, int is_on);
STATIC int mcd_gadget_ioctl(struct usb_gadget *gadget, unsigned code, unsigned long param);

STATIC void mcd_gadget_dma_set_default(struct mcd_gadget *mgadget);

static struct usb_ep_ops mcd_gadget_endpoint_operations = {
	.enable = mcd_gadget_enable_end,
	.disable = mcd_gadget_disable_end,
	.alloc_request = mcd_gadget_alloc_request,
	.free_request = mcd_gadget_free_request,
	.alloc_buffer = mcd_gadget_alloc_buffer,
	.free_buffer = mcd_gadget_free_buffer,
	.queue = mcd_gadget_queue,
	.dequeue = mcd_gadget_dequeue,
	.set_halt = mcd_gadget_set_halt,
	.fifo_status = mcd_gadget_fifo_status,
	.fifo_flush = mcd_gadget_fifo_flush
};

static struct usb_gadget_ops mcd_gadget_operations = {
	.get_frame = mcd_gadget_get_frame,
	.wakeup = mcd_gadget_wakeup,
	.set_selfpowered = mcd_gadget_set_self_powered,
	.vbus_session = mcd_gadget_vbus_session,
	.vbus_draw = mcd_gadget_vbus_draw,
	.pullup = mcd_gadget_pullup,
	.ioctl = mcd_gadget_ioctl
};

#ifdef MCD_DMA
void mcd_gadget_dma_tx_cb(struct mcd_dma_channel *pChannel, uint8_t bEnd, uint8_t bTransmit, uint8_t bProtocol, void *pData);

void mcd_gadget_dma_rx_cb(struct mcd_dma_channel *pChannel, uint8_t bEnd, uint8_t bTransmit, uint8_t bProtocol, void *pData);
#endif				/* MCD_DMA */

/**************************************************************************
Gadget End points
**************************************************************************/

struct mcd_gadget *mcd_first_gadget;

/* counters and queues */
static struct {
	spinlock_t lock;
	struct list_head req_list;
} SchedulerQueue;

static char aGadgetDefaultEndNames[MCD_NUM_EPS][6];	/* the singleton */
static int iGadgetEndNamesLen = -1;
static char *aGadgetEndNames[MCD_NUM_EPS];

#ifdef MCD_GADGET_PARAMS
module_param_array(aGadgetEndNames, charp, iGadgetEndNamesLen, S_IRUGO);
#endif

struct mcd_gadget_request {
	struct usb_request request;
#if MCD_DEBUG > 0
	uint32_t dwPadFront;
#endif
	uint8_t bEnd;		/* will be used by the completition routine */
	struct list_head completition_list;	/* will be used by the completition routine */
#if MCD_DEBUG > 0
	uint8_t bQueued;
	uint32_t dwPadBack;
#endif
};

/**************************************************************************
Diagnostic
**************************************************************************/

inline int is_corrupted_request(void *req)
{
	int rc = 0;
	const struct mcd_gadget_request *pRequest = (struct mcd_gadget_request *)req;

	if (!pRequest) {
		ERR("pRequest is NULL\n");
		rc = 1;
	}
#if MCD_DEBUG >0
	if (pRequest->dwPadFront != MCD_PAD_FRONT) {
		ERR("pRequest->dwPadFront corrupted in pRequest=%p\n", pRequest);
		rc = 1;
	}
	if (pRequest->dwPadBack != MCD_PAD_BACK) {
		ERR("pRequest->dwPadBack corrupted in pRequest=%p\n", pRequest);
		rc = 1;
	}
#endif

	return rc;		/* not corrupted */
}

#if MCD_DEBUG > 0
#define is_queued_for_completition( _req ) (((struct mcd_gadget_request*)(_req))->bQueued)
#else
#define is_queued_for_completition( _req ) 0
#endif

static void stop_dma(struct mcd_gadget *mgadget, uint8_t bEnd)
{
#ifdef MCD_DMA
	uint16_t wCsrVal;
	struct mcd_device *mdev = mgadget_to_mdev(mgadget);
	struct mcd_dma_channel *pDmaChannel = mgadget->ep[bEnd].pDmaChannel;
	void *regs = mdev_link_reg(mdev);

	if (pDmaChannel) {
		mcd_dma_stop_channel(pDmaChannel);
		mcd_dma_disable_int_channel(pDmaChannel);
		mcd_dma_release_channel(pDmaChannel);
		mgadget->ep[bEnd].pDmaChannel = pDmaChannel = NULL;

		mcd_select_end(regs, bEnd);
		wCsrVal = MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd);
		wCsrVal &= ~(MCD_M_TXCSR_DMAENAB);
		wCsrVal &= ~(MCD_M_TXCSR_TXPKTRDY);
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd, wCsrVal);
		wCsrVal &= ~(MCD_M_TXCSR_MODE);
		wCsrVal &= ~(MCD_M_TXCSR_AUTOSET | MCD_M_TXCSR_DMAMODE);
		wCsrVal &= ~(MCD_M_TXCSR_TXPKTRDY);
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd, wCsrVal);

		wCsrVal = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd);
		wCsrVal &= ~(MCD_M_RXCSR_AUTOCLEAR | MCD_M_RXCSR_DMAMODE);
		wCsrVal &= ~(MCD_M_RXCSR_RXPKTRDY);
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd, wCsrVal);
	}
#endif
}

inline void set_enabledma_bulkout(struct mcd_gadget *mgadget, uint8_t bEnd)
{
	uint16_t wCsrVal;
	void *regs = mdev_link_reg(mgadget_to_mdev(mgadget));
	struct mcd_gep *pEnd = &mgadget->ep[bEnd];

	if (pEnd->bTrafficType == USB_ENDPOINT_XFER_BULK && !pEnd->bIsTx && pEnd->bDma_ena == USB_MCD_DMA_SETON) {
		/* BulkOutEP ShortPacketꥯdequeueDMA mode1 */
		mcd_select_end(regs, bEnd);
		wCsrVal = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd);
		wCsrVal |= MCD_M_RXCSR_DMAENAB;
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd, wCsrVal);
	}
}

static int complete_request_immediate(struct usb_ep *ep, struct usb_request *pRequest, int status)
{
	unsigned long flags;
	struct usb_request *req;
	struct mcd_gep *pEnd = (struct mcd_gep *)ep;
	const uint8_t bEnd = EP_NUMBER(ep);
	struct mcd_gadget *mgadget = mcd_first_gadget;

	DBG(2, "<==\n");
	DBG(3, "dequeuing from nEnd=0x%x, pEnd=%p, pRequest=%p\n", EP_NUMBER(pEnd), pEnd, pRequest);

	EP_SPIN_LOCK_IRQSAVE(pEnd, flags);

	req = mcd_get_current_request(pEnd);

	if (NULL == req) {
		EP_SPIN_UNLOCK_IRQRESTORE(pEnd, flags);
		return -EINVAL;
	}

	if (!bEnd) {
		req->status = status;
		mcd_gadget_complete_ep0_request();
		EP_SPIN_UNLOCK_IRQRESTORE(pEnd, flags);
		mcd_gadget_fifo_flush_flush_only(ep);
	} else if (req == pRequest) {
		list_del(&req->list);
		if (-EINPROGRESS != req->status) {
			/* list_del() SchedulerQueue. */
			list_del(&((struct mcd_gadget_request *)req)->completition_list);
		}
		req->status = status;
		stop_dma(mgadget, bEnd);
		set_enabledma_bulkout(mgadget, bEnd);
		mcd_gadget_fifo_flush_flush_only(ep);
		EP_SPIN_UNLOCK_IRQRESTORE(pEnd, flags);
		if (req->complete) {
			req->complete(&pEnd->end_point, req);
		}
	} else {
		list_for_each_entry(req, &pEnd->req_list, list) {
			if (req == pRequest) {
				list_del(&req->list);
				if (-EINPROGRESS != req->status) {
					/* list_del() SchedulerQueue. */
					list_del(&((struct mcd_gadget_request *)req)->completition_list);
				}
				req->status = status;
				EP_SPIN_UNLOCK_IRQRESTORE(pEnd, flags);
				if (req->complete) {
					req->complete(&pEnd->end_point, req);
				}
				goto OUT_ELSE_DEQUE;
			}
		}
		EP_SPIN_UNLOCK_IRQRESTORE(pEnd, flags);

		return -EINVAL;
	}
      OUT_ELSE_DEQUE:

	return 0;
}

/*============================================================================*/
/* DMA-Coherent Buffers                                                       */
/*============================================================================*/
/* FIXME tune these based on pool statistics ... */
static const size_t pool_max[HCD_BUFFER_POOLS] = {
	/* platforms without dma-friendly caches might need to
	 * prevent cacheline sharing...
	 */
	32,
	128,
	512,
	PAGE_SIZE / 2
	    /* bigger --> allocate pages */
};

/*============================================================================*/
/* mcd_base_buffer_destroy                                                    */
/* release pools to allocate a buffer for DMA                                 */
/* @param pThis struct mcd_gadget* to hold areas of dma_pool                  */
/*============================================================================*/
static void pool_buffer_destroy(struct mcd_gadget *mgadget)
{
	int i;

	DBG(1, "<==\n");

	for (i = 0; i < HCD_BUFFER_POOLS; i++) {
		struct dma_pool *pool = mgadget->pool[i];
		if (pool) {
			dma_pool_destroy(pool);
			mgadget->pool[i] = NULL;
			DBG(1, "mgadget->pool [%d] =%p \n", i, pool);
		}
	}
}

/*============================================================================*/
/* mcd_base_buffer_create                                                     */
/* create pools to allocate a buffer for DMA                                  */
/* @param mgadget struct mcd_gadget* to hold areas of dma_pool                */
/*============================================================================*/
static uint8_t pool_buffer_create(struct mcd_gadget *mgadget)
{
	char name[16];
	int i, size;
	struct mcd_device *mdev = mgadget_to_mdev(mgadget);

	DBG(1, "<==\n");

	for (i = 0; i < HCD_BUFFER_POOLS; i++) {
		if (!(size = pool_max[i]))
			continue;
		snprintf(name, sizeof name, "buffer-%d", size);
		mgadget->pool[i] = dma_pool_create(name, mdev->dev, size, size, 0);
		DBG(1, "pThis->pool [%d] =%p size=%d \n", i, mgadget->pool[i], size);
		if (!mgadget->pool[i]) {
			DBG(1, "can't get a pool-area!" " call pool_buffer_destroy \n");
			pool_buffer_destroy(mgadget);
			return -ENOMEM;
		}
	}
	return 0;
}

/*============================================================================*/
/* Allocate memory for a buffer that might use DMA.                           */
/*                                                                            */
/* retrun addr:         pointer of a virtual address                          */
/* @param pThis:        struct mcd_base_cd* to hold areas of dma_pool         */
/* @param size:         the length of memory to get                           */
/* @param mem_flags:    the same Flags for memory area as kmalloc's flags     */
/* @param dma:	        pointer of a bus address. NULL when DMAble memeory is */
/*                      not requested.                                        */
/*==============================================================================*/
static void *pool_alloc_buffer(struct mcd_gadget *mgadget, size_t size, int mem_flags, dma_addr_t * dma)
{
	int i;
	struct mcd_device *mdev = mgadget_to_mdev(mgadget);
	void *addr = NULL;

	DBG(1, "size =%d \n", size);
	/* some USB hosts just use PIO */

	for (i = 0; i < HCD_BUFFER_POOLS; i++) {
		if (size <= pool_max[i]) {
			addr = dma_pool_alloc(mgadget->pool[i], mem_flags, dma);
			DBG(1, "dma_pool_alloc " "pool[%d] pAddr=%p pDma=%p mem_flags=%d \n", i, addr, dma, mem_flags);
			return addr;
		}
	}
	addr = dma_alloc_coherent(mdev->dev, size, dma, 0);
	DBG(1, "dma_alloc_coherent pAddr=%p, pDma=%p mem_flags=0 <-Force \n", addr, (void *)*dma);
	return addr;

}

/*============================================================================*/
/* Free memory previously allocated with AllocBufferMemory.                   */
/* @param mgadget       struct mcd_gadget* to hold areas of dma_pool          */
/* @param size:         the length of memory to free                          */
/* @param addr:         virtual address for tha bus address(dma)              */
/* @param dma:          the pointer of a bus address                          */
/*============================================================================*/
static void pool_free_buffer(struct mcd_gadget *mgadget, size_t size, void *addr, dma_addr_t dma)
{
	int i;
	struct mcd_device *mdev = mgadget_to_mdev(mgadget);

	DBG(1, "<== freeing size=%d, pAddr=%p, pDma=%p\n", size, addr, (void *)dma);
	if (!addr) {
		DBG(1, "address NULL!\n");
		return;
	}

	for (i = 0; i < HCD_BUFFER_POOLS; i++) {
		if (size <= pool_max[i]) {
			dma_pool_free(mgadget->pool[i], addr, dma);
			DBG(1, "dma_pool_free pool[%d] pAddr=%p pDma=%p \n", i, addr, (void *)dma);
			return;
		}
	}

	dma_free_coherent(mdev->dev, size, addr, dma);

	DBG(1, "dma_free_coherent \n");
}

/**************************************************************************
Tasklet
**************************************************************************/

static void gadget_complete_requests(unsigned long data)
{
#if MCD_DEBUG>0
	int success = 0, fail = 0;
#endif
	struct list_head local_list;
	struct mcd_gadget *mgadget = mcd_first_gadget;
	struct mcd_gadget_request *req = NULL;
	struct usb_request *pRequest;
	unsigned long flags;

	INIT_LIST_HEAD(&local_list);

	spin_lock_irqsave(&SchedulerQueue.lock, flags);	/* Modyfied for request queue restart */
	while (!list_empty(&SchedulerQueue.req_list)) {
		req = list_entry(SchedulerQueue.req_list.next, struct mcd_gadget_request, completition_list);
		list_move_tail(&req->completition_list, &local_list);
	}

	/* complete the request first */
	while (!list_empty(&local_list)) {
		req = list_entry(local_list.next, struct mcd_gadget_request, completition_list);
		pRequest = (struct usb_request *)req;
		list_del(&pRequest->list);
		pRequest = mcd_get_current_request(&mgadget->ep[req->bEnd]);
		if (pRequest) {
			mcd_gadget_exec_request((struct usb_ep *)&mgadget->ep[req->bEnd], pRequest, GFP_KERNEL);
		}
		list_del(&req->completition_list);
#if MCD_DEBUG>0
		req->bQueued = 0;
#endif

		DBG(4, "completing pRequest=%p on bEnd=%d\n", req, req->bEnd);

		if (!is_corrupted_request(req)) {
			if (req->request.complete) {
				req->request.complete(&mgadget->ep[req->bEnd].end_point, (struct usb_request *)req);
			}
#if MCD_DEBUG>0
			success++;
		} else {
			fail++;
#endif
			DBG(4, "pRequest=%p on bEnd=%d is corrupted!!\n", req, req->bEnd);

		}
	}
	spin_unlock_irqrestore(&SchedulerQueue.lock, flags);
	PRLOCK();
}

DECLARE_TASKLET(mcd_gadget_complete_requests, gadget_complete_requests, 0);

/* 
 * @pre spinlock_is_locked( &mgadget->ep[bEnd] );
 */
int mcd_gadget_complete_request(struct usb_request *pRequest, int status)
{
	struct mcd_gadget_request *req = (struct mcd_gadget_request *)pRequest;

	DBG(2, "<==\n");
	if (!req) {
		ERR("scheduling completition a null request\n");
		DBG(2, "==>\n");
		return -EINVAL;
	}

	DBG(3, "scheduling completition of %p on bEnd=0x%x\n", req, (req) ? req->bEnd : -1);

#ifdef MCD_PARANOID
	ASSERT_SPINLOCK_LOCKED(&pEnd->Lock);
#endif

	if (is_corrupted_request(req)) {
		ERR("skipping completition of a corrupted request %p on bEnd=%d\n", req, req->bEnd);
		return -EINVAL;
	}

	if (is_queued_for_completition(req)) {
		ERR("skipping completition of queued request %p on bEnd=%d\n", req, req->bEnd);
		return -EINVAL;
	}

	pRequest->status = status;

	spin_lock(&SchedulerQueue.lock);
	INIT_LIST_HEAD(&req->completition_list);
	list_add_tail(&(req->completition_list), &(SchedulerQueue.req_list));
#if MCD_DEBUG >0
	req->bQueued = 1;
#endif
	spin_unlock(&SchedulerQueue.lock);
	DBG(3, "%p queued to completition list\n", req);

	tasklet_schedule(&mcd_gadget_complete_requests);
	DBG(2, "==>\n");
	return 0;
}

/*
 * @pre the end should be locked
 */
void mcd_gadget_flush_ep_requests(struct mcd_gep *pEnd, const int status)
{
	unsigned long flags;
	struct usb_request *req;
	uint8_t bEnd = EP_NUMBER(pEnd);
	struct mcd_gadget *mgadget = mcd_first_gadget;

	DBG(-1, "<== ep%d, pEnd=%p\n", bEnd, pEnd);

	spin_lock_irqsave(&pEnd->Lock, flags);

	req = mcd_get_current_request(pEnd);

	if (!bEnd) {
		if (req) {
			req->status = status;
			mcd_gadget_complete_ep0_request();
		}
	} else {
		if (req) {
			stop_dma(mgadget, bEnd);
		}
		list_for_each_entry(req, &pEnd->req_list, list) {
			if (-EINPROGRESS == req->status) {
				mcd_gadget_complete_request(req, status);
			}
		}
	}

	spin_unlock_irqrestore(&pEnd->Lock, flags);

	mcd_gadget_fifo_flush_flush_only((struct usb_ep *)pEnd);

	DBG(2, "==>\n");
}

/**************************************************************************
Tx/Rx Data
**************************************************************************/

inline int get_ep_packet_size(struct mcd_gadget *mgadget, uint8_t bEnd)
{
	int size = mgadget->ep[bEnd].wPacketSize;
	return size;
}

/*
 * @pre spin_is_locked(mgadget->ep[bEnd].Lock)
 */
static void txstate(struct mcd_gadget *mgadget, struct mcd_gadget_request *req)
{
	unsigned long flags;
	uint32_t dwFifoCount = 0;
	uint16_t wCsrVal;

	struct mcd_device *mdev = mgadget_to_mdev(mgadget);
	void *regs = mdev_link_reg(mdev);
	const uint8_t bEnd = req->bEnd;
	struct usb_request *pRequest = &req->request;
	struct mcd_gep *pEnd = &mgadget->ep[bEnd];
#ifdef MCD_DMA
	struct mcd_dma_controller *pDmaController = mdev->dmac;
	struct mcd_dma_channel *pDmaChannel = mgadget->ep[bEnd].pDmaChannel;
	uint8_t bAllocChannel;
	uint8_t bDmaOk = false;
	uint8_t bReqDmaState = USB_REQ_CPU;
	uint8_t bMode = 0;
	void *pData = NULL;
#endif

#ifdef MCD_PARANOID
	if (!pRequest) {
		ERR("cannot call txstate without request\n");
		return;
	}

	ASSERT_SPINLOCK_UNLOCKED(&mgadget->Lock);
	ASSERT_SPINLOCK_UNLOCKED(&pEnd->Lock);
#endif

	DBG(2, "TestBulk txstate 1\n");
#ifdef MCD_DMA
	/* usb_request¤dma != NULL 
	   ġpEnd->bTrafficType "ȥž"Ǥʤ 
	   ġžĹ0Ǥʤ 0 != pRequest->length */
	if (((dma_addr_t) NULL != pRequest->dma) && (USB_ENDPOINT_XFER_CONTROL != pEnd->bTrafficType) && (0 != pRequest->length)
	    && (0 != (pRequest->length - pRequest->actual))) {
		/* pEnd->bTrafficType  "Bulkž"ξ硢ĺǸΥ硼ȥѥåȤǤʤ  Mode1 */
		if ((USB_ENDPOINT_XFER_BULK == pEnd->bTrafficType)
		    && ((pRequest->length - pRequest->actual) >= (get_ep_packet_size(mgadget, bEnd)))) {
			if (mgadget->ep[bEnd].bDma_ena == USB_MCD_DMA_SETON) {
				/* DMA transfer */
				bReqDmaState = USB_REQ_DMA_MODE1;
				DBG(1, "txstate[1]: bReqDmaState %d \n", bReqDmaState);
			} else {
				/* Forced CPU transfer */
				bReqDmaState = USB_REQ_CPU;
				DBG(1, "txstate[2]: bReqDmaState %d \n", bReqDmaState);
			}
#ifdef MCD_DMA_FEWBYTE_USING_CPU_TRANSFER
			if (MCD_DMA_FEWBYTE >= (pRequest->length - pRequest->actual)) {
				bReqDmaState = USB_REQ_CPU;
				DBG(1, "txstate[4]: bReqDmaState %d \n", bReqDmaState);
			}
#endif
		} else {	/* ¾ξ Mode0 */
			if (mgadget->ep[bEnd].bDma_ena == USB_MCD_DMA_SETON) {
				/* DMA transfer */
				bReqDmaState = USB_REQ_DMA_MODE0;
				DBG(1, "txstate[5]: bReqDmaState %d \n", bReqDmaState);
			} else {
				/* Forced CPU transfer */
				bReqDmaState = USB_REQ_CPU;
				DBG(1, "txstate[6]: bReqDmaState %d \n", bReqDmaState);
			}

#ifdef USB_DMA_SHORTPKT_USING_CPU_TRANSFER	/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
			if ((pRequest->length - pRequest->actual) < (get_ep_packet_size(mgadget, bEnd))) {
				bReqDmaState = USB_REQ_CPU;
				DBG(1, "txstate[8]: bReqDmaState %d \n", bReqDmaState);
			}
#endif
#ifdef MCD_DMA_FEWBYTE_USING_CPU_TRANSFER
			if (MCD_DMA_FEWBYTE >= (pRequest->length - pRequest->actual)) {
				bReqDmaState = USB_REQ_CPU;
				DBG(1, "txstate[9]: bReqDmaState %d \n", bReqDmaState);
			}
#endif
		}

	} else {
		bReqDmaState = USB_REQ_CPU;	/* DMAʤ CPUž */
		DBG(1, "txstate[10]: bReqDmaState %d \n", bReqDmaState);
	}
#endif				/* MCD_DMA */
	DBG(2, "TestBulk txstate 2\n");

#ifdef MCD_DMA
	switch (bReqDmaState) {
	case USB_REQ_DMA_MODE1:	/* Mode1ǺǸΥ硼ȥѥåȰʳΥǡžԤ */
		DBG(1, "txstate[11]: bReqDmaState %d \n", bReqDmaState);
		dwFifoCount = (pRequest->length - pRequest->actual) - ((pRequest->length - pRequest->actual) % get_ep_packet_size(mgadget, bEnd));
		DBG(2, "txstate: length=%d actual=%d dwFifoCount=%d \n", pRequest->length, pRequest->actual, dwFifoCount);
		DBG(2, "TestBulk USB_REQ_DMA_MODE1\n");
		break;

	case USB_REQ_DMA_MODE0:
		DBG(2, "TestBulk USB_REQ_DMA_MODE0\n");
		DBG(1, "txstate[12]: bReqDmaState %d \n", bReqDmaState);

	case USB_REQ_CPU:
		DBG(1, "txstate[13]: bReqDmaState %d \n", bReqDmaState);
		/* ǡĹDmaMode0/CPU žǡĹ(MAXPʲ)˥å */
		dwFifoCount = (uint32_t) min(get_ep_packet_size(mgadget, bEnd), (int)(pRequest->length - pRequest->actual));
		break;
	default:		/* 顼 */
		break;
	}
#else
	dwFifoCount = (uint32_t) min(get_ep_packet_size(mgadget, bEnd), (int)(pRequest->length - pRequest->actual));
#endif				/* MCD_DMA */

	if (pRequest->length - pRequest->actual < 0) {
		ERR("NEGATIVE FIFOCOUNT! pRequest->actual=%d, dwFifoCount=%d, pRequest->length=%d\n", pRequest->actual, dwFifoCount, pRequest->length);
		dwFifoCount = 0;
	}

	if (pRequest->actual + dwFifoCount > pRequest->length) {
		ERR("trying to write PAST length! pRequest->actual=%d, dwFifoCount=%d, pRequest->length=%d\n", pRequest->actual, dwFifoCount, pRequest->length);
		dwFifoCount = pRequest->length - pRequest->actual;
	}
	DBG(3, "bEnd=0x%x, pEnd->wPacketSize=0x%x, dwFifoCount=%d, %s\n", bEnd, pEnd->wPacketSize, dwFifoCount, dump_usb_request(pRequest));
	spin_lock_irqsave(&mdev->lock, flags);
	mcd_select_end(regs, bEnd);
	wCsrVal = MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd);
	DBG(2, "TestBulk MCD_O_HDRC_TXCSR=%x \n", wCsrVal);
	if (pDmaChannel) {	/* DMA꤬Ƥ */
		if (MCD_DMA_STATUS_FREE !=	/* DMAžνλǧ */
		    mcd_dma_get_channel_status(pDmaChannel)) {
			spin_unlock_irqrestore(&mdev->lock, flags);
			return;
		}
	}
	if ((MCD_M_TXCSR_FIFONOTEMPTY == (wCsrVal & MCD_M_TXCSR_FIFONOTEMPTY))
	    || (MCD_M_TXCSR_TXPKTRDY == (wCsrVal & MCD_M_TXCSR_TXPKTRDY))) {
		spin_unlock_irqrestore(&mdev->lock, flags);
		return;
	}
	if (USB_ENDPOINT_XFER_ISOC == pEnd->bTrafficType) {
		wCsrVal |= MCD_M_TXCSR_ISO;
	}

	DBG(2, "TestBulk txstate 3\n");

#ifdef MCD_DMA
	if (0 != dwFifoCount) {
		switch (bReqDmaState) {
		case USB_REQ_DMA_MODE1:
			DBG(1, "txstate[14]: bReqDmaState %d \n", bReqDmaState);
			bMode = 1;
		case USB_REQ_DMA_MODE0:
			DBG(1, "txstate[15]: bReqDmaState %d \n", bReqDmaState);
			bAllocChannel = false;

			if (!pDmaController) {
				bReqDmaState = USB_REQ_CPU;
				DBG(1, "txstate[16]: bReqDmaState %d \n", bReqDmaState);
				break;
			}
			if (!pDmaChannel) {
				pDmaChannel = mgadget->ep[bEnd].pDmaChannel =
				    mcd_dma_allocate_channel(pDmaController->pPrivateData, bEnd, true, pEnd->bTrafficType, pEnd->wPacketSize);
				bAllocChannel = true;
			}
			if (pDmaChannel) {
				pEnd->dwRequestSize = min(pRequest->length, pDmaChannel->dwMaxLength);
				bDmaOk = mcd_dma_program_channel(pDmaChannel, mcd_gadget_dma_tx_cb, pData,
								 pEnd->wPacketSize, bMode, (uint8_t *) (pRequest->dma + (void *)pRequest->actual), dwFifoCount);
				DBG(2, "TestBulk dwFifoCount=%d \n", dwFifoCount);
				if (bDmaOk) {
					pDmaChannel->dwActualLength = 0L;
					dwFifoCount = 0;
					DBG(2, "TestBulk USB_REQ_DMA_MODE1/0 set success\n");
				} else {
					/* ǡĹDmaMode0/CPU žǡĹ(MAXPʲ)˥å */
					dwFifoCount = (uint32_t) min(get_ep_packet_size(mgadget, bEnd), (int)(pRequest->length - pRequest->actual));
					bReqDmaState = USB_REQ_CPU;
					DBG(1, "txstate[17]: bReqDmaState %d \n", bReqDmaState);

					if (bAllocChannel) {
						mcd_dma_release_channel(pDmaChannel);
						mgadget->ep[bEnd].pDmaChannel = pDmaChannel = NULL;
					}
					break;
				}
			} else {
				/* ǡĹDmaMode0/CPU žǡĹ(MAXPʲ)˥å */
				dwFifoCount = (uint32_t) min(get_ep_packet_size(mgadget, bEnd), (int)(pRequest->length - pRequest->actual));
				bReqDmaState = USB_REQ_CPU;
				DBG(1, "txstate[18]: bReqDmaState %d \n", bReqDmaState);
				break;
			}
		default:
			break;
		}
	} else {
		bReqDmaState = USB_REQ_CPU;
		DBG(1, "txstate[19]: bReqDmaState %d \n", bReqDmaState);
	}
#endif				/* MCD_DMA */

#ifdef MCD_DMA

	wCsrVal |= MCD_M_TXCSR_MODE;

	switch (bReqDmaState) {
	case USB_REQ_DMA_MODE1:
		DBG(1, "txstate[20]: bReqDmaState %d \n", bReqDmaState);
		wCsrVal |= (MCD_M_TXCSR_AUTOSET | MCD_M_TXCSR_DMAENAB | MCD_M_TXCSR_DMAMODE);
		MCD_WRITE_CSR8(regs, MCD_O_HDRC_TXCSR + 1, bEnd, (0x00ff & (wCsrVal >> 8)));
		bDmaOk = mcd_dma_enable_int_channel(pDmaChannel);
		if (!bDmaOk) {
			printk("DMA_TEST " "DmaEnableIntChannel Error %d\n", bDmaOk);
			spin_unlock_irqrestore(&mdev->lock, flags);
			return;
		}
		/* mcd_gadget_enable_end()Ƚʣ(MAXP) */
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXMAXP, bEnd, get_ep_packet_size(mgadget, bEnd));
		bDmaOk = mcd_dma_start_channel(pDmaChannel);
		if (!bDmaOk) {
			printk("DMA_TEST " "pfDmaStartChannel Error %d\n", bDmaOk);
			spin_unlock_irqrestore(&mdev->lock, flags);
			return;
		}
		DBG(2, "TestBulk CSR=%x \n", MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd));
		DBG(2, "TestBulk USB_REQ_DMA_MODE1 set end\n");
		break;
	case USB_REQ_DMA_MODE0:
		DBG(1, "txstate[21]: bReqDmaState %d \n", bReqDmaState);
		wCsrVal |= (MCD_M_TXCSR_DMAENAB);
		wCsrVal &= (~MCD_M_TXCSR_AUTOSET);
		MCD_WRITE_CSR8(regs, MCD_O_HDRC_TXCSR + 1, bEnd, (0x00ff & (wCsrVal >> 8)));
		bDmaOk = mcd_dma_enable_int_channel(pDmaChannel);
		if (!bDmaOk) {
			printk("DMA_TEST " "DmaEnableIntChannel Error %d\n", bDmaOk);
			spin_unlock_irqrestore(&mdev->lock, flags);
			return;
		}
		/* mcd_gadget_enable_end()Ƚʣ(MAXP) */
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXMAXP, bEnd, get_ep_packet_size(mgadget, bEnd));
		bDmaOk = mcd_dma_start_channel(pDmaChannel);
		if (!bDmaOk) {
			printk("DMA_TEST " "pfDmaStartChannel Error %d\n", bDmaOk);
			spin_unlock_irqrestore(&mdev->lock, flags);
			return;
		}
		DBG(2, "TestBulk pEnd->wPacketSize=%d \n", pEnd->wPacketSize);
		DBG(2, "TestBulk get_ep_packet_size(mgadget, bEnd)=%d \n", get_ep_packet_size(mgadget, bEnd));
		DBG(2, "TestBulk pDmaChannel=%u \n", (uint32_t) pDmaChannel);
		DBG(2, "TestBulk CSR=%x \n", MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd));
		DBG(2, "TestBulk USB_REQ_DMA_MODE0 set end\n");
		break;
	case USB_REQ_CPU:
		DBG(1, "txstate[22]: bReqDmaState %d \n", bReqDmaState);
		wCsrVal |= (MCD_M_TXCSR_TXPKTRDY);
		wCsrVal &= ~(MCD_M_TXCSR_DMAENAB);
		if (dwFifoCount > MCD_FIFO_MAXSIZE) {
			/* check the upper 2Byte */
			/* length of packet err */
			DBG(2, "DMA_TEST " "TX dwFifoCount Error Size:%d\n", dwFifoCount);
			spin_unlock_irqrestore(&mdev->lock, flags);
			return;
		}
		mcd_device_load_fifo(regs, bEnd, (uint16_t) dwFifoCount, (uint8_t *) (pRequest->buf + pRequest->actual));
		pRequest->actual += dwFifoCount;
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd, wCsrVal);
		DBG(2, "TestBulk USB_REQ_CPU set end\n");
		break;
	}
#else

	wCsrVal = MCD_M_TXCSR_MODE | MCD_M_TXCSR_TXPKTRDY;

	mcd_device_load_fifo(regs, bEnd, (uint16_t) dwFifoCount, (uint8_t *) (pRequest->buf + pRequest->actual));
	pRequest->actual += dwFifoCount;
	MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd, wCsrVal);

#endif				/* MCD_DMA */

	DBG(3, "bEnd=0x%x, pEnd->wPacketSize=0x%x, dwFifoCount=%d, %s\n", bEnd, pEnd->wPacketSize, dwFifoCount, dump_usb_request(pRequest));
	spin_unlock_irqrestore(&mdev->lock, flags);
}

/* Data ready for a request; called from IRQ; no need to disable the
 * IRQS bc they are already disabled.
 */
void mcd_gadget_service_tx_avail(struct mcd_gadget *mgadget, uint8_t bEnd)
{
	uint16_t wCsrVal;
	struct usb_request *pRequest;
	struct mcd_device *mdev = mgadget_to_mdev(mgadget);
	struct mcd_gep *pEnd = &mgadget->ep[bEnd];
	void *regs = mdev_link_reg(mdev);
#ifdef MCD_DMA
	struct mcd_dma_channel *pDmaChannel = mgadget->ep[bEnd].pDmaChannel;
	uint8_t bReqDmaState = USB_REQ_CPU;
	void *pData = NULL;
#endif				/* MCD_DMA */
	DBG(2, "<==\n");

#ifdef MCD_PARANOID
	ASSERT_SPINLOCK_UNLOCKED(&mdev->lock);
	ASSERT_SPINLOCK_UNLOCKED(&pEnd->Lock);
#endif

	DBG(2, "TestBulk mcd_gadget_service_tx_avail  1\n");
	do {
		spin_lock(&mdev->lock);
		mcd_select_end(regs, bEnd);
		wCsrVal = MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd);

		if (wCsrVal & MCD_M_TXCSR_P_SENTSTALL) {
			wCsrVal &= ~MCD_M_TXCSR_P_SENTSTALL;
			wCsrVal &= ~MCD_M_TXCSR_TXPKTRDY;
			MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd, wCsrVal);

			/* wait for clear! */
			spin_unlock(&mdev->lock);
			break;
		}
		DBG(2, "TestBulk mcd_gadget_service_tx_avail  1.1\n");

		if (wCsrVal & MCD_M_TXCSR_P_UNDERRUN) {
			wCsrVal &= ~MCD_M_TXCSR_P_UNDERRUN;
			wCsrVal &= ~MCD_M_TXCSR_TXPKTRDY;
			MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd, wCsrVal);
			DBG(3, "underrun on ep%d\n", bEnd);
		}
		spin_unlock(&mdev->lock);

		DBG(2, "TestBulk mcd_gadget_service_tx_avail  1.2\n");
		spin_lock(&pEnd->Lock);
		pRequest = mcd_get_current_request(pEnd);
		DBG(2, "TestBulk mcd_gadget_service_tx_avail  1.3\n");
		if (pRequest) {
			DBG(2, "TestBulk mcd_gadget_service_tx_avail  1.4\n");

			if (pRequest->status != -EINPROGRESS) {
				spin_unlock(&pEnd->Lock);
				break;
			}
#ifdef MCD_DMA
			DBG(2, "TestBulk mcd_gadget_service_tx_avail  2\n");

			/* usb_request¤dma != NULL 
			   ġpEnd->bTrafficType "ȥž"Ǥʤ 
			   ġžĹ0Ǥʤ 0 != pRequest->length */
			if (((dma_addr_t) NULL != pRequest->dma) && (USB_ENDPOINT_XFER_CONTROL != pEnd->bTrafficType) && (0 != pRequest->length)
			    && (0 != (pRequest->length - pRequest->actual))) {
				/* pEnd->bTrafficType  "Bulkž"ξ硢ĺǸΥ硼ȥѥåȤǤʤ  Mode1 */
				if ((USB_ENDPOINT_XFER_BULK == pEnd->bTrafficType) &&
				    ((pRequest->length - pRequest->actual) >= (get_ep_packet_size(mgadget, bEnd)))) {
					if (mgadget->ep[bEnd].bDma_ena == USB_MCD_DMA_SETON) {
						/* DMA transfer */
						bReqDmaState = USB_REQ_DMA_MODE1;
						DBG(1, "TxAvail[1]: bReqDmaState %d \n", bReqDmaState);
					} else {
						/* CPU transfer */
						bReqDmaState = USB_REQ_CPU;
						DBG(1, "TxAvail[2]: bReqDmaState %d \n", bReqDmaState);
					}

#ifdef MCD_DMA_FEWBYTE_USING_CPU_TRANSFER
					if (MCD_DMA_FEWBYTE >= (pRequest->length - pRequest->actual)) {
						bReqDmaState = USB_REQ_CPU;
						DBG(1, "TxAvail[4]: bReqDmaState %d \n", bReqDmaState);
					}
#endif
				} else {	/* ¾ξ Mode0 */
					if (mgadget->ep[bEnd].bDma_ena == USB_MCD_DMA_SETON) {
						/* DMA transfer */
						bReqDmaState = USB_REQ_DMA_MODE0;
						DBG(1, "TxAvail[5]: bReqDmaState %d \n", bReqDmaState);
					} else {
						/* CPU transfer */
						bReqDmaState = USB_REQ_CPU;
						DBG(1, "TxAvail[6]: bReqDmaState %d \n", bReqDmaState);
					}

#ifdef USB_DMA_SHORTPKT_USING_CPU_TRANSFER	/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
					if ((pRequest->length - pRequest->actual) < (get_ep_packet_size(mgadget, bEnd))) {
						bReqDmaState = USB_REQ_CPU;
						DBG(1, "TxAvail[8]: bReqDmaState %d \n", bReqDmaState);
					}
#endif
#ifdef MCD_DMA_FEWBYTE_USING_CPU_TRANSFER
					if (MCD_DMA_FEWBYTE >= (pRequest->length - pRequest->actual)) {
						bReqDmaState = USB_REQ_CPU;
						DBG(1, "TxAvail[9]: bReqDmaState %d \n", bReqDmaState);
					}
#endif
				}

			} else {
				bReqDmaState = USB_REQ_CPU;	/* DMAʤ CPUž */
				DBG(1, "TxAvail[10]: bReqDmaState %d \n", bReqDmaState);
			}

			DBG(2, "TestBulk mcd_gadget_service_tx_avail  3\n");
			if (pDmaChannel) {	/* DMA꤬Ƥ */
				DBG(2, "TestBulk pfDmaGetChannelStatus  %d\n", (mcd_dma_get_channel_status(pDmaChannel)));
				DBG(2, "TestBulk MCD_DMA_STATUS_FREE=%d \n", MCD_DMA_STATUS_FREE);
				DBG(2, "TestBulk MCD_DMA_STATUS_BUSY=%d \n", MCD_DMA_STATUS_BUSY);
				DBG(2, "TestBulk MCD_DMA_STATUS_BUS_ABORT=%d \n", MCD_DMA_STATUS_BUS_ABORT);
				DBG(2, "TestBulk pDmaChannel->dwActualLength=%d \n", pDmaChannel->dwActualLength);
				if (MCD_DMA_STATUS_FREE ==	/* DMAžνλǧ */
				    mcd_dma_get_channel_status(pDmaChannel)) {
					DBG(2, "TestBulk mcd_gadget_service_tx_avail  3.1\n");
					DBG(2, "TestBulk wCsrVal=%x  \n", wCsrVal);
					if ((NULL != mgadget->ep[bEnd].pDmaChannel->pData)
					    && (true == *((uint8_t *) (mgadget->ep[bEnd].pDmaChannel->pData)))) {
						DBG(2, "TestBulk mcd_gadget_service_tx_avail  3.2\n");
						switch (bReqDmaState) {
						case USB_REQ_DMA_MODE1:
							DBG(1, "TxAvail[11]: bReqDmaState %d \n", bReqDmaState);

							pRequest->actual += pDmaChannel->dwActualLength;
							/* Mode1ξϡDMAλ */
							mcd_dma_stop_channel(pDmaChannel);
							mcd_dma_disable_int_channel(pDmaChannel);
							mcd_dma_program_channel(pDmaChannel,
										mcd_gadget_dma_tx_cb,
										pData, pEnd->wPacketSize, 0, (uint8_t *) (pRequest->dma + (void *)
															  pRequest->actual), 0);
							wCsrVal &= ~(MCD_M_TXCSR_DMAENAB);
							MCD_WRITE_CSR8(regs, MCD_O_HDRC_TXCSR + 1, bEnd, (0x00ff & (wCsrVal >> 8)));
							wCsrVal &= ~(MCD_M_TXCSR_AUTOSET | MCD_M_TXCSR_DMAMODE);	/* AUTOSETΥꥢ */
							MCD_WRITE_CSR8(regs, MCD_O_HDRC_TXCSR + 1, bEnd, (0x00ff & (wCsrVal >> 8)));

							wCsrVal = MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd);
							DBG(2, "TestBulk wCsrVal=%x \n", wCsrVal);
							if (MCD_M_TXCSR_FIFONOTEMPTY == (wCsrVal & MCD_M_TXCSR_FIFONOTEMPTY)) {
								spin_unlock(&pEnd->Lock);
								return;
							}
							DBG(2, "TestBulk CSR=%x \n", MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd));
							DBG(2, "TestBulk mcd_gadget_service_tx_avail 4 Mode1\n");
							break;
						case USB_REQ_DMA_MODE0:
							DBG(1, "TxAvail[12]: bReqDmaState %d \n", bReqDmaState);

							DBG(2, "TestBulk wCsrVal=%x \n", wCsrVal);
							DBG(2, "TestBulk CSR    =%x \n", MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd));
							mcd_dma_stop_channel(pDmaChannel);
							mcd_dma_disable_int_channel(pDmaChannel);
							wCsrVal &= ~MCD_M_TXCSR_DMAENAB;
							/* Mode0ǤϡFIFOؤDMAžäƤUSB¹ԡ */
							wCsrVal |= (MCD_M_TXCSR_TXPKTRDY);
							MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd, wCsrVal);
							DBG(2, "TestBulk mcd_gadget_service_tx_avail  4 Mode0\n");
							DBG(2, "TestBulk wCsrVal=%x \n", wCsrVal);
							DBG(2, "TestBulk CSR    =%x \n", MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd));
							if ((USB_ENDPOINT_XFER_BULK != pEnd->bTrafficType)
							    || ((USB_ENDPOINT_XFER_BULK == pEnd->bTrafficType)
								&& ((pRequest->length - pRequest->actual) < get_ep_packet_size(mgadget, bEnd)))) {
								spin_unlock(&pEnd->Lock);
								return;
							}
						}
					} else {
						DBG(2, "TestBulk mcd_gadget_service_tx_avail  4.1 \n");
						DBG(2, "TestBulk CSR=%x \n", MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd));
						pRequest->actual += pDmaChannel->dwActualLength;
						DBG(2, "TestBulk pRequest->actual%d  \n", pRequest->actual);
						if (pDmaChannel) {
							/* Mode1ξϡDMAλ */
							mcd_dma_stop_channel(pDmaChannel);
							mcd_dma_disable_int_channel(pDmaChannel);
							mcd_dma_program_channel(pDmaChannel,
										mcd_gadget_dma_tx_cb,
										pData, pEnd->wPacketSize, 0, (uint8_t *) (pRequest->dma + (void *)
															  pRequest->actual), 0);
							wCsrVal &= ~(MCD_M_TXCSR_DMAENAB);
							MCD_WRITE_CSR8(regs, MCD_O_HDRC_TXCSR + 1, bEnd, (0x00ff & (wCsrVal >> 8)));
							wCsrVal &= ~(MCD_M_TXCSR_AUTOSET | MCD_M_TXCSR_DMAMODE);	/* AUTOSETΥꥢ */
							MCD_WRITE_CSR8(regs, MCD_O_HDRC_TXCSR + 1, bEnd, (0x00ff & (wCsrVal >> 8)));
						}
					}
				} else {
					spin_unlock(&pEnd->Lock);
					return;
				}
			} else {
				if ((MCD_M_TXCSR_FIFONOTEMPTY == (wCsrVal & MCD_M_TXCSR_FIFONOTEMPTY)) ||
				    (MCD_M_TXCSR_TXPKTRDY == (wCsrVal & MCD_M_TXCSR_TXPKTRDY))) {
					spin_unlock(&pEnd->Lock);
					return;
				}
				/* ʤˤ⤷ʤ */
			}

#endif				/* MCD_DMA */

			if ((pRequest->actual == pRequest->length)) {	/* žλ */
				DBG(3, "completing pRequest=%p on bEnd=0x%x\n", pRequest, bEnd);
#ifdef MCD_DMA
				if (pDmaChannel) {
					mcd_dma_stop_channel(pDmaChannel);
					mcd_dma_disable_int_channel(pDmaChannel);
					/* (pfDmaReleaseBuffer();) */
					mcd_dma_release_channel(pDmaChannel);
					mgadget->ep[bEnd].pDmaChannel = pDmaChannel = false;
				}
				/* use the zero flag to request the transmission setting it to
				 * zero, after the zero packet has been submitted */
				if (pRequest->zero
				    && pRequest->status == -EINPROGRESS
				    && pRequest->length && pRequest->length % get_ep_packet_size(mgadget, bEnd) == 0 && !(wCsrVal & MCD_M_TXCSR_P_SENTSTALL)) {
					pRequest->zero = 0;
					wCsrVal |= MCD_M_TXCSR_TXPKTRDY;
					MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd, wCsrVal);
					DBG(2, "TestBulk Zero set PktRdy\n");
					spin_unlock(&pEnd->Lock);
					break;
				}
#endif				/* MCD_DMA */
				wCsrVal &= ~(MCD_M_TXCSR_MODE);
				wCsrVal &= ~MCD_M_TXCSR_TXPKTRDY;
				/* EP¦ˤǽˤƤ */
				MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd, wCsrVal);
				DBG(2, "TestBulk mcd_gadget_service_tx_avail() 5\n");
				if ((NULL != pRequest) && (pRequest->status == -EINPROGRESS)) {
					pRequest->status = 0;	/* ꥯȾ֤λˤ done */
					/* λ ComplitionХå */
					mcd_gadget_complete_request(pRequest, pRequest->status);
				}
			} else {
				/* 󥻥뽪λƤʤСν */
				if ((NULL != pRequest) && (pRequest->status != -ECONNRESET)) {
					spin_unlock(&pEnd->Lock);
					txstate(mgadget, (struct mcd_gadget_request *)pRequest);
					spin_lock(&pEnd->Lock);
				}
			}
		} else {
		}
		spin_unlock(&pEnd->Lock);

	} while (false);

	DBG(2, "==>\n");
}

/* ------------------------------------------------------------ */

/*
 * @pre spinlock_is_locked(&mgadget->ep[bEnd])
 */
static void rxstate(struct mcd_gadget *mgadget, struct mcd_gadget_request *req)
{
	uint16_t wCsrVal = 0;
	unsigned long flags;
	const uint8_t bEnd = req->bEnd;
	struct usb_request *pRequest = &req->request;
	struct mcd_device *mdev = mgadget_to_mdev(mgadget);
	void *regs = mdev_link_reg(mdev);
	struct mcd_gep *pEnd = &mgadget->ep[bEnd];
	uint32_t dwFifoCount = 0;

#ifdef MCD_DMA
	struct mcd_dma_controller *pDmaController = mdev->dmac;
	struct mcd_dma_channel *pDmaChannel = mgadget->ep[bEnd].pDmaChannel;
	uint8_t bAllocChannel;
	uint8_t bDmaOk = false;
	uint8_t bReqDmaState = USB_REQ_CPU;
	uint8_t bMode = 0;
	void *pData = NULL;
#endif				/* MCD_DMA */

#ifdef MCD_PARANOID
	ASSERT_SPINLOCK_UNLOCKED(&mdev->lock);
#endif

	spin_lock_irqsave(&mdev->lock, flags);
	mcd_select_end(regs, bEnd);
	wCsrVal = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd);

	DBG(2, "TestBulk rxstate() 1\n");
#ifdef MCD_DMA
	/* usb_request¤dma != NULL 
	   ġpEnd->bTrafficType "ȥž"Ǥʤ 
	   ġžĹ0Ǥʤ 0 != pRequest->length */
	if (((dma_addr_t) NULL != pRequest->dma) && (USB_ENDPOINT_XFER_CONTROL != pEnd->bTrafficType) && (0 != pRequest->length)
	    && (0 != (pRequest->length - pRequest->actual))) {
		/* pEnd->bTrafficType  "Bulkž"ξ硢
		   ĺǸΥ硼ȥѥåȤǤʤ  Mode1 */
		if ((USB_ENDPOINT_XFER_BULK == pEnd->bTrafficType)
		    && ((pRequest->length - pRequest->actual) >= (get_ep_packet_size(mgadget, bEnd)))) {
			if (mgadget->ep[bEnd].bDma_ena == USB_MCD_DMA_SETON) {
				/* DMA transfer */
				bReqDmaState = USB_REQ_DMA_MODE1;
				DBG(1, "rxstate[1]: bReqDmaState %d \n", bReqDmaState);
			} else {
				/* CPU transfer */
				bReqDmaState = USB_REQ_CPU;
				DBG(1, "rxstate[2]: bReqDmaState %d \n", bReqDmaState);
			}

#ifdef MCD_DMA_FEWBYTE_USING_CPU_TRANSFER
			if (MCD_DMA_FEWBYTE >= (pRequest->length - pRequest->actual)) {
				bReqDmaState = USB_REQ_CPU;
				DBG(1, "rxstate[4]: bReqDmaState %d \n", bReqDmaState);
			}
#endif
		} else {	/* ¾ξ Mode0 */
			if (mgadget->ep[bEnd].bDma_ena == USB_MCD_DMA_SETON) {
				/* DMA transfer */
				bReqDmaState = USB_REQ_DMA_MODE0;
				DBG(1, "rxstate[5]: bReqDmaState %d \n", bReqDmaState);
			} else {
				/* CPU transfer */
				bReqDmaState = USB_REQ_CPU;
				DBG(1, "rxstate[6]: bReqDmaState %d \n", bReqDmaState);
			}

#ifdef USB_DMA_SHORTPKT_USING_CPU_TRANSFER	/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
			if ((pRequest->length - pRequest->actual) < (get_ep_packet_size(mgadget, bEnd))) {
				bReqDmaState = USB_REQ_CPU;
				DBG(1, "rxstate[8]: bReqDmaState %d \n", bReqDmaState);
			}
#endif
#ifdef MCD_DMA_FEWBYTE_USING_CPU_TRANSFER
			if (MCD_DMA_FEWBYTE >= (pRequest->length - pRequest->actual)) {
				bReqDmaState = USB_REQ_CPU;
				DBG(1, "rxstate[9]: bReqDmaState %d \n", bReqDmaState);
			}
#endif
		}

	} else {
		bReqDmaState = USB_REQ_CPU;	/* DMAʤ CPUž */
		DBG(1, "rxstate[10]: bReqDmaState %d \n", bReqDmaState);

	}

	DBG(2, "TestBulk bReqDmaState=%d \n", bReqDmaState);
	DBG(2, "TestBulk RXCOUNT=%d \n", (MCD_READ_CSR16(regs, MCD_O_HDRC_RXCOUNT, bEnd)));

	if (USB_INIT_REQ_PHASE1 == pEnd->bIsInitRequest) {
		DBG(2, "TestBulk USB_INIT_REQ_PHASE1 \n");
		if (USB_REQ_DMA_MODE1 == bReqDmaState) ;
		else {
			wCsrVal &= (~MCD_M_RXCSR_DMAENAB);
			MCD_WRITE_CSR8(regs, MCD_O_HDRC_RXCSR + 1, bEnd, (0x00ff & (wCsrVal >> 8)));
			goto exit_rxstate;
		}
	} else if (USB_INIT_REQ_PHASE2 == pEnd->bIsInitRequest) {
		if (pDmaChannel) {
			wCsrVal = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd);
			dwFifoCount = (uint32_t) MCD_READ_CSR16(regs, MCD_O_HDRC_RXCOUNT, bEnd);
			if ((wCsrVal & MCD_M_RXCSR_RXPKTRDY)
			    && (dwFifoCount < (uint32_t) get_ep_packet_size(mgadget, bEnd))) {
				if ((1 == pRequest->zero) && (0 == pRequest->actual) && (0 == dwFifoCount)) {
					mcd_dma_start_channel(pDmaChannel);
				}
				DBG(2, "TestBulk USB_INIT_REQ_PHASE2.1 \n");
				;
			} else {
				DBG(2, "TestBulk USB_INIT_REQ_PHASE2.2 \n");
				mcd_dma_start_channel(pDmaChannel);
				goto exit_rxstate;
			}
		} else {
			DBG(2, "TestBulk USB_INIT_REQ_PHASE2.3 \n");
			if ((uint16_t) NULL == (wCsrVal & MCD_M_RXCSR_RXPKTRDY))
				goto exit_rxstate;
		}
	} else if ((USB_REQ_DMA_MODE1 == bReqDmaState) && (NULL == pDmaChannel)) {
		/* USB_REQ_DMA_MODE1 ν DMA ch ˶ʤǤCPUž륱 */
		bReqDmaState = USB_REQ_CPU;
	}
#endif				/* MCD_DMA */

#ifdef MCD_DMA
	if ((wCsrVal & MCD_M_RXCSR_RXPKTRDY) || (USB_REQ_DMA_MODE1 == bReqDmaState)) {
#else
	if (wCsrVal & MCD_M_RXCSR_RXPKTRDY) {	/* } */
#endif

		DBG(2, "TestBulk rxstate() 2\n");
#ifdef MCD_DMA
		/* ⤷DMA¹ʤ (DMA Mode1硼ȥѥåȤˤǤξ) */
		if ((MCD_M_RXCSR_DMAMODE == (wCsrVal & MCD_M_RXCSR_DMAMODE)) &&
		    (false != pDmaChannel) && (MCD_DMA_STATUS_FREE != mcd_dma_get_channel_status(pDmaChannel))) {

			DBG(2, "TestBulk RX ShortPacket \n");
			DBG(2, "TestBulk pDmaChannel->dwActualLength = %d \n", pDmaChannel->dwActualLength);
			DBG(2, "TestBulk wCsrVal = %x \n", MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd));
			DBG(2, "TestBulk DmaGetChannelStatus(pDmaChannel) = %d \n", mcd_dma_get_channel_status(pDmaChannel));
			DBG(2, "TestBulk RXCOUNT = %d \n", MCD_READ_CSR16(regs, MCD_O_HDRC_RXCOUNT, bEnd));

			pRequest->actual += pDmaChannel->dwActualLength;	/* ޤǤDMAžʬ򥫥ȥå */

			/* ǽ0ѥåȤ̵뤹ϡDMA Mode1ONΤޤޡ */
			dwFifoCount = (uint32_t) MCD_READ_CSR16(regs, MCD_O_HDRC_RXCOUNT, bEnd);
			if (!((1 == pRequest->zero) && (0 == pRequest->actual) && (0 == dwFifoCount))) {

				mcd_dma_stop_channel(pDmaChannel);	/* DmaߡDma */
				mcd_dma_disable_int_channel(pDmaChannel);
				mcd_dma_program_channel(pDmaChannel, mcd_gadget_dma_rx_cb, pData,
							pEnd->wPacketSize, 0, (uint8_t *) (pRequest->dma + (void *)pRequest->actual), 0);
				/* (pfDmaReleaseBuffer();) */
				mcd_dma_release_channel(pDmaChannel);
				mgadget->ep[bEnd].pDmaChannel = pDmaChannel = false;
				DBG(2, "TestBulk pDmaChannel=%u \n", (uint32_t) pDmaChannel);

				wCsrVal &= ~(MCD_M_RXCSR_AUTOCLEAR | MCD_M_RXCSR_DMAMODE);	/* AUTOSETΥꥢ */
				MCD_WRITE_CSR8(regs, MCD_O_HDRC_RXCSR + 1, bEnd, (0x00ff & (wCsrVal >> 8)));
			}

			if (mgadget->ep[bEnd].bDma_ena == USB_MCD_DMA_SETON) {
				/* DMA transfer */
				bReqDmaState = USB_REQ_DMA_MODE0;
				DBG(1, "rxstate[11]: bReqDmaState %d \n", bReqDmaState);
			} else {
				/* CPU transfer */
				bReqDmaState = USB_REQ_CPU;
				DBG(1, "rxstate[12]: bReqDmaState %d \n", bReqDmaState);
			}

#ifdef USB_DMA_SHORTPKT_USING_CPU_TRANSFER	/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
			dwFifoCount = (uint32_t) MCD_READ_CSR16(regs, MCD_O_HDRC_RXCOUNT, bEnd);
			if (dwFifoCount < (uint32_t) (get_ep_packet_size(mgadget, bEnd))) {
				bReqDmaState = USB_REQ_CPU;
				DBG(1, "rxstate[14]: bReqDmaState %d \n", bReqDmaState);
			}
#endif
#ifdef MCD_DMA_FEWBYTE_USING_CPU_TRANSFER
			dwFifoCount = (uint32_t) MCD_READ_CSR16(regs, MCD_O_HDRC_RXCOUNT, bEnd);
			if (MCD_DMA_FEWBYTE >= dwFifoCount) {
				bReqDmaState = USB_REQ_CPU;
				DBG(1, "rxstate[15]: bReqDmaState %d \n", bReqDmaState);
			}
#endif
		}

		switch (bReqDmaState) {
		case USB_REQ_DMA_MODE1:
			DBG(1, "rxstate[16]: bReqDmaState %d \n", bReqDmaState);
			dwFifoCount = (pRequest->length - pRequest->actual) - ((pRequest->length - pRequest->actual) % get_ep_packet_size(mgadget, bEnd));
			pEnd->dwRxOverrun = 0;
			pEnd->bRxMaxpOverrun = false;

			DBG(2, "TestBulk USB_REQ_DMA_MODE1 dwFifoCount=%d\n", dwFifoCount);
			break;

		case USB_REQ_DMA_MODE0:
			DBG(1, "rxstate[17]: bReqDmaState %d \n", bReqDmaState);

		case USB_REQ_CPU:
			DBG(1, "rxstate[18]: bReqDmaState %d \n", bReqDmaState);
			dwFifoCount = (uint32_t) MCD_READ_CSR16(regs, MCD_O_HDRC_RXCOUNT, bEnd);
			/* Mode0/CPURXCOUNTʬǡžԤ */
#ifdef USB_DMA_SHORTPKT_USING_CPU_TRANSFER	/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
			if (dwFifoCount < (uint32_t) (get_ep_packet_size(mgadget, bEnd))) {
				bReqDmaState = USB_REQ_CPU;
				DBG(1, "rxstate[19]: bReqDmaState %d \n", bReqDmaState);
			}
#endif
#ifdef MCD_DMA_FEWBYTE_USING_CPU_TRANSFER
			if (MCD_DMA_FEWBYTE >= dwFifoCount) {
				bReqDmaState = USB_REQ_CPU;
				DBG(1, "rxstate[20]: bReqDmaState %d \n", bReqDmaState);
			}
#endif
			if (dwFifoCount > (get_ep_packet_size(mgadget, bEnd))) {
				bReqDmaState = USB_REQ_CPU;
				/* MAXPʾMAXPĹޤǡΤFlush롣 */
				pEnd->dwRxOverrun = dwFifoCount - ((uint32_t) min(get_ep_packet_size(mgadget, bEnd),
										  (int)(pRequest->length - pRequest->actual)));
				pEnd->bRxMaxpOverrun = true;

				dwFifoCount = (uint32_t) min(get_ep_packet_size(mgadget, bEnd), (int)(pRequest->length - pRequest->actual));
			} else if ((dwFifoCount > (pRequest->length - pRequest->actual))) {
				/* ׵Ĺʾ׵ĹޤǡΤFlush롣 */
				/* OUT_BULK  СĹ򲱤Ƥ */
				pEnd->dwRxOverrun = dwFifoCount - (pRequest->length - pRequest->actual);
				pEnd->bRxMaxpOverrun = false;

				dwFifoCount = pRequest->length - pRequest->actual;
			}
			DBG(2, "TestBulk USB_REQ_DMA_MODE0/CPU wFifoCount=%d\n", dwFifoCount);
			break;
		default:
			/* 顼 */
			break;
		}
#else
		/* CPURXCOUNTʬΥǡžԤ */
		dwFifoCount = (uint32_t) MCD_READ_CSR16(regs, MCD_O_HDRC_RXCOUNT, bEnd);
#endif				/* MCD_DMA */

		DBG(2, "TestBulk rxstate() 3\n");
		DBG(2, "TestBulk rxstate() 4\n");
#ifdef MCD_DMA
		if (0 != dwFifoCount) {
			switch (bReqDmaState) {
			case USB_REQ_DMA_MODE1:
				DBG(1, "rxstate[21]: bReqDmaState %d \n", bReqDmaState);
				bMode = 1;

			case USB_REQ_DMA_MODE0:
				DBG(1, "rxstate[22]: bReqDmaState %d \n", bReqDmaState);
				bAllocChannel = false;

				if (!pDmaController) {
					DBG(1, "rxstate[23]: bReqDmaState %d \n", bReqDmaState);
					bReqDmaState = USB_REQ_CPU;
					break;
				}
				if (!pDmaChannel) {
					pDmaChannel = mgadget->ep[bEnd].pDmaChannel =
					    mcd_dma_allocate_channel(pDmaController->pPrivateData, bEnd, false, pEnd->bTrafficType, pEnd->wPacketSize);
					bAllocChannel = true;
				}
				if (pDmaChannel) {
					pEnd->dwRequestSize = min(pRequest->length, pDmaChannel->dwMaxLength);
					bDmaOk = mcd_dma_program_channel(pDmaChannel,
									 mcd_gadget_dma_rx_cb, pData,
									 pEnd->wPacketSize, bMode,
									 (uint8_t *) (pRequest->dma + (void *)pRequest->actual), dwFifoCount);
					if (bDmaOk) {
						pDmaChannel->dwActualLength = 0L;
						dwFifoCount = 0;
					} else {
						/* USB_REQ_DMA_MODE1 ν DMA ch 
						   ʤǤCPUž륱 */
						if (bAllocChannel) {
							mcd_dma_release_channel(pDmaChannel);
							mgadget->ep[bEnd].pDmaChannel = pDmaChannel = NULL;
						}
						if (USB_INIT_REQ_PHASE1 == pEnd->bIsInitRequest) {
							wCsrVal &= (~MCD_M_RXCSR_DMAENAB);
							MCD_WRITE_CSR8(regs, MCD_O_HDRC_RXCSR + 1, bEnd, (0x00ff & (wCsrVal >> 8)));
							goto exit_rxstate;
						}
						/* CPUǥѥåȤȤ˥ǡžԤ */
						dwFifoCount = (uint32_t) MCD_READ_CSR16(regs, MCD_O_HDRC_RXCOUNT, bEnd);
						bReqDmaState = USB_REQ_CPU;
						DBG(1, "rxstate[24]: bReqDmaState %d \n", bReqDmaState);

						break;
					}
				} else {
					/* USB_REQ_DMA_MODE1 ν DMA ch 
					   ʤǤCPUž륱 */
					if (USB_INIT_REQ_PHASE1 == pEnd->bIsInitRequest) {
						wCsrVal &= (~MCD_M_RXCSR_DMAENAB);
						MCD_WRITE_CSR8(regs, MCD_O_HDRC_RXCSR + 1, bEnd, (0x00ff & (wCsrVal >> 8)));
						goto exit_rxstate;
					}
					/* CPUǥѥåȤȤ˥ǡžԤ */
					dwFifoCount = (uint32_t) MCD_READ_CSR16(regs, MCD_O_HDRC_RXCOUNT, bEnd);
					bReqDmaState = USB_REQ_CPU;
					DBG(1, "rxstate[25]: bReqDmaState %d \n", bReqDmaState);
					break;
				}

			default:	/* CPUžξϤʤˤ⤷ʤ */
				break;
			}
		} else {
			bReqDmaState = USB_REQ_CPU;	/* ǡĹ0ξϡCUPž */
			DBG(1, "rxstate[26]: bReqDmaState %d \n", bReqDmaState);
		}
#endif				/* MCD_DMA */

		DBG(2, "TestBulk rxstate() 5\n");
#ifdef MCD_DMA
		switch (bReqDmaState) {
		case USB_REQ_DMA_MODE1:
			DBG(1, "rxstate[27]: bReqDmaState %d \n", bReqDmaState);
			wCsrVal |= (MCD_M_RXCSR_AUTOCLEAR | MCD_M_RXCSR_DMAENAB | MCD_M_RXCSR_DMAMODE);
			MCD_WRITE_CSR8(regs, MCD_O_HDRC_RXCSR + 1, bEnd, (0x00ff & (wCsrVal >> 8)));
			bDmaOk = mcd_dma_enable_int_channel(pDmaChannel);
			if (!bDmaOk) {
				DBG(2, "TestBulk DmaEnableIntChannel() Error \n");
			}
			/* mcd_gadget_enable_end()Ƚʣ(MAXP) */
			MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXMAXP, bEnd, get_ep_packet_size(mgadget, bEnd));
			DBG(2, "TestBulk RXMAXP = %d \n", MCD_READ_CSR16(regs, MCD_O_HDRC_RXMAXP, bEnd));
			if ((USB_INIT_REQ_PHASE1 != pEnd->bIsInitRequest)) {
				DBG(2, "TestBulk DmaStartChannel() \n");
				DBG(2, "TestBulk (pEnd->bIsInitRequest)=%d \n", (pEnd->bIsInitRequest));
				bDmaOk = mcd_dma_start_channel(pDmaChannel);
				if (!bDmaOk) {
					DBG(2, "TestBulk DmaStartChannel() Error \n");
				}
			}
			DBG(2, "TestBulk USB_REQ_DMA_MODE1\n");
			DBG(2, "TestBulk wCsrVal = %x \n", MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd));
			break;
		case USB_REQ_DMA_MODE0:
			DBG(1, "rxstate[28]: bReqDmaState %d \n", bReqDmaState);
			wCsrVal |= (MCD_M_RXCSR_DMAENAB);
			wCsrVal &= (~(MCD_M_RXCSR_AUTOCLEAR | MCD_M_RXCSR_DMAMODE));
			MCD_WRITE_CSR8(regs, MCD_O_HDRC_RXCSR + 1, bEnd, (0x00ff & (wCsrVal >> 8)));
			bDmaOk = mcd_dma_enable_int_channel(pDmaChannel);
			if (!bDmaOk) {
				DBG(2, "TestBulk DmaEnableIntChannel() Error \n");
			}
			if ((USB_INIT_REQ_PHASE1 != pEnd->bIsInitRequest)) {
				DBG(2, "TestBulk DmaStartChannel() \n");
				bDmaOk = mcd_dma_start_channel(pDmaChannel);
				if (!bDmaOk) {
					DBG(2, "TestBulk DmaStartChannel() Error \n");
				}
			}
			DBG(2, "TestBulk USB_REQ_DMA_MODE0\n");
			break;
		case USB_REQ_CPU:
			DBG(1, "rxstate[29]: bReqDmaState %d \n", bReqDmaState);
#endif				/* MCD_DMA */
			DBG(2, "TestBulk USB_REQ_CPU 1\n");
			if ((NULL != pRequest) && (pRequest->status != -ECONNRESET)
			    && (pRequest->actual <= pRequest->length)) {
				if (dwFifoCount > MCD_FIFO_MAXSIZE) {
					/* check the upper 2Byte */
					/* length of packet err */
					DBG(2, "DMA_TEST " "RX dwFifoCount Error Size:%d\n", dwFifoCount);
					goto exit_rxstate;
				}
				mcd_device_unload_fifo(regs, bEnd, (uint16_t) dwFifoCount, (uint8_t *) (pRequest->buf + pRequest->actual));
				DBG(2, "TestBulk USB_REQ_CPU 2\n");
				pRequest->actual += dwFifoCount;
				/* DBG(3, "wFifoCount=%d, bEnd=0x%x, pEnd->wPacketSize=0x%x, wCount=0x%x, %s", \
				   wFifoCount, bEnd, pEnd->wPacketSize, wCount, dump_usb_request(pRequest));  */

				DBG(2, "TestBulk USB_REQ_CPU 3\n");
/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
#if  defined(USB_DMA_SHORTPKT_USING_CPU_TRANSFER) || defined(MCD_DMA_FEWBYTE_USING_CPU_TRANSFER)
				if ((pRequest->actual == pRequest->length) || (dwFifoCount < (uint32_t) get_ep_packet_size(mgadget, bEnd))
				    || (true == pEnd->bRxMaxpOverrun)) {

					/* ǽ0ѥåȼξDMAENABӥåȤϤΤޤ */
					if ((USB_ENDPOINT_XFER_BULK == pEnd->bTrafficType)
					    && (!((pRequest->zero == 1) && (pRequest->actual == 0)))) {
						if (mgadget->ep[bEnd].bDma_ena == USB_MCD_DMA_SETON) {
							/* DMA transfer */
							/* Modyfied for AUTOCLEAR Clear *//* AUTOSETΥꥢ */
							wCsrVal &= ~(MCD_M_RXCSR_AUTOCLEAR | MCD_M_RXCSR_DMAMODE);
							MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd, wCsrVal);
							wCsrVal |= MCD_M_RXCSR_DMAENAB;
						}
					}
					if (0 == pEnd->dwRxOverrun || true == pEnd->bRxMaxpOverrun) {
						wCsrVal &= ~MCD_M_RXCSR_RXPKTRDY;
					}
				} else {
					if ((USB_ENDPOINT_XFER_BULK == pEnd->bTrafficType)
					    && ((dma_addr_t) NULL != pRequest->dma)) {
						if (mgadget->ep[bEnd].bDma_ena == USB_MCD_DMA_SETON) {
							/* DMA transfer */
							/* Modyfied for AUTOCLEAR Clear *//* AUTOSETΥꥢ */
							wCsrVal &= ~(MCD_M_RXCSR_AUTOCLEAR | MCD_M_RXCSR_DMAMODE);
							MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd, wCsrVal);
						}
					}
					wCsrVal &= ~MCD_M_RXCSR_RXPKTRDY;
				}
#endif
				MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd, wCsrVal);

				DBG(2, "TestBulk USB_REQ_CPU 4\n");
				DBG(2, "TestBulk pRequest->actual=%u \n", pRequest->actual);
				DBG(2, "TestBulk pRequest->length=%u \n", pRequest->length);
				DBG(2, "TestBulk get_ep_packet_size(mgadget, bEnd)=%u \n", get_ep_packet_size(mgadget, bEnd));
				DBG(2, "TestBulk dwFifoCount=%u \n", dwFifoCount);
				/* reach the end or short packet detected */
				if ((pRequest->actual == pRequest->length) || (dwFifoCount < (uint32_t) get_ep_packet_size(mgadget, bEnd))
				    || (true == pEnd->bRxMaxpOverrun)) {
					if (pRequest->zero && pRequest->status == -EINPROGRESS && pRequest->actual == 0) {
						pRequest->zero = 0;
						goto exit_rxstate;
					}
					spin_lock(&pEnd->Lock);
#ifdef USB_DMA_SHORTPKT_USING_CPU_TRANSFER	/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
					if (pDmaChannel) {	/* DMA꤬Ƥ */
						if ((NULL != mgadget->ep[bEnd].pDmaChannel->pData) &&
						    (true != *((uint8_t *) (mgadget->ep[bEnd].pDmaChannel->pData)))) {
							spin_unlock(&pEnd->Lock);
							goto exit_rxstate;
						}
						mcd_dma_stop_channel(pDmaChannel);
						mcd_dma_disable_int_channel(pDmaChannel);
						/* (pfDmaReleaseBuffer();) */
						mcd_dma_release_channel(pDmaChannel);
						mgadget->ep[bEnd].pDmaChannel = pDmaChannel = false;
						DBG(2, "TestBulk pDmaChannel=%u \n", (uint32_t) pDmaChannel);
					}
#endif
					DBG(3, "completing pRequest=%p on bEnd=0x%x\n", pRequest, bEnd);
					/* OUT_BULK ХξΥ顼ȡshort_not_okΥå */
					if ((NULL != pRequest) && (pRequest->status == -EINPROGRESS)) {
						if (0 == pEnd->dwRxOverrun) {
							if (!((1 == pRequest->short_not_ok) &&
							      ((0 != ((pRequest->actual) % (get_ep_packet_size(mgadget, bEnd))))
							       || (0 == dwFifoCount)))) {
								pRequest->status = 0;
							} else {
								pRequest->status = -EREMOTEIO;
							}
							/* async complete */
							mcd_gadget_complete_request(pRequest, pRequest->status);
						} else {
							if (true == pEnd->bRxMaxpOverrun) {
								pRequest->status = -ECOMM;
							} else {
								pRequest->status = -EOVERFLOW;
							}
							pRequest->actual = pRequest->actual + pEnd->dwRxOverrun;
							pEnd->dwRxOverrun = 0;
							pEnd->bRxMaxpOverrun = false;
							mcd_gadget_complete_request(pRequest, pRequest->status);
						}

					}
					spin_unlock(&pEnd->Lock);
				}
				DBG(2, "TestBulk USB_REQ_CPU\n");
			}
#ifdef MCD_DMA
			break;
		default:
			/* 顼 */
			break;
		}
#endif				/* MCD_DMA */

	}
	DBG(2, "TestBulk rxstate() 6\n");
      exit_rxstate:
	spin_unlock_irqrestore(&mdev->lock, flags);
}

/* Data ready for a request; called from IRQ
 */
void mcd_gadget_service_rx_ready(struct mcd_gadget *mgadget, uint8_t bEnd)
{
	uint16_t wCsrVal;
	struct mcd_device *mdev = mgadget_to_mdev(mgadget);
	void *regs = mdev_link_reg(mdev);
	struct mcd_gep *pEnd = &mgadget->ep[bEnd];
	struct usb_request *pRequest = NULL;
#ifdef MCD_DMA
	struct mcd_dma_channel *pDmaChannel = mgadget->ep[bEnd].pDmaChannel;
	uint8_t bReqDmaState = USB_REQ_CPU;
	void *pData = NULL;
	uint32_t dwFifoCount = 0;
#endif				/* MCD_DMA */

	DBG(2, "<==\n");

#ifdef MCD_PARANOID
	ASSERT_SPINLOCK_UNLOCKED(&mdev->lock);
	ASSERT_SPINLOCK_LOCKED(&pEnd->Lock);
#endif

	DBG(2, "TestBulk mcd_gadget_service_rx_ready() 1\n");
	do {
		spin_lock(&mdev->lock);	/* executed from interrupt no need to disable them */
		mcd_select_end(regs, bEnd);
		wCsrVal = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd);

		if (wCsrVal & MCD_M_RXCSR_P_SENTSTALL) {
			wCsrVal &= ~MCD_M_RXCSR_P_SENTSTALL;
			MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd, wCsrVal);
			DBG(3, "received stall on ep%d\n", bEnd);
			spin_unlock(&mdev->lock);
			break;
		}

		if (wCsrVal & MCD_M_RXCSR_P_OVERRUN) {
			wCsrVal &= ~MCD_M_RXCSR_P_OVERRUN;
			MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd, wCsrVal);
			DBG(3, "received overrun on ep%d\n", bEnd);
		}

		spin_unlock(&mdev->lock);

		DBG(2, "TestBulk mcd_gadget_service_rx_ready() 2\n");
		DBG(2, "TestBulk (wCsrVal&MCD_M_RXCSR_RXPKTRDY)=%x \n", (wCsrVal & MCD_M_RXCSR_RXPKTRDY));
		/* analyze request if the ep is not inactive */
		spin_lock(&pEnd->Lock);
		pRequest = (pEnd->bInactive) ? NULL : mcd_get_current_request(pEnd);
		DBG(2, "TestBulk mcd_gadget_service_rx_ready() 3 pRequest=%u \n", (uint32_t) pRequest);
		DBG(2, "TestBulk (wCsrVal&MCD_M_RXCSR_RXPKTRDY)=%x \n", (wCsrVal & MCD_M_RXCSR_RXPKTRDY));

		if (pRequest) {

#ifdef MCD_DMA
			if (pRequest->status != -EINPROGRESS) {
				/* CurrentRequestꥯȥХåԤξ硢 */
				/* ߤ⤷;ʬߤäƤⲿ߽ʤ */
				spin_unlock(&pEnd->Lock);
				/* breakwhileȴƤǽ֥årxstate뤿return */
				return;
			}

			/* usb_request¤dma != NULL 
			   ġpEnd->bTrafficType "ȥž"Ǥʤ 
			   ġžĹ0Ǥʤ 0 != pRequest->length */
			if (((dma_addr_t) NULL != pRequest->dma) && (USB_ENDPOINT_XFER_CONTROL != pEnd->bTrafficType) && (0 != pRequest->length)
			    && (0 != (pRequest->length - pRequest->actual))) {
				/* pEnd->bTrafficType  "Bulkž"ξ硢
				   ĺǸΥ硼ȥѥåȤǤʤ  Mode1,
				   ġǤ˥ѥåȤƤʤ */
				if ((USB_ENDPOINT_XFER_BULK == pEnd->bTrafficType) &&
				    ((pRequest->length - pRequest->actual) >= (get_ep_packet_size(mgadget, bEnd)))) {
					if (mgadget->ep[bEnd].bDma_ena == USB_MCD_DMA_SETON) {
						/* DMA transfer */
						bReqDmaState = USB_REQ_DMA_MODE1;
						DBG(1, "RxReady[1]: bReqDmaState %d \n", bReqDmaState);
					} else {
						/* Forced CPU transfer */
						bReqDmaState = USB_REQ_CPU;
						DBG(1, "RxReady[2]: bReqDmaState %d \n", bReqDmaState);
					}

#ifdef MCD_DMA_FEWBYTE_USING_CPU_TRANSFER
					if (MCD_DMA_FEWBYTE >= (pRequest->length - pRequest->actual)) {
						bReqDmaState = USB_REQ_CPU;
						DBG(1, "RxReady[4]: bReqDmaState %d \n", bReqDmaState);
					}
#endif
				} else {	/* ¾ξ Mode0 */
					if (mgadget->ep[bEnd].bDma_ena == USB_MCD_DMA_SETON) {
						/* DMA transfer */
						bReqDmaState = USB_REQ_DMA_MODE0;
						DBG(1, "RxReady[5]: bReqDmaState %d \n", bReqDmaState);
					} else {
						/* CPU transfer */
						bReqDmaState = USB_REQ_CPU;
						DBG(1, "RxReady[6]: bReqDmaState %d \n", bReqDmaState);
					}

#ifdef USB_DMA_SHORTPKT_USING_CPU_TRANSFER	/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
					if ((pRequest->length - pRequest->actual) < (get_ep_packet_size(mgadget, bEnd))) {
						bReqDmaState = USB_REQ_CPU;
						DBG(1, "RxReady[8]: bReqDmaState %d \n", bReqDmaState);
					}
#endif
#ifdef MCD_DMA_FEWBYTE_USING_CPU_TRANSFER
					if (MCD_DMA_FEWBYTE >= (pRequest->length - pRequest->actual)) {
						bReqDmaState = USB_REQ_CPU;
						DBG(1, "RxReady[9]: bReqDmaState %d \n", bReqDmaState);
					}
#endif
				}

			} else {
				bReqDmaState = USB_REQ_CPU;	/* DMAʤ CPUž */
				DBG(1, "RxReady[10]: bReqDmaState %d \n", bReqDmaState);

			}

			if (pDmaChannel) {	/* DMA꤬Ƥ */
				if ((MCD_DMA_STATUS_FREE ==	/* DMAžνλǧ */
				     mcd_dma_get_channel_status(pDmaChannel))) {
					DBG(2, "TestBulk RxReady pDmaChannel->dwActualLength=%d \n", pDmaChannel->dwActualLength);
					if (0 == pDmaChannel->dwActualLength) {	/* ʤˤ⤷ʤ  (RxPktRdyΤ߼) */
					} else {
						/* DMAžĹ򡢼žĹ­ */
						pRequest->actual += pDmaChannel->dwActualLength;
						dwFifoCount = pDmaChannel->dwActualLength;
						pDmaChannel->dwActualLength = 0;

						/* Mode1ξϡDMAλ */
						if (MCD_M_RXCSR_DMAMODE == (wCsrVal & MCD_M_RXCSR_DMAMODE)) {
							mcd_dma_stop_channel(pDmaChannel);
							mcd_dma_disable_int_channel(pDmaChannel);
							mcd_dma_program_channel(pDmaChannel,
										mcd_gadget_dma_rx_cb,
										pData, pEnd->wPacketSize, 0, (uint8_t *) (pRequest->dma + (void *)
															  pRequest->actual), 0);
							/* ž̤λĺǸ˥硼ȥѥåȤ */
							if ((NULL != pRequest)
							    && (((pRequest->actual) < (pRequest->length))
								&& (0 == ((pRequest->actual) % (get_ep_packet_size(mgadget, bEnd)))))) {
								wCsrVal &= ~(MCD_M_RXCSR_DMAENAB);
								MCD_WRITE_CSR8(regs, MCD_O_HDRC_RXCSR + 1, bEnd, (0x00ff & (wCsrVal >> 8)));
								spin_unlock(&pEnd->Lock);
								return;	/* Modyfied for AUTOCLEAR Clear */
							} else {
								/* Modyfied for AUTOCLEAR Clear *//* AUTOSETΥꥢ */
								wCsrVal &= ~(MCD_M_RXCSR_AUTOCLEAR | MCD_M_RXCSR_DMAMODE);
							}

							MCD_WRITE_CSR8(regs, MCD_O_HDRC_RXCSR + 1, bEnd, (0x00ff & (wCsrVal >> 8)));
						}
						/* Mode0ξϡRXPKTRDYΥꥢ  */
						else if (MCD_M_RXCSR_DMAMODE != (wCsrVal & MCD_M_RXCSR_DMAMODE)) {
							mcd_dma_stop_channel(pDmaChannel);
							mcd_dma_disable_int_channel(pDmaChannel);
							mcd_dma_program_channel(pDmaChannel,
										mcd_gadget_dma_rx_cb,
										pData, pEnd->wPacketSize, 0, (uint8_t *) (pRequest->dma + (void *)
															  pRequest->actual), 0);
							if ((USB_ENDPOINT_XFER_BULK == pEnd->bTrafficType)) {
								wCsrVal |= MCD_M_RXCSR_DMAENAB;
								/* Modyfied for AUTOCLEAR Clear *//* AUTOSETΥꥢ */
								wCsrVal &= ~(MCD_M_RXCSR_AUTOCLEAR | MCD_M_RXCSR_DMAMODE);
								MCD_WRITE_CSR8(regs, MCD_O_HDRC_RXCSR + 1, bEnd, (0x00ff & (wCsrVal >> 8)));
								/* OverrunǤMAXPʲξϡ
								   MCD_M_RXCSR_RXPKTRDY򤪤Ȥʤ */
								if (0 == pEnd->dwRxOverrun || true == pEnd->bRxMaxpOverrun) {

									wCsrVal &= ~(MCD_M_RXCSR_RXPKTRDY);
								}
								MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd, wCsrVal);
							} else {
								wCsrVal &= ~(MCD_M_RXCSR_DMAENAB);
								/* OverrunǤMAXPʲξϡ
								   MCD_M_RXCSR_RXPKTRDY򤪤Ȥʤ */
								if (0 == pEnd->dwRxOverrun || true == pEnd->bRxMaxpOverrun) {

									wCsrVal &= ~(MCD_M_RXCSR_RXPKTRDY);
								}
								MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd, wCsrVal);
							}
						}
					}
				} else {
					/* DMAžλRxPtkRdyߤΤϡDmaMode1
					   硼ȥѥåȽλξ */
					if (MCD_M_RXCSR_DMAMODE == (wCsrVal & MCD_M_RXCSR_DMAMODE)) {
						if (0 == pEnd->bIsInitRequest) {
							/* Mode1ξϡ硼ȥѥåȤˤǤξν */
							if ((pRequest->status == -ECONNRESET)) {
								spin_unlock(&pEnd->Lock);
								return;
							}
							DBG(2, "TestBulk Short Packet! \n");
							DBG(2, "TestBulk wCsrVal = %x \n", MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd));
							DBG(2, "TestBulk DmaGetChannelStatus(pDmaChannel) = %d \n", mcd_dma_get_channel_status(pDmaChannel));
							DBG(2, "TestBulk RXMAXP = %d \n", MCD_READ_CSR16(regs, MCD_O_HDRC_RXMAXP, bEnd));

							spin_unlock(&pEnd->Lock);

							if ((1 == pRequest->zero)) {
								rxstate(mgadget, (struct mcd_gadget_request *)pRequest);	/* NULLѥåȼ */
								if ((0 == pRequest->zero)) {
									return;
								}
							} else {
								/* 硼ȥѥåȼ */
								rxstate(mgadget, (struct mcd_gadget_request *)pRequest);
							}
							spin_lock(&pEnd->Lock);
							MCD_WRITE_CSR8(regs, MCD_O_HDRC_RXCSR, bEnd, (0x00ff & wCsrVal));
#ifdef USB_DMA_SHORTPKT_USING_CPU_TRANSFER	/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
							bReqDmaState = USB_REQ_CPU;
							DBG(1, "RxReady[11]: bReqDmaState %d \n", bReqDmaState);
#else
							if (mgadget->ep[bEnd].bDma_ena == USB_MCD_DMA_SETON) {
								/* DMA transfer */
								bReqDmaState = USB_REQ_DMA_MODE0;
								DBG(1, "RxReady[12]: bReqDmaState %d \n", bReqDmaState);
							} else {
								/* CPU transfer */
								bReqDmaState = USB_REQ_CPU;
								DBG(1, "RxReady[13]: bReqDmaState %d \n", bReqDmaState);
							}
#endif
						}
					}
				}
			}
			spin_unlock(&pEnd->Lock);
#else
			spin_unlock(&pEnd->Lock);
			rxstate(mgadget, (struct mcd_gadget_request *)pRequest);
#endif				/* MCD_DMA */
		} else {
			spin_unlock(&pEnd->Lock);
			DBG(3, "Rx: bytes waiting on ep=0x%x\n", bEnd);
			return;
		}
	} while (false);

#ifdef MCD_DMA
	DBG(2, "TestBulk mcd_gadget_service_rx_ready() 4\n");
	DBG(2, "TestBulk (wCsrVal&MCD_M_RXCSR_RXPKTRDY)=%x \n", (wCsrVal & MCD_M_RXCSR_RXPKTRDY));
	/* žλޤϥ硼ȥѥåȤǴλ (DMAžξ) */
	/* MaxpOverData DmaMode0 is not used. */
	if ((((NULL != pRequest) && (((pRequest->actual) >= (pRequest->length)) || (0 != ((pRequest->actual) % (get_ep_packet_size(mgadget, bEnd))))))
	     && (pDmaChannel)) || (true == pEnd->bRxMaxpOverrun)) {
		if (pRequest->zero && pRequest->status == -EINPROGRESS && pRequest->actual == 0) {
			pRequest->zero = 0;
			return;
		}
		DBG(2, "TestBulk mcd_gadget_service_rx_ready() 5\n");
		DBG(2, "TestBulk (wCsrVal&MCD_M_RXCSR_RXPKTRDY)=%x \n", (wCsrVal & MCD_M_RXCSR_RXPKTRDY));
		spin_lock(&pEnd->Lock);
		if ((USB_REQ_DMA_MODE1 == bReqDmaState) || (USB_REQ_DMA_MODE0 == bReqDmaState)) {

			DBG(1, "RxReady[15]: bReqDmaState %d \n", bReqDmaState);
			pRequest = (pEnd->bInactive) ? NULL : mcd_get_current_request(pEnd);
			pDmaChannel = mgadget->ep[bEnd].pDmaChannel;
			if (pDmaChannel) {	/* DMA꤬Ƥ */
				if ((NULL != mgadget->ep[bEnd].pDmaChannel->pData)
				    && (true != *((uint8_t *) (mgadget->ep[bEnd].pDmaChannel->pData)))) {
					return;
				}
				mcd_dma_stop_channel(pDmaChannel);	/* ɲ */
				mcd_dma_disable_int_channel(pDmaChannel);
				/* (pfDmaReleaseBuffer();) */
				mcd_dma_release_channel(pDmaChannel);
				mgadget->ep[bEnd].pDmaChannel = pDmaChannel = NULL;
			}
		}
		/* ꥯȾ֤λˤ done */
		DBG(2, "TestBulk mcd_gadget_service_rx_ready() 6\n");
		/* OUT_BULK ХξΥ顼ȡshort_not_okΥå */
		if ((NULL != pRequest) && (pRequest->status == -EINPROGRESS)) {
			if (0 == pEnd->dwRxOverrun) {
				if (!((1 == pRequest->short_not_ok)
				      && ((0 != ((pRequest->actual) % (get_ep_packet_size(mgadget, bEnd))))
					  || (0 == dwFifoCount)))) {
					pRequest->status = 0;
					mcd_gadget_complete_request(pRequest, pRequest->status);	/* async complete */
				} else {
					pRequest->status = -EREMOTEIO;
					mcd_gadget_complete_request(pRequest, pRequest->status);
				}
			} else {
				if (true == pEnd->bRxMaxpOverrun) {
					pRequest->status = -ECOMM;
				} else {
					pRequest->status = -EOVERFLOW;
				}
				pRequest->actual = pRequest->actual + pEnd->dwRxOverrun;
				pEnd->dwRxOverrun = 0;
				pEnd->bRxMaxpOverrun = false;
				mcd_gadget_complete_request(pRequest, pRequest->status);
			}

		}
		spin_unlock(&pEnd->Lock);
		DBG(2, "TestBulk mcd_gadget_service_rx_ready() 7\n");
	} else {
		DBG(2, "TestBulk mcd_gadget_service_rx_ready() 5.1\n");
		DBG(2, "TestBulk (wCsrVal&MCD_M_RXCSR_RXPKTRDY)=%x \n", (wCsrVal & MCD_M_RXCSR_RXPKTRDY));
		/* 󥻥뽪λƤʤ,RXPKTRDYäƤ */
		if ((NULL != pRequest) && (pRequest->status != -ECONNRESET)
		    && ((wCsrVal & MCD_M_RXCSR_RXPKTRDY) || (USB_REQ_DMA_MODE1 == bReqDmaState))) {
			DBG(1, "RxReady[16]: bReqDmaState %d \n", bReqDmaState);

			spin_lock(&pEnd->Lock);
			rxstate(mgadget, (struct mcd_gadget_request *)pRequest);	/* μ */
			spin_unlock(&pEnd->Lock);
			DBG(2, "TestBulk pRequest->actual=%d \n", pRequest->actual);
		}
	}
#endif				/* MCD_DMA */
	DBG(2, "TestBulk mcd_gadget_service_rx_ready() 8\n");

	DBG(2, "==>\n");
}

int mcd_gadget_restart_request(struct mcd_gadget *mgadget, uint8_t bEnd, struct usb_request *pRequest)
{
	struct mcd_gadget_request *req = (struct mcd_gadget_request *)pRequest;

	DBG(2, "TestBulk mcd_gadget_restart_request 1\n");

	if (!pRequest) {
		DBG(3, "nothing to execute on nEnd=0x%x, (tx=%d)\n", bEnd, mgadget->ep[bEnd].bIsTx);
		return -EINVAL;
	}

	DBG(2, "TestBulk mcd_gadget_restart_request 2\n");
	DBG(3, "executing pRequest=%p on nEnd=0x%x, (tx=%d), pRequest->length=%d, "
	    "pRequest->actual=0x%x\n", pRequest, bEnd, mgadget->ep[bEnd].bIsTx, pRequest->length, pRequest->actual);

	if (mgadget->ep[bEnd].bIsTx) {
		DBG(2, "TestBulk mcd_gadget_restart_request 31\n");
		txstate(mgadget, req);
		DBG(2, "TestBulk mcd_gadget_restart_request 32\n");
	} else {
		rxstate(mgadget, req);
		DBG(2, "TestBulk mcd_gadget_restart_request 4\n");
	}
	DBG(2, "TestBulk mcd_gadget_restart_request 5\n");

	return 0;
}

/**************************************************************************
Gadget Functions
**************************************************************************/

/* new interface */
extern uint8_t mcd_gadget_service_ep0(struct mcd_gadget *mgadget);

void mcd_gadget_service_default_end(struct mcd_gadget *mgadget)
{
	mcd_gadget_service_ep0(mgadget);
}

/* ------------------------------------------------------------ */

/**
 * The endpoint selection policy should have already selected
 * an appropriate usb_ep from us, since we follow the naming convention.
 */
STATIC int mcd_gadget_enable_end(struct usb_ep *ep, const struct usb_endpoint_descriptor *desc)
{
	unsigned long flags;
	uint8_t bEnd = EP_NUMBER(ep);
	struct mcd_gadget *mgadget = mcd_first_gadget;
	uint8_t *regs = mdev_link_reg(mgadget_to_mdev(mgadget));
	uint16_t wCsrVal;
	uint16_t wIntrTxE;
	uint16_t wIntrRxE;
	struct mcd_gep *pEnd = (struct mcd_gep *)ep;

	DBG(2, "<==\n ep=%p, desc=%p, nEnd=0x%x\n", ep, desc, EP_NUMBER(ep));

#ifdef MCD_PARANOID
	ASSERT_SPINLOCK_UNLOCKED(&pEnd->Lock);

	if (!mcd_first_gadget) {
		ERR("Gadget not initialized, mgadget=NULL\n");
		return -EINVAL;
	}

	if (mcd_gadget_find_end(ep) < 0) {
		ERR("invalid endpoint ep=%p\n", ep);
		return -EINVAL;
	}

	if ((EP_NUMBER(ep) != (desc->bEndpointAddress & USB_ENDPOINT_NUMBER_MASK))) {
		return -EINVAL;
	}

	if (!mcd_gadget_is_idle(ep)) {
		WARN("Enabling ep%d with request queued!\n", EP_NUMBER(ep));
	};

#endif

	spin_lock_irqsave(&pEnd->Lock, flags);

	ep->maxpacket = desc->wMaxPacketSize;
	pEnd->wPacketSize = desc->wMaxPacketSize;
	pEnd->bTrafficType = desc->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK;
	pEnd->bIsTx = (desc->bEndpointAddress & USB_ENDPOINT_DIR_MASK) ? true : false;
	pEnd->bInactive = 0;

	INIT_LIST_HEAD(&pEnd->req_list);

	/* register setting */
	mcd_select_end(regs, bEnd);

	/* USB is not configured with High-bandwidth or split transaction */
	if (pEnd->bIsTx) {
		/*  */
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXMAXP, bEnd, pEnd->wPacketSize);
		/* եåFIFO */
		mcd_gadget_fifo_flush_flush_only(ep);
		/* ǡȥ򥯥ꥢ */
		/* STALLꥢ */
		mcd_gadget_set_halt(ep, 0);

		/* EPγߵ */
		wIntrTxE = MCD_READ16(regs, MCD_O_HDRC_INTRTXE);
		MCD_WRITE16(regs, MCD_O_HDRC_INTRTXE, wIntrTxE | (1 << bEnd));
	} else {
		/*  */
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXMAXP, bEnd, pEnd->wPacketSize);

		if (USB_ENDPOINT_XFER_BULK == pEnd->bTrafficType) {
			wCsrVal = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd);
			wCsrVal |= (MCD_M_RXCSR_DMAENAB);
			wCsrVal &= ~MCD_M_RXCSR_DISNYET;
			MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd, wCsrVal);
		} else if (USB_ENDPOINT_XFER_INT == pEnd->bTrafficType) {
			wCsrVal = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd);
			wCsrVal |= MCD_M_RXCSR_DISNYET;
			MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd, wCsrVal);
		} else {	/* USB_ENDPOINT_XFER_ISOC == pEnd->bTrafficType */
			wCsrVal = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd);
			wCsrVal &= ~MCD_M_RXCSR_DISNYET;
			MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd, wCsrVal);
		}

		/* եåFIFO */
		mcd_gadget_fifo_flush_flush_only(ep);
		/* ǡȥ򥯥ꥢ */
		/* STALLꥢ */
		mcd_gadget_set_halt(ep, 0);

		/* EPγߵ */
		wIntrRxE = MCD_READ16(regs, MCD_O_HDRC_INTRRXE);
		MCD_WRITE16(regs, MCD_O_HDRC_INTRRXE, wIntrRxE | (1 << bEnd));
	}

	spin_unlock_irqrestore(&pEnd->Lock, flags);

#if MCD_DEBUG > 0
	memcpy(&pEnd->epDescriptor, desc, sizeof(struct usb_endpoint_descriptor));
#endif

	DBG(2, "==>\n");
	return 0;
}

int mcd_gadget_enable_end_point(uint8_t bEnd, const struct usb_endpoint_descriptor *desc)
{
	struct mcd_gadget *mgadget = mcd_first_gadget;
#if MCD_DEBUG > 0
	if (!desc) {
		desc = &mgadget->ep[bEnd].epDescriptor;
	}
#endif
	if (!desc) {
		ERR("invalid description activating ep%d\n", bEnd);
		return -EINVAL;
	}
	return mcd_gadget_enable_end(&mgadget->ep[bEnd].end_point, desc);
}

/* Disable an endpoint.
 * @param struct usb_ep *ep the endpoint to disable.
 */
STATIC int mcd_gadget_disable_end(struct usb_ep *ep)
{
	unsigned long flags;
	uint8_t bEnd = EP_NUMBER(ep);
	struct mcd_gadget *mgadget = mcd_first_gadget;
	uint8_t *regs = mdev_link_reg(mgadget_to_mdev(mgadget));
	uint16_t wIntrTxE;
	uint16_t wIntrRxE;
	struct mcd_gep *pEnd = (struct mcd_gep *)ep;
	struct usb_request *req;
	int status;

	DBG(2, "<==\n");
	DBG(-1, "disabling ep %d\n", EP_NUMBER(ep));

#ifdef MCD_PARANOID
	ASSERT_SPINLOCK_UNLOCKED(&pEnd->Lock);

	if (!mcd_first_gadget) {
		ERR("Gadget not initialized, mgadget=NULL\n");
		return -EINVAL;
	}

	if (mcd_gadget_find_end(ep) < 0) {
		ERR("invalid endpoint ep=%p\n", ep);
		return -EINVAL;
	}
#endif

	spin_lock_irqsave(&pEnd->Lock, flags);

	((struct mcd_gep *)ep)->bInactive = 1;

	/* EPγԵ */
	if (pEnd->bIsTx) {
		/*  */
		wIntrTxE = MCD_READ16(regs, MCD_O_HDRC_INTRTXE);
		MCD_WRITE16(regs, MCD_O_HDRC_INTRTXE, wIntrTxE & ~(1 << bEnd));
	} else {
		/*  */
		wIntrRxE = MCD_READ16(regs, MCD_O_HDRC_INTRRXE);
		MCD_WRITE16(regs, MCD_O_HDRC_INTRRXE, wIntrRxE & ~(1 << bEnd));
	}

	spin_unlock_irqrestore(&pEnd->Lock, flags);

	req = mcd_get_current_request(pEnd);

	spin_lock_irqsave(&pEnd->Lock, flags);

	status = -ESHUTDOWN;

	if (!bEnd) {
		if (req) {
			req->status = status;
			mcd_gadget_complete_ep0_request();
		}
	} else if (req && req->status != -ESHUTDOWN) {
		struct usb_request *tmp;
		stop_dma(mgadget, bEnd);
		list_for_each_entry_safe(req, tmp, &pEnd->req_list, list) {
			if (-EINPROGRESS == req->status) {
				req->status = status;
				list_del(&req->list);
				spin_unlock_irqrestore(&pEnd->Lock, flags);
				if (req->complete) {
					req->complete(&pEnd->end_point, req);
				}
				spin_lock_irqsave(&pEnd->Lock, flags);
			}
		}
	}
	spin_unlock_irqrestore(&pEnd->Lock, flags);

	mcd_gadget_fifo_flush_flush_only(ep);

	DBG(2, "==>\n");
	return 0;
}

int mcd_gadget_disable_end_point(uint8_t bEnd)
{
	struct mcd_gadget *mgadget = mcd_first_gadget;
	return mcd_gadget_disable_end(&mgadget->ep[bEnd].end_point);
}

/**
 */
STATIC struct usb_request *mcd_gadget_alloc_request(struct usb_ep *ep, gfp_t gfp_flags)
{
	struct mcd_gadget_request *pRequest = NULL;

	DBG(2, "<==\n");

#ifdef MCD_PARANOID
	if (mcd_gadget_find_end(ep) < 0) {
		ERR("invalid endpoint ep=%p\n", ep);
		return NULL;
	}
#endif

	KMALLOC(pRequest, sizeof(struct mcd_gadget_request), gfp_flags);
	if (pRequest) {
		memset(pRequest, 0, sizeof(struct mcd_gadget_request));

		/* make sure it's fine! */
		INIT_LIST_HEAD(&pRequest->request.list);
		pRequest->completition_list.prev = 0;
		pRequest->completition_list.next = 0;
		pRequest->bEnd = EP_NUMBER(ep);

#if MCD_DEBUG > 0
		pRequest->bQueued = 0;
		pRequest->dwPadFront = MCD_PAD_FRONT;
		pRequest->dwPadBack = MCD_PAD_BACK;
#endif
	}

	DBG(2, "==> allocated request at %p for ep %d\n", pRequest, EP_NUMBER(ep));
	return (struct usb_request *)pRequest;
}

/**
 */
STATIC void mcd_gadget_free_request(struct usb_ep *ep, struct usb_request *req)
{
#if MCD_DEBUG > 0
	is_corrupted_request(req);
	if (req)
#endif
	{
		KFREE(req);
	}
	DBG(2, "freed request at %p\n", req);
}

/**
 */
STATIC void *mcd_gadget_alloc_buffer(struct usb_ep *ep, unsigned bytes, dma_addr_t * dma, gfp_t gfp_flags)
{
#ifdef MCD_PARANOID
	if (!mcd_first_gadget) {
		ERR("Gadget not initialized, mgadget=NULL\n");
		return 0;
	}

	if (mcd_gadget_find_end(ep) < 0) {
		ERR("invalid endpoint ep=%p\n", ep);
		return 0;
	}

	if (dma) {
		WARN("requested DMA for ep%d\n", EP_NUMBER(ep));
	}
#endif

	if (!EP_NUMBER(ep) && dma) {
		*dma = 0;
		dma = 0L;
	}

	/* PIO when DMA buffer pointer is NULL */
	if (!dma) {
		void *vadr;
		/* in case without DMAC */
		KMALLOC(vadr, bytes, gfp_flags);
		DBG(4, "malloced pAddr=%p, pDma=%p\n", vadr, dma);
		return vadr;
	}

	/* DMA is controlled in the endpoint initialization; no DMA for now... */
	return pool_alloc_buffer(mcd_first_gadget, bytes, gfp_flags, dma);
}

/**
 */
STATIC void mcd_gadget_free_buffer(struct usb_ep *ep, void *address, dma_addr_t dma, unsigned bytes)
{
	if (!dma) {
		KFREE(address);
	} else {
		pool_free_buffer(mcd_first_gadget, bytes, address, dma);
	}
}

/* Queue requests to endpoints for execution.
 * 
 * @param struct usb_ep *ep
 * @param struct usb_request *pRequest
 * @param gfp_t gfp_flags
 */

STATIC int mcd_gadget_queue(struct usb_ep *ep, struct usb_request *pRequest, gfp_t gfp_flags)
{
	const uint8_t bEnd = EP_NUMBER(ep);
	struct mcd_gep *pEnd = (struct mcd_gep *)ep;
	struct mcd_gadget *mgadget = mcd_first_gadget;
#ifdef MCD_DMA
	unsigned long flags;
#endif				/* MCD_DMA */
	DBG(2, "<==\n");

#ifdef MCD_PARANOID
	if (!mgadget) {
		ERR("Gadget not initialized, mgadget=NULL\n");
		return -EINVAL;
	}

	if (!pRequest) {
		ERR("pRequest=%p\n", pRequest);
		return -EINVAL;
	}

	if (mcd_gadget_find_end(ep) < 0) {
		ERR("invalid endpoint ep=%p\n", ep);
		return -EINVAL;
	}

	ASSERT_SPINLOCK_UNLOCKED(&pEnd->Lock);

	if (is_queued_for_completition(pRequest)) {
		ERR("pRequest=%p is in the completion list %s!\\n", pRequest, dump_node(&((struct mcd_gadget_request *)pRequest)->completition_list));
		return -EINVAL;
	}
#endif

	DBG(2, "TestBulk mcd_gadget_queue() 1 EP=%d \n", bEnd);
	pRequest->actual = 0;
	pRequest->status = -EINPROGRESS;
	((struct mcd_gadget_request *)pRequest)->bEnd = bEnd;
	/* Request was length=0,So it must be zero=0. */
	if (0 == pRequest->length)
		pRequest->zero = 0;
	INIT_LIST_HEAD(&pRequest->list);

	DBG(2, "TestBulk mcd_gadget_queue() 2\n");
	/* list count==1 pRequest is the head of the list */
	EP_SPIN_LOCK_IRQSAVE(ep, flags);
	list_add_tail(&(pRequest->list), &(pEnd->req_list));
	if (!pEnd->bInactive && (pRequest == mcd_get_current_request(pEnd))) {
		EP_SPIN_UNLOCK_IRQRESTORE(ep, flags);
		DBG(2, "TestBulk mcd_gadget_queue() 3\n");

		if (bEnd == 0) {
			DBG(-6, "executing on bEnd=0x%x, (%s), pRequest->length=%d\n", bEnd, pEnd->bIsTx ? "tx" : "rx", pRequest->length);

			DBG(2, "TestBulk mcd_gadget_queue() 4\n");
			switch (mgadget->bEnd0Stage) {
				/* sequence #1, TX State */
			case MCD_END0_STAGE_TX:
				mcd_gadget_ep0_txstate(mgadget);
				break;

				/* sequence #2, RX State */
			case MCD_END0_STAGE_RX:
				mcd_gadget_ep0_rxstart(mgadget);
				break;

				/* certain gadged may keep ep0 busy; g_file_storage
				   does it after a set_config command. */
			default:{
					const uint8_t code = (pRequest->actual == pRequest->length)
					    ? 0 : -EINVAL;
					pRequest->status = code;
					mcd_gadget_complete_ep0_request();
				} break;
			}

		} else {
			mcd_gadget_exec_request(ep, pRequest, gfp_flags);
		}
	} else {
		EP_SPIN_UNLOCK_IRQRESTORE(ep, flags);
	}

	DBG(2, "==>\n");
	return 0;
}

STATIC int mcd_gadget_exec_request(struct usb_ep *ep, struct usb_request *pRequest, gfp_t gfp_flags)
{				/* Modyfied for request queue restart */
	struct mcd_gadget *mgadget = mcd_first_gadget;
	const uint8_t bEnd = EP_NUMBER(ep);
	struct mcd_gep *pEnd = (struct mcd_gep *)ep;
#ifdef MCD_DMA
	unsigned long flags;
	uint16_t wIntrRxE;
	void *regs = mdev_link_reg(mgadget_to_mdev(mgadget));
#endif				/* MCD_DMA */

	if (pEnd->bIsTx) {
		mcd_gadget_restart_request(mgadget, EP_NUMBER(ep), pRequest);
	} else {

		DBG(2, "TestBulk mcd_gadget_queue() 5\n");

		spin_lock_irqsave(&mgadget_to_mdev(mgadget)->lock, flags);
		pEnd->bIsInitRequest = USB_INIT_REQ_PHASE1;
		wIntrRxE = MCD_READ16(regs, MCD_O_HDRC_INTRRXE);
		MCD_WRITE16(regs, MCD_O_HDRC_INTRRXE, wIntrRxE & ~(1 << bEnd));
		DBG(2, "TestBulk mcd_gadget_queue() " "wIntrRxE=%x  wIntrRxE & ~(1 << bEnd)=%x \n", wIntrRxE, MCD_READ16(regs, MCD_O_HDRC_INTRRXE));
		DBG(2, "TestBulk EP%d CSR=%x 2 INTNG \n", bEnd, MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd));

		mcd_select_end(regs, bEnd);
		mcd_gadget_restart_request(mgadget, EP_NUMBER(ep), pRequest);

		DBG(2, "TestBulk EP%d CSR=%x 3 INTNG \n", bEnd, MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd));
		DBG(2, "TestBulk mcd_gadget_queue() 5.1\n");
		MCD_WRITE16(regs, MCD_O_HDRC_INTRRXE, wIntrRxE);
		DBG(2, "TestBulk mcd_gadget_queue() wIntrRx =%x  \n", wIntrRxE);

		pEnd->bIsInitRequest = USB_INIT_REQ_PHASE2;

		mcd_gadget_restart_request(mgadget, EP_NUMBER(ep), pRequest);

		pEnd->bIsInitRequest = 0;

		DBG(2, "TestBulk mcd_gadget_queue() 5.2\n");
		DBG(2, "TestBulk EP%d CSR=%x 4 INTE \n", bEnd, MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd));
		spin_unlock_irqrestore(&mgadget_to_mdev(mgadget)->lock, flags);
	}

	return 0;
}

STATIC int mcd_gadget_dequeue(struct usb_ep *ep, struct usb_request *pRequest)
{
	DBG(2, "<==\n");
	DBG(3, "dequeuing from nEnd=0x%x, pEnd=%p, pRequest=%p\n", EP_NUMBER((struct mcd_gep *)ep), (struct mcd_gep *)ep, pRequest);

#ifdef MCD_PARANOID
	if (!mcd_first_gadget) {
		ERR("Gadget not initialized, mgadget=NULL\n");
		return -EINVAL;
	}

	if (mcd_gadget_find_end(ep) < 0) {
		ERR("invalid endpoint ep=%p\n", ep);
		return -EINVAL;
	}

	if (!pRequest) {
		return -EINVAL;	/* abort */
	}

	ASSERT_SPINLOCK_UNLOCKED(&((struct mcd_gep *)ep)->Lock);
#endif

	return complete_request_immediate(ep, pRequest, -ECONNRESET);
}

int mcd_gadget_set_halt(struct usb_ep *ep, int value)
{
	uint16_t wCsr;
	unsigned long flags;
	const uint8_t bEnd = EP_NUMBER(ep);
	struct mcd_gadget *mgadget = mcd_first_gadget;
	struct mcd_device *mdev = mgadget_to_mdev(mgadget);
	uint8_t *regs = mdev_link_reg(mdev);
	struct mcd_gep *pEnd = (struct mcd_gep *)ep;
	struct mcd_gadget_request *req = (struct mcd_gadget_request *)mcd_get_current_request(pEnd);

	DBG(3, "<== end=%d, value=%d\n", bEnd, value);

#ifdef MCD_PARANOID
	if (mcd_gadget_find_end(ep) < 0) {
		ERR("invalid endpoint ep=%p\n", ep);
		return -EINVAL;
	}
#endif

	spin_lock_irqsave(&mdev->lock, flags);
	mcd_select_end(regs, bEnd);

	if (0 == bEnd) {
		((struct mcd_gep *)ep)->bInactive = value;

		/* end 0 should never stall! */
		wCsr = MCD_READ_CSR16(regs, MCD_O_HDRC_CSR0, 0);
		if (value) {
			wCsr |= MCD_M_CSR0_P_SENDSTALL;
		} else {
			wCsr &= ~MCD_M_CSR0_P_SENDSTALL;
		}
		wCsr &= ~MCD_M_CSR0_TXPKTRDY;
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_CSR0, 0, wCsr);
		if (value) {
			mcd_gadget_fifo_flush_flush_only(ep);
		}
		spin_unlock_irqrestore(&mdev->lock, flags);
		return 0;
	}

	/* prevent further request to be executed */
	pEnd->bInactive = value;

	/* abort the current request? */
	if (value && !pEnd->bInactive && req && ((struct mcd_gep *)ep)->bIsTx) {

		wCsr = MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd);
		if (wCsr & MCD_M_TXCSR_FIFONOTEMPTY) {
			DBG(-1, "ep=%d: %d request(s), %s cannot halt it\n",
			    bEnd, mcd_gadget_queue_length(&pEnd->req_list), dump_usb_request(mcd_get_current_request(pEnd)));

			spin_unlock_irqrestore(&mdev->lock, flags);
			return -EAGAIN;
		}

	}

	if (((struct mcd_gep *)ep)->bIsTx) {
		wCsr = MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd);
		if (value) {
			wCsr |= MCD_M_TXCSR_P_SENDSTALL;
		} else {
			wCsr &= ~MCD_M_TXCSR_P_SENDSTALL;
			wCsr |= MCD_M_TXCSR_CLRDATATOG;
		}

		wCsr &= ~MCD_M_TXCSR_TXPKTRDY;
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, bEnd, wCsr);
		if (value) {
			mcd_gadget_fifo_flush_flush_only(ep);
		}
	} else {
		wCsr = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd);
		if (value) {
			wCsr |= MCD_M_RXCSR_P_SENDSTALL;
		} else {
			wCsr &= ~MCD_M_RXCSR_P_SENDSTALL;
			wCsr |= MCD_M_RXCSR_CLRDATATOG;
		}

		MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, bEnd, wCsr);
		if (value) {
			mcd_gadget_fifo_flush_flush_only(ep);
		}
	}

	spin_unlock_irqrestore(&mdev->lock, flags);
	DBG(2, "==>\n");
	return 0;
}

STATIC int mcd_gadget_fifo_status(struct usb_ep *ep)
{
	unsigned long flags;
	int nResult = 0;
	struct mcd_gadget *mgadget = mcd_first_gadget;
	struct mcd_device *mdev = mgadget_to_mdev(mgadget);
	void *regs = mdev_link_reg(mdev);
	int bEnd = EP_NUMBER(ep);

	DBG(2, "<==\n");

#ifdef MCD_PARANOID
	ASSERT_SPINLOCK_UNLOCKED(&pEnd->Lock);
	ASSERT_SPINLOCK_UNLOCKED(&mdev->lock);

	if (mcd_gadget_find_end(ep) < 0) {
		ERR("invalid endpoint ep=%p\n", ep);
		return -EINVAL;
	}
#endif

	spin_lock_irqsave(&mdev->lock, flags);
	mcd_select_end(regs, bEnd);
	nResult = MCD_READ_CSR16(regs, (bEnd) ? MCD_O_HDRC_RXCOUNT : MCD_O_HDRC_COUNT0, bEnd);
	spin_unlock_irqrestore(&mdev->lock, flags);

	DBG(2, "==>\n");
	return nResult;
}

STATIC void mcd_gadget_fifo_flush(struct usb_ep *ep)
{
	uint16_t wCsr;
	unsigned long flags;
	struct mcd_gadget *mgadget = mcd_first_gadget;
	struct mcd_device *mdev = mgadget_to_mdev(mgadget);
	void *regs = mdev_link_reg(mdev);
	int nEnd = EP_NUMBER(ep);
	struct usb_request *req;

	DBG(2, "<==\n");

#ifdef MCD_PARANOID
	ASSERT_SPINLOCK_UNLOCKED(&pEnd->Lock);
	ASSERT_SPINLOCK_UNLOCKED(&mdev->lock);
#endif

	spin_lock_irqsave(&mdev->lock, flags);
	mcd_select_end(regs, (uint8_t) nEnd);

	if (nEnd) {
		/* EP0ʳ */
		struct timeval tv;
		unsigned long susec;
		unsigned long check_time;
		/* TX */
		/* ϻּ */
		do_gettimeofday(&tv);
		susec = tv.tv_usec;
		/* bogoMIPSˤ5usȤлֳǧ */
		check_time = loops_per_jiffy / (200000 / HZ);
		do {
			wCsr = MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, (uint8_t) nEnd);
			if (!(wCsr & MCD_M_TXCSR_FIFONOTEMPTY)) {
				break;
			}
			if (check_time == 0) {
				do_gettimeofday(&tv);
				if (tv.tv_usec - susec > 200) {	/* Ϥ200ustimeout */
					break;
				}
				check_time = loops_per_jiffy / (200000 / HZ);	/* 5us */
			}
			check_time--;
			wCsr = (wCsr | MCD_M_TXCSR_FLUSHFIFO) & ~MCD_M_TXCSR_TXPKTRDY;
			MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, (uint8_t) nEnd, wCsr);
		} while (1);
		/* RX */
		wCsr = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, (uint8_t) nEnd);
		wCsr |= MCD_M_RXCSR_FLUSHFIFO;
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, (uint8_t) nEnd, wCsr);
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, (uint8_t) nEnd, wCsr);
		mcd_gadget_flush_ep_requests((struct mcd_gep *)ep, -ECONNRESET);
	} else {
		/* EP0 */
		wCsr = MCD_READ_CSR16(regs, MCD_O_HDRC_CSR0, (uint8_t) nEnd);
		wCsr |= (MCD_M_CSR0_FLUSHFIFO | MCD_M_CSR0_P_SENTSTALL);
		wCsr &= ~MCD_M_CSR0_TXPKTRDY;
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_CSR0, 0, wCsr);
		req = mcd_get_current_request(&mgadget->ep[0]);
		if (req) {
			req->status = -ECONNRESET;
			mcd_gadget_complete_ep0_request();
		}
	}

	/* re-enable interrupt */
	spin_unlock_irqrestore(&mdev->lock, flags);
}

STATIC void mcd_gadget_fifo_flush_flush_only(struct usb_ep *ep)
{
	uint16_t wCsr;
	unsigned long flags;
	struct mcd_gadget *mgadget = mcd_first_gadget;
	struct mcd_device *mdev = mgadget_to_mdev(mgadget);
	void *regs = mdev_link_reg(mdev);
	int nEnd = EP_NUMBER(ep);

	DBG(2, "<==\n");

#ifdef MCD_PARANOID
	ASSERT_SPINLOCK_UNLOCKED(&pEnd->Lock);
	ASSERT_SPINLOCK_UNLOCKED(&mdev->lock);
#endif

	spin_lock_irqsave(&mdev->lock, flags);
	mcd_select_end(regs, (uint8_t) nEnd);

	if (nEnd) {
		/* EP0ʳ */
		struct timeval tv;
		unsigned long susec;
		unsigned long check_time;
		/* TX */
		/* ϻּ */
		do_gettimeofday(&tv);
		susec = tv.tv_usec;
		/* bogoMIPSˤ5usȤлֳǧ */
		check_time = loops_per_jiffy / (200000 / HZ);
		do {
			wCsr = MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, (uint8_t) nEnd);
			if (!(wCsr & MCD_M_TXCSR_FIFONOTEMPTY)) {
				break;
			}
			if (check_time == 0) {
				do_gettimeofday(&tv);
				if (tv.tv_usec - susec > 200) {	/* Ϥ200ustimeout */
					break;
				}
				check_time = loops_per_jiffy / (200000 / HZ);	/* 5us */
			}
			check_time--;
			wCsr = (wCsr | MCD_M_TXCSR_FLUSHFIFO) & ~MCD_M_TXCSR_TXPKTRDY;
			MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, (uint8_t) nEnd, wCsr);
		} while (1);
		/* RX */
		wCsr = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, (uint8_t) nEnd);
		wCsr |= MCD_M_RXCSR_FLUSHFIFO;
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, (uint8_t) nEnd, wCsr);
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, (uint8_t) nEnd, wCsr);
	} else {
		/* EP0 */
		wCsr = MCD_READ_CSR16(regs, MCD_O_HDRC_CSR0, (uint8_t) nEnd);
		wCsr |= (MCD_M_CSR0_FLUSHFIFO | MCD_M_CSR0_P_SENTSTALL);
		wCsr &= ~MCD_M_CSR0_TXPKTRDY;
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_CSR0, 0, wCsr);
	}

	/* re-enable interrupt */
	spin_unlock_irqrestore(&mdev->lock, flags);
}

/** general operations */
STATIC int mcd_gadget_get_frame(struct usb_gadget *gadget)
{
	struct mcd_gadget *mgadget = mcd_first_gadget;
#ifdef MCD_PARANOID
	if (!mgadget) {
		ERR("Gadget not initialized, mgadget=NULL\n");
		return -EINVAL;
	}
#endif
	return (int)MCD_READ16(mdev_link_reg(mgadget_to_mdev(mgadget)), MCD_O_HDRC_FRAME);
}

/** */
STATIC int mcd_gadget_wakeup(struct usb_gadget *gadget)
{
	uint8_t power;
	struct mcd_gadget *mgadget = mcd_first_gadget;
	struct mcd_device *mdev = mgadget_to_mdev(mgadget);
	void *regs = mdev_link_reg(mdev);

	DBG(2, "<==\n");

#ifdef MCD_PARANOID
	if (!mcd_first_gadget) {
		ERR("Gadget not initialized, mgadget=NULL\n");
		return -EINVAL;
	}
#endif

	power = MCD_READ8(regs, MCD_O_HDRC_POWER);
	power |= MCD_M_POWER_RESUME;
	MCD_WRITE8(regs, MCD_O_HDRC_POWER, power);

	return 0;
}

/** */
STATIC int mcd_gadget_set_self_powered(struct usb_gadget *gadget, int is_selfpowered)
{
	struct mcd_gadget *mgadget = mcd_first_gadget;
	DBG(2, "<==\n");

#ifdef MCD_PARANOID
	if (!mgadget) {
		ERR("Gadget not initialized, mgadget=NULL\n");
		return -EINVAL;
	}
#endif

	mgadget->bIsSelfPowered = (uint8_t) is_selfpowered;
	return 0;
}

/** */
STATIC int mcd_gadget_vbus_session(struct usb_gadget *gadget, int is_active)
{
	DBG(2, "<==>\n");
	return -EINVAL;
}

/** */
STATIC int mcd_gadget_vbus_draw(struct usb_gadget *gadget, unsigned mA)
{
	struct mcd_gadget *mgadget = mcd_first_gadget;

#ifdef MCD_PARANOID
	if (!mgadget) {
		ERR("Gadget not initialized, mgadget=NULL\n");
		return -EINVAL;
	}
#endif
	mgadget->bIsSelfPowered = (mA) ? true : false;
	return 0;
}

/** */
STATIC int mcd_gadget_pullup(struct usb_gadget *gadget, int is_on)
{
	unsigned long flags;
	uint8_t power;
	uint8_t intrusbe;
	struct mcd_gadget *mgadget = mcd_first_gadget;
	struct mcd_device *mdev = mgadget_to_mdev(mgadget);
	void *regs = mdev_link_reg(mdev);

	DBG(2, "<==\n");

#ifdef MCD_PARANOID
	if (!mcd_first_gadget) {
		ERR("Gadget not initialized, mgadget=NULL\n");
		return -EINVAL;
	}
#endif

	spin_lock_irqsave(&mdev->lock, flags);
	power = MCD_READ8(regs, MCD_O_HDRC_POWER);
	if (mdev->attribute & MCD_DEVICE_ATTR_HS_ENABLE) {
		power |= MCD_M_POWER_HSENAB;
	} else {
		power &= ~MCD_M_POWER_HSENAB;
	}

	if (is_on) {
		mgadget->bSuspended = false;
		/* Suspend/Resume/Resetߤε */
		intrusbe = (MCD_M_INTR_SUSPEND | MCD_M_INTR_RESUME |
			    MCD_M_INTR_RESET);
		MCD_WRITE8(regs, MCD_O_HDRC_INTRUSBE, intrusbe);
		DBG(3, "gadget %s connecting to host\n",
		    mgadget->pGadgetDriver->function);
		power |= MCD_M_POWER_SOFTCONN;
		MCD_WRITE8(regs, MCD_O_HDRC_POWER, power);
#if defined(MCD_AVAIL_WA_SOFTCON) && !defined(CONFIG_USB_MCD_HOST)
		/* Vbus ON */
		mcd_otg_control_force_vbus(&mdev->otgi.otgc, USB_OTG_ON);
#endif				/* defined(MCD_AVAIL_WA_SOFTCON) &&
                                   !defined(CONFIG_USB_MCD_HOST) */
	} else {
		mgadget->bSuspended = true;
		DBG(3, "gadget %s disconnecting from host\n",
		    mgadget->pGadgetDriver->function);
		/* Suspend/Resume/ResetߤԵ */
		MCD_WRITE8(regs, MCD_O_HDRC_INTRUSBE, 0);
#if defined(MCD_AVAIL_WA_SOFTCON)
		/* SoftConơȽ */
# if defined(CONFIG_USB_MCD_HOST)
		/* VbusΤ */
		mcd_otg_control_stop_detector(&mdev->otgi.otgc);
# endif				/* defined(CONFIG_USB_MCD_HOST) */
		/* Vbus OFF */
		mcd_otg_control_force_vbus(&mdev->otgi.otgc, USB_OTG_VBUS_OFF);
		/* SoftConӥڥɵĥӥåȤΥꥢ */
		power &= ~MCD_M_POWER_ENSUSPEND;
#endif				/* defined(MCD_AVAIL_WA_SOFTCON) */
		power &= ~MCD_M_POWER_SOFTCONN;
		MCD_WRITE8(regs, MCD_O_HDRC_POWER, power);
#if defined(MCD_AVAIL_WA_SOFTCON)
		spin_unlock_irqrestore(&mdev->lock, flags);
		msleep(25);	/* ChirpλԤ25msWAIT */
		spin_lock_irqsave(&mdev->lock, flags);
# if defined(CONFIG_USB_MCD_HOST)
		/* Vbus OFFβ */
		mcd_otg_control_force_vbus(&mdev->otgi.otgc, USB_OTG_THRU);
# endif				/* defined(CONFIG_USB_MCD_HOST) */
		power = MCD_READ8(regs, MCD_O_HDRC_POWER);
		power |= MCD_M_POWER_ENSUSPEND;
		MCD_WRITE8(regs, MCD_O_HDRC_POWER, power);
# if defined(CONFIG_USB_MCD_HOST)
		/* VbusΤκƳ */
		mcd_otg_control_start_detector(&mdev->otgi.otgc);
# endif				/* defined(CONFIG_USB_MCD_HOST) */
#endif				/* defined(MCD_AVAIL_WA_SOFTCON) */
	}

	spin_unlock_irqrestore(&mdev->lock, flags);

	DBG(2, "==>\n");
	return 0;
}

STATIC int mcd_gadget_ioctl(struct usb_gadget *gadget, unsigned code, unsigned long param)
{
	struct mcd_gadget *mgadget = mcd_first_gadget;
	struct mcd_device *mdev = mgadget_to_mdev(mgadget);

	DBG(2, "<==>\n");

	switch (code) {
	case USB_IOCTL_FIFO_RESET:{
			mcd_fifo_reset(mdev);
			mcd_gadget_dma_set_default(mgadget);
			return 0;
		}
	case USB_IOCTL_FIFO_MAP:{
			void *adr;
			uint8_t bEnd;
			struct usb_ioctl_fifo_param *fifo_param = (struct usb_ioctl_fifo_param *)param;
			adr = mcd_fifo_map(mdev, fifo_param->bEpAdr, fifo_param->wSize, fifo_param->bType,
					   /* Double:2, Single:1 */
					   false);	/* Gadget */
			if (adr) {
				bEnd = fifo_param->bEpAdr & USB_ENDPOINT_NUMBER_MASK;
				mgadget->ep[bEnd].end_point.maxpacket = fifo_param->wSize;
			}
			return (adr ? 0 : -EINVAL);
		}
	case USB_IOCTL_FIFO_DUMP:{
			mcd_fifo_dump(mdev);
			return 0;
		}
	case USB_IOCTL_TESTMODE:{
			if (NULL == mgadget)
				return -EAGAIN;
			mgadget->bTestMode = true;
			switch (param) {
			case USB_IOCTL_TESTMODE_NORMAL:
				DBG(3, "TESTMODE_NORMAL\n");
#ifdef DUMMY_USB_IOCTL_TESTMODE_NORMAL
				mgadget->bTestMode = false;
#else
				/* TESTMODE_CLEAR */
				mgadget->bTestModeValue = (uint8_t) NULL;
#endif				/* DUMMY_USB_IOCTL_TESTMODE_NORMAL */
				break;
			case USB_IOCTL_TESTMODE_TEST_J:
				DBG(3, "TEST_J\n");
				/* TEST_J */
				mgadget->bTestModeValue = MCD_M_TEST_J;
				break;
			case USB_IOCTL_TESTMODE_TEST_K:
				/* TEST_K */
				DBG(3, "TEST_K\n");
				mgadget->bTestModeValue = MCD_M_TEST_K;
				break;
			case USB_IOCTL_TESTMODE_SE0_NAK:
				/* TEST_SE0_NAK */
				DBG(3, "TEST_SE0_NAK\n");
				mgadget->bTestModeValue = MCD_M_TEST_SE0_NAK;
				break;
			case USB_IOCTL_TESTMODE_TEST_PACKET:
				/* TEST_PACKET */
				DBG(3, "TEST_PACKET\n");
				mgadget->bTestModeValue = MCD_M_TEST_PACKET;
				break;
			default:
				/* my gadget might know what to do with it */
				return -EINVAL;
			}
			return 0;
		}
	case USB_IOCTL_DMA_SET:{
			struct usb_ioctl_dmaset_param *dmaset_param;
			uint8_t bEnd;
#if defined(CONFIG_ARCH_CXD4108)
			unsigned short uidtype0;
			extern unsigned short mcd_cxd4108_get_uidtype0(void);

			uidtype0 = mcd_cxd4108_get_uidtype0();
#endif				/* defined(CONFIG_ARCH_CXD4108) */

			dmaset_param = (struct usb_ioctl_dmaset_param *)param;
			bEnd = dmaset_param->bEp_num & USB_ENDPOINT_NUMBER_MASK;

			if (dmaset_param->bDma_ena == USB_MCD_DMA_SETON) {
				/* DMA transfer enable */
				DBG(1, "DMA_SET ON bDma_ena:%d \n", dmaset_param->bDma_ena);
#if defined(CONFIG_ARCH_CXD4108)
				if (uidtype0) {
					mgadget->ep[bEnd].bDma_ena = USB_MCD_DMA_SETON;
				} else {
					/* always DMA OFF */
					mgadget->ep[bEnd].bDma_ena = USB_MCD_DMA_SETOFF;
				}
#else
				mgadget->ep[bEnd].bDma_ena = USB_MCD_DMA_SETON;
#endif				/* defined(CONFIG_ARCH_CXD4108) */
			} else if (dmaset_param->bDma_ena == USB_MCD_DMA_SETOFF) {
				/* DMA transfer disable */
				DBG(1, "DMA_SET OFF bDma_ena:%d \n", dmaset_param->bDma_ena);
				mgadget->ep[bEnd].bDma_ena = USB_MCD_DMA_SETOFF;
			} else {
				DBG(1, "DMA_SET error INVALID param bDma_ena:%d \n", dmaset_param->bDma_ena);
				return -EINVAL;
			}
			return 0;
		}
	}

	return -EINVAL;
}

/**
 * Start Gadget HW
 * @param mgadget controller driver
 */
void mcd_gadget_start(struct mcd_gadget *mgadget)
{
	void *regs = mdev_link_reg(mgadget_to_mdev(mgadget));

	MCD_WRITE16(regs, MCD_O_HDRC_INTRTXE, 1);
	MCD_WRITE16(regs, MCD_O_HDRC_INTRRXE, 0);
	MCD_WRITE8(regs, MCD_O_HDRC_INTRUSBE, 0);
	MCD_WRITE8(regs, MCD_O_HDRC_TESTMODE, 0);
	MCD_WRITE8(regs, MCD_O_HDRC_POWER, MCD_M_POWER_ENSUSPEND);

	pool_buffer_create(mgadget);

	mcd_dma_start_controller(mgadget_to_mdev(mgadget)->dmac);
}

/**
 * Stop Gadget HW
 * @param mgadget controller driver
 */
void mcd_gadget_stop(struct mcd_gadget *mgadget)
{
	void *regs = mdev_link_reg(mgadget_to_mdev(mgadget));

	MCD_WRITE16(regs, MCD_O_HDRC_INTRTXE, 0);
	MCD_WRITE16(regs, MCD_O_HDRC_INTRRXE, 0);
	MCD_WRITE8(regs, MCD_O_HDRC_INTRUSBE, 0);
	MCD_WRITE8(regs, MCD_O_HDRC_TESTMODE, 0);
	MCD_WRITE8(regs, MCD_O_HDRC_POWER, MCD_M_POWER_ENSUSPEND);

	pool_buffer_destroy(mgadget);

	mcd_dma_stop_controller(mgadget_to_mdev(mgadget)->dmac);
}

/*
 *	Interrupt handling
 */

/**
 * Interrupt Service Routine to record USB "global" interrupts.
 * Since these do not happen often and signify things of 
 * paramount importance, it seems OK to check them individually.
 *
 * @param pThis instance pointer
 * @param reg IntrUSB register contents
 */
static void service_intrusb(struct mcd_gadget *mgadget, u8 intrusb)
{
	DBG(1, "INTR" "%x\n", intrusb);

	if (intrusb & MCD_M_INTR_RESUME) {
		DBG(3, "RESUME\n");
		mcd_gadget_resume(mgadget);
	}

	if (intrusb & MCD_M_INTR_SUSPEND) {
		DBG(3, "SUSPEND\n");
		mcd_gadget_suspend(mgadget);
	}

	if (intrusb & MCD_M_INTR_RESET) {
		DBG(3, "BUS RESET\n");
		mcd_gadget_reset(mgadget);
	}
}

irqreturn_t mcd_gadget_link_irq(int irq, void *data)
{
	struct mcd_device *mdev = (struct mcd_device *)data;
	struct mcd_gadget *mgadget = &mdev->gadget;
	void *regs = mdev_link_reg(mdev);
	irqreturn_t status = IRQ_NONE;
	u8 intrusb;
	u16 intrtx;
	u16 intrrx;
	u8 i;

	intrusb = MCD_READ8(regs, MCD_O_HDRC_INTRUSB);
	intrusb &= MCD_READ8(regs, MCD_O_HDRC_INTRUSBE);
	if (intrusb) {
		service_intrusb(mgadget, intrusb);
		status = IRQ_HANDLED;
	}

	intrtx = MCD_READ16(regs, MCD_O_HDRC_INTRTX);
	intrtx &= MCD_READ16(regs, MCD_O_HDRC_INTRTXE);
	if (intrtx) {
		if (intrtx & 1) {
			mcd_gadget_service_default_end(mgadget);
		}
		intrtx >>= 1;
		i = 0;
		while (intrtx && i < MCD_NUM_EPS - 1) {
			if (intrtx & 1) {
				mcd_gadget_service_tx_avail(mgadget, i + 1);
			}
			intrtx >>= 1;
			i++;
		}
		status = IRQ_HANDLED;
	}

	intrrx = MCD_READ16(regs, MCD_O_HDRC_INTRRX);
	intrrx &= MCD_READ16(regs, MCD_O_HDRC_INTRRXE);
	if (intrrx) {
		intrrx >>= 1;
		i = 0;
		while (intrrx && i < MCD_NUM_EPS - 1) {
			if (intrrx & 1) {
				mcd_gadget_service_rx_ready(mgadget, i + 1);
			}
			intrrx >>= 1;
			i++;
		}
		status = IRQ_HANDLED;
	}

	return status;
}

#ifdef MCD_DMA

static uint8_t mcd_gadget_is_dma_interrupt[MCD_NUM_EPS];

void mcd_gadget_dma_tx_cb(struct mcd_dma_channel *pChannel, uint8_t bEnd, uint8_t bTransmit, uint8_t bProtocol, void *pData)
{
	struct mcd_gadget *mgadget = mcd_first_gadget;

	DBG(2, "TestBulk DMA CallBack mcd_gadget_service_tx_avail\n");
	mcd_gadget_is_dma_interrupt[bEnd] = true;
	if (NULL != mgadget->ep[bEnd].pDmaChannel)
		mgadget->ep[bEnd].pDmaChannel->pData = &(mcd_gadget_is_dma_interrupt[bEnd]);

	mcd_gadget_service_tx_avail(mgadget, bEnd);

	mcd_gadget_is_dma_interrupt[bEnd] = false;
	DBG(2, "TestBulk DMA CallBack mcd_gadget_service_tx_avail end \n");
}

void mcd_gadget_dma_rx_cb(struct mcd_dma_channel *pChannel, uint8_t bEnd, uint8_t bTransmit, uint8_t bProtocol, void *pData)
{
	struct mcd_gadget *mgadget = mcd_first_gadget;

	DBG(2, "TestBulk DMA CallBack mcd_gadget_service_rx_ready\n");
	mcd_gadget_is_dma_interrupt[bEnd] = true;
	if (NULL != mgadget->ep[bEnd].pDmaChannel)
		mgadget->ep[bEnd].pDmaChannel->pData = &(mcd_gadget_is_dma_interrupt[bEnd]);

	mcd_gadget_service_rx_ready(mgadget, bEnd);

	mcd_gadget_is_dma_interrupt[bEnd] = false;
	DBG(2, "TestBulk DMA CallBack mcd_gadget_service_rx_ready end \n");
}
#endif				/* MCD_DMA */

/****************************************************************
 * Registration operations 
 ****************************************************************/

/*----------------------------------------------------------
 name:mcd_gadget_dma_set_default
 param: struct mcd_gadget *mgadget
 content: Set a Default Value for bDma_ena 
 			in mgadget->ep
----------------------------------------------------------*/
STATIC void mcd_gadget_dma_set_default(struct mcd_gadget *mgadget)
{
	uint8_t i;
#if defined(CONFIG_ARCH_CXD4108)
	unsigned short uidtype0;
	extern unsigned short mcd_cxd4108_get_uidtype0(void);

	uidtype0 = mcd_cxd4108_get_uidtype0();
#endif				/* defined(CONFIG_ARCH_CXD4108) */

	for (i = 0; i < MCD_NUM_EPS; i++) {
		/* set a Default value(DMA enable) for bDma_ena */
#if defined(CONFIG_ARCH_CXD4108)
		if (uidtype0) {
			mgadget->ep[i].bDma_ena = USB_MCD_DMA_SETON;
		} else {
			mgadget->ep[i].bDma_ena = USB_MCD_DMA_SETOFF;
		}
#else
		mgadget->ep[i].bDma_ena = USB_MCD_DMA_SETON;
#endif				/* defined(CONFIG_ARCH_CXD4108) */
		DBG(3, "Init mgadget->ep[%d].bDma_ena:%d \n", i, mgadget->ep[i].bDma_ena);
	}
}

/**
 *
 * -EINVAL
 * -EBUSY
 * -ENOMEM
 */
int usb_gadget_register_driver(struct usb_gadget_driver *driver)
{
	int ret;
	uint8_t bEnd, power;
	unsigned long flags;
	struct mcd_gadget *mgadget = mcd_first_gadget;
	struct mcd_device *mdev = mgadget_to_mdev(mgadget);
	void *regs = mdev_link_reg(mdev);

	DBG(2, "<==\n");

#ifdef MCD_PARANOID
	if (!mgadget) {
		ERR("Gadget not initialized, mgadget=NULL\n");
		return -EINVAL;
	}
#endif

	/* this one is checked to verify that no Gadget is registered */
	if (mgadget->pGadgetDriver) {
		ERR("musb-gcd is alredy used by %s\n", mgadget->pGadget->name);
		return -EBUSY;
	}

	INFO("registering gadget %s\n", driver->function);

	spin_lock_irqsave(&mdev->lock, flags);
	/* allocate memory for gadget structure */
	KMALLOC(mgadget->pGadget, sizeof(struct usb_gadget), GFP_KERNEL);
	if (!mgadget->pGadget) {
		spin_unlock_irqrestore(&mdev->lock, flags);
		ERR("out of memory\n");
		return -ENOMEM;
	}
	memset(mgadget->pGadget, 0, sizeof(struct usb_gadget));

	/* allocate real memory for ep0 buffer, when MCD_TEST_GADGET 
	 * is set, I'll be using the static buffer up there */
	KMALLOC(mgadget->pEnd0Buffer, sizeof(struct mcd_end0_buffer), GFP_KERNEL);
	if (!mgadget->pEnd0Buffer) {
		KFREE(mgadget->pGadget);
		spin_unlock_irqrestore(&mdev->lock, flags);
		ERR("out of memory\n");
		return -ENOMEM;
	}
	memset(mgadget->pEnd0Buffer, 0, sizeof(struct mcd_end0_buffer));

    /** Init completition scheduler */
	spin_lock_init(&SchedulerQueue.lock);
	INIT_LIST_HEAD(&SchedulerQueue.req_list);

    /** Init local, and gadget endpoints */
	INIT_LIST_HEAD(&mgadget->pGadget->ep_list);
	for (bEnd = 0; bEnd < MCD_NUM_EPS; bEnd++) {

		memset(&mgadget->ep[bEnd], 0, sizeof(mgadget->ep[bEnd]));
#if MCD_DEBUG>0
		mgadget->ep[bEnd].wPadFront = MCD_LOCAL_PAD;
#endif
		mgadget->ep[bEnd].bEndNumber = bEnd;
		spin_lock_init(&mgadget->ep[bEnd].Lock);
		INIT_LIST_HEAD(&mgadget->ep[bEnd].req_list);

		if (bEnd < iGadgetEndNamesLen || !aGadgetEndNames[bEnd]) {
			snprintf(aGadgetDefaultEndNames[bEnd],
				 sizeof(aGadgetDefaultEndNames[bEnd]),
				 "ep%d", bEnd);
			aGadgetEndNames[bEnd] = aGadgetDefaultEndNames[bEnd];
		}

		mgadget->ep[bEnd].end_point.name = aGadgetEndNames[bEnd];
		mgadget->ep[bEnd].end_point.ops = &mcd_gadget_endpoint_operations;
		INIT_LIST_HEAD(&mgadget->ep[bEnd].end_point.ep_list);

		if (bEnd) {
			mgadget->ep[bEnd].end_point.maxpacket = 0;
			list_add_tail(&mgadget->ep[bEnd].end_point.ep_list, &(mgadget->pGadget->ep_list));
		} else {
			mgadget->ep[0].end_point.maxpacket = MCD_END0_FIFOSIZE;
			mgadget->ep[0].wPacketSize = MCD_END0_FIFOSIZE;
		}

	}

	mgadget->pGadget->ops = &mcd_gadget_operations;
	mgadget->pGadget->ep0 = &mgadget->ep[0].end_point;

	GADGET_SET_IS_DUALSPEED(mgadget->pGadget, 1);
	GADGET_SET_IS_OTG(mgadget->pGadget, 1);
	GADGET_SET_A_HNP_SUPPORT(mgadget->pGadget, 1);

	/* it takes the name of the driver */
	mgadget->pGadget->name = driver->function;

	mgadget->pGadget->speed = USB_SPEED_FULL;
	power = MCD_READ8(regs, MCD_O_HDRC_POWER);

	if (power & MCD_M_POWER_HSMODE) {
		mgadget->pGadget->speed = USB_SPEED_HIGH;
	}

	/* this one is checked to verify that a Gadget registered */
	mgadget->pGadgetDriver = driver;

	mgadget->pGadget->dev.driver = &driver->driver;
	driver->driver.bus = NULL;

	/* KLUDGE for weird gadgets */
	kobject_init(&driver->driver.kobj);
	//kobject_set_name(&driver->driver.kobj, "kludge");
	//kobject_register(&driver->driver.kobj);

	spin_unlock_irqrestore(&mdev->lock, flags);
	if ((ret = driver->bind(mgadget->pGadget))) {
		ERR("gadget %s not registered. fail bind\n", mgadget->pGadget->name);

		spin_lock_irqsave(&mdev->lock, flags);
		KFREE(mgadget->pGadget);
		KFREE(mgadget->pEnd0Buffer);
		mgadget->pGadgetDriver = NULL;
		mgadget->pGadget = NULL;
		mgadget->pEnd0Buffer = NULL;
		spin_unlock_irqrestore(&mdev->lock, flags);

		return ret;
	}

	INFO("gadget %s registered\n", mgadget->pGadget->name);

	return 0;
}

EXPORT_SYMBOL(usb_gadget_register_driver);

/**
 *
 *
 */
int usb_gadget_unregister_driver(struct usb_gadget_driver *driver)
{
	unsigned long flags;
	struct mcd_gadget *mgadget = mcd_first_gadget;
	struct mcd_device *mdev = mgadget_to_mdev(mgadget);

	DBG(2, "<==\n");

#ifdef MCD_PARANOID
	if (!mcd_first_gadget) {
		ERR("Gadget not initialized, mgadget=NULL\n");
		return -EINVAL;
	}
#endif

	INFO("unregistering gadget %s\n", driver->function);

	spin_lock_irqsave(&mdev->lock, flags);

	/* shall I flush/abort the pending requests?? */
	if (mgadget->pGadgetDriver) {
		driver->unbind(mgadget->pGadget);

		if (mgadget->pGadget) {
			KFREE(mgadget->pGadget);

			if (mgadget->pEnd0Buffer) {
				KFREE(mgadget->pEnd0Buffer);
			}
		}
	}

	mgadget->pGadgetDriver = NULL;
	mgadget->pGadget = NULL;
	mgadget->pEnd0Buffer = NULL;

#if !defined(CONFIG_ARCH_INTEGRATOR)
	mcd_board_phy_reset();
#endif
	spin_unlock_irqrestore(&mdev->lock, flags);

	return 0;
}

EXPORT_SYMBOL(usb_gadget_unregister_driver);
