/*-----------------------------------------------------------------------------
 * drivers/usb/mcd/usb_mcd_device.h
 * 
 * USB controller driver
 * 
 * Copyright 2009 Sony Corporation
 *
 * This file is part of the Inventra Controller Driver for Linux.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *---------------------------------------------------------------------------*/

#ifndef _USB_MCD_DEVICE_H_
#define _USB_MCD_DEVICE_H_

#define _MCD2_LF1_		/* MCD2.0 LF1 release */

#define MCD_DEVICE_ATTR_HS_ENABLE      (1<<0)	/* HSEnable */

typedef s32 mcd_time_t;

/* mcd_timer */
struct mcd_timer {
	void *data;
};

/* mcd_device */

struct mcd_device {
	struct device *dev;

	spinlock_t lock;
	u16 fifo_head;
	u16 attribute;
	unsigned int irqs[MCD_NUM_IRQS];
	irq_handler_t irq_hdlrs[MCD_NUM_IRQS];
	void __iomem *regs[MCD_NUM_REGS];
	struct list_head mdev_list;

	struct mcd_timer timer;
	struct mcd_otg_intdat otgi;
	struct mcd_gadget gadget;

	struct usb_hcd *hcd;
	struct mcd_dma_controller *dmac;
	struct proc_dir_entry *proc_entry;
};

/* inline functions */

static inline struct mcd_otg_intdat *otgc_to_otgi(struct usb_otg_control* otgc)
{
	return container_of(otgc, struct mcd_otg_intdat, otgc);
}

static inline struct mcd_device *otgi_to_mdev(struct mcd_otg_intdat *otgi)
{
	return container_of((void *)otgi, struct mcd_device, otgi);
}

static inline void *otgi_regs(struct mcd_otg_intdat *otgi)
{
	return otgi_to_mdev(otgi)->regs[MCD_LINK_REG];
}

static inline struct mcd_device *mgadget_to_mdev(struct mcd_gadget *mgadget)
{
	return container_of((void *)mgadget, struct mcd_device, gadget);
}

static inline void *mdev_link_reg(struct mcd_device *mdev)
{
	return mdev->regs[MCD_LINK_REG];
}
static inline bool time_gt(mcd_time_t t1, mcd_time_t t2)
{
	return (u32)t1 > (u32)t2;
}

/* proto types */
/* bus */
extern irqreturn_t mcd_request_irq(struct mcd_device *);
extern void mcd_free_irq(struct mcd_device *);

/* control */
extern void mcd_device_init(struct mcd_device *mdev);
extern struct mcd_device *mcd_device_probe(struct device *dev, void **regs,
					   int *irqs);
extern void mcd_device_remove(struct mcd_device *mdev);
extern int mcd_device_shutdown(struct mcd_device *mdev);
extern int mcd_device_suspend(struct mcd_device *mdev, pm_message_t message);
extern int mcd_device_resume(struct mcd_device *mdev);
extern int mcd_device_attach_host(const struct mcd_host_ops *ops);
extern void mcd_device_detach_host(void);

extern int mcd_device_request_session(struct mcd_device *mdev);
extern int mcd_device_end_dpulse(struct mcd_device *mdev);
extern int mcd_device_start(struct mcd_device *mdev);
extern int mcd_device_stop(struct mcd_device *mdev);
extern int mcd_device_start_gadget(struct mcd_device *mdev);
extern int mcd_device_stop_gadget(struct mcd_device *mdev);
extern int mcd_device_start_host(struct mcd_device *mdev);
extern int mcd_device_stop_host(struct mcd_device *mdev);
extern int mcd_device_notify_vbus_change(struct mcd_device *mdev,
					 u8 old_vbus, u8 new_vbus);
extern enum usb_otg_vbus_stat mcd_device_check_vbus(struct mcd_device *mdev);
extern int mcd_device_set_speed(struct mcd_device *mdev, unsigned int speed);
extern enum usb_otg_control_speed mcd_device_get_speed(struct mcd_device *mdev);
extern int mcd_device_set_test_mode(struct mcd_device *mdev,
				    enum usb_otg_control_test_mode test_mode);
extern enum usb_otg_control_test_mode mcd_device_get_test_mode(struct mcd_device
							       *mdev);

/* service */
extern void mcd_device_load_fifo(void *regs, u8 index, u32 len, u8* src);
extern void mcd_device_unload_fifo(void *regs, u8 index, u32 len, u8* dst);

/*
 *	Board
 */
/* phy timeout */
extern unsigned long mcd_board_timeout_phy(void);
/* phy */
extern void mcd_board_phy_reset(void);
/* phy suspend on/off */
extern void mcd_board_control_suspend_off(void);
extern void mcd_board_control_suspend_on(void);
/* linestate */
extern void mcd_board_linestate_set(void);
/* max port number */
extern u32 mcd_board_get_max_port_num(void);
/* timer */
extern int mcd_board_init_timer(struct mcd_device *mdev);
extern int mcd_board_exit_timer(struct mcd_device *mdev);
extern mcd_time_t mcd_board_get_time(struct mcd_device *mdev);
extern u32 mcd_board_time2usec(mcd_time_t time);
extern mcd_time_t mcd_board_usec2time(u32 usec);
extern int mcd_board_start_timer(struct mcd_device *mdev, mcd_time_t expire,
				 void (*func)(struct mcd_device *, void *),
				 void *data);
extern void mcd_board_stop_timer(struct mcd_device *mdev, void *);

/* debug */
extern void mcd_board_dump(struct mcd_device *mdev);
#if defined(CONFIG_ARCH_CXD4105)
extern int mcd_board_cid_pullup(int on);
#endif			/* defined(CONFIG_ARCH_CXD4105) */
extern struct proc_dir_entry *mcd_proc_create_fs(struct mcd_device *mdev,
						 char *name);
extern void mcd_proc_delete_fs(struct mcd_device *mdev);

#endif /* _USB_MCD_DEVICE_H_ */
