/*--------------------------------------------------------------------------*
 * drivers/usb/mcd/usb_mcd_fifo.c
 * 
 * USB controller driver
 * 
 * Copyright 2005,2006,2009 Sony Corporation
 *
 * This file is part of the Inventra Controller Driver for Linux.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *---------------------------------------------------------------------------*/

#include <linux/kernel.h>
#include <linux/usb.h>
#include <linux/usb/ch9.h>
#include <linux/usb_gadget.h>
#include <linux/usb/mcd/usb_otg_control.h>
#include <linux/usb/mcd/usb_mcd_ioctl.h>
#include "../core/hub.h"
#include "../core/hcd.h"

#include "usb_mcd_cnf.h"
#include "usb_mcd_debug.h"
#include "usb_mcd_dma.h"
#include "usb_mcd_otg.h"
#include "usb_mcd_gadget.h"
#include "usb_mcd_host.h"
#include "usb_mcd_device.h"
#include "usb_mcd_access.h"
#include "usb_mcd_musbhdrc.h"
#include "usb_mcd_fifo.h"

#define FIFO_END 4224

/**
 * Reset dynamic FIFO size
 *
 * @param mdev controller driver instance pointer
 */
void mcd_fifo_reset(struct mcd_device *mdev)
{
	u8 i;

	mdev->fifo_head = FIFO_END - 1;	/* safe region */
	for (i = 1; i < MCD_NUM_EPS; i++) {
		mcd_fifo_map(mdev, 0x80 | i, 0, 1, false);	/* IN */
		mcd_fifo_map(mdev, 0x00 | i, 0, 1, false);	/* OUT */
	}
	mdev->fifo_head = 64;	/* 64byte offset for EP0 */
}
EXPORT_SYMBOL(mcd_fifo_reset);

/**
 * map dynamic FIFO for specified ep
 *
 * @param mdev controller driver instance pointer
 * @param epadr endpoint address described in endpoint descriptor
 * @param size specify FIFO size
 * @param type specify 2 when double buffer or 1 when single buffer
 * @param is_host specify 1 when use as a host, 0 as a gadget.
 * @return FIFO address mapped in FIFO map
 */
void *mcd_fifo_map(struct mcd_device *mdev,
		   uint8_t epadr,
		   u16 size, uint8_t type, uint8_t is_host)
{
	u8 i;
	u16 cal_size;
	u8 size_val;
	u16 adr_val;
	u16 adr_ret;
	u8 epnum = epadr & USB_ENDPOINT_NUMBER_MASK;
	u8 is_tx;
	unsigned long flags;
	void *regs = mdev_link_reg(mdev);


	if (type < 1 || type > 2) {
		ERR("Unknown FIFO type\n");
		return NULL;
	}

	if (mdev->fifo_head + size * type >= FIFO_END) {
		ERR("FIFO overflow\n");
		return NULL;
	}

	is_tx = (epadr & USB_ENDPOINT_DIR_MASK) ? !is_host : is_host;

	i = 0;
	/* size_val = ROUNDUP(log2(size)) - 3 */
	if (size == 0) {
		size_val = 0;
	}
	else {
		cal_size = (size - 1) >> 3;
		while (1) {
			if (!cal_size) {
				size_val = i;
				break;
			}
			cal_size >>= 1;
			i++;
			if (i > 9) {
				ERR("FIFO size is too big\n");
				return NULL;
			}
		}
	}
	size_val |= (type - 1) << 4;	/* Double buffe or not? */

	/* adr_val unit is 8Byte */
	adr_val = mdev->fifo_head >> 3;

	spin_lock_irqsave(&mdev->lock, flags);

	mcd_select_end(regs, epnum);

	/* EP0ξTXRXɤFIFOꤹ */
	if (!epnum || is_tx) {
		MCD_WRITE8(regs, MCD_O_HDRC_TXFIFOSZ, size_val);
		MCD_WRITE16(regs, MCD_O_HDRC_TXFIFOADD, adr_val);
	}
	if (!epnum || !is_tx) {
		MCD_WRITE8(regs, MCD_O_HDRC_RXFIFOSZ, size_val);
		MCD_WRITE16(regs, MCD_O_HDRC_RXFIFOADD, adr_val);
	}

	spin_unlock_irqrestore(&mdev->lock, flags);

	adr_ret = mdev->fifo_head;
	mdev->fifo_head += size * type;
	DBG(3, "FifoHead is 0x%x\n", mdev->fifo_head);

	return (void *)(0x0 + adr_ret);
}
EXPORT_SYMBOL(mcd_fifo_map);

void mcd_fifo_dump(struct mcd_device *mdev)
{
	void *regs = mdev_link_reg(mdev);
	u8 epnum;
	u8 tx_size;
	u8 rx_size;
	u16 tx_adr;
	u16 rx_adr;
	const char *type[] = { "S", "D" };

	/* print config */
	for (epnum = 0; epnum < MCD_NUM_EPS; epnum++) {
		mcd_select_end(regs, epnum);
		tx_size = MCD_READ8(regs, MCD_O_HDRC_TXFIFOSZ);
		tx_adr = MCD_READ16(regs, MCD_O_HDRC_TXFIFOADD);
		rx_size = MCD_READ8(regs, MCD_O_HDRC_RXFIFOSZ);
		rx_adr = MCD_READ16(regs, MCD_O_HDRC_RXFIFOADD);
		printk
		    ("FIFO DUMP : EP=%d "
                     "TX(0x%x,%s,Sz=%d,Adr=0x%x) "
                     "RX(0x%x,%s,Sz=%d,Adr=0x%x)\n",
		     epnum, tx_size, type[(0x10 & tx_size) >> 4],
		     1 << ((0x0f & tx_size) + 3), tx_adr, rx_size,
		     type[(0x10 & rx_size) >> 4], 1 << ((0x0f & rx_size) + 3),
		     rx_adr);
	}
}
