/*-----------------------------------------------------------------------------
 * drivers/usb/mcd/usb_mcd_hep.c
 * 
 * USB controller driver
 * 
 * Copyright 2009 Sony Corporation
 *
 * This file is part of the Inventra Controller Driver for Linux.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *---------------------------------------------------------------------------*/

#include <linux/module.h>
#include <linux/device.h>
#include <linux/memory.h>
#include <linux/interrupt.h>
#include <linux/dma-mapping.h>
#include <asm/io.h>
#include <linux/usb.h>
#include <linux/usb/ch9.h>
#include <linux/usb_gadget.h>
#include <linux/usb/mcd/usb_otg_control.h>
#include "../core/hub.h"
#include "../core/hcd.h"

#include "usb_mcd_cnf.h"
#include "usb_mcd_debug.h"
#include "usb_mcd_dma.h"
#include "usb_mcd_otg.h"
#include "usb_mcd_gadget.h"
#include "usb_mcd_host.h"
#include "usb_mcd_device.h"
#include "usb_mcd_access.h"
#include "usb_mcd_musbhdrc.h"
#include "usb_mcd_fifo.h"

/* hep states and triggers */
enum {
	MCD_HEP_S_CNTL_IDLE = 0,
	MCD_HEP_S_CNTL_WAIT_TX_SETUP,
	MCD_HEP_S_CNTL_WAIT_TX_DATA,
	MCD_HEP_S_CNTL_WAIT_RX_DATA,
	MCD_HEP_S_CNTL_WAIT_RX_STATUS,
	MCD_HEP_S_CNTL_WAIT_TX_STATUS,
};

enum {
	MCD_HEP_S_TX_IDLE = 0,
	MCD_HEP_S_TX_WAIT_DMA_MULTI,
	MCD_HEP_S_TX_WAIT_SHORT,
	MCD_HEP_S_TX_WAIT_DMA_SINGLE,
	MCD_HEP_S_TX_WAIT_SINGLE,
	MCD_HEP_S_TX_WAIT_CPU,
};

enum {
	MCD_HEP_S_RX_IDLE = 0,
	MCD_HEP_S_RX_WAIT_DMA_MULTI,
	MCD_HEP_S_RX_WAIT_SINGLE,
	MCD_HEP_S_RX_WAIT_DMA_SINGLE,
	MCD_HEP_S_RX_WAIT_CPU,
};

enum {
	MCD_HEP_T_START,
	MCD_HEP_T_RESET,
	MCD_HEP_T_LINK_INTR,
	MCD_HEP_T_DMAC_INTR,
};

enum {
	MCD_HEP_CNTL_NODATA,
	MCD_HEP_CNTL_IN,
	MCD_HEP_CNTL_OUT,
};

/* error */
#define MCD_HEP_E_NOBUFFER	(1 << 0)
#define MCD_HEP_E_MAXPOVER	(1 << 1)
#define MCD_HEP_E_NORESP	(1 << 2)
#define MCD_HEP_E_RXSTALL	(1 << 3)
#define MCD_HEP_E_OVERFLOW	(1 << 4)
#define MCD_HEP_E_SHORT 	(1 << 5)

/* proto define of internal functions */

static int hep_cntl_trans_state(struct mcd_device *mdev, struct mcd_hep *ep,
				u32 trigger);
static int hep_tx_trans_state(struct mcd_device *mdev, struct mcd_hep *ep,
			      u32 trigger);
static int hep_rx_trans_state(struct mcd_device *mdev, struct mcd_hep *ep,
			      u32 trigger);

/* internal functions */

/* common internal functions  */

static void hep_disable_tx(struct mcd_device *mdev, u8 index)
{
	void *regs = mdev_link_reg(mdev);
	u16 intrtxe;

	intrtxe = MCD_READ16(regs, MCD_O_HDRC_INTRTXE);
	intrtxe &= ~(1 << index);
	MCD_WRITE16(regs, MCD_O_HDRC_INTRTXE, intrtxe);
}

static void hep_disable_rx(struct mcd_device *mdev, u8 index)
{
	void *regs = mdev_link_reg(mdev);
	u16 intrrxe;

	intrrxe = MCD_READ16(regs, MCD_O_HDRC_INTRRXE);
	intrrxe &= ~(1 << index);
	MCD_WRITE16(regs, MCD_O_HDRC_INTRRXE, intrrxe);
}

static void hep_enable_tx(struct mcd_device *mdev, u8 index)
{
	void *regs = mdev_link_reg(mdev);
	u16 intrtxe;

	intrtxe = MCD_READ16(regs, MCD_O_HDRC_INTRTXE);
	intrtxe |= 1 << index;
	MCD_WRITE16(regs, MCD_O_HDRC_INTRTXE, intrtxe);
}

static void hep_enable_rx(struct mcd_device *mdev, u8 index)
{
	void *regs = mdev_link_reg(mdev);
	u16 intrrxe;

	intrrxe = MCD_READ16(regs, MCD_O_HDRC_INTRRXE);
	intrrxe |= 1 << index;
	MCD_WRITE16(regs, MCD_O_HDRC_INTRRXE, intrrxe);
}

inline static u8 hep_get_tx_index(struct mcd_device *mdev, struct mcd_hep *ep)
{
	struct mcd_host *mhost = hcd_to_mhost(mdev->hcd);

	return ep - &mhost->ep_out[0] + 1;
}

inline static u8 hep_get_rx_index(struct mcd_device *mdev, struct mcd_hep *ep)
{
	struct mcd_host *mhost = hcd_to_mhost(mdev->hcd);

	return ep - &mhost->ep_in[0] + 1;
}

inline u32 hep_get_remain(struct urb *urb)
{
	return urb->transfer_buffer_length - urb->actual_length;
}

static bool hep_assign_dmach(struct mcd_device *mdev, struct mcd_hep *ep)
{
	u32 remain;
	bool assigned;
	bool is_tx;
	u8 index;
	u8 xfer_type;
	u16 maxp;
	struct usb_endpoint_descriptor *epd;

#if defined(CONFIG_ARCH_INTEGRATOR) /* DMAǡк */
	return false;
#endif
	remain = hep_get_remain(ep->urb);

	if (ep->dmach) {
		assigned = true;
		goto done;
	}
	if (ep->attributes & MCD_HEP_ATR_NO_DMA || remain == 0) {
		assigned = false;
		goto done;
	}

	epd = &ep->hep->desc;
	is_tx = usb_endpoint_dir_out(epd);
	xfer_type = epd->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK;
	maxp = epd->wMaxPacketSize;

	index = is_tx ? hep_get_tx_index(mdev, ep) : hep_get_rx_index(mdev, ep);
	ep->dmach = mcd_dma_allocate_channel(mdev->dmac, index, is_tx, xfer_type,
					     maxp);
	assigned  = ep->dmach ? true : false;
done:
	return assigned;
}

static void hep_release_dmach(struct mcd_hep *ep)
{
	if (!ep->dmach) {
		return;
	}

	mcd_dma_disable_int_channel(ep->dmach);
	mcd_dma_stop_channel(ep->dmach);
	mcd_dma_release_channel(ep->dmach);
	ep->dmach = NULL;
}

inline static u8 hep_hash(struct usb_host_endpoint *hep, int devnum, u8 epnum)
{
	return epnum - 1;
}

inline static int hep_hep2epnum(struct usb_host_endpoint *hep)
{
	return hep->desc.bEndpointAddress & USB_ENDPOINT_NUMBER_MASK;
}

inline static u8 hep_get_fifo_type(struct usb_endpoint_descriptor *epd)
{
	bool is_bulk = usb_endpoint_xfer_bulk(epd);
	int is_in = usb_endpoint_dir_in(epd);

	return is_bulk && is_in ? 2 : 1;
}

static struct mcd_hep *hep_inquiry_ep(struct mcd_device *mdev,
				      struct usb_device *udev,
				      struct usb_host_endpoint *hep,
				      struct mcd_hep *ep_table, u8 *nep)
{
	struct mcd_hep *ep;
	struct usb_endpoint_descriptor *epd = &hep->desc;
	u8 epnum = hep_hep2epnum(hep);
	u8 cur_epnum;
	u8 hidx;
	u8 i;

	hidx = hep_hash(hep, udev->devnum, epnum);

	i = 0;
	do {
		ep = &ep_table[hidx];
		if (!ep->dev && !ep->hep) {
			void *chk;
			/* assign new ep */
			u8 epdir = epd->bEndpointAddress & USB_ENDPOINT_DIR_MASK;
			u8 epadr = epdir | (epdir == USB_DIR_IN ?
					    hep_get_rx_index(mdev, ep) :
					    hep_get_tx_index(mdev, ep));
			u16 maxp = epd->wMaxPacketSize;
			u8 type = hep_get_fifo_type(epd);
			chk = mcd_fifo_map(mdev, epadr, maxp, type, true);
			if (!chk) {
				return NULL;
			}
			(*nep)++;
			ep->dev = udev;
			ep->hep = hep;
			break;
		}

		cur_epnum = hep_hep2epnum(ep->hep);
		if (cur_epnum == epnum && ep->dev &&
		    ep->dev->devnum == udev->devnum) {
			/* match! */
			break;
		}

		i++;
		hidx++;
		if (i == MCD_NUM_EPS) {
			/* no match and no space for new ep */
			ep = NULL;
			break;
		}
		if (hidx == MCD_NUM_EPS) {
			/* reset hidx to search from top of ep table */
			hidx = 0;
		}
	} while (1);

	return ep;
}

static void hep_release_ep(struct mcd_device *mdev,
			   struct usb_host_endpoint *hep,
			   struct mcd_hep *ep_table, u8 *nep, u8 other_nep)
{
	struct mcd_hep *ep;
	u8 hidx;

	hidx = 0;

	do {
		ep = &ep_table[hidx];
		if (ep->hep == hep) {
			/* match! */
			ep->dev = NULL;
			ep->hep = NULL;
			(*nep)--;
			if (*nep + other_nep == 0) {
				mcd_fifo_reset(mdev);
			}
			break;
		}

		hidx++;

	} while (hidx != MCD_NUM_EPS);
}

static struct mcd_hep *hep_search_ep(struct mcd_port *port, u8 epadr)
{
	struct mcd_host *mhost = hub_to_mhost(port_to_hub(port));
	bool is_in =
		(epadr & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN ? true : false;
	struct mcd_hep *ep;
	struct mcd_hep *last_ep;

	ep = is_in ? &mhost->ep_in[0] : &mhost->ep_out[0];
	last_ep = &ep[MCD_NUM_EPS - 1];
	do {
		if (ep->hep && ep->hep->desc.bEndpointAddress == epadr)
			return ep;
		ep++;
	} while (ep != last_ep);

	return NULL;
}

inline static void hep_set_tx_target(struct mcd_device *mdev, u8 index,
				     struct usb_device *udev)
{
	void *regs = mdev_link_reg(mdev);

	MCD_WRITE8(regs, MCD_BUSCTL_OFFSET(index, MCD_O_HDRC_TXFUNCADDR),
		   udev->state > USB_STATE_DEFAULT ? udev->devnum : 0);
	MCD_WRITE8(regs, MCD_BUSCTL_OFFSET(index, MCD_O_HDRC_TXHUBADDR),
		   udev->parent->devnum);
	MCD_WRITE8(regs, MCD_BUSCTL_OFFSET(index, MCD_O_HDRC_TXHUBPORT),
		   udev->portnum);
}


inline static void hep_set_rx_target(struct mcd_device *mdev, u8 index,
				     struct usb_device *udev)
{
	void *regs = mdev_link_reg(mdev);

	MCD_WRITE8(regs, MCD_BUSCTL_OFFSET(index, MCD_O_HDRC_RXFUNCADDR),
		   udev->state > USB_STATE_DEFAULT ? udev->devnum : 0);
	MCD_WRITE8(regs, MCD_BUSCTL_OFFSET(index, MCD_O_HDRC_RXHUBADDR),
		   udev->parent->devnum);
	MCD_WRITE8(regs, MCD_BUSCTL_OFFSET(index, MCD_O_HDRC_RXHUBPORT),
		   udev->portnum);
}


static void hep_set_type_reg(struct mcd_device *mdev, u32 type_reg, u8 index,
			     struct usb_endpoint_descriptor *epd)
{
	void *regs = mdev_link_reg(mdev);
	u32 data;

	data = (epd->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
		<< MCD_S_TYPE_PROTO;
	data |= epd->bEndpointAddress & USB_ENDPOINT_NUMBER_MASK;

	MCD_WRITE_CSR16(regs, type_reg, index, data);
}

static void hep_set_interval_reg(struct mcd_device *mdev, u32 intv_reg, u8 index,
			     struct usb_endpoint_descriptor *epd)
{
	void *regs = mdev_link_reg(mdev);
	bool is_bulk = usb_endpoint_xfer_bulk(epd);
	u32 data;

	data = is_bulk ? 0 : epd->bInterval;
	MCD_WRITE_CSR16(regs, intv_reg, index, data);
}

static void hep_get_buf(struct urb *urb, struct usb_endpoint_descriptor *epd,
			struct mcd_dma_channel *dmach, char **buf, u32 *len)
{
	bool is_bulk = usb_endpoint_xfer_bulk(epd);
	u16 maxp = epd->wMaxPacketSize;
	u32 remain;

	*buf = (dmach ?
		(char *)urb->transfer_dma + urb->actual_length :
		(char *)urb->transfer_buffer + urb->actual_length);

	remain = hep_get_remain(urb);
	*len = dmach && is_bulk ? remain : min((u32)maxp, remain);
}

inline static void hep_add_actual(struct urb *urb, u32 len)
{
	urb->actual_length += len;
}

static long hep_check_bandwidth(struct usb_device *udev, struct mcd_hep *ep)
{
	long bustime;
	bool is_in = usb_endpoint_dir_in(&ep->hep->desc);
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	int old_alloc;
	int new_alloc;

	old_alloc = udev->bus->bandwidth_allocated;
	bustime = NS_TO_US(usb_calc_bus_time(udev->speed, is_in, false, maxp));
	new_alloc = old_alloc + (int)bustime;

	if (new_alloc > FRAME_TIME_MAX_USECS_ALLOC) {
		bustime = -ENOSPC;
	}

	return bustime;
}

static void hep_claim_bandwidth(struct usb_device *udev, struct mcd_hep *ep,
				long bustime)
{
	udev->bus->bandwidth_allocated += bustime;
	udev->bus->bandwidth_int_reqs++;
}

static void hep_release_bandwidth(struct usb_device *udev, struct mcd_hep *ep)
{
	udev->bus->bandwidth_allocated -= hep_check_bandwidth(udev, ep);
	udev->bus->bandwidth_int_reqs--;
}

static int hep_assign_bandwidth(struct usb_device *udev, struct mcd_hep *ep)
{
	long bustime;
	bustime = hep_check_bandwidth(udev, ep);
	if (bustime < 0) {
		return bustime;
	}

	hep_claim_bandwidth(udev, ep, bustime);

	return 0;
}

static void hep_complete(struct mcd_device *mdev, struct mcd_hep *ep,
			 struct urb *urb, int status)
{
	hep_release_bandwidth(urb->dev, ep);
	urb->hcpriv = NULL;
	urb->status = status;

	if (status || urb->actual_length != urb->transfer_buffer_length) {
		dev_dbg(mdev->dev, "hep_complete: error complete "
			"(devnum:%d,epadr:%x,status:%d,act:%d,expect:%d,"
			"flags:%x)\n",
			ep->dev->devnum, ep->hep->desc.bEndpointAddress,
			status, urb->actual_length, urb->transfer_buffer_length,
			urb->transfer_flags);
	}

	usb_hcd_giveback_urb(mdev->hcd, urb);
	if (ep->urb == urb)
		ep->urb = NULL;
}

static void hep_error(struct mcd_device *mdev, struct mcd_hep *ep, u32 error)
{
	int status;

	if (error & MCD_HEP_E_NOBUFFER) {
		status = -ENOMEM;
	}
	else if (error & MCD_HEP_E_MAXPOVER) {
		status = -EOVERFLOW;
	}
	else if (error & MCD_HEP_E_NORESP) {
		status = -ETIME;
	}
	else if (error & MCD_HEP_E_RXSTALL) {
		status = -EPIPE;
	}
	else if (error & MCD_HEP_E_OVERFLOW) {
		status = -EOVERFLOW;
	}
	else if (error & MCD_HEP_E_SHORT) {
		status = -EREMOTEIO;
	}
	else {
		status = -ERANGE; /* unknown */
	}

	hep_complete(mdev, ep, ep->urb, status);
}

static u32 hep_check_rx_len_error(u16 maxp, u32 expected, u16 count,
				  bool short_not_ok)
{
	u32 error = 0;

	if (short_not_ok && count != 0 && count < maxp) {
		error |= MCD_HEP_E_SHORT;
	}
	if (count > maxp) {
		error |= MCD_HEP_E_MAXPOVER;
	}
	if (count > expected) {
		error |= MCD_HEP_E_OVERFLOW;
	}

	return error;
}

inline bool hep_check_bulk(struct mcd_hep *ep)
{
	return usb_endpoint_xfer_bulk(&ep->hep->desc);
}

inline bool hep_check_short(struct mcd_hep *ep)
{
	struct urb *urb = ep->urb;
	u16 maxp = ep->hep->desc.wMaxPacketSize;

	return hep_get_remain(urb) < maxp ? true : false;
}

/* control transaction */

inline static void hep_cntl_disable(struct mcd_device *mdev)
{
	hep_disable_tx(mdev, MCD_HEP_EP0_IDX);
}

inline static void hep_cntl_enable(struct mcd_device *mdev)
{
	hep_enable_tx(mdev, MCD_HEP_EP0_IDX);
}

inline static void hep_cntl_reset(struct mcd_device *mdev)
{
	mcd_hep_cntl_init(mdev);
}

static bool hep_rx_check_short(struct mcd_device *mdev, struct mcd_hep *ep)
{
	void *regs = mdev_link_reg(mdev);
	u8 index;
	u16 rxcount;
	u16 maxp = ep->hep->desc.wMaxPacketSize;

	index = hep_get_rx_index(mdev, ep);
	rxcount = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCOUNT, index);

	return (maxp - rxcount) ? true : false;
}

static void hep_cntl_self_start(struct mcd_device *mdev, struct mcd_hep *ep)
{
	struct urb *urb;
	struct usb_host_endpoint *hep = ep->hep;
	
	if (ep->urb || !hep || list_empty(&hep->urb_list)) {
		return;
	}
	urb = list_entry(hep->urb_list.next, struct urb, urb_list);
	ep->urb = urb;

	hep_cntl_trans_state(mdev, ep, MCD_HEP_T_START);
}

static void hep_cntl_send_setup(struct mcd_device *mdev, struct mcd_hep *ep)
{
	struct urb *urb = ep->urb;
	void *regs = mdev_link_reg(mdev);

	mcd_device_load_fifo(regs, MCD_HEP_EP0_IDX,
			     sizeof(struct usb_ctrlrequest), urb->setup_packet);
	MCD_WRITE_CSR16(regs, MCD_O_HDRC_CSR0, MCD_HEP_EP0_IDX,
			MCD_M_CSR0_H_DISPING |
			MCD_M_CSR0_H_SETUPPKT |	MCD_M_CSR0_TXPKTRDY);
}

inline static u8 hep_cntl_get_type(struct mcd_hep *ep)
{
	u8 type;
	struct urb *urb = ep->urb;
	struct usb_ctrlrequest *request =
		(struct usb_ctrlrequest *)urb->setup_packet;

	if (request->wLength == 0) {
		type = MCD_HEP_CNTL_NODATA;
	}
	else if (request->bRequestType & USB_DIR_IN) {
		type = MCD_HEP_CNTL_IN;
	}
	else {
		type = MCD_HEP_CNTL_OUT;
	}

	return type;
}

static u32 hep_cntl_check_error(struct mcd_device *mdev)
{
	u16 csr0;
	u32 error = 0;
	void *regs = mdev_link_reg(mdev);

	csr0 = MCD_READ_CSR16(regs, MCD_O_HDRC_CSR0, MCD_HEP_EP0_IDX);

	if (csr0 & MCD_M_CSR0_H_RXSTALL) {
		error |= MCD_HEP_E_RXSTALL;
	}
	if (csr0 & MCD_M_CSR0_H_ERROR) {
		error |= MCD_HEP_E_NORESP;
	}

	return error;
}

static void hep_cntl_send_data(struct mcd_device *mdev, struct mcd_hep *ep)
{
	char* buf;
	u32 len;
	void *regs = mdev_link_reg(mdev);

	hep_get_buf(ep->urb, &ep->hep->desc, NULL, &buf, &len);

	mcd_device_load_fifo(regs, MCD_HEP_EP0_IDX, len, buf);
	MCD_WRITE_CSR16(regs, MCD_O_HDRC_CSR0, MCD_HEP_EP0_IDX,
			MCD_M_CSR0_H_DISPING | MCD_M_CSR0_TXPKTRDY);
}

static bool hep_cntl_check_tx_remain(struct mcd_hep *ep)
{
	bool remain;
	char* buf;
	u32 len;
	
	hep_get_buf(ep->urb, &ep->hep->desc, NULL, &buf, &len);
	hep_add_actual(ep->urb, len);
	remain = hep_get_remain(ep->urb);

	return remain;
}

static void hep_cntl_request_data(struct mcd_device *mdev)
{
	void *regs = mdev_link_reg(mdev);

	MCD_WRITE_CSR16(regs, MCD_O_HDRC_CSR0, MCD_HEP_EP0_IDX,
			MCD_M_CSR0_H_DISPING | MCD_M_CSR0_H_REQPKT);
}

static bool hep_cntl_get_rx_data(struct mcd_device *mdev, struct mcd_hep *ep,
				u32 *error)
{
	bool remain;
	char* buf;
	u32 expected;
	u32 len;
	void *regs = mdev_link_reg(mdev);
	u8 count0;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	struct urb *urb = ep->urb;

	hep_get_buf(urb, &ep->hep->desc, NULL, &buf, &expected);
	count0 = MCD_READ_CSR16(regs, MCD_O_HDRC_COUNT0, MCD_HEP_EP0_IDX);
	len = min(expected, (u32)count0);
	hep_add_actual(urb, len);

	mcd_device_unload_fifo(regs, MCD_HEP_EP0_IDX, len, buf);
	MCD_WRITE_CSR16(regs, MCD_O_HDRC_CSR0, MCD_HEP_EP0_IDX,
			MCD_M_CSR0_H_DISPING);

	remain = count0 < maxp ? false
		: urb->transfer_buffer_length > urb->actual_length ? true
			: false;

	*error |= hep_check_rx_len_error(maxp, expected, count0, false);

	return remain;
}

static void hep_cntl_request_status(struct mcd_device *mdev)
{
	void *regs = mdev_link_reg(mdev);

	MCD_WRITE_CSR16(regs, MCD_O_HDRC_CSR0, MCD_HEP_EP0_IDX,
			MCD_M_CSR0_H_DISPING |
			MCD_M_CSR0_H_STATUSPKT | MCD_M_CSR0_H_REQPKT);
}

static void hep_cntl_send_status(struct mcd_device *mdev)
{
	void *regs = mdev_link_reg(mdev);

	MCD_WRITE_CSR16(regs, MCD_O_HDRC_CSR0, MCD_HEP_EP0_IDX,
			MCD_M_CSR0_H_DISPING |
			MCD_M_CSR0_H_STATUSPKT | MCD_M_CSR0_TXPKTRDY);
}

static void hep_cntl_dma_sync_start(struct mcd_device *mdev, struct mcd_hep *ep)
{
	struct urb *urb = ep->urb;
	u8 type;
	enum dma_data_direction dir;

	type = hep_cntl_get_type(ep);
	if (type == MCD_HEP_CNTL_OUT) {
		dir = DMA_FROM_DEVICE;
	}
	else if (type == MCD_HEP_CNTL_IN) {
		dir = DMA_TO_DEVICE;
	}
	else {
		goto no_data;
	}

	if (!urb->transfer_buffer) {
		urb->transfer_buffer = ioremap(urb->transfer_dma,
					       urb->transfer_buffer_length);
		if (!urb->transfer_buffer) {
			goto no_data;
		}
		ep->attributes |= MCD_HEP_ATR_IOREMAP;
	}

	if (urb->transfer_dma) {
		dma_sync_single_for_cpu(mdev->dev, urb->transfer_dma,
					urb->transfer_buffer_length, dir);
	}

no_data:
	if (urb->setup_dma) {
		dma_sync_single_for_cpu(mdev->dev, urb->setup_dma, 8,
					DMA_FROM_DEVICE);
	}
}

static void hep_cntl_dma_sync_end(struct mcd_device *mdev, struct mcd_hep *ep)
{
	struct urb *urb = ep->urb;
	u8 type;
	enum dma_data_direction dir;

	type = hep_cntl_get_type(ep);
	if (type == MCD_HEP_CNTL_OUT) {
		dir = DMA_FROM_DEVICE;
	}
	else if (type == MCD_HEP_CNTL_IN) {
		dir = DMA_TO_DEVICE;
	}
	else {
		goto no_data;
	}

	if (urb->transfer_dma) {
		dma_sync_single_for_device(mdev->dev, urb->transfer_dma,
					   urb->transfer_buffer_length, dir);
	}

	if (ep->attributes & MCD_HEP_ATR_IOREMAP) {
		iounmap(urb->transfer_buffer);
		urb->transfer_buffer = NULL;
		ep->attributes &= ~MCD_HEP_ATR_IOREMAP;
	}

no_data:
	if (urb->setup_dma) {
		dma_sync_single_for_device(mdev->dev, urb->setup_dma, 8,
					   DMA_FROM_DEVICE);
	}
}

static int hep_cntl_trans_state(struct mcd_device *mdev, struct mcd_hep *ep,
				u32 trigger)
{
	unsigned long flags;
	void *regs = mdev_link_reg(mdev);
	u32 error = 0;
	bool reset = trigger == MCD_HEP_T_RESET ? true : false;
	bool remain = false;
	u8 type = MCD_HEP_CNTL_NODATA;

	spin_lock_irqsave(&mdev->lock, flags);
	HDBG("HEP\tcntl exit state=%d trig=%d\n", ep->state, trigger);

	mcd_select_end(regs, MCD_HEP_EP0_IDX);
	hep_set_tx_target(mdev, MCD_HEP_EP0_IDX, ep->dev);

	/* exit state */
	switch (ep->state) {
	case MCD_HEP_S_CNTL_IDLE:
		if (!reset) {
			hep_cntl_enable(mdev);
		}
		hep_cntl_dma_sync_start(mdev, ep);
		break;
	case MCD_HEP_S_CNTL_WAIT_TX_SETUP:
		if (!reset) {
			error = hep_cntl_check_error(mdev);
			if (!error) {
				type = hep_cntl_get_type(ep);
			}
		}
		break;
	case MCD_HEP_S_CNTL_WAIT_TX_DATA:
		if (!reset) {
			error = hep_cntl_check_error(mdev);
			remain = hep_cntl_check_tx_remain(ep);
		}
		break;
	case MCD_HEP_S_CNTL_WAIT_RX_DATA:
		if (!reset) {
			error = hep_cntl_check_error(mdev);
			remain = hep_cntl_get_rx_data(mdev, ep, &error);
		}
		break;
	case MCD_HEP_S_CNTL_WAIT_TX_STATUS:
		if (!reset) {
			error = hep_cntl_check_error(mdev);
		}
		break;
	case MCD_HEP_S_CNTL_WAIT_RX_STATUS:
		if (!reset) {
			error = hep_cntl_check_error(mdev);
		}
		break;
	}

	/* transition state */
	if (trigger == MCD_HEP_T_RESET ||
	    (trigger == MCD_HEP_T_LINK_INTR && error)) {
		ep->state = MCD_HEP_S_CNTL_IDLE;
		goto enter_state;
	}
	switch (ep->state) {
	case MCD_HEP_S_CNTL_IDLE:
		if (trigger == MCD_HEP_T_START) {
			ep->state = MCD_HEP_S_CNTL_WAIT_TX_SETUP;
			goto enter_state;
		}
		break;
	case MCD_HEP_S_CNTL_WAIT_TX_SETUP:
		if (trigger == MCD_HEP_T_LINK_INTR) {
			if (type == MCD_HEP_CNTL_OUT && !error) {
				ep->state = MCD_HEP_S_CNTL_WAIT_TX_DATA;
				goto enter_state;
			}
			else if (type == MCD_HEP_CNTL_IN && !error) {
				ep->state = MCD_HEP_S_CNTL_WAIT_RX_DATA;
				goto enter_state;
			}
			else if (type == MCD_HEP_CNTL_NODATA && !error) {
				ep->state = MCD_HEP_S_CNTL_WAIT_RX_STATUS;
				goto enter_state;
			}
		}
		break;
	case MCD_HEP_S_CNTL_WAIT_TX_DATA:
		if (trigger == MCD_HEP_T_LINK_INTR) {
			if (remain && !error) {
				ep->state = MCD_HEP_S_CNTL_WAIT_TX_DATA;
				goto enter_state;
			}
			else if (!remain && !error) {
				ep->state = MCD_HEP_S_CNTL_WAIT_RX_STATUS;
				goto enter_state;
			}
		}
		break;
	case MCD_HEP_S_CNTL_WAIT_RX_DATA:
		if (trigger == MCD_HEP_T_LINK_INTR) {
			if (remain && !error) {
				ep->state = MCD_HEP_S_CNTL_WAIT_RX_DATA;
				goto enter_state;
			}
			else if (!remain && !error) {
				ep->state = MCD_HEP_S_CNTL_WAIT_TX_STATUS;
				goto enter_state;
			}
		}
		break;
	case MCD_HEP_S_CNTL_WAIT_RX_STATUS:
		if (trigger == MCD_HEP_T_LINK_INTR) {
			if (!error) {
				ep->state = MCD_HEP_S_CNTL_IDLE;
				goto enter_state;
			}
		}
		break;
	case MCD_HEP_S_CNTL_WAIT_TX_STATUS:
		if (trigger == MCD_HEP_T_LINK_INTR) {
			if (!error) {
				ep->state = MCD_HEP_S_CNTL_IDLE;
				goto enter_state;
			}
		}
		break;
	}
	goto error_end;

	/* enter state */
enter_state:
	switch (ep->state) {
	case MCD_HEP_S_CNTL_IDLE:
		hep_cntl_disable(mdev);
		hep_cntl_reset(mdev);
		hep_cntl_dma_sync_end(mdev, ep);
		if (!reset) {
			if (!error) {
				hep_complete(mdev, ep, ep->urb, 0);
				hep_cntl_self_start(mdev, ep);
			}
			else {
				hep_error(mdev, ep, error);
			}
		}
		break;
	case MCD_HEP_S_CNTL_WAIT_TX_SETUP:
		hep_cntl_send_setup(mdev, ep);
		break;
	case MCD_HEP_S_CNTL_WAIT_TX_DATA:
		hep_cntl_send_data(mdev, ep);
		break;
	case MCD_HEP_S_CNTL_WAIT_RX_DATA:
		hep_cntl_request_data(mdev);
		break;
	case MCD_HEP_S_CNTL_WAIT_TX_STATUS:
		hep_cntl_send_status(mdev);
		break;
	case MCD_HEP_S_CNTL_WAIT_RX_STATUS:
		hep_cntl_request_status(mdev);
		break;
	}

	HDBG("HEP\t\tenter state=%d\n", ep->state);
	spin_unlock_irqrestore(&mdev->lock, flags);
	return 0;

error_end:
	HDBG("HEP\t\tenter state=%d(ERROR)\n", ep->state);
	spin_unlock_irqrestore(&mdev->lock, flags);
	return -EINVAL;
}

/* tx transaction */

inline static void hep_tx_disable(struct mcd_device *mdev, struct mcd_hep *ep)
{
	u8 index;

	index = hep_get_tx_index(mdev, ep);
	hep_disable_tx(mdev, index);
}

inline static void hep_tx_enable(struct mcd_device *mdev, struct mcd_hep *ep)
{
	u8 index;

	index = hep_get_tx_index(mdev, ep);
	hep_enable_tx(mdev, index);
}

inline static void hep_tx_reset(struct mcd_device *mdev, struct mcd_hep *ep)
{
	u8 index;

	index = hep_get_tx_index(mdev, ep);
	mcd_hep_tx_init(mdev, index);
}

static void hep_tx_self_start(struct mcd_device *mdev, struct mcd_hep *ep)
{
	struct urb *urb;
	struct usb_host_endpoint *hep = ep->hep;

	if (ep->urb || !hep || list_empty(&hep->urb_list)) {
		return;
	}
	urb = list_entry(hep->urb_list.next, struct urb, urb_list);
	ep->urb = urb;

	hep_tx_trans_state(mdev, ep, MCD_HEP_T_START);
}

static void hep_tx_dma_cb(struct mcd_dma_channel *dmach, u8 index, u8 is_tx,
			  u8 xfer_type, void * data)
{
	struct mcd_device *mdev = (struct mcd_device *)data;
	struct mcd_host *mhost = hcd_to_mhost(mdev->hcd);
	struct mcd_hep *ep;

	ep = &mhost->ep_out[index - 1];
	hep_tx_trans_state(mdev, ep, MCD_HEP_T_DMAC_INTR);
}

static void hep_tx_start_dma(struct mcd_device *mdev,
			     struct mcd_dma_channel* dmach,
			     struct usb_endpoint_descriptor *epd,
			     char* buf, u32 len)
{
	u16 maxp = epd->wMaxPacketSize;
	bool is_bulk = usb_endpoint_xfer_bulk(epd);
	mcd_dma_cb_t cb = hep_tx_dma_cb;

	mcd_dma_program_channel(dmach, cb, mdev, maxp, is_bulk, buf, len);
	mcd_dma_enable_int_channel(dmach);
	mcd_dma_start_channel(dmach);
}

static void hep_tx_set_epd(struct mcd_device *mdev, u8 index,
			   struct usb_endpoint_descriptor *epd,
			   struct mcd_dma_channel* dmach)
{
	void *regs = mdev_link_reg(mdev);
	u16 maxp = epd->wMaxPacketSize;
	bool is_bulk = usb_endpoint_xfer_bulk(epd);
	u16 txcsr;

	hep_set_type_reg(mdev, MCD_O_HDRC_TXTYPE, index, epd);
	hep_set_interval_reg(mdev, MCD_O_HDRC_TXINTERVAL, index, epd);

	MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXMAXP, index, maxp);

	if (dmach && is_bulk) {
		txcsr = (MCD_M_TXCSR_AUTOSET | MCD_M_TXCSR_DMAENAB |
			 MCD_M_TXCSR_DMAMODE);
	}
	else if (dmach && !is_bulk) {
		txcsr = MCD_M_TXCSR_DMAENAB;
	}
	else {
		txcsr = 0;
	}
	txcsr &= ~MCD_M_TXCSR_TXPKTRDY;
	txcsr |= (MCD_M_TXCSR_H_ERROR | MCD_M_TXCSR_H_RXSTALL );
	MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, index, txcsr);
}

static int hep_tx_dma_data(struct mcd_device *mdev, struct mcd_hep *ep)
{
	char* buf;
	u32 len;
	u8 index;

	index = hep_get_tx_index(mdev, ep);
	hep_get_buf(ep->urb, &ep->hep->desc, ep->dmach, &buf, &len);
	hep_tx_set_epd(mdev, index, &ep->hep->desc, ep->dmach);
	hep_tx_start_dma(mdev, ep->dmach, &ep->hep->desc, buf, len);

	return 0;
}

static int hep_tx_write_data(struct mcd_device *mdev, struct mcd_hep *ep)
{
	char* buf;
	u32 len;
	void *regs = mdev_link_reg(mdev);
	u8 index;

	index = hep_get_tx_index(mdev, ep);
	hep_get_buf(ep->urb, &ep->hep->desc, NULL, &buf, &len);
	hep_tx_set_epd(mdev, index, &ep->hep->desc, NULL);
	mcd_device_load_fifo(regs, index, len, buf);

	return 0;
}

static int hep_tx_set_pktrdy(struct mcd_device *mdev, struct mcd_hep *ep)
{
	void *regs = mdev_link_reg(mdev);
	u8 index;
	u16 txcsr;

	index = hep_get_tx_index(mdev, ep);
	txcsr = MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, index);

	if (txcsr & MCD_M_TXCSR_DMAMODE) {
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, index,
			MCD_M_TXCSR_DMAMODE | 
			MCD_M_TXCSR_H_ERROR | MCD_M_TXCSR_H_RXSTALL );
	}

	MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, index, MCD_M_TXCSR_TXPKTRDY | 
		MCD_M_TXCSR_H_ERROR | MCD_M_TXCSR_H_RXSTALL );
	return 0;
}

static u32 hep_tx_check_error(struct mcd_device *mdev, struct mcd_hep *ep)
{
	void *regs = mdev_link_reg(mdev);
	u8 index;
	u32 error;
	u16 txcsr;

	index = hep_get_tx_index(mdev, ep);
	txcsr = MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, index);

	error = 0;
	if (txcsr & MCD_M_TXCSR_H_RXSTALL) {
		error |= MCD_HEP_E_RXSTALL;
	}
	if (txcsr & MCD_M_TXCSR_H_ERROR) {
		error |= MCD_HEP_E_NORESP;
	}

	return error;
}

static bool hep_tx_update_actual_multi(struct mcd_hep *ep)
{
	char* buf;
	u32 len;
	u32 remain;
	bool shortp;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	struct urb *urb = ep->urb;

	hep_get_buf(ep->urb, &ep->hep->desc, ep->dmach, &buf, &len);
	len -= len % maxp;
	hep_add_actual(urb, len);

	remain = hep_get_remain(urb);
	shortp = remain || urb->transfer_flags & URB_ZERO_PACKET ? true : false;

	return shortp;
}

inline static void hep_tx_update_actual_short(struct mcd_hep *ep)
{
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	struct urb *urb = ep->urb;

	hep_add_actual(urb, urb->transfer_buffer_length % maxp);
}

static u32 hep_tx_get_nextsz(struct mcd_hep *ep)
{
	char* buf;
	u32 len;
	struct urb *urb = ep->urb;

	hep_get_buf(urb, &ep->hep->desc, ep->dmach, &buf, &len);
	return urb->transfer_buffer_length + len - urb->actual_length;
}

static bool hep_tx_check_remain(struct mcd_hep *ep, u32 *nextsz)
{
	char* buf;
	u32 len;
	u32 lastsz;
	struct urb *urb = ep->urb;

	hep_get_buf(ep->urb, &ep->hep->desc, ep->dmach, &buf, &len);
	lastsz = hep_get_remain(urb);
	hep_add_actual(urb, len);
	*nextsz = hep_get_remain(urb);
	return (lastsz && urb->transfer_flags & URB_ZERO_PACKET) ||
		*nextsz ? true : false;
}

static void hep_tx_set_last_toggle(struct mcd_device *mdev, struct mcd_hep *ep)
{
	void *regs = mdev_link_reg(mdev);
	u8 index;
	u8 epnum;
	u16 txcsr;
	bool toggle;

	index = hep_get_rx_index(mdev, ep);
	txcsr = MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, index);
	toggle = txcsr & MCD_M_TXCSR_H_DATATOGGLE ? true : false;
	epnum = hep_hep2epnum(ep->hep);
	usb_settoggle(ep->dev, epnum, true, toggle);
}

static void hep_tx_set_new_toggle(struct mcd_device *mdev, struct mcd_hep *ep)
{
	void *regs = mdev_link_reg(mdev);
	u8 index;
	u8 epnum;
	bool toggle;

	epnum = hep_hep2epnum(ep->hep);
	toggle = usb_gettoggle(ep->dev, epnum, true);

	if (!toggle) {
		index = hep_get_rx_index(mdev, ep);
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, index,
				MCD_M_TXCSR_CLRDATATOG | MCD_M_TXCSR_H_ERROR | 
				MCD_M_TXCSR_H_RXSTALL );
	}
}

static void hep_tx_dma_sync_start(struct mcd_device *mdev, struct mcd_hep *ep)
{
	struct urb *urb = ep->urb;

	if (!urb->transfer_buffer) {
		urb->transfer_buffer = ioremap(urb->transfer_dma,
					       urb->transfer_buffer_length);
		if (!urb->transfer_buffer) {
			return;
		}
		ep->attributes |= MCD_HEP_ATR_IOREMAP;
	}

	if (urb->transfer_dma) {
		dma_sync_single_for_cpu(mdev->dev, urb->transfer_dma,
					urb->transfer_buffer_length,
					DMA_FROM_DEVICE);
	}
}

static void hep_tx_dma_sync_end(struct mcd_device *mdev, struct mcd_hep *ep)
{
	struct urb *urb = ep->urb;

	if (urb->transfer_dma) {
		dma_sync_single_for_device(mdev->dev, urb->transfer_dma,
					   urb->transfer_buffer_length,
					   DMA_FROM_DEVICE);
	}

	if (ep->attributes & MCD_HEP_ATR_IOREMAP) {
		iounmap(urb->transfer_buffer);
		urb->transfer_buffer = NULL;
		ep->attributes &= ~MCD_HEP_ATR_IOREMAP;
	}
}

static int hep_tx_trans_state(struct mcd_device *mdev, struct mcd_hep *ep,
			      u32 trigger)
{
	unsigned long flags;
	void *regs = mdev_link_reg(mdev);
	u8 index;

	bool reset = trigger == MCD_HEP_T_RESET ? true : false;
	bool remain = false;
	bool shortp = false; 
	bool assigned = false;
	bool is_bulk = false;
	u32 nextsz = 0;
	u32 error = 0;

	spin_lock_irqsave(&mdev->lock, flags);

	index = hep_get_tx_index(mdev, ep);
	mcd_select_end(regs, index);
	hep_set_tx_target(mdev, index, ep->dev);

	HDBG("HEP\ttx[%d] exit state=%d trig=%d\n", index, ep->state, trigger);

	/* exit state */
	switch (ep->state) {
	case MCD_HEP_S_TX_IDLE:
		hep_tx_set_new_toggle(mdev, ep);
		if (!reset) {
			hep_tx_enable(mdev, ep);
			nextsz = hep_tx_get_nextsz(ep);
			if (nextsz) {
				assigned = hep_assign_dmach(mdev, ep);
			}
			if (!assigned) {
				hep_tx_dma_sync_start(mdev, ep);
			}
			is_bulk = hep_check_bulk(ep);
		}
		break;
	case MCD_HEP_S_TX_WAIT_DMA_MULTI:
		if (!reset) {
			error = hep_tx_check_error(mdev, ep);
			shortp = hep_tx_update_actual_multi(ep);
		}
		break;
	case MCD_HEP_S_TX_WAIT_SHORT:
		if (!reset) {
			error = hep_tx_check_error(mdev, ep);
			if (!error) {
				hep_tx_update_actual_short(ep);
			}
		}
		break;
	case MCD_HEP_S_TX_WAIT_SINGLE:
		if (!reset) {
			error = hep_tx_check_error(mdev, ep);
			if (!error) {
				remain = hep_tx_check_remain(ep, &nextsz);
			}
		}
		break;
	case MCD_HEP_S_TX_WAIT_CPU:
		if (!reset) {
			error = hep_tx_check_error(mdev, ep);
			if (!error) {
				remain = hep_tx_check_remain(ep, &nextsz);
			}
		}
		if (reset || error || !remain) {
			hep_tx_dma_sync_end(mdev, ep);
		}
		break;
	}

	/* transition state */
	if (trigger == MCD_HEP_T_RESET ||
	    (trigger == MCD_HEP_T_LINK_INTR && error)) {
		ep->state = MCD_HEP_S_TX_IDLE;
		goto enter_state;
	}
	switch (ep->state) {
	case MCD_HEP_S_TX_IDLE:
		if (trigger == MCD_HEP_T_START) {
			if (!assigned) {
				ep->state = MCD_HEP_S_TX_WAIT_CPU;
				goto enter_state;
			}
			else if (assigned && is_bulk) {
				ep->state = MCD_HEP_S_TX_WAIT_DMA_MULTI;
				goto enter_state;
			}
			else if (assigned && !is_bulk) {
				ep->state = MCD_HEP_S_TX_WAIT_DMA_SINGLE;
				goto enter_state;
			}
		}
		break;
	case MCD_HEP_S_TX_WAIT_DMA_MULTI:
		if (trigger == MCD_HEP_T_DMAC_INTR) {
			if (!shortp) {
				ep->state = MCD_HEP_S_TX_IDLE;
				goto enter_state;
			}
			else if (shortp) {
				ep->state = MCD_HEP_S_TX_WAIT_SHORT;
				goto enter_state;
			}
		}
		break;
	case MCD_HEP_S_TX_WAIT_SHORT:
		if (trigger == MCD_HEP_T_LINK_INTR) {
			if (!error) {
				ep->state = MCD_HEP_S_TX_IDLE;
				goto enter_state;
			}
		}
		break;
	case MCD_HEP_S_TX_WAIT_DMA_SINGLE:
		if (trigger == MCD_HEP_T_DMAC_INTR) {
			ep->state = MCD_HEP_S_TX_WAIT_SINGLE;
			goto enter_state;
		}
		break;
	case MCD_HEP_S_TX_WAIT_SINGLE:
		if (trigger == MCD_HEP_T_LINK_INTR) {
			if (!remain && !error) {
				ep->state = MCD_HEP_S_TX_IDLE;
				goto enter_state;
			}
			else if (remain && nextsz && !error) {
				ep->state = MCD_HEP_S_TX_WAIT_DMA_SINGLE;
				goto enter_state;
			}
			else if (remain && !nextsz && !error) {
				ep->state = MCD_HEP_S_TX_WAIT_SINGLE;
				goto enter_state;
			}
		}
		break;
	case MCD_HEP_S_TX_WAIT_CPU:
		if (trigger == MCD_HEP_T_LINK_INTR) {
			if (!remain && !error) {
				ep->state = MCD_HEP_S_TX_IDLE;
				goto enter_state;
			}
			else if (remain && !error) {
				ep->state = MCD_HEP_S_TX_WAIT_CPU;
				goto enter_state;
			}
		}
		break;
	}
	goto error_end;

	/* enter state */
enter_state:
	switch (ep->state) {
	case MCD_HEP_S_TX_IDLE:
		hep_release_dmach(ep);
		hep_tx_disable(mdev, ep);
		hep_tx_set_last_toggle(mdev, ep);
		hep_tx_reset(mdev, ep);
		if (!reset) {
			if (!error) {
				hep_complete(mdev, ep, ep->urb, 0);
				hep_tx_self_start(mdev, ep);
			}
			else {
				hep_error(mdev, ep, error);
			}
		}
		break;
	case MCD_HEP_S_TX_WAIT_DMA_MULTI:
		hep_tx_dma_data(mdev, ep);
		break;
	case MCD_HEP_S_TX_WAIT_SHORT:
		hep_tx_set_pktrdy(mdev, ep);
		break;
	case MCD_HEP_S_TX_WAIT_DMA_SINGLE:
		hep_tx_dma_data(mdev, ep);
		break;
	case MCD_HEP_S_TX_WAIT_SINGLE:
		hep_tx_set_pktrdy(mdev, ep);
		break;
	case MCD_HEP_S_TX_WAIT_CPU:
		hep_tx_write_data(mdev, ep);
		hep_tx_set_pktrdy(mdev, ep);
		break;
	}

	HDBG("HEP\t\tenter state=%d\n", ep->state);
	spin_unlock_irqrestore(&mdev->lock, flags);
	return 0;

error_end:
	HDBG("HEP\t\tenter state=%d(ERROR)\n", ep->state);
	spin_unlock_irqrestore(&mdev->lock, flags);
	return -EINVAL;
}

/* rx transaction */

inline static void hep_rx_disable(struct mcd_device *mdev, struct mcd_hep *ep)
{
	u8 index;

	index = hep_get_rx_index(mdev, ep);
	hep_disable_rx(mdev, index);
}

inline static void hep_rx_enable(struct mcd_device *mdev, struct mcd_hep *ep)
{
	u8 index;

	index = hep_get_rx_index(mdev, ep);
	hep_enable_rx(mdev, index);
}

inline static void hep_rx_reset(struct mcd_device *mdev, struct mcd_hep *ep)
{
	u8 index;

	index = hep_get_rx_index(mdev, ep);
	mcd_hep_rx_init(mdev, index);
}

static void hep_rx_self_start(struct mcd_device *mdev, struct mcd_hep *ep)
{
	struct urb *urb;
	struct usb_host_endpoint *hep = ep->hep;
	
	if (ep->urb || !hep || list_empty(&hep->urb_list)) {
		return;
	}
	urb = list_entry(hep->urb_list.next, struct urb, urb_list);
	ep->urb = urb;

	hep_rx_trans_state(mdev, ep, MCD_HEP_T_START);
}

static void hep_rx_dma_cb(struct mcd_dma_channel *dmach, u8 index, u8 is_tx,
			  u8 xfer_type, void * data)
{
	struct mcd_device *mdev = (struct mcd_device *)data;
	struct mcd_host *mhost = hcd_to_mhost(mdev->hcd);
	struct mcd_hep *ep;

	ep = &mhost->ep_in[index - 1];
	hep_rx_trans_state(mdev, ep, MCD_HEP_T_DMAC_INTR);
}

static void hep_rx_start_dma(struct mcd_device *mdev,
			     struct mcd_dma_channel* dmach,
			     struct usb_endpoint_descriptor *epd,
			     char* buf, u32 len)
{
	u16 maxp = epd->wMaxPacketSize;
	bool is_bulk = usb_endpoint_xfer_bulk(epd);
	mcd_dma_cb_t cb = hep_rx_dma_cb;

	mcd_dma_program_channel(dmach, cb, mdev, maxp, is_bulk, buf, len);
	mcd_dma_enable_int_channel(dmach);
	mcd_dma_start_channel(dmach);
}

static void hep_rx_set_epd(struct mcd_device *mdev, u8 index,
			   struct usb_endpoint_descriptor *epd,
			   struct mcd_dma_channel* dmach, u32 multi_len)
{
	void *regs = mdev_link_reg(mdev);
	u16 maxp = epd->wMaxPacketSize;
	bool is_bulk = usb_endpoint_xfer_bulk(epd);
	u16 rxcsr;
	u16 npkt;

	hep_set_type_reg(mdev, MCD_O_HDRC_RXTYPE, index, epd);
	hep_set_interval_reg(mdev, MCD_O_HDRC_RXINTERVAL, index, epd);

	MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXMAXP, index, maxp);

	if (dmach && is_bulk) {
		rxcsr = (MCD_M_RXCSR_AUTOCLEAR | MCD_M_RXCSR_H_AUTOREQ |
			 MCD_M_RXCSR_DMAENAB | MCD_M_RXCSR_DMAMODE);
	}
	else if (!dmach && !is_bulk) {
		rxcsr = MCD_M_RXCSR_DISNYET;
	}
	else if (dmach && !is_bulk) {
		rxcsr = MCD_M_RXCSR_DISNYET | MCD_M_RXCSR_DMAENAB;
	}
	else {
		rxcsr = 0;
	}
	MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, index,
			rxcsr | MCD_M_RXCSR_H_ERROR | 
			MCD_M_RXCSR_H_RXSTALL | MCD_M_RXCSR_RXPKTRDY);

	npkt = dmach && is_bulk ? (multi_len + maxp - 1) / maxp : 0;
	MCD_WRITE16(regs, MCD_O_HDRC_RQPKTCOUNT(index), npkt);
}

static int hep_rx_dma_multi(struct mcd_device *mdev, struct mcd_hep *ep)
{
	void *regs = mdev_link_reg(mdev);
	char* buf;
	u32 len;
	u8 index;
	u16 rxcsr;

	index = hep_get_rx_index(mdev, ep);
	hep_get_buf(ep->urb, &ep->hep->desc, ep->dmach, &buf, &len);
	hep_rx_set_epd(mdev, index, &ep->hep->desc, ep->dmach, len);

	rxcsr = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, index);
	rxcsr |= (MCD_M_RXCSR_H_REQPKT | MCD_M_RXCSR_H_ERROR | 
			MCD_M_RXCSR_H_RXSTALL | MCD_M_RXCSR_RXPKTRDY);
	MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, index, rxcsr);
	hep_rx_start_dma(mdev, ep->dmach, &ep->hep->desc, buf, len);

	return 0;
}

static int hep_rx_dma_single(struct mcd_device *mdev, struct mcd_hep *ep)
{
	void *regs = mdev_link_reg(mdev);
	char* buf;
	u32 len;
	u32 expected;
	u16 rxcount;
	u8 index;

	index = hep_get_rx_index(mdev, ep);
	hep_get_buf(ep->urb, &ep->hep->desc, ep->dmach, &buf, &expected);
	rxcount = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCOUNT, index);
	len = min(expected, (u32)rxcount);
	hep_rx_set_epd(mdev, index, &ep->hep->desc, ep->dmach, 0);
	hep_rx_start_dma(mdev, ep->dmach, &ep->hep->desc, buf, len);

	return 0;
}

static int hep_rx_request_data(struct mcd_device *mdev, struct mcd_hep *ep)
{
	void *regs = mdev_link_reg(mdev);
	u8 index;
	u16 rxcsr;

	index = hep_get_rx_index(mdev, ep);
	hep_rx_set_epd(mdev, index, &ep->hep->desc, NULL, 0);
	rxcsr = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, index);
	rxcsr |= (MCD_M_RXCSR_H_REQPKT | MCD_M_RXCSR_H_ERROR | 
			MCD_M_RXCSR_H_RXSTALL | MCD_M_RXCSR_RXPKTRDY);
	MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, index, rxcsr);

	return 0;
}

static u32 hep_rx_check_error(struct mcd_device *mdev, struct mcd_hep *ep)
{
	void *regs = mdev_link_reg(mdev);
	u8 index;
	u32 error;
	u16 rxcsr;

	index = hep_get_rx_index(mdev, ep);
	rxcsr = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, index);

	error = 0;
	if (rxcsr & MCD_M_RXCSR_H_RXSTALL) {
		error |= MCD_HEP_E_RXSTALL;
	}
	if (rxcsr & MCD_M_RXCSR_H_ERROR) {
		error |= MCD_HEP_E_NORESP;
	}

	return error;
}

static void hep_rx_update_actual_multi(struct mcd_hep *ep)
{
	mcd_dma_get_channel_status(ep->dmach);
	hep_add_actual(ep->urb, ep->dmach->dwActualLength);
}

static int hep_rx_check_len(struct mcd_device *mdev, struct mcd_hep *ep,
			    u32 *error)
{
	void *regs = mdev_link_reg(mdev);
	char* buf;
	u32 expected;
	u16 rxcount;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u8 index;
	struct urb *urb = ep->urb;

	index = hep_get_rx_index(mdev, ep);
	hep_get_buf(urb, &ep->hep->desc, ep->dmach, &buf, &expected);
	rxcount = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCOUNT, index);
	*error |= hep_check_rx_len_error(maxp, expected, rxcount,
					 urb->transfer_flags & URB_SHORT_NOT_OK);

	return rxcount;
}

inline static void hep_rx_clear_fifo(struct mcd_device *mdev, u8 index)
{
	void *regs = mdev_link_reg(mdev);
	u16 rxcount;
	u8 offset = MCD_FIFO_OFFSET(index);

	rxcount = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCOUNT, index);
	while (rxcount >= 4) {
		MCD_READ32(regs, offset);
		rxcount -= 4;
	}
	while (rxcount > 0) {
		MCD_READ8(regs, offset);
		rxcount--;
	}
}

static void hep_rx_clear_pktrdy(struct mcd_device *mdev, struct mcd_hep *ep)
{
	void *regs = mdev_link_reg(mdev);
	u8 index;

	index = hep_get_rx_index(mdev, ep);
	hep_rx_clear_fifo(mdev, index);
	MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, index,
			MCD_M_RXCSR_H_ERROR | MCD_M_RXCSR_H_RXSTALL);
}

static bool hep_rx_get_data(struct mcd_device *mdev, struct mcd_hep *ep,
			    u32 *error)
{
	void *regs = mdev_link_reg(mdev);
	char* buf;
	u32 expected;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u16 rxcount;
	u32 len;
	u8 index;
	bool remain;
	struct urb *urb= ep->urb;

	index = hep_get_rx_index(mdev, ep);
	hep_get_buf(urb, &ep->hep->desc, NULL, &buf, &expected);
	if (!urb->transfer_buffer) {
		*error |= MCD_HEP_E_NOBUFFER;
		return false;
	}
	rxcount = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCOUNT, index);
	len = min(expected, (u32)rxcount);
	hep_add_actual(urb, rxcount);
	mcd_device_unload_fifo(regs, index, len, buf);
	hep_rx_clear_fifo(mdev, index);
	MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, index,
			MCD_M_RXCSR_H_ERROR | MCD_M_RXCSR_H_RXSTALL);
	remain = urb->transfer_buffer_length > urb->actual_length ? true : false;
	*error |= hep_check_rx_len_error(maxp, expected, rxcount,
					 urb->transfer_flags & URB_SHORT_NOT_OK);

	return remain;
}

static bool hep_rx_check_remain(struct mcd_hep *ep)
{
	char* buf;
	u32 len;
	u32 nextsz;
	struct urb *urb = ep->urb;

	hep_get_buf(urb, &ep->hep->desc, ep->dmach, &buf, &len);
	hep_add_actual(urb, len);
	nextsz = hep_get_remain(urb);
	return nextsz ? true : false;
}

static void hep_rx_set_last_toggle(struct mcd_device *mdev, struct mcd_hep *ep)
{
	void *regs = mdev_link_reg(mdev);
	u8 index;
	u8 epnum;
	u16 rxcsr;
	bool toggle;

	index = hep_get_rx_index(mdev, ep);
	rxcsr = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, index);
	toggle = rxcsr & MCD_M_RXCSR_H_DATATOGGLE ? true : false;
	epnum = hep_hep2epnum(ep->hep);
	usb_settoggle(ep->dev, epnum, false, toggle);
}

static void hep_rx_set_new_toggle(struct mcd_device *mdev, struct mcd_hep *ep)
{
	void *regs = mdev_link_reg(mdev);
	u8 index;
	u8 epnum;
	bool toggle;

	epnum = hep_hep2epnum(ep->hep);
	toggle = usb_gettoggle(ep->dev, epnum, false);

	if (!toggle) {
		index = hep_get_rx_index(mdev, ep);
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, index,
			MCD_M_RXCSR_CLRDATATOG | MCD_M_RXCSR_H_ERROR | 
			MCD_M_RXCSR_H_RXSTALL | MCD_M_RXCSR_RXPKTRDY);
	}
}

static void hep_rx_dma_sync_start(struct mcd_device *mdev, struct mcd_hep *ep)
{
	struct urb *urb = ep->urb;

	if (!urb->transfer_buffer) {
		urb->transfer_buffer = ioremap(urb->transfer_dma,
					       urb->transfer_buffer_length);
		if (!urb->transfer_buffer) {
			return;
		}
		ep->attributes |= MCD_HEP_ATR_IOREMAP;
	}

	if (urb->transfer_dma) {
		dma_sync_single_for_cpu(mdev->dev, urb->transfer_dma,
					urb->transfer_buffer_length,
					DMA_TO_DEVICE);
	}
}

static void hep_rx_dma_sync_end(struct mcd_device *mdev, struct mcd_hep *ep)
{
	struct urb *urb = ep->urb;

	if (urb->transfer_dma) {
		dma_sync_single_for_device(mdev->dev, urb->transfer_dma,
					   urb->transfer_buffer_length,
					   DMA_TO_DEVICE);
	}

	if (ep->attributes & MCD_HEP_ATR_IOREMAP) {
		iounmap(urb->transfer_buffer);
		urb->transfer_buffer = NULL;
		ep->attributes &= ~MCD_HEP_ATR_IOREMAP;
	}
}

static int hep_rx_trans_state(struct mcd_device *mdev, struct mcd_hep *ep,
			      u32 trigger)
{
	unsigned long flags;
	void *regs = mdev_link_reg(mdev);
	u8 index;

	bool reset = trigger == MCD_HEP_T_RESET ? true : false;
	bool remain = false;
	bool first_shortp = false; 
	bool assigned = false;
	bool is_bulk = false;
	bool shortp = false;
	u32 rxcount = 0;
	u32 error = 0;

	spin_lock_irqsave(&mdev->lock, flags);

	index = hep_get_rx_index(mdev, ep);
	mcd_select_end(regs, index);
	hep_set_rx_target(mdev, index, ep->dev);

	HDBG("HEP\trx[%d] exit state=%d trig=%d\n", index, ep->state, trigger);

	/* exit state */
	switch (ep->state) {
	case MCD_HEP_S_RX_IDLE:
		hep_rx_set_new_toggle(mdev, ep);
		if (!reset) {
			hep_rx_enable(mdev, ep);
			is_bulk = hep_check_bulk(ep);
			if (is_bulk) {
				first_shortp = hep_check_short(ep);
			}
			if (!first_shortp) {
				assigned = hep_assign_dmach(mdev, ep);
			}
			if (!assigned) {
				hep_rx_dma_sync_start(mdev, ep);
			}
		}
		break;
	case MCD_HEP_S_RX_WAIT_DMA_MULTI:
		if (!reset) {
			error = hep_rx_check_error(mdev, ep);
			hep_rx_update_actual_multi(ep);
			rxcount = hep_rx_check_len(mdev, ep, &error);
			if (rxcount) {
				hep_rx_dma_sync_start(mdev, ep);
				hep_rx_get_data(mdev, ep, &error);
				hep_rx_dma_sync_end(mdev, ep);
			}
		}
		break;
	case MCD_HEP_S_RX_WAIT_SINGLE:
		if (!reset) {
			error = hep_rx_check_error(mdev, ep);
			rxcount = hep_rx_check_len(mdev, ep, &error);
			if (!error) {
				shortp = hep_rx_check_short(mdev, ep);
			}
			if (error || shortp) {
				hep_rx_dma_sync_start(mdev, ep);
				hep_rx_get_data(mdev, ep, &error);
				hep_rx_dma_sync_end(mdev, ep);
			}
		}
		break;
	case MCD_HEP_S_RX_WAIT_DMA_SINGLE:
		if (!reset) {
			hep_rx_clear_pktrdy(mdev, ep);
			remain = hep_rx_check_remain(ep);
		}
		break;
	case MCD_HEP_S_RX_WAIT_CPU:
		if (!reset) {
			error = hep_rx_check_error(mdev, ep);
			rxcount = hep_rx_check_len(mdev, ep, &error);
			shortp = hep_rx_check_short(mdev, ep);
			remain = hep_rx_get_data(mdev, ep, &error);
		}
		if (reset || error || !remain || shortp) {
			hep_rx_dma_sync_end(mdev, ep);
		}
		break;
	}

	/* transition state */
	if (trigger == MCD_HEP_T_RESET ||
	    (trigger == MCD_HEP_T_LINK_INTR && error)) {
		ep->state = MCD_HEP_S_RX_IDLE;
		goto enter_state;
	}
	switch (ep->state) {
	case MCD_HEP_S_RX_IDLE:
		if (trigger == MCD_HEP_T_START) {
			if (!assigned) {
				ep->state = MCD_HEP_S_RX_WAIT_CPU;
				goto enter_state;
			}
			else if (assigned && is_bulk) {
				ep->state = MCD_HEP_S_RX_WAIT_DMA_MULTI;
				goto enter_state;
			}
			else if (assigned && !is_bulk) {
				ep->state = MCD_HEP_S_RX_WAIT_SINGLE;
				goto enter_state;
			}
		}
		break;
	case MCD_HEP_S_RX_WAIT_DMA_MULTI:
		if (trigger == MCD_HEP_T_DMAC_INTR) {
			ep->state = MCD_HEP_S_RX_IDLE;
			goto enter_state;
		}
		else if (trigger == MCD_HEP_T_LINK_INTR) {
			if (!error) {
				ep->state = MCD_HEP_S_RX_IDLE;
				goto enter_state;
			}
		}
		break;
	case MCD_HEP_S_RX_WAIT_SINGLE:
		if (trigger == MCD_HEP_T_LINK_INTR) {
			if (shortp && !error) {
				ep->state = MCD_HEP_S_RX_IDLE;
				goto enter_state;
			}
			else if (!shortp && !error) {
				ep->state = MCD_HEP_S_RX_WAIT_DMA_SINGLE;
				goto enter_state;
			}
		}
		break;
	case MCD_HEP_S_RX_WAIT_DMA_SINGLE:
		if (trigger == MCD_HEP_T_DMAC_INTR) {
			if (remain) {
				ep->state = MCD_HEP_S_RX_WAIT_SINGLE;
				goto enter_state;
			}
			else if (!remain) {
				ep->state = MCD_HEP_S_RX_IDLE;
				goto enter_state;
			}
		}
		break;
	case MCD_HEP_S_RX_WAIT_CPU:
		if (trigger == MCD_HEP_T_LINK_INTR) {
			if ((!remain || shortp) && !error) {
				ep->state = MCD_HEP_S_RX_IDLE;
				goto enter_state;
			}
			else if ((remain && !shortp) && !error) {
				ep->state = MCD_HEP_S_RX_WAIT_CPU;
				goto enter_state;
			}
		}
		break;
	}
	goto error_end;

	/* enter state */
enter_state:
	switch (ep->state) {
	case MCD_HEP_S_RX_IDLE:
		hep_release_dmach(ep);
		hep_rx_disable(mdev, ep);
		hep_rx_set_last_toggle(mdev, ep);
		hep_rx_reset(mdev, ep);
		if (!reset) {
			if (!error) {
				hep_complete(mdev, ep, ep->urb, 0);
				hep_rx_self_start(mdev, ep);
			}
			else {
				hep_error(mdev, ep, error);
			}
		}
		break;
	case MCD_HEP_S_RX_WAIT_DMA_MULTI:
		hep_rx_dma_multi(mdev, ep);
		break;
	case MCD_HEP_S_RX_WAIT_SINGLE:
		hep_rx_request_data(mdev, ep);
		break;
	case MCD_HEP_S_RX_WAIT_DMA_SINGLE:
		hep_rx_dma_single(mdev, ep);
		break;
	case MCD_HEP_S_RX_WAIT_CPU:
		hep_rx_request_data(mdev, ep);
		break;
	}

	HDBG("HEP\t\tenter state=%d\n", ep->state);
	spin_unlock_irqrestore(&mdev->lock, flags);
	return 0;

error_end:
	HDBG("HEP\t\tenter state=%d(ERROR)\n", ep->state);
	spin_unlock_irqrestore(&mdev->lock, flags);
	return -EINVAL;
}

/* global functions */

void mcd_hep_clear_atr(struct mcd_port *port, u8 epadr, u32 attribute)
{
	struct mcd_hep *ep;

	ep = hep_search_ep(port, epadr);
	if (ep) {
		ep->attributes &= ~attribute;
	}
}

void mcd_hep_set_atr(struct mcd_port *port, u8 epadr, u32 attribute)
{
	struct mcd_hep *ep;

	ep = hep_search_ep(port, epadr);
	if (ep) {
		ep->attributes |= attribute;
	}
}

void mcd_hep_cntl_init(struct mcd_device *mdev)
{
	void *regs = mdev_link_reg(mdev);

	MCD_WRITE_CSR16(regs, MCD_O_HDRC_CSR0, MCD_HEP_EP0_IDX,
			MCD_M_CSR0_H_DISPING | MCD_M_CSR0_FLUSHFIFO);
}

void mcd_hep_tx_init(struct mcd_device *mdev, u8 index)
{
	mcd_time_t timeout;
	mcd_time_t now;
	u16 txcsr;
	void *regs = mdev_link_reg(mdev);

	mcd_board_start_timer(mdev, mcd_board_usec2time(1000), NULL, NULL);
	timeout = mcd_board_get_time(mdev) + mcd_board_usec2time(200);

	txcsr = MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, index);
	if (txcsr & MCD_M_TXCSR_DMAMODE) {
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, index,
				MCD_M_TXCSR_DMAMODE);
	}

	do {
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_TXCSR, index,
				MCD_M_TXCSR_FLUSHFIFO);
		txcsr = MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, index);
		now = mcd_board_get_time(mdev);
	} while (txcsr & MCD_M_TXCSR_FIFONOTEMPTY && time_gt(timeout, now));

	mcd_board_stop_timer(mdev, NULL);
}

void mcd_hep_rx_init(struct mcd_device *mdev, u8 index)
{
	mcd_time_t timeout;
	mcd_time_t now;
	u16 rxcsr;
	void *regs = mdev_link_reg(mdev);

	mcd_board_start_timer(mdev, mcd_board_usec2time(1000), NULL, NULL);
	timeout = mcd_board_get_time(mdev) + mcd_board_usec2time(200);

	do {
		MCD_WRITE_CSR16(regs, MCD_O_HDRC_RXCSR, index,
				MCD_M_RXCSR_FLUSHFIFO);
		rxcsr = MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, index);
		now = mcd_board_get_time(mdev);
	} while (rxcsr & MCD_M_RXCSR_RXPKTRDY && time_gt(timeout, now));

	mcd_board_stop_timer(mdev, NULL);
}

void mcd_hep_cntl_irq(struct mcd_device *mdev)
{
	struct mcd_host *mhost = hcd_to_mhost(mdev->hcd);

	hep_cntl_trans_state(mdev, &mhost->ep0, MCD_HEP_T_LINK_INTR);
}

void mcd_hep_tx_irq(struct mcd_device *mdev, struct mcd_hep *ep)
{
	hep_tx_trans_state(mdev, ep, MCD_HEP_T_LINK_INTR);
}

void mcd_hep_rx_irq(struct mcd_device *mdev, struct mcd_hep *ep)
{
	hep_rx_trans_state(mdev, ep, MCD_HEP_T_LINK_INTR);
}

/* operation functions */

int mcd_hep_urb_enqueue(struct usb_hcd *hcd, struct usb_host_endpoint *hep,
			struct urb *urb, gfp_t mem_flags)
{
	struct mcd_host *mhost;
	struct mcd_device *mdev;

	struct usb_device *udev;
	struct mcd_hep *ep_table;
	u8 *nep;
	struct mcd_hep *ep;
	unsigned long flags;

	int status = 0;
	unsigned int pipe = urb->pipe;
	int is_out = !usb_pipein(pipe);
	int type = usb_pipetype(pipe);

	mhost = hcd_to_mhost(hcd);
	mdev = mhost->mdev;
	udev = urb->dev;

	spin_lock_irqsave(&mdev->lock, flags);

	if (type == PIPE_CONTROL) {
		ep = &mhost->ep0;
	}
	else {
		ep_table = is_out ? &mhost->ep_out[0] : &mhost->ep_in[0];
		nep = is_out ? &mhost->nep_out : &mhost->nep_in;
		ep = hep_inquiry_ep(mdev, udev, hep, ep_table, nep);
		if (!ep) {
			dev_warn(mdev->dev, "no pipe resource"
				 "(devnum%d,epadr:%.2x)", udev->devnum,
				 hep->desc.bEndpointAddress);
			status = -ENOMEM;
			goto error_end;
		}
		if (type != PIPE_BULK) {
			status = hep_assign_bandwidth(udev, ep);
			if (status < 0) {
				dev_warn(mdev->dev,
					 "no bandwdith resource"
					 "(devnum:%d,epadr:%.2x,sts=%d)",
					 udev->devnum,
					 hep->desc.bEndpointAddress, status);
				goto error_end;
			}
		}
	}

	if (ep->urb != NULL) {
		status = 0;
		urb->hcpriv = (void *)ep;
		goto done;
	}

	ep->urb = urb;
	ep->hep = hep;
	ep->dev = udev;
	urb->hcpriv = (void *)ep;

	/* hep_enqueue */
	switch (type) {
	case PIPE_CONTROL:
		status = hep_cntl_trans_state(mdev, ep, MCD_HEP_T_START);
		break;
	case PIPE_BULK:
	case PIPE_INTERRUPT:
		if (is_out) {
			status = hep_tx_trans_state(mdev, ep, MCD_HEP_T_START);
		}
		else {
			status = hep_rx_trans_state(mdev, ep, MCD_HEP_T_START);
		}
		break;
	default:
		status = -ENOSPC;
		break;
	}

done:
error_end:
	spin_unlock_irqrestore(&mdev->lock, flags);
	return status;
}

int mcd_hep_urb_dequeue(struct usb_hcd *hcd, struct urb *urb)
{
	struct mcd_host *mhost;
	struct mcd_device *mdev;

	struct usb_device *udev;
	struct mcd_hep *ep;
	unsigned long flags;
	struct urb *ep_urb;

	int status = 0;
	unsigned int pipe = urb->pipe;
	int is_out = !usb_pipein(pipe);
	int type = usb_pipetype(pipe);

	mhost = hcd_to_mhost(hcd);
	mdev = mhost->mdev;
	udev = urb->dev;

	ep = (struct mcd_hep *)urb->hcpriv;
	ep_urb = ep->urb;

	spin_lock_irqsave(&mdev->lock, flags);

	if (ep_urb != urb) {
		goto done;
	}

	/* hep_enqueue */
	switch (type) {
	case PIPE_CONTROL:
		status = hep_cntl_trans_state(mdev, ep, MCD_HEP_T_RESET);
		break;
	case PIPE_BULK:
	case PIPE_INTERRUPT:
		if (is_out) {
			status = hep_tx_trans_state(mdev, ep, MCD_HEP_T_RESET);
		}
		else {
			status = hep_rx_trans_state(mdev, ep, MCD_HEP_T_RESET);
		}
		break;
	default:
		status = -ENOSPC;
		break;
	}

done:
	hep_complete(mdev, ep, urb, urb->status);

	spin_unlock_irqrestore(&mdev->lock, flags);

	return status;
}

void mcd_hep_endpoint_disable(struct usb_hcd *hcd, struct usb_host_endpoint *hep)
{
	struct mcd_host *mhost;
	struct mcd_device *mdev;

	struct mcd_hep *ep_table;
	u8 *nep;
	u8 other_nep;
	struct mcd_hep *ep;
	unsigned long flags;

	struct usb_endpoint_descriptor *epd = &hep->desc;
	int is_out = usb_endpoint_dir_out(epd);
	int epnum = hep_hep2epnum(hep);

	mhost = hcd_to_mhost(hcd);
	mdev = mhost->mdev;

	spin_lock_irqsave(&mdev->lock, flags);


	if (epnum == 0) {
		ep = &mhost->ep0;
		ep->dev = NULL;
		ep->hep = NULL;
	}
	else {
		ep_table = is_out ? &mhost->ep_out[0] : &mhost->ep_in[0];
		nep = is_out ? &mhost->nep_out : &mhost->nep_in;
		other_nep = !is_out ? mhost->nep_out : mhost->nep_in;
		hep_release_ep(mdev, hep, ep_table, nep, other_nep);
	}
	spin_unlock_irqrestore(&mdev->lock, flags);
}
