/*-----------------------------------------------------------------------------
 * drivers/usb/mcd/usb_mcd_host.c
 * 
 * USB controller driver
 * 
 * Copyright 2009 Sony Corporation
 *
 * This file is part of the Inventra Controller Driver for Linux.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *---------------------------------------------------------------------------*/

#include <linux/module.h>
#include <linux/device.h>
#include <linux/memory.h>
#include <linux/interrupt.h>
#include <linux/usb.h>
#include <linux/usb/ch9.h>
#include <linux/usb_gadget.h>
#include <linux/usb/mcd/usb_otg_control.h>
#include "../core/hub.h"
#include "../core/hcd.h"

#include "usb_mcd_version.h"
#include "usb_mcd_cnf.h"
#include "usb_mcd_debug.h"
#include "usb_mcd_dma.h"
#include "usb_mcd_otg.h"
#include "usb_mcd_gadget.h"
#include "usb_mcd_host.h"
#include "usb_mcd_device.h"
#include "usb_mcd_access.h"
#include "usb_mcd_musbhdrc.h"
#include "usb_mcd_fifo.h"

/* internal functions */

static void host_init_hub(struct mcd_host *mhost)
{
	mcd_port_init(&mhost->hub.port[0]);
}

static void host_init_hep(struct mcd_host *mhost)
{
	struct mcd_device *mdev;
	void *regs;

	u8 i;

	mdev = mhost->mdev;
	regs = mdev_link_reg(mdev);

	mcd_select_end(regs, 0);
	mcd_hep_cntl_init(mdev);
	spin_lock_init(&mhost->ep0.lock);

	for (i = 1; i < MCD_NUM_EPS - 1; i++) {
		mcd_select_end(regs, i);
		mcd_hep_tx_init(mdev, i);
		mcd_hep_rx_init(mdev, i);
		spin_lock_init(&mhost->ep_in[i].lock);
		spin_lock_init(&mhost->ep_out[i].lock);
	}
}

static inline void host_init_regs(struct usb_hcd *hcd)
{
	struct mcd_host *mhost;
	struct mcd_device *mdev;
	void *regs;
	u8 power;

	mhost = hcd_to_mhost(hcd);
	mdev = mhost->mdev;
	regs = mdev_link_reg(mdev);

	MCD_WRITE8(regs, MCD_O_HDRC_FADDR, 0);
	power = MCD_M_POWER_ENSUSPEND;
	if (mdev->attribute & MCD_DEVICE_ATTR_HS_ENABLE) {
		power |= MCD_M_POWER_HSENAB;
	}
	MCD_WRITE8(regs, MCD_O_HDRC_POWER, power);
	MCD_WRITE8(regs, MCD_O_HDRC_INTRTXE, 0);
	MCD_WRITE8(regs, MCD_O_HDRC_INTRRXE, 0);
	MCD_WRITE8(regs, MCD_O_HDRC_INTRUSBE, 0);
	MCD_WRITE8(regs, MCD_O_HDRC_FRAME, 0);
	MCD_WRITE8(regs, MCD_O_HDRC_TESTMODE, 0);

}

static inline void host_run_regs(struct usb_hcd *hcd)
{
	struct mcd_host *mhost;
	struct mcd_device *mdev;
	void *regs;

	mhost = hcd_to_mhost(hcd);
	mdev = mhost->mdev;
	regs = mdev_link_reg(mdev);

	MCD_WRITE8(regs, MCD_O_HDRC_INTRUSBE, (MCD_M_INTR_RESUME |
					       MCD_M_INTR_CONNECT |
					       MCD_M_INTR_VBUSERROR |
					       MCD_M_INTR_BABBLE |
					       MCD_M_INTR_DISCONNECT));
}

/* operation functions */

static int mcd_host_reset(struct usb_hcd *hcd)
{
	struct mcd_host *mhost;
	struct mcd_device *mdev;

	mhost = hcd_to_mhost(hcd);
	mdev = mhost->mdev;

	host_init_regs(hcd);

	mcd_fifo_reset(mdev);
	host_init_hub(mhost);
	host_init_hep(mhost);

	memset(mhost, sizeof(*mhost), 0);

	return 0;
}

static int mcd_host_start(struct usb_hcd *hcd)
{
	struct mcd_host *mhost;
	struct mcd_device *mdev;
	void *regs;

	unsigned char status;

	mhost = hcd_to_mhost(hcd);
	mdev = mhost->mdev;
	regs = mdev_link_reg(mdev);

	status = mcd_dma_start_controller(mdev->dmac);
	if (!status) {
		return -EBUSY;
	}

	hcd->state = HC_STATE_RUNNING;

	host_run_regs(hcd);

	return 0;
}

static int mcd_host_suspend(struct usb_hcd *hcd, pm_message_t message)
{
	struct mcd_host *mhost;

	mhost = hcd_to_mhost(hcd);

	if (hcd->state != HC_STATE_SUSPENDED) {
		return -EINVAL;
	}

	if (message.event == PM_EVENT_FREEZE) {
		host_init_hub(mhost);
	}

	clear_bit(HCD_FLAG_HW_ACCESSIBLE, &hcd->flags);

	return 0;
}

static int mcd_host_resume(struct usb_hcd *hcd)
{
	set_bit(HCD_FLAG_HW_ACCESSIBLE, &hcd->flags);

	host_init_regs(hcd);
	host_run_regs(hcd);

	/* 顼꥿ܴؿƤӽФ¦usb_hc_diedԤΤ˾ޤ */

	return 0;
}

static void mcd_host_stop(struct usb_hcd *hcd)
{
	struct mcd_host *mhost;
	struct mcd_device *mdev;
	void *regs;
	u8 power;

	mhost = hcd_to_mhost(hcd);
	mdev = mhost->mdev;
	regs = mdev_link_reg(mdev);

	mcd_dma_stop_controller(mdev->dmac);

	MCD_WRITE8(regs, MCD_O_HDRC_FADDR, 0);
	power = MCD_M_POWER_ENSUSPEND;
	if (mdev->attribute & MCD_DEVICE_ATTR_HS_ENABLE) {
		power |= MCD_M_POWER_HSENAB;
	}
	MCD_WRITE8(regs, MCD_O_HDRC_POWER, power);
	MCD_WRITE8(regs, MCD_O_HDRC_DEVCTL, 0);
	MCD_WRITE8(regs, MCD_O_HDRC_INTRUSBE, 0);
	MCD_WRITE8(regs, MCD_O_HDRC_INTRTXE, 0);
	MCD_WRITE8(regs, MCD_O_HDRC_INTRRXE, 0);

	hcd->state = HC_STATE_HALT;
}

static void mcd_host_shutdown(struct usb_hcd *hcd)
{
	/* nothing */
}

static int mcd_host_get_frame_number(struct usb_hcd *hcd)
{
	struct mcd_host *mhost;
	struct mcd_device *mdev;
	void *regs;

	int frame;

	mhost = hcd_to_mhost(hcd);
	mdev = mhost->mdev;
	regs = mdev_link_reg(mdev);

	frame = MCD_READ16(regs, MCD_O_HDRC_FRAME);

	return frame;
}

static struct hc_driver mcd_hc_driver = {
	.description = "mcd_hcd",
	.product_desc = NULL,
	.hcd_priv_size = sizeof(struct mcd_host),
	.irq = NULL,
	.flags = HCD_USB2,
	/* host operations */
	.reset = mcd_host_reset,
	.start = mcd_host_start,
	.suspend = mcd_host_suspend,
	.resume = mcd_host_resume,
	.stop = mcd_host_stop,
	.shutdown = mcd_host_shutdown,
	.get_frame_number = mcd_host_get_frame_number,
	/* hep operations */
	.urb_enqueue = mcd_hep_urb_enqueue,
	.urb_dequeue = mcd_hep_urb_dequeue,
	.endpoint_disable = mcd_hep_endpoint_disable,
	/* hub operations */
	.hub_status_data = mcd_hub_status_data,
	.hub_control = mcd_hub_control,
	.bus_suspend = mcd_hub_bus_suspend,
	.bus_resume = mcd_hub_bus_resume,
	/* unsupport operations */
	.start_port_reset = NULL,
	.hub_irq_enable = NULL,
};

static struct usb_hcd *mcd_host_create_hcd(struct mcd_device *mdev)
{
	struct usb_hcd *hcd;

	hcd = usb_create_hcd(&mcd_hc_driver, mdev->dev, "MUSBMHDRC");
	hcd_to_mhost(hcd)->mdev = mdev;

	return hcd;
}

irqreturn_t mcd_host_link_irq(int irq, void *data)
{
	struct mcd_device *mdev = (struct mcd_device *)data;
	struct usb_hcd *hcd = mdev->hcd;
	struct mcd_host *mhost = hcd_to_mhost(mdev->hcd);
	void *regs = mdev_link_reg(mdev);
	irqreturn_t status = IRQ_NONE;
	u8 intrusb;
	u16 intrtx;
	u16 intrrx;
	u8 i;

	if (unlikely(hcd->state == HC_STATE_HALT ||
	    !test_bit(HCD_FLAG_HW_ACCESSIBLE, &hcd->flags)))
		return IRQ_NONE;

	intrusb = MCD_READ8(regs, MCD_O_HDRC_INTRUSB);
	intrusb &= MCD_READ8(regs, MCD_O_HDRC_INTRUSBE);
	if (intrusb) {
		mcd_hub_irq(&mhost->hub, intrusb);
		status = IRQ_HANDLED;
	}

	intrtx = MCD_READ16(regs, MCD_O_HDRC_INTRTX);
	intrtx &= MCD_READ16(regs, MCD_O_HDRC_INTRTXE);
	if (intrtx) {
		if (intrtx & 1) {
			mcd_hep_cntl_irq(mdev);
		}
		intrtx >>= 1;
		i = 0;
		while (intrtx && i < MCD_NUM_EPS - 1) {
			if (intrtx & 1) {
				mcd_hep_tx_irq(mdev, &mhost->ep_out[i]);
			}
			intrtx >>= 1;
			i++;
		}
		status = IRQ_HANDLED;
	}

	intrrx = MCD_READ16(regs, MCD_O_HDRC_INTRRX);
	intrrx &= MCD_READ16(regs, MCD_O_HDRC_INTRRXE);
	if (intrrx) {
		intrrx >>= 1;
		i = 0;
		while (intrrx && i < MCD_NUM_EPS - 1) {
			if (intrrx & 1) {
				mcd_hep_rx_irq(mdev, &mhost->ep_in[i]);
			}
			intrrx >>= 1;
			i++;
		}
		status = IRQ_HANDLED;
	}

	set_bit(HCD_FLAG_SAW_IRQ, &hcd->flags);

	if (unlikely(hcd->state == HC_STATE_HALT))
		usb_hc_died (hcd);

	return status;
}

const static struct mcd_host_ops mhost_ops = {
	.create_hcd = mcd_host_create_hcd,
	.add_hcd = usb_add_hcd,
	.remove_hcd = usb_remove_hcd,
	.put_hcd = usb_put_hcd,
	.notify_over_current = mcd_port_notify_over_current,
	.link_irq = mcd_host_link_irq,
};

int __init mcd_host_init(void)
{
	return mcd_device_attach_host(&mhost_ops);
}
module_init(mcd_host_init);

void __exit mcd_host_exit(void)
{
	mcd_device_detach_host();
}
module_exit(mcd_host_exit);

MODULE_DESCRIPTION(DRIVER_HOST_INFO);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_LICENSE("GPL");
