/*-----------------------------------------------------------------------------
 * drivers/usb/mcd/usb_mcd_hub.c
 * 
 * USB controller driver
 * 
 * Copyright 2009 Sony Corporation
 *
 * This file is part of the Inventra Controller Driver for Linux.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *---------------------------------------------------------------------------*/

#include <linux/module.h>
#include <linux/device.h>
#include <linux/memory.h>
#include <linux/interrupt.h>
#include <linux/usb.h>
#include <linux/usb/ch9.h>
#include <linux/usb_gadget.h>
#include <linux/usb/mcd/usb_otg_control.h>
#include <linux/usb/mcd/usb_mcd_hub.h>
#include "../core/hub.h"
#include "../core/hcd.h"

#include "usb_mcd_cnf.h"
#include "usb_mcd_debug.h"
#include "usb_mcd_dma.h"
#include "usb_mcd_otg.h"
#include "usb_mcd_gadget.h"
#include "usb_mcd_host.h"
#include "usb_mcd_device.h"
#include "usb_mcd_access.h"
#include "usb_mcd_musbhdrc.h"

/* proto type of internal functions */
static int port_trans_state(struct mcd_port *port, u32 trigger);

/* hub states and triggers */
enum {
	MCD_PORT_S_POWER_OFF = 0,
	MCD_PORT_S_DISCONNECTED,
	MCD_PORT_S_DISABLED,
	MCD_PORT_S_RESETTING,
	MCD_PORT_S_ENABLED,
	MCD_PORT_S_SUSPENDED,
	MCD_PORT_S_RESUMING,
};

enum {
	MCD_PORT_T_CLEAR_PORT_ENABLE,
	MCD_PORT_T_CLEAR_PORT_SUSPEND,
	MCD_PORT_T_CLEAR_PORT_POWER,
	MCD_PORT_T_SET_PORT_SUSPEND,
	MCD_PORT_T_SET_PORT_POWER,
	MCD_PORT_T_SET_PORT_RESET,
	MCD_PORT_T_RESET_END,
	MCD_PORT_T_RESUME_END,
	MCD_PORT_T_DETECT_REMOTEWAKEUP,
	MCD_PORT_T_DETECT_CONNECT,
	MCD_PORT_T_DETECT_DISCONNECT,
};

/* descriptor for root hub */
static const struct usb_hub_descriptor hub_desc = {
	.bDescLength = 9,
	.bDescriptorType = 0x29,
	.bNbrPorts = 1,
	.wHubCharacteristics = (0x1 << 0 | 0x0 << 2 | 0x1 << 3 |
				0x0 << 5 | 0x0 << 7),
	.bPwrOn2PwrGood = 10,
	.bHubContrCurrent = 0,
	.DeviceRemovable[0] = 0 << 1,
	.PortPwrCtrlMask[0] = 0xff,
};

/* internal functions */

static enum usb_device_speed port_get_speed(struct mcd_port *port)
{
	struct mcd_hub *hub;
	struct mcd_host *mhost;
	struct mcd_device *mdev;
	void *regs;

	u8 devctl;
	u8 power;

	hub = port_to_hub(port);
	mhost = hub_to_mhost(hub);
	mdev = mhost->mdev;
	regs = mdev_link_reg(mdev);

	devctl = MCD_READ8(regs, MCD_O_HDRC_DEVCTL);
	power = MCD_READ8(regs, MCD_O_HDRC_POWER);

	if (power & MCD_M_POWER_HSMODE) {
		return USB_SPEED_HIGH;
	}
	else if (devctl & MCD_M_DEVCTL_FSDEV) {
		return USB_SPEED_FULL;
	}
	else if (devctl & MCD_M_DEVCTL_LSDEV) {
		return USB_SPEED_LOW;
	}

	return USB_SPEED_UNKNOWN;
}

static void port_clear_power(struct mcd_port *port)
{
	struct mcd_hub *hub;
	struct mcd_host *mhost;
	struct mcd_device *mdev;
	void *regs;

	u8 devctl;

	hub = port_to_hub(port);
	mhost = hub_to_mhost(hub);
	mdev = mhost->mdev;
	regs = mdev_link_reg(mdev);

	devctl = MCD_READ8(regs, MCD_O_HDRC_DEVCTL);
	devctl &= ~MCD_M_DEVCTL_SESSION;
	MCD_WRITE8(regs, MCD_O_HDRC_DEVCTL, devctl);

	mcd_otg_stop_vbusout(&mdev->otgi.otgc);
}

static void port_set_power(struct mcd_port *port)
{
	struct mcd_hub *hub;
	struct mcd_host *mhost;
	struct mcd_device *mdev;
	void *regs;

	u8 devctl;

	hub = port_to_hub(port);
	mhost = hub_to_mhost(hub);
	mdev = mhost->mdev;
	regs = mdev_link_reg(mdev);

	devctl = MCD_READ8(regs, MCD_O_HDRC_DEVCTL);
	devctl |= MCD_M_DEVCTL_SESSION;
	MCD_WRITE8(regs, MCD_O_HDRC_DEVCTL, devctl);

	mcd_otg_start_vbusout(&mdev->otgi.otgc);
}

static void port_expire_reset(struct mcd_device *mdev, void *data)
{
	struct mcd_port *port = (struct mcd_port *)data;

	port_trans_state(port, MCD_PORT_T_RESET_END);
}

static void port_start_reset(struct mcd_port *port)
{
	struct mcd_hub *hub;
	struct mcd_host *mhost;
	struct mcd_device *mdev;
	void *regs;

	u8 power;
#if defined(CONFIG_ARCH_CXD4105)
#if defined(CONFIG_ARCH_MB8AC102)
	enum usb_device_speed speed;
	speed = port_get_speed(port);
#endif /* CONFIG_ARCH_MB8AC102 */
#else /* CONFIG_ARCH_CXD4115 etc. */
	enum usb_device_speed speed;
	speed = port_get_speed(port);
#endif /* CONFIG_ARCH_CXD4105 */


	hub = port_to_hub(port);
	mhost = hub_to_mhost(hub);
	mdev = mhost->mdev;
	regs = mdev_link_reg(mdev);

	power = MCD_READ8(regs, MCD_O_HDRC_POWER);

#if !defined(CONFIG_ARCH_CXD4105) || defined(CONFIG_ARCH_MB8AC102)
	/* Driver transmits Resume of short time 
	   before the Reset transmission, if it is Suspend mode. */
	if(power & MCD_M_POWER_SUSPENDM){
		power |= MCD_M_POWER_RESUME;
		MCD_WRITE8(regs, MCD_O_HDRC_POWER, power);
		udelay(mcd_board_timeout_phy());

		power = MCD_READ8(regs, MCD_O_HDRC_POWER);
		power &= ~MCD_M_POWER_RESUME;
		MCD_WRITE8(regs, MCD_O_HDRC_POWER, power);
	}
#endif /* !defined(CONFIG_ARCH_CXD4105) || defined(CONFIG_ARCH_MB8AC102) */

	power |= MCD_M_POWER_RESET;
	MCD_WRITE8(regs, MCD_O_HDRC_POWER, power);
#if defined(CONFIG_ARCH_CXD4105)
#if defined(CONFIG_ARCH_MB8AC102)
	if (speed == USB_SPEED_HIGH) {
		mcd_board_linestate_set();
	}
#endif /* CONFIG_ARCH_MB8AC102 */
#else /* CONFIG_ARCH_CXD4115 etc. */
	if (speed == USB_SPEED_HIGH) {
		mcd_board_linestate_set();
	}
#endif /* CONFIG_ARCH_CXD4105 */

	mcd_board_start_timer(mdev, mcd_board_usec2time(MCD_RESET_TIME),
			      port_expire_reset, (void *)port);
}

static void port_stop_reset(struct mcd_port *port)
{
	struct mcd_hub *hub;
	struct mcd_host *mhost;
	struct mcd_device *mdev;
	void *regs;

	u8 power;

	hub = port_to_hub(port);
	mhost = hub_to_mhost(hub);
	mdev = mhost->mdev;
	regs = mdev_link_reg(mdev);

	power = MCD_READ8(regs, MCD_O_HDRC_POWER);
	power &= ~MCD_M_POWER_RESET;
	MCD_WRITE8(regs, MCD_O_HDRC_POWER, power);

	mcd_board_stop_timer(mdev, port);
}

static void port_set_suspend(struct mcd_port *port)
{
	struct mcd_hub *hub;
	struct mcd_host *mhost;
	struct mcd_device *mdev;
	void *regs;

	u8 power;

	hub = port_to_hub(port);
	mhost = hub_to_mhost(hub);
	mdev = mhost->mdev;
	regs = mdev_link_reg(mdev);

	/* PHY is operated on/off, bofore suspend mode */
	mcd_board_control_suspend_on();
	mcd_board_control_suspend_off();

	power = MCD_READ8(regs, MCD_O_HDRC_POWER);
	power |= MCD_M_POWER_SUSPENDM;
	MCD_WRITE8(regs, MCD_O_HDRC_POWER, power);
}

static void port_expire_resume(struct mcd_device *mdev, void *data)
{
	struct mcd_port *port = (struct mcd_port *)data;

	port_trans_state(port, MCD_PORT_T_RESUME_END);
}

static void port_start_resume(struct mcd_port *port)
{
	struct mcd_hub *hub;
	struct mcd_host *mhost;
	struct mcd_device *mdev;
	void *regs;

	u8 power;

	hub = port_to_hub(port);
	mhost = hub_to_mhost(hub);
	mdev = mhost->mdev;
	regs = mdev_link_reg(mdev);

	power = MCD_READ8(regs, MCD_O_HDRC_POWER);
	power |= MCD_M_POWER_RESUME;
	MCD_WRITE8(regs, MCD_O_HDRC_POWER, power);

	mcd_board_start_timer(mdev, mcd_board_usec2time(MCD_RESUME_TIME),
			      port_expire_resume, (void *)port);
}

static void port_stop_resume(struct mcd_port *port)
{
	struct mcd_hub *hub;
	struct mcd_host *mhost;
	struct mcd_device *mdev;
	void *regs;

	u8 power;

	hub = port_to_hub(port);
	mhost = hub_to_mhost(hub);
	mdev = mhost->mdev;
	regs = mdev_link_reg(mdev);

	power = MCD_READ8(regs, MCD_O_HDRC_POWER);
	power &= ~MCD_M_POWER_RESUME;
	MCD_WRITE8(regs, MCD_O_HDRC_POWER, power);

	mcd_board_stop_timer(mdev, port);
}

static inline void port_clear_c_flag(struct mcd_port *port, u16 flag)
{
	port->status.wPortChange &= ~(cpu_to_le16(flag));
}

static void port_set_flag(struct mcd_port *port, u16 flag)
{
	u16 le16_flag = cpu_to_le16(flag);

	if (!(le16_to_cpu(port->status.wPortStatus) & le16_flag) &&
	    flag <= USB_PORT_STAT_RESET) {
		port->status.wPortChange |= le16_flag;
	}
	port->status.wPortStatus |= le16_flag;
}

static void port_clear_flag(struct mcd_port *port, u16 flag)
{
	u16 le16_flag = cpu_to_le16(flag);

	if ((le16_to_cpu(port->status.wPortStatus)) & le16_flag &&
	    flag <= USB_PORT_STAT_RESET) {
		port->status.wPortChange |= le16_flag;
	}
	port->status.wPortStatus &= ~le16_flag;
}

static int port_trans_state(struct mcd_port *port, u32 trigger)
{
	unsigned long flags;
	enum usb_device_speed speed;
	HDBG("HUB\tport exit state=%d trig=%d\n", port->state, trigger);

	spin_lock_irqsave(&port->lock, flags);

	/* exit state */
	switch (port->state) {
	case MCD_PORT_S_RESETTING:
		switch (trigger) {
		case MCD_PORT_T_RESET_END:
		case MCD_PORT_T_CLEAR_PORT_ENABLE:
		case MCD_PORT_T_SET_PORT_RESET:
		case MCD_PORT_T_DETECT_DISCONNECT:
		case MCD_PORT_T_CLEAR_PORT_POWER:
			port_clear_flag(port, USB_PORT_STAT_RESET);
			port_stop_reset(port);
			break;
		}
		break;
	case MCD_PORT_S_RESUMING:
		switch (trigger) {
		case MCD_PORT_T_RESUME_END:
		case MCD_PORT_T_CLEAR_PORT_ENABLE:
		case MCD_PORT_T_SET_PORT_RESET:
		case MCD_PORT_T_DETECT_DISCONNECT:
		case MCD_PORT_T_CLEAR_PORT_POWER:
			port_stop_resume(port);
			break;
		}
		break;
	}

	/* transition state */
	if (trigger == MCD_PORT_T_CLEAR_PORT_POWER) {
		port->state = MCD_PORT_S_POWER_OFF;
		goto enter_state;
	}
	switch (port->state) {
	case MCD_PORT_S_POWER_OFF:
		if (trigger == MCD_PORT_T_SET_PORT_POWER) {
			port->state = MCD_PORT_S_DISCONNECTED;
			goto enter_state;
		}
		break;
	case MCD_PORT_S_DISCONNECTED:
		if (trigger == MCD_PORT_T_DETECT_CONNECT) {
			port->state = MCD_PORT_S_DISABLED;
			goto enter_state;
		}
		break;
	case MCD_PORT_S_DISABLED:
	case MCD_PORT_S_RESETTING:
	case MCD_PORT_S_ENABLED:
	case MCD_PORT_S_SUSPENDED:
	case MCD_PORT_S_RESUMING:
		if (trigger == MCD_PORT_T_DETECT_DISCONNECT) {
			port->state = MCD_PORT_S_DISCONNECTED;
			goto enter_state;
		}
		else if (trigger == MCD_PORT_T_SET_PORT_RESET) {
			port->state = MCD_PORT_S_RESETTING;
			goto enter_state;
		}
		else if (trigger == MCD_PORT_T_CLEAR_PORT_ENABLE) {
			port->state = MCD_PORT_S_DISABLED;
			goto enter_state;
		}
		switch (port->state) {
		case MCD_PORT_S_RESETTING:
			if (trigger == MCD_PORT_T_RESET_END) {
				port->state = MCD_PORT_S_ENABLED;
				goto enter_state;
			}
			break;
		case MCD_PORT_S_ENABLED:
			if (trigger == MCD_PORT_T_SET_PORT_SUSPEND) {
				port->state = MCD_PORT_S_SUSPENDED;
				goto enter_state;
			}
			break;
		case MCD_PORT_S_SUSPENDED:
			if (trigger == MCD_PORT_T_DETECT_REMOTEWAKEUP) {
				port->state = MCD_PORT_S_RESUMING;
				goto enter_state;
			}
			else if (trigger == MCD_PORT_T_CLEAR_PORT_SUSPEND) {
				port->state = MCD_PORT_S_RESUMING;
				goto enter_state;
			}
			break;
		case MCD_PORT_S_RESUMING:
			if (trigger == MCD_PORT_T_RESUME_END) {
				port->state = MCD_PORT_S_ENABLED;
				goto enter_state;
			}
			break;
		}
		break;
	}
	goto error_end;

	/* enter state */
enter_state:
	switch (port->state) {
	case MCD_PORT_S_POWER_OFF:
		port_clear_flag(port, USB_PORT_STAT_CONNECTION);
		port_clear_flag(port, USB_PORT_STAT_ENABLE);
		port_clear_flag(port, USB_PORT_STAT_SUSPEND);
		port_clear_power(port);
		port_clear_flag(port, USB_PORT_STAT_POWER);
		break;
	case MCD_PORT_S_DISCONNECTED:
		port_clear_flag(port, USB_PORT_STAT_CONNECTION);
		port_clear_flag(port, USB_PORT_STAT_ENABLE);
		port_clear_flag(port, USB_PORT_STAT_SUSPEND);
		port_clear_flag(port, USB_PORT_STAT_OVERCURRENT);
		port_set_power(port);
		port_set_flag(port, USB_PORT_STAT_POWER);
		break;
	case MCD_PORT_S_DISABLED:
		port_set_flag(port, USB_PORT_STAT_CONNECTION);
		port_clear_flag(port, USB_PORT_STAT_ENABLE);
		port_clear_flag(port, USB_PORT_STAT_SUSPEND);
		break;
	case MCD_PORT_S_RESETTING:
		port_clear_flag(port, USB_PORT_STAT_ENABLE);
		port_clear_flag(port, USB_PORT_STAT_SUSPEND);
		port_start_reset(port);
		port_set_flag(port, USB_PORT_STAT_RESET);
		break;
	case MCD_PORT_S_ENABLED:
		port_set_flag(port, USB_PORT_STAT_ENABLE);
		port_clear_flag(port, USB_PORT_STAT_SUSPEND);
		speed = port_get_speed(port);
		if (speed == USB_SPEED_LOW) {
			port_set_flag(port, USB_PORT_STAT_LOW_SPEED);
			port_clear_flag(port, USB_PORT_STAT_HIGH_SPEED);
		}
		else if (speed == USB_SPEED_HIGH) {
			port_clear_flag(port, USB_PORT_STAT_LOW_SPEED);
			port_set_flag(port, USB_PORT_STAT_HIGH_SPEED);
		}
		else { /* speed == USB_SPEED_FULL or other */
			port_clear_flag(port, USB_PORT_STAT_LOW_SPEED);
			port_clear_flag(port, USB_PORT_STAT_HIGH_SPEED);
		}
		break;
	case MCD_PORT_S_SUSPENDED:
		port_set_flag(port, USB_PORT_STAT_SUSPEND);
		port_set_suspend(port);
		break;
	case MCD_PORT_S_RESUMING:
		port_start_resume(port);
		break;
	default:
		goto error_end;
	}

	spin_unlock_irqrestore(&port->lock, flags);
	HDBG("HUB\t     enter state=%d\n", port->state);
	return 0;

error_end:
	spin_unlock_irqrestore(&port->lock, flags);
	HDBG("HUB\t     exit state=%d (ERROR)\n", port->state);
	return -EINVAL;
}

/* operation functions */

void mcd_port_init(struct mcd_port* port)
{
	spin_lock_init(&port->lock);
	port_trans_state(port, MCD_PORT_T_CLEAR_PORT_POWER);
}

void mcd_port_notify_over_current(struct mcd_port* port)
{
	port_set_flag(port, USB_PORT_STAT_OVERCURRENT);
	port_trans_state(port, MCD_PORT_T_CLEAR_PORT_POWER);
}

void mcd_hub_irq(struct mcd_hub *hub, u8 intrusb)
{
	struct mcd_device *mdev = hub_to_mhost(hub)->mdev;
	struct mcd_port *port = &hub->port[0];
	union usb_otg_event otg_event;

	if (intrusb & MCD_M_INTR_RESUME) {
		port_trans_state(port, MCD_PORT_T_DETECT_REMOTEWAKEUP);
	}
	if (intrusb & MCD_M_INTR_CONNECT) {
		otg_event.type = USB_OTG_EVENT_TYPE_CON;
		otg_event.con.value = 1;
		mcd_otg_control_signal_interrupt(&mdev->otgi.otgc, &otg_event);
		port_trans_state(port, MCD_PORT_T_DETECT_CONNECT);
	}
	if (intrusb & MCD_M_INTR_DISCONNECT) {
		otg_event.type = USB_OTG_EVENT_TYPE_CON;
		otg_event.con.value = 0;
		mcd_otg_control_signal_interrupt(&mdev->otgi.otgc, &otg_event);
		port_trans_state(port, MCD_PORT_T_DETECT_DISCONNECT);
	}
	if (intrusb & MCD_M_INTR_VBUSERROR) {
		printk(KERN_WARNING "usb_mcd VbusError\n");
	}
	if (intrusb & MCD_M_INTR_BABBLE) {
		otg_event.type = USB_OTG_EVENT_TYPE_CON;
		otg_event.con.value = 0;
		mcd_otg_control_signal_interrupt(&mdev->otgi.otgc, &otg_event);
		port_trans_state(port, MCD_PORT_T_DETECT_DISCONNECT);
		port_trans_state(port, MCD_PORT_T_CLEAR_PORT_POWER);
		port_trans_state(port, MCD_PORT_T_SET_PORT_POWER);
		printk(KERN_WARNING "usb_mcd Babble\n");
	}
}

int mcd_hub_status_data(struct usb_hcd *hcd, char *buf)
{
	struct mcd_host *mhost;
	struct mcd_port *port;

	mhost = hcd_to_mhost(hcd);
	port = &mhost->hub.port[0];

	if (port->status.wPortChange) {
		buf[0] = 1 << 1;
	}

	return 1;		/* return byte */
}

int mcd_hub_control(struct usb_hcd *hcd, u16 typeReq, u16 wValue, u16 wIndex,
		    char *buf, u16 wLength)
{
	struct mcd_host *mhost;
	struct mcd_device *mdev;
	struct mcd_port *port;

	int status;
	unsigned long flags;
	u8 testmode;
	const u8 ports = MCD_NUM_HUB_PORT;

	mhost = hcd_to_mhost(hcd);
	mdev = mhost->mdev;
	port = &mhost->hub.port[0];

	spin_lock_irqsave(&mdev->lock, flags);
	switch (typeReq) {
	case ClearHubFeature:
		switch (wValue) {
		case C_HUB_LOCAL_POWER:
		case C_HUB_OVER_CURRENT:
			/* none */
			break;
		default:
			goto error_end;
		}
		break;
	case ClearPortFeature:
		if (!wIndex || wIndex > ports)
			goto error_end;
		switch (wValue) {
		case USB_PORT_FEAT_ENABLE:
			status = port_trans_state(port,
						  MCD_PORT_T_CLEAR_PORT_ENABLE);
			break;
		case USB_PORT_FEAT_C_ENABLE:
			port_clear_c_flag(port, USB_PORT_STAT_C_ENABLE);
			break;
		case USB_PORT_FEAT_SUSPEND:
			status = port_trans_state(port,
						  MCD_PORT_T_CLEAR_PORT_SUSPEND);
			break;
		case USB_PORT_FEAT_C_SUSPEND:
			port_clear_c_flag(port, USB_PORT_STAT_C_SUSPEND);
			break;
		case USB_PORT_FEAT_POWER:
			status = port_trans_state(port,
						  MCD_PORT_T_CLEAR_PORT_POWER);
			break;
		case USB_PORT_FEAT_C_CONNECTION:
			port_clear_c_flag(port, USB_PORT_STAT_C_CONNECTION);
			break;
		case USB_PORT_FEAT_OVER_CURRENT:
			/* none */
			break;
		case USB_PORT_FEAT_C_OVER_CURRENT:
			port_clear_c_flag(port, USB_PORT_STAT_C_OVERCURRENT);
			break;
		case USB_PORT_FEAT_RESET:
			/* none */
			break;
		case USB_PORT_FEAT_C_RESET:
			port_clear_c_flag(port, USB_PORT_STAT_C_RESET);
			break;
		case USB_PORT_FEAT_INDICATOR:
			port_clear_flag(port, USB_PORT_STAT_INDICATOR);
			break;
		case USB_PORT_FEAT_LOWSPEED:
			/* none */
			break;
		case USB_PORT_FEAT_HIGHSPEED:
			/* none */
			break;
		default:
			goto error_end;
		}
		break;
	case GetHubDescriptor:
		memcpy(buf, &hub_desc, sizeof(hub_desc));
		break;
	case GetHubStatus:
		memset(buf, 0, 4);
		break;
	case GetPortStatus:
		if (!wIndex || wIndex > ports)
			goto error_end;
		memcpy(buf, &port->status, sizeof(port->status));
		break;
	case SetHubFeature:
		goto error_end;
	case SetPortFeature:
		if (!(wIndex & 0xff) || (wIndex & 0xff) > ports)
			goto error_end;
		switch (wValue) {
		case USB_PORT_FEAT_SUSPEND:
			status = port_trans_state(port,
						  MCD_PORT_T_SET_PORT_SUSPEND);
			break;
		case USB_PORT_FEAT_POWER:
			status = port_trans_state(port,
						  MCD_PORT_T_SET_PORT_POWER);
			break;
		case USB_PORT_FEAT_RESET:
			status = port_trans_state(port,
						  MCD_PORT_T_SET_PORT_RESET);
			break;
		case USB_PORT_FEAT_OVER_CURRENT:
			/* none */
			break;
		case USB_PORT_FEAT_INDICATOR:
			port_set_flag(port, USB_PORT_STAT_INDICATOR);
			break;
		case USB_PORT_FEAT_LOWSPEED:
			/* none */
			break;
		case USB_PORT_FEAT_HIGHSPEED:
			/* none */
			break;
		case USB_PORT_FEAT_TEST:
			switch (wIndex >> 8) {
			case 0x1: /* Test_J */
				testmode = USB_OTG_TEST_MODE_H_J;
				break;
			case 0x2: /* Test_K */
				testmode = USB_OTG_TEST_MODE_H_K;
				break;
			case 0x3: /* Test_SE0_NAK */
				testmode = USB_OTG_TEST_MODE_H_SE0_NAK;
				break;
			case 0x4: /* Test_Packet */
				testmode = USB_OTG_TEST_MODE_H_PACKET;
				break;
			case 0x5: /* Test_Force_Enable */
				testmode = USB_OTG_TEST_MODE_H_HS;
				break;
			default:
				goto error_end;
			}
			usb_otg_control_set_test_mode(&mdev->otgi.otgc, testmode);
		}
		break;
	case McdClearPortFeature:
		if (!(wIndex & 0xff) || (wIndex & 0xff) > ports)
			goto error_end;
		switch (wValue) {
		case MCD_PORT_FEAT_EP_NO_DMA:
			mcd_hep_clear_atr(port, wIndex & 0xff00 >> 8,
					  MCD_HEP_ATR_NO_DMA);
			break;
		default:
			goto error_end;
		}
		break;
	case McdSetPortFeature:
		if (!(wIndex & 0xff) || (wIndex & 0xff) > ports)
			goto error_end;
		switch (wValue) {
		case MCD_PORT_FEAT_EP_NO_DMA:
			mcd_hep_set_atr(port, wIndex & 0xff00 >> 8,
					MCD_HEP_ATR_NO_DMA);
			break;
		case MCD_PORT_FEAT_TEST_SE0:
			testmode = USB_OTG_TEST_MODE_H_SE0_NAK;
			goto set_testmode;
		case MCD_PORT_FEAT_TEST_J:
			testmode = USB_OTG_TEST_MODE_H_J;
			goto set_testmode;
		case MCD_PORT_FEAT_TEST_K:
			testmode = USB_OTG_TEST_MODE_H_K;
			goto set_testmode;
		case MCD_PORT_FEAT_TEST_PACKET:
			testmode = USB_OTG_TEST_MODE_H_PACKET;
			goto set_testmode;
		case MCD_PORT_FEAT_TEST_SETUP:
			testmode = USB_OTG_TEST_MODE_H_SETUP;
			goto set_testmode;
		case MCD_PORT_FEAT_TEST_IN:
			testmode = USB_OTG_TEST_MODE_H_IN;
set_testmode:
			usb_otg_control_set_test_mode(&mdev->otgi.otgc, testmode);
			break;
		default:
			goto error_end;
		}
		break;
	default:
		goto error_end;
	}
	spin_unlock_irqrestore(&mdev->lock, flags);
	return 0;

error_end:
	spin_unlock_irqrestore(&mdev->lock, flags);
	return -EPIPE;
}

int mcd_hub_bus_suspend(struct usb_hcd *hcd)
{
	return 0;
}

int mcd_hub_bus_resume(struct usb_hcd *hcd)
{
	struct mcd_host *mhost;
	struct mcd_port *port;

	mhost = hcd_to_mhost(hcd);
	port = &mhost->hub.port[0];

	port_trans_state(port, MCD_PORT_T_CLEAR_PORT_POWER);

	usb_hcd_resume_root_hub(hcd);

	return 0;
}
