/*-----------------------------------------------------------------------------
 * drivers/usb/mcd/usb_mcd_platform.c
 * 
 * USB controller driver
 * 
 * Copyright 2009 Sony Corporation
 *
 * This file is part of the Inventra Controller Driver for Linux.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *---------------------------------------------------------------------------*/

#include <linux/module.h>
#include <linux/device.h>
#include <linux/io.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <linux/usb.h>
#include <linux/usb/ch9.h>
#include <linux/usb_gadget.h>
#include <linux/usb/mcd/usb_otg_control.h>
#include "../core/hub.h"
#include "../core/hcd.h"

#include "usb_mcd_cnf.h"
#include "usb_mcd_debug.h"
#include "usb_mcd_dma.h"
#include "usb_mcd_otg.h"
#include "usb_mcd_gadget.h"
#include "usb_mcd_host.h"
#include "usb_mcd_device.h"
#include "usb_mcd_access.h"
#include "usb_mcd_musbhdrc.h"

static int mcd_platform_map_mem(struct platform_device *pdev, void **regs)
{
	unsigned int start;
	unsigned int len;
	void *ireg;
	struct resource *rsc;

	rsc = platform_get_resource(pdev, IORESOURCE_IO, 0);
	start = rsc->start;
	len = rsc->end - rsc->start;

	ireg = ioremap_nocache(start, len);
	*regs = ireg;

	return 0;
}

static int mcd_platform_release_mem(struct platform_device *pdev, void **regs)
{
	void *ireg;

	ireg = regs[MCD_LINK_REG];
	if (ireg) {
		iounmap(ireg);
	}

	return 0;
}

static int mcd_platform_probe(struct platform_device *pdev)
{
	int status;
	void* regs[MCD_NUM_REGS];
	int irqs[MCD_NUM_IRQS];
	struct mcd_device *mdev;

	/* ֹ */
	irqs[MCD_LINK_IRQ] = platform_get_irq_byname(pdev, "USB_HS");
	if (!irqs[MCD_LINK_IRQ]) {
		return -ENOMEM;
	}
	irqs[MCD_DMAC_IRQ] = platform_get_irq_byname(pdev, "USB_DMAC");
	if (!irqs[MCD_DMAC_IRQ]) {
		return -ENOMEM;
	}

	status = mcd_platform_map_mem(pdev, &regs[MCD_LINK_REG]);
	if (status < 0) {
		mcd_platform_release_mem(pdev, regs);
		return status;
	}

	mdev = mcd_device_probe(&pdev->dev, regs, irqs);
	if (!mdev) {
		mcd_platform_release_mem(pdev, regs);
		return -ENODEV;
	}

	platform_set_drvdata(pdev, mdev);
	mdev->dev = &pdev->dev;

	return 0;
}

static int mcd_platform_remove(struct platform_device *pdev)
{
	struct mcd_device *mdev;

	mdev = platform_get_drvdata(pdev);
	mcd_device_remove(mdev);
	mcd_platform_release_mem(pdev, mdev->regs);
	platform_set_drvdata(pdev, NULL);

	return 0;
}

static int mcd_platform_suspend(struct platform_device *pdev,
				pm_message_t message)
{
	struct mcd_device *mdev;

	mdev = platform_get_drvdata(pdev);
	return mcd_device_suspend(mdev, message);
}

static int mcd_platform_resume(struct platform_device *pdev)
{
	struct mcd_device *mdev;

	mdev = platform_get_drvdata(pdev);
	return mcd_device_resume(mdev);
}

static void mcd_platform_shutdown(struct platform_device *pdev)
{
	struct mcd_device *mdev;

	mdev = platform_get_drvdata(pdev);
	mcd_device_shutdown(mdev);
}

static struct platform_driver mcd_platform_driver = {
	.probe = mcd_platform_probe,
	.remove = mcd_platform_remove,
	.suspend = mcd_platform_suspend,
	.resume = mcd_platform_resume,
	.shutdown = mcd_platform_shutdown,
	.driver = {
		.name = "usb_mcd",
		.owner = THIS_MODULE,
	},
};

int mcd_request_irq(struct mcd_device *mdev)
{
	int ret;
	ret = request_irq(mdev->irqs[MCD_LINK_IRQ], mdev->irq_hdlrs[MCD_LINK_IRQ],
			  IRQF_DISABLED, "mcd_platform_link", (void *)mdev);
	if (ret < 0) {
		return ret;
	}

	ret = request_irq(mdev->irqs[MCD_DMAC_IRQ], mdev->irq_hdlrs[MCD_DMAC_IRQ],
			  IRQF_DISABLED, "mcd_platform_dmac", (void *)mdev);
	if (ret < 0) {
		free_irq(mdev->irqs[MCD_LINK_IRQ], (void *)mdev);
	}

	return ret;
}

void mcd_free_irq(struct mcd_device *mdev)
{
	free_irq(mdev->irqs[MCD_LINK_IRQ], (void *)mdev);
	free_irq(mdev->irqs[MCD_DMAC_IRQ], (void *)mdev);
}

int __init usb_mcd_module_init(void)
{
	extern struct platform_device mcd_platform_device;

	/* Platform deviceȤϿ */
	platform_device_register(&mcd_platform_device);

	return platform_driver_register(&mcd_platform_driver);
}
#if !defined(MCD_STATIC_LINK)
module_init(usb_mcd_module_init);
#endif

void __exit usb_mcd_module_exit(void)
{
	extern struct platform_device mcd_platform_device;

	/* ץåȥեǥХȤϿ */
	platform_device_unregister(&mcd_platform_device);

	platform_driver_unregister(&mcd_platform_driver);
}
module_exit(usb_mcd_module_exit);
