/*--------------------------------------------------------------------------*
 * drivers/usb/mcd/usb_mcd_procfs.c
 * 
 * USB controller driver
 * 
 * Copyright 2005,2006,2008,2009 Sony Corporation
 * 
 * This code is based on musb_procfs.c,
 * Copyright 2005 Mentor Graphics Corporation
 *
 * This file is part of the Inventra Controller Driver for Linux.
 *
 * The Inventra Controller Driver for Linux is free software; you
 * can redistribute it and/or modify it under the terms of the GNU
 * General Public License version 2 as published by the Free Software
 * Foundation.
 *
 * The Inventra Controller Driver for Linux is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with The Inventra Controller Driver for Linux ; if not,
 * write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA  02111-1307  USA
 *
 * ANY DOWNLOAD, USE, REPRODUCTION, MODIFICATION OR DISTRIBUTION
 * OF THIS DRIVER INDICATES YOUR COMPLETE AND UNCONDITIONAL ACCEPTANCE
 * OF THOSE TERMS.THIS DRIVER IS PROVIDED "AS IS" AND MENTOR GRAPHICS
 * MAKES NO WARRANTIES, EXPRESS OR IMPLIED, RELATED TO THIS DRIVER.
 * MENTOR GRAPHICS SPECIFICALLY DISCLAIMS ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY; FITNESS FOR A PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT.  MENTOR GRAPHICS DOES NOT PROVIDE SUPPORT
 * SERVICES OR UPDATES FOR THIS DRIVER, EVEN IF YOU ARE A MENTOR
 * GRAPHICS SUPPORT CUSTOMER.
 *---------------------------------------------------------------------------*/

#include <asm/uaccess.h>
#include <asm/io.h>

#include <linux/kernel.h>
#include <linux/proc_fs.h>

#include <linux/pci.h>
#include <linux/usb.h>
#include <linux/usb/ch9.h>
#include <linux/usb_gadget.h>
#include <linux/usb/mcd/usb_otg_control.h>
#include "../core/hub.h"
#include "../core/hcd.h"

#include "usb_mcd_cnf.h"
#include "usb_mcd_debug.h"
#include "usb_mcd_dma.h"
#include "usb_mcd_otg.h"
#include "usb_mcd_gadget.h"
#include "usb_mcd_host.h"
#include "usb_mcd_device.h"
#include "usb_mcd_access.h"
#include "usb_mcd_musbhdrc.h"
#include "usb_mcd_version.h"

int mcd_proc_cid = 1;

#if MCD_DEBUG > 0
static int atoi(char *buffer, int base, int len)
{
	int result = 0, digit = 0;

	while (len-- > 0 && (*buffer)) {
		digit = ((*buffer >= '0') && (*buffer <= '9'))
		    ? *buffer - '0' : ((*buffer >= 'a') && (*buffer <= 'f'))
		    ? *buffer - 'a' : -1;

		if (digit < 0) {
			break;
		}

		buffer++;
		result = result * base + digit;
	}

	return result;
}

static const char *decode_address(int index)
{
	static const char *COMMON_REGISTER_MAP[] = {
		"FAddr", "Power", "IntrTx", "IntrRx",
		"IntrTxE", "IntrRxE", "IntrUSB", "IntrUSBE",
		"Frame", "Index", "TestMode"
	};
	return (index < 11) ? COMMON_REGISTER_MAP[index] : NULL;
}
#endif

static int
uname(void *tgt,
      int (*func) (void *tgt, const char *fmt, ...), struct mcd_device *mdev)
{
	void *regs = mdev_link_reg(mdev);

	return func(tgt,
		    "mcd " DRIVER_VERSION
		    " %sDRC, Mode=%d [debug=%d][eps=%d(%p)] "
		    "(Power=%02x, DevCtl=%02x)\n",
		    "MH", mdev->otgi.state,
#if MCD_DEBUG>0
		    mcd_debug_level,
#else
		    -1,
#endif
		    MCD_NUM_EPS, mdev,
		    MCD_READ8(regs, MCD_O_HDRC_POWER),
		    MCD_READ8(regs, MCD_O_HDRC_DEVCTL));
}

static int report_hep(void *tgt,
		     int (*func) (void *tgt, const char *fmt, ...),
		     struct mcd_hep *ep)
{
	u8 ttype;
	int code;
	const char *type_str[] = { "CTRL", "ISOC", "BULK", "INTR", "NOUSE" };

	code = func(tgt, "st=%d, ", ep->state);
	if (code < 0) return -1;

	if (ep->hep) {
		struct usb_endpoint_descriptor *desc;
		desc = &ep->hep->desc;
		ttype = desc->bmAttributes &
			USB_ENDPOINT_XFERTYPE_MASK;
		code = func(tgt, "%s, maxp=%d, epadr=0x%x, ",
			    type_str[ttype], desc->wMaxPacketSize,
			    desc->bEndpointAddress);
		if (code < 0) return -1;
	}
	else {
		code = func(tgt, "%s", type_str[4]);
		if (code < 0) return -1;
	}
	code = func(tgt, "\n");

	return code;
}

static int report_gep(void *tgt,
		     int (*func) (void *tgt, const char *fmt, ...),
		     struct mcd_gep *ep)
{
	u8 ttype;
	int code;
	const char *type_str[] = { "CTRL", "ISOC", "BULK", "INTR", "NOUSE" };
	u8 index;

	index = ep->bEndNumber;
	if (index == 0 || ep->bInactive) {
		ttype = ep->bTrafficType & USB_ENDPOINT_XFERTYPE_MASK;
		if (ttype > sizeof(type_str)/sizeof(type_str[0])) {
			ttype = 4;
		}
		code = func(tgt, "%s, maxp=%d, epnum=%d, is_tx=%d",
			    type_str[ttype], ep->wPacketSize,
			    ep->bEndNumber, ep->bIsTx);
		if (code < 0) return -1;
	}
	else {
		code = func(tgt, "%s", type_str[4]);
		if (code < 0) return -1;
	}
	code = func(tgt, "\n");

	return code;
}

/* Read the registers and generate the report.
 */
static int report(void *tgt,
		  int (*func) (void *tgt, const char *fmt, ...),
		  struct mcd_device *mdev)
{
	int code;
	unsigned int i;
	void *regs = mdev_link_reg(mdev);

	code = uname(tgt, func, mdev);
	if (code < 0) {
		goto report_error;
	}

	if (mdev->otgi.state == MCD_OTGC_S_A_HOST && mdev->hcd) {
		struct mcd_host *mhost = hcd_to_mhost(mdev->hcd);
		code = func(tgt, "HOST\n");
		if (code < 0) goto report_error;
		/* report ep0 */
		code = func(tgt, "EP0     ");
		if (code < 0) goto report_error;
		code = report_hep(tgt, func, &mhost->ep0);
		if (code < 0) goto report_error;

		/* report EP-IN */
		for (i = 0; i < MCD_NUM_EPS; i++) {
			code = func(tgt, "EP%d-IN  ", i);
			if (code < 0) goto report_error;
			code = report_hep(tgt, func, &mhost->ep_in[i]);
			if (code < 0) goto report_error;
		}

		/* report EP-OUT */
		for (i = 0; i < MCD_NUM_EPS; i++) {
			code = func(tgt, "EP%d-OUT ", i);
			if (code < 0) goto report_error;
			code = report_hep(tgt, func, &mhost->ep_out[i]);
			if (code < 0) goto report_error;
		}
	}
	else if (mdev->otgi.state == MCD_OTGC_S_B_PERIPH) {
		struct mcd_gadget *mgadget = &mdev->gadget;
		code = func(tgt, "GADGET\n");
		if (code < 0) goto report_error;
		/* report EP */
		for (i = 0; i < MCD_NUM_EPS; i++) {
			code = func(tgt, "EP%d  ", i);
			if (code < 0) goto report_error;
			code = report_gep(tgt, func, &mgadget->ep[i]);
			if (code < 0) goto report_error;
			if (i == 0) continue;
			mcd_select_end(regs, i);
			code = func(tgt, "\t\trxcsr=0x%x txcsr=0x%x\n",
				    MCD_READ_CSR16(regs, MCD_O_HDRC_RXCSR, i),
				    MCD_READ_CSR16(regs, MCD_O_HDRC_TXCSR, i));
		}
	}

	return 0;

      report_error:
	WARN("A problem generating the report\n");
	return code;
}

struct print_tgt {
	int size;
};

static int print_func(void *tgt, const char *fmt, ...)
{
	int code;
	va_list args;

	va_start(args, fmt);
	code = vprintk(fmt, args);
	va_end(args);
	if (code >= 0) {
		((struct print_tgt *)tgt)->size += code;
	}

	return code;
}

struct sprint_tgt {
	int size;
	int limit;
	int off;
	char *buf;
};

static int sprint_func(void *tgt, const char *fmt, ...)
{
	int code;
	struct sprint_tgt *stgt = (struct sprint_tgt *)tgt;
	va_list args;

	if (!stgt->buf) {
		return -1;
	}

	va_start(args, fmt);
	code = vsnprintf(stgt->buf, stgt->limit, fmt, args);
	va_end(args);
	if (code >= 0) {
		/* not support offset */
		stgt->buf += code;
		stgt->size += code;
		stgt->limit -= code;
	}

	return code;
}

/* Write to ProcFS
 *
 * a dummy input for a-device as CID=0
 * b dummy input for b-device as CID=1
 * C soft-connect
 * c soft-disconnect
 * I enable HS
 * i disable HS
 * R resume bus
 * S start session (OTG-friendly when OTG-compiled)
 * s stop session
 * F force session (OTG-unfriendly)
 * N CID not pullup
 * n CID pullup
 * H request host mode
 * P disable the low-power mode that kills us in peripheral mode
 * D<num> set/query the debug level
 * Z zap
 */
static int mcd_proc_write(struct file *file, const char *buffer,
			  unsigned long count, void *data)
{
	char cmd;
	uint8_t bReg;
	struct mcd_device *mdev = (struct mcd_device *)data;
	void *regs = mdev_link_reg(mdev);
        unsigned long sts;

	/* MOD_INC_USE_COUNT; */

        sts = copy_from_user(&cmd, buffer, 1);
	switch (cmd) {
	case ':':
		if (regs) {
			mcd_select_end(regs, 0);
			printk("" " HWVERS   %x\n", MCD_READ16(regs, 0x6c));
			printk("" " CFGDATA  %x\n", MCD_READ8(regs, 0x1f));
			printk("" " FADDR    %x\n",
			       MCD_READ8(regs, MCD_O_HDRC_FADDR));
			printk("" " INTRUSBE %x\n",
			       MCD_READ8(regs, MCD_O_HDRC_INTRUSBE));
			printk("" " INTRTXE %x\n",
			       MCD_READ8(regs, MCD_O_HDRC_INTRTXE));
			printk("" " INTRRXE %x\n",
			       MCD_READ8(regs, MCD_O_HDRC_INTRRXE));
			printk("" " POWER    %x\n",
			       MCD_READ8(regs, MCD_O_HDRC_POWER));
			printk("" " DEVCTL   %x\n",
			       MCD_READ8(regs, MCD_O_HDRC_DEVCTL));
			printk("" " TESTMODE %x\n",
			       MCD_READ8(regs, MCD_O_HDRC_TESTMODE));
			printk("" " FRAME    %x\n",
			       MCD_READ8(regs, MCD_O_HDRC_FRAME));
			mcd_board_dump(mdev);
		}
		break;
	case 'a':
		mcd_proc_cid = 0;
		break;
	case 'b':
		mcd_proc_cid = 1;
		break;
	case 'v':
		{
#if defined(CONFIG_ARCH_INTEGRATOR)
			extern void mcd_otg_integrator_notify_vbuserror(void);
			mcd_otg_integrator_notify_vbuserror();
#endif
		}
		break;

	case 'C':
		/* D+ Pullup ON */
		if (regs) {
			struct usb_gadget *gadget = mdev->gadget.pGadget;
			usb_gadget_connect(gadget);
		}
		break;
	case 'c':
		/* D+ Pullup OFF */
		if (regs) {
			struct usb_gadget *gadget = mdev->gadget.pGadget;
			usb_gadget_disconnect(gadget);
		}
		break;

	case 'I':
		if (regs) {
			bReg =
			    MCD_READ8(regs,
				      MCD_O_HDRC_POWER) | MCD_M_POWER_HSENAB;
			MCD_WRITE8(regs, MCD_O_HDRC_POWER, bReg);
		}
		break;

	case 'i':
		if (regs) {
			bReg =
			    MCD_READ8(regs,
				      MCD_O_HDRC_POWER) & ~MCD_M_POWER_HSENAB;
			MCD_WRITE8(regs, MCD_O_HDRC_POWER, bReg);
		}
		break;

	case 'R':
		if (regs) {
			bReg = MCD_READ8(regs, MCD_O_HDRC_POWER);
			MCD_WRITE8(regs, MCD_O_HDRC_POWER,
				   bReg | MCD_M_POWER_RESUME);
			msleep(10);
			MCD_WRITE8(regs, MCD_O_HDRC_POWER, bReg);
			WARN("Power Resumed\n");
		}
		break;
	case 'S':
		if (regs) {
			/* might not b necessary */
			bReg = MCD_READ8(regs, MCD_O_HDRC_DEVCTL);
			MCD_WRITE8(regs, MCD_O_HDRC_DEVCTL,
				   bReg | MCD_M_DEVCTL_SESSION);
			WARN("SESSION MODE\n");
		}
		break;

	case 's':
		bReg = MCD_READ8(regs, MCD_O_HDRC_DEVCTL);
		bReg &= ~MCD_M_DEVCTL_SESSION;
		MCD_WRITE8(regs, MCD_O_HDRC_DEVCTL, bReg);
		break;

	case 'F':
		bReg = MCD_READ8(regs, MCD_O_HDRC_DEVCTL);
		bReg |= MCD_M_DEVCTL_SESSION;
		MCD_WRITE8(regs, MCD_O_HDRC_DEVCTL, bReg);
		break;

#if defined(CONFIG_ARCH_CXD4105)
	case 'N':
		/* CID pullup off */
		{
			int ret = 0;

			ret = mcd_board_cid_pullup(0);
			if (ret < 0) {
				printk(" cid_pullup off error ret=%d\n", ret);
			}
		}
		break;

	case 'n':
		/* CID pullup on */
		{
			int ret = 0;

			ret = mcd_board_cid_pullup(1);
			if (ret < 0) {
				printk(" cid_pullup on error ret=%d\n", ret);
			}
		}
		break;
#endif			/* defined(CONFIG_ARCH_CXD4105) */

	case 't':
		{
			char command[3];
			unsigned int pn;
			int ret;

                        sts = copy_from_user(command, buffer, 2);
			command[2] = '\0';
			pn = (unsigned int)(command[1] - '0');
			ret =
			    usb_otg_control_select_port(&mdev->otgi.otgc, pn);
			if (ret < 0) {
				printk(" select_port error pn=%d, ret=%d\n", pn,
				       ret);
			}
		}
		break;

	case 'T':
		{
			int ret;
			struct usb_otg_control_port_info info;

			ret =
			    usb_otg_control_get_port_info(&mdev->otgi.otgc,
							  &info);
			if (ret < 0) {
				printk(" get_port_info error ret=%d\n", ret);
			}
			printk
			    (" get_port_info : port_info nr=%d,"
			     " current_port=%d\n",
			     info.nr, info.current_port);
		}
		break;

	case 'U':
		{
			int code;
			struct print_tgt tgt;
			tgt.size = 0;
			code = uname((void *)&tgt, print_func, mdev);
			if (code < 0) {
				WARN("A problem generating the report\n");
			}
		}
		break;

	case 'p':
		/* D+ Pullup ON (direct) */
		bReg =
		    MCD_READ8(regs, MCD_O_HDRC_POWER) | MCD_M_POWER_SOFTCONN;
		MCD_WRITE8(regs, MCD_O_HDRC_POWER, bReg);
		break;
	case 'P':
		/* D+ Pullup OFF (direct) */
		bReg =
		    MCD_READ8(regs, MCD_O_HDRC_POWER) & ~MCD_M_POWER_SOFTCONN;
		MCD_WRITE8(regs, MCD_O_HDRC_POWER, bReg);
		break;

	case '@':
		/* PHY reset */
		{
			extern void mcd_board_phy_reset(void);
			mcd_board_phy_reset();
		}
		break;

#if (MCD_DEBUG>0)
		/* read & write registers */
	case 'r':
	case 'w':{
			uint8_t index = 0;
			uint32_t value = 0;
			char command[64];

			memset(command, 0, sizeof(command));
			copy_from_user(command, buffer,
				       min(count, (unsigned long)63));

			/* detrermine the index, 
			 * only the adrress now */
			index = atoi(&command[2], 16, count - 2);
			if (index > 0 && regs) {
				const char *address = decode_address(index);

				if (buffer[0] == 'r') {
					value = (command[1] == '8')
					    ? MCD_READ8(regs, index)
					    : (command[1] == 'f')
					    ? MCD_READ16(regs, index)
					    : 0;
				} else {
					// not write, not yet...
					index = -1;
				}

				if (address) {
					printk(KERN_INFO "%s=0x%x\n", address,
					       value);
				} else {
					printk(KERN_INFO "0x%x=0x%x\n", index,
					       value);
				}
			}
		}
		break;

		/* set/read debug level */
	case 'D':{
			if (count > 1) {
				char digits[8], *p = digits;
				int i = 0, level = 0, sign = 1, len =
				    min(count - 1, (unsigned long)8);

				copy_from_user(&digits, &buffer[1], len);

				/* optional sign */
				if (*p == '-') {
					len -= 1;
					sign = -sign;
					p++;
				}

				/* read it */
				while (i++ < len && *p > '0' && *p < '9') {
					level = level * 10 + (*p - '0');
					p++;
				}

				mcd_debug_level = sign * level;
				INFO("mcd_debug_level=%d\n", mcd_debug_level);
			} else {
				INFO("mcd_debug_level=%d\n", mcd_debug_level);
			}

		}
		break;

		/* flush */
#if defined(MCD_GADGET)
	case 'd':
	case 'e':{
			int index = 0;
			char command[64];

			memset(command, 0, sizeof(command));
			copy_from_user(command, buffer,
				       min(count, (unsigned long)63));

			/* detrermine the index, 
			 * only the adrress now */
			index = atoi(&command[1], 10, count - 1);
			if (index == 0) {
				ERR("you can't oprate on ep0\n");
				break;
			}

			if (index < 0 || index > MCD_NUM_EPS) {
				ERR("ep%d does not exist\n", index);
				break;
			}

			if (cmd == 'd') {
				mcd_gadget_disable_end_point(index);
			} else {
				mcd_gadget_enable_end_point(index, NULL);
			}
		} break;
#endif				/* defined(MCD_GADGET) */

	case '?':
		INFO("?: D'ho\n");
		INFO(":: show registers\n");
		INFO("C/c: soft connect enable/disable\n");
		INFO("I/i: hispeed enable/disable\n");
		INFO("S/s: session set/clear\n");
		INFO("F: \n");
		INFO("r/w: read write register\n");
		INFO("U: uname\n");
		INFO("D: set/read dbug level\n");
#if defined(MCD_GADGET)
		INFO("d/e: disable/enable end point (only new gadgets)\n");
		INFO("p: force D+ Pullup ON\n");
		INFO("P: force D+ Pullup OFF\n");
#endif				/* defined(MCD_GADGET) */
		break;
#endif				/* (MCD_DEBUG>0) */

	default:
		ERR("Command %c not implemented\n", cmd);
		break;
	}

	/* MOD_DEC_USE_COUNT; */
	return count;
}

/*
 *
 *
 *
 *
 */
static int
mcd_proc_read(char *page,
	      char **start, off_t off, int count, int *eof, void *data)
{
	struct sprint_tgt tgt;
	unsigned long flags;
	struct mcd_device *mdev = (struct mcd_device *)data;

	tgt.buf = page;
	tgt.size = 0;
	tgt.off = off;
	tgt.limit = count;

	spin_lock_irqsave(&mdev->lock, flags);

	/* off0ʳξưʤmore/lessˤ̤б(catOK) */
	report((void *)&tgt, sprint_func, mdev);

	spin_unlock_irqrestore(&mdev->lock, flags);

	*eof = 1;

	return tgt.size;
}

/**
 */
void mcd_proc_delete_fs(struct mcd_device *mdev)
{
	remove_proc_entry(mdev->proc_entry->name, NULL);
}

/**
 */
struct proc_dir_entry *mcd_proc_create_fs(struct mcd_device *mdev, char *name)
{
	mdev->proc_entry = create_proc_entry(name,
					     S_IFREG | S_IRUGO | S_IWUSR, NULL);
	if (mdev->proc_entry) {
		mdev->proc_entry->data = mdev;
		mdev->proc_entry->owner = THIS_MODULE;
		mdev->proc_entry->read_proc = mcd_proc_read;
		mdev->proc_entry->write_proc = mcd_proc_write;
		mdev->proc_entry->size = 0;

		dbg("Registered /proc/%s\n", name);
	} else {
		dbg("Cannot create a valid proc file entry");
	}

	mcd_proc_cid = 1;

	return mdev->proc_entry;
}
