/* 2008-10-21: File changed by Sony Corporation */
/*
 *  linux/fs/proc/proc_misc.c
 *
 *  linux/fs/proc/array.c
 *  Copyright (C) 1992  by Linus Torvalds
 *  based on ideas by Darren Senn
 *
 *  This used to be the part of array.c. See the rest of history and credits
 *  there. I took this into a separate file and switched the thing to generic
 *  proc_file_inode_operations, leaving in array.c only per-process stuff.
 *  Inumbers allocation made dynamic (via create_proc_entry()).  AV, May 1999.
 *
 * Changes:
 * Fulton Green      :  Encapsulated position metric calculations.
 *			<kernel@FultonGreen.com>
 */

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/time.h>
#include <linux/kernel.h>
#include <linux/kernel_stat.h>
#include <linux/fs.h>
#include <linux/tty.h>
#include <linux/string.h>
#include <linux/mman.h>
#include <linux/proc_fs.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/mmzone.h>
#include <linux/pagemap.h>
#include <linux/swap.h>
#include <linux/slab.h>
#include <linux/smp.h>
#include <linux/signal.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/seq_file.h>
#include <linux/times.h>
#include <linux/profile.h>
#include <linux/utsname.h>
#include <linux/blkdev.h>
#include <linux/hugetlb.h>
#include <linux/jiffies.h>
#include <linux/sysrq.h>
#include <linux/vmalloc.h>
#include <linux/crash_dump.h>
#include <linux/pid_namespace.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/io.h>
#include <asm/tlb.h>
#include <asm/div64.h>
#include "internal.h"

#define LOAD_INT(x) ((x) >> FSHIFT)
#define LOAD_FRAC(x) LOAD_INT(((x) & (FIXED_1-1)) * 100)
/*
 * Warning: stuff below (imported functions) assumes that its output will fit
 * into one page. For some of those functions it may be wrong. Moreover, we
 * have a way to deal with that gracefully. Right now I used straightforward
 * wrappers, but this needs further analysis wrt potential overflows.
 */
extern int get_hardware_list(char *);
extern int get_stram_list(char *);
extern int get_filesystem_list(char *);
extern int get_exec_domain_list(char *);
extern int get_dma_list(char *);
extern int get_locks_status (char *, char **, off_t, int);
#ifdef CONFIG_SNSC
extern int __init nlver_init(void);
#endif

static int proc_calc_metrics(char *page, char **start, off_t off,
				 int count, int *eof, int len)
{
	if (len <= off+count) *eof = 1;
	*start = page + off;
	len -= off;
	if (len>count) len = count;
	if (len<0) len = 0;
	return len;
}

static int loadavg_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int a, b, c;
	int len;

	a = avenrun[0] + (FIXED_1/200);
	b = avenrun[1] + (FIXED_1/200);
	c = avenrun[2] + (FIXED_1/200);
	len = sprintf(page,"%d.%02d %d.%02d %d.%02d %ld/%d %d\n",
		LOAD_INT(a), LOAD_FRAC(a),
		LOAD_INT(b), LOAD_FRAC(b),
		LOAD_INT(c), LOAD_FRAC(c),
		nr_running(), nr_threads, current->nsproxy->pid_ns->last_pid);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

static int uptime_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	struct timespec uptime;
	struct timespec idle;
	int len;
	cputime_t idletime = cputime_add(init_task.utime, init_task.stime);

	do_posix_clock_monotonic_gettime(&uptime);
	cputime_to_timespec(idletime, &idle);
	len = sprintf(page,"%lu.%02lu %lu.%02lu\n",
			(unsigned long) uptime.tv_sec,
			(uptime.tv_nsec / (NSEC_PER_SEC / 100)),
			(unsigned long) idle.tv_sec,
			(idle.tv_nsec / (NSEC_PER_SEC / 100)));

	return proc_calc_metrics(page, start, off, count, eof, len);
}

static int meminfo_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	struct sysinfo i;
	int len;
	unsigned long committed;
	unsigned long allowed;
	struct vmalloc_info vmi;
	long cached;

/*
 * display in kilobytes.
 */
#define K(x) ((x) << (PAGE_SHIFT - 10))
	si_meminfo(&i);
	si_swapinfo(&i);
	committed = atomic_read(&vm_committed_space);
	allowed = ((totalram_pages - hugetlb_total_pages())
		* sysctl_overcommit_ratio / 100) + total_swap_pages;

	cached = global_page_state(NR_FILE_PAGES) -
			total_swapcache_pages - i.bufferram;
	if (cached < 0)
		cached = 0;

	get_vmalloc_info(&vmi);

	/*
	 * Tagged format, for easy grepping and expansion.
	 */
	len = sprintf(page,
		"MemTotal:     %8lu kB\n"
		"MemFree:      %8lu kB\n"
		"Buffers:      %8lu kB\n"
		"Cached:       %8lu kB\n"
		"SwapCached:   %8lu kB\n"
		"Active:       %8lu kB\n"
		"Inactive:     %8lu kB\n"
#ifdef CONFIG_HIGHMEM
		"HighTotal:    %8lu kB\n"
		"HighFree:     %8lu kB\n"
		"LowTotal:     %8lu kB\n"
		"LowFree:      %8lu kB\n"
#endif
		"SwapTotal:    %8lu kB\n"
		"SwapFree:     %8lu kB\n"
		"Dirty:        %8lu kB\n"
		"Writeback:    %8lu kB\n"
		"AnonPages:    %8lu kB\n"
		"Mapped:       %8lu kB\n"
		"Slab:         %8lu kB\n"
		"SReclaimable: %8lu kB\n"
		"SUnreclaim:   %8lu kB\n"
		"PageTables:   %8lu kB\n"
		"NFS_Unstable: %8lu kB\n"
		"Bounce:       %8lu kB\n"
		"CommitLimit:  %8lu kB\n"
		"Committed_AS: %8lu kB\n"
		"VmallocTotal: %8lu kB\n"
		"VmallocUsed:  %8lu kB\n"
		"VmallocChunk: %8lu kB\n",
		K(i.totalram),
		K(i.freeram),
		K(i.bufferram),
		K(cached),
		K(total_swapcache_pages),
		K(global_page_state(NR_ACTIVE)),
		K(global_page_state(NR_INACTIVE)),
#ifdef CONFIG_HIGHMEM
		K(i.totalhigh),
		K(i.freehigh),
		K(i.totalram-i.totalhigh),
		K(i.freeram-i.freehigh),
#endif
		K(i.totalswap),
		K(i.freeswap),
		K(global_page_state(NR_FILE_DIRTY)),
		K(global_page_state(NR_WRITEBACK)),
		K(global_page_state(NR_ANON_PAGES)),
		K(global_page_state(NR_FILE_MAPPED)),
		K(global_page_state(NR_SLAB_RECLAIMABLE) +
				global_page_state(NR_SLAB_UNRECLAIMABLE)),
		K(global_page_state(NR_SLAB_RECLAIMABLE)),
		K(global_page_state(NR_SLAB_UNRECLAIMABLE)),
		K(global_page_state(NR_PAGETABLE)),
		K(global_page_state(NR_UNSTABLE_NFS)),
		K(global_page_state(NR_BOUNCE)),
		K(allowed),
		K(committed),
		(unsigned long)VMALLOC_TOTAL >> 10,
		vmi.used >> 10,
		vmi.largest_chunk >> 10
		);

		len += hugetlb_report_meminfo(page + len);

	return proc_calc_metrics(page, start, off, count, eof, len);
#undef K
}

#if defined(CONFIG_PROC_PPGSTAT) || defined(CONFIG_PROC_MEMMAP)
static unsigned long calc_min_pfn(void)
{
	unsigned long result = NODE_DATA(0)->node_start_pfn;
	int node;
	for_each_online_node(node) {
		unsigned long start_pfn =NODE_DATA(node)->node_start_pfn;
		if (start_pfn < result)
			result = start_pfn;
	}
	return result;
}

static unsigned long calc_max_pfn(void)
{
	unsigned long result = NODE_DATA(0)->node_start_pfn
			     + NODE_DATA(0)->node_spanned_pages;
	int node;
	for_each_online_node(node) {
		pg_data_t *ndat = NODE_DATA(node);
		unsigned long end_pfn = ndat->node_start_pfn
				      + ndat->node_spanned_pages;
		if (end_pfn > result)
			result = end_pfn;
	}
	return result;
}
#endif

#ifdef CONFIG_PROC_PPGSTAT
/*
 * This is a system wide /proc entry.
 * This entry shows the status of each page of a process in detail,
 * such as  active/inactive, shared/non-shared, slab/normal,
 * reserved/not.
 *
 */
static unsigned int ppgstat_total_memsize;

#define PPGSTAT_PAGES_PER_LINE	(16)

static unsigned int base64(unsigned int x) {

	x &= 63;
	if (x < 26) return x + 'A';
	if (x < 52) return x + 'a' - 26;
	if (x < 62) return x + '0' - 52;
	if (x < 63) return '+';
	return '/';
}

static void *ppgstat_start(struct seq_file *m, loff_t *pos)
{
	unsigned int *bitmap = m->private;
	int n;

	n = *pos * PPGSTAT_PAGES_PER_LINE;
	if (n < ppgstat_total_memsize) {
		return bitmap + n;
	}

	return 0;
}

static void ppgstat_stop(struct seq_file *m, void *p)
{
}

static void *ppgstat_next(struct seq_file *m, void *p, loff_t *pos)
{
	unsigned int *bitmap = m->private;
	int n;

	++*pos;
	n = *pos * PPGSTAT_PAGES_PER_LINE;
	if (n < ppgstat_total_memsize) {
		return bitmap + n;
	}

	return 0;
}

static int ppgstat_show(struct seq_file *m, void *p)
{
	int i, d;
	unsigned int *bitmap = p;

	for ( i = 0; i < PPGSTAT_PAGES_PER_LINE; i++) {
		d = *bitmap++;
		seq_putc(m, base64(d));
		seq_putc(m, base64(d >>6));
		seq_putc(m, base64(d >>12));
		seq_putc(m, base64(d >>18));
	}
	seq_putc(m, '\n');

	return 0;
}

struct seq_operations ppgstatinfo_op = {
        .start  = ppgstat_start,
        .next   = ppgstat_next,
        .stop   = ppgstat_stop,
        .show   = ppgstat_show,
};

static int ppgstat_file_open(struct inode *inode, struct file *file)
{
	int ret, index;
	struct page *p;
	unsigned int *bitmap = NULL;
	struct seq_file *m;
	int pfn, d;
	unsigned long max_pfn, min_pfn;

	ret = seq_open(file, &ppgstatinfo_op);

	if(ret) return ret;

	ppgstat_total_memsize = 0;

	min_pfn = calc_min_pfn();
	max_pfn = calc_max_pfn();
	ppgstat_total_memsize = max_pfn - min_pfn;

	bitmap = vmalloc(ppgstat_total_memsize * sizeof(int));
	if (!bitmap) {
		return -ENOMEM;
	}
	for ( index = 0; index < ppgstat_total_memsize; index++ ) {
		bitmap[index] = 0;
	}

	index = 0;
	for (pfn = min_pfn; pfn < max_pfn; pfn++) {
		if (!pfn_valid(pfn)) {
			d = 0;
			d |= (1 << 20);
			bitmap[index++] = d;
			continue;
		}
		else if (!pfn_valid(pfn))
			continue;

		p=pfn_to_page(pfn);
		d = 0;
		if (PageReserved(p))    d |= (1 << 0);
		if (page_count(p) == 0) d |= (1 << 1);
		if (PageSlab(p))        d |= (1 << 2);
		if (PageLRU(p))         d |= (1 << 3);
		if (PageReferenced(p))  d |= (1 << 4);
		if (PageActive(p))      d |= (1 << 5);
		if (page_count(p) > 1)  d |= (1 << 6);
		if (PagePrivate(p))     d |= (1 << 7);
		if (PageMappedToDisk(p)) d |= (1 << 8);
		if (PageChecked(p))     d |= (1 << 9);
		if (PageDirty(p))       d |= (1 << 10);
		if (PageUptodate(p))    d |= (1 << 11);
		if (PageLocked(p))      d |= (1 << 12);
		if (PageWriteback(p))   d |= (1 << 13);
		if (PageCompound(p))    d |= (1 << 14);
		if (PageBuddy(p))       d |= (1 << 15);
		if (PageSwapCache(p))   d |= (1 << 16);
		if (PageError(p))       d |= (1 << 17);
		if (PageReclaim(p))     d |= (1 << 18);
		if (PageHighMem(p))     d |= (1 << 19);

		bitmap[index++] = d;
	}
	m = file->private_data;
	m->private = bitmap;
	return 0;
}

static int ppgstat_file_release (struct inode *inode, struct file *file)
{
	struct seq_file *m = file->private_data;

	if (m->private) {
		vfree(m->private);
	}

	return seq_release(inode, file);
}

static struct file_operations proc_ppgstat_file_operations = {
	.open           = ppgstat_file_open,
	.read           = seq_read,
	.llseek         = seq_lseek,
	.release        = ppgstat_file_release,
};
#endif

#ifdef CONFIG_PROC_MEMMAP

/*
 * This is a system wide /proc entry.
 * This entry provides a global view of the free and allocated physical
 * memory pages. It provides data in a bit map format, one bit represents
 * a single page frame. A one means the page is allocated, a zero means
 * the page is free.
 */
#define pfn_to_pte(pfn) (__pte(pfn << PAGE_SHIFT))
#define MEMMAP_OUTPUT_PER_LINE	8
#define BITS_PER_INT 32
#define ASCII_VALUE_DOT 0x2e
#define MEMORY_LARGE_THAN_1MB (1024*1024/PAGE_SIZE)+1

static int lowwater = 0;
static int highwater = 0;
static int mem_acc_debug = 0;
static int memmap_total_memsize ;

static void *memmap_start(struct seq_file *m, loff_t *pos)
{
	unsigned int *bitmap = m->private;
	int n;

	if(!*pos)
		return SEQ_START_TOKEN;

	n = *pos * MEMMAP_OUTPUT_PER_LINE;
	if (n < memmap_total_memsize) {
		return bitmap + n;
	}
	return 0;
}

static void memmap_stop(struct seq_file *m, void *p)
{
}

static void *memmap_next(struct seq_file *m, void *p, loff_t *pos)
{
	unsigned int *bitmap = m->private;
	int n;

	n = *pos * MEMMAP_OUTPUT_PER_LINE;
	++ *pos;

	if (n < memmap_total_memsize) {
		return bitmap + n;
	}

	return 0;
}

static int memmap_show(struct seq_file *m, void *p)
{
	int i, d, index;
	unsigned int *bitmap = p;

	if ( p == SEQ_START_TOKEN ) {
		seq_printf(m, "High water used pages =%d\t Low water used pages %d\n", highwater, lowwater);
	} else {
		for (i = 0; i < MEMMAP_OUTPUT_PER_LINE; i++) {
			d = *bitmap++;
			if(d ==  ASCII_VALUE_DOT){
				for(index = 0 ; index < MEMMAP_OUTPUT_PER_LINE ; index++)
					seq_printf(m, "%c", d);
			}
			else
				seq_printf(m, "%08x", d);
		}
		seq_printf(m, "\n");
	}

	return 0;
}

struct seq_operations memmapinfo_op = {
	.start	= memmap_start,
	.next	= memmap_next,
	.stop	= memmap_stop,
	.show	= memmap_show,
};

/*
 * We walk all the page structs and translate them to pfns
 * and then to ptes.  We get the free pages by the page count, count == 0.
 *
 * We have an array of unsigned ints of which each bit in an int
 * represents one page in memory.
 *
 * The pfn is the index into the array, pfn 0 == index 0 (index = pfn / 32).
 * We figure out which bit of the unsigned int to set by the following:
 * bit = pfn % 32, we then set that bit to zero to denote it's free via:
 * array[index] &= ~(1<<bit);   then we pump the data to /proc
 */

static int memmap_file_open(struct inode *inode, struct file *file)
{
	struct page *p;
	unsigned int *bitmap = NULL;
	int bit, pfn;
	int mark, ret;
	int free, total;
	struct seq_file *m;
	static int index;
	pte_t pte;
	unsigned long  mem_hole_start_addr = 0;
	unsigned long max_pfn, min_pfn;
	int mem_hole_size;
	int pfn_count;

	ret = seq_open(file, &memmapinfo_op);
	if(ret)
		return ret;

	memmap_total_memsize = 0;
	min_pfn = calc_min_pfn();
	max_pfn = calc_max_pfn();
	memmap_total_memsize = (max_pfn - min_pfn) / BITS_PER_INT;
	bitmap = vmalloc(memmap_total_memsize * sizeof(int));

	if (!bitmap) {
		return -ENOMEM;
	}

	/*
	 * one means used, zero means free, set them all to 1,
	 * clear them as we find them free.
	 */
	for (index = 0; index < memmap_total_memsize; index++) {
		bitmap[index] = 0xffffffff;
	}

	total = free = 0;
	index = 0;
	mem_hole_size = 0;
	pfn_count  = 0;

	for (pfn = min_pfn; pfn < max_pfn; pfn++)
	{
		if (!pfn_valid(pfn))
		{
			if(mem_hole_size == 0)
				mem_hole_start_addr = pfn << PAGE_SHIFT;
			mem_hole_size ++;
			if(mem_hole_size == MEMORY_LARGE_THAN_1MB)
				printk("Starting page address of the Memory hole : 0x%lx\n",mem_hole_start_addr);
			pfn_count++;
			index = pfn_count;
			index /= BITS_PER_INT;
			bitmap[index] = ASCII_VALUE_DOT;
			continue;
		}
		else if (!pfn_valid(pfn))
		continue;

		p=pfn_to_page(pfn);
		total++;
		if (!PageReserved(p) && page_count(p) == 0
			&& pfn_valid(pfn)) {
			index = pfn_count;
			pfn_count++;
			bit = index % BITS_PER_INT;
			index /= BITS_PER_INT;
			free++;
			bitmap[index] &= ~(1 << bit);
			mem_hole_size = 0;

		}
		else if (pfn_valid(pfn))
			pfn_count++;

		if (mem_acc_debug) {
			pte = pfn_to_pte(pfn);
			index = pfn_count;
			pfn_count++;
			bit = index % BITS_PER_INT;
			index /= BITS_PER_INT;
			printk("i %d bit %d pfn %d pa 0x%lx va 0x%lx "
				"c %d\n", index, bit, pfn, pte_val(pte),
				(unsigned long)__va(pte_val(pte)),
				page_count(p));
		}
	}
	m = file->private_data;
	m->private = bitmap;
	mark = total - free;

	if (mark < lowwater) {
		lowwater = mark;
	}
	if (mark > highwater) {
		highwater = mark;
	}

	return 0;
}

static int memmap_file_release (struct inode *inode, struct file *file)
{
	struct seq_file *m = file->private_data;

	if (m->private) {
		vfree(m->private);
	}

	return seq_release(inode, file);
}

static struct file_operations proc_memmap_file_operations = {
	.open		= memmap_file_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= memmap_file_release,
};

#endif

extern struct seq_operations fragmentation_op;
static int fragmentation_open(struct inode *inode, struct file *file)
{
	(void)inode;
	return seq_open(file, &fragmentation_op);
}

static const struct file_operations fragmentation_file_operations = {
	.open		= fragmentation_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
};

extern struct seq_operations pagetypeinfo_op;
static int pagetypeinfo_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &pagetypeinfo_op);
}

static const struct file_operations pagetypeinfo_file_ops = {
	.open		= pagetypeinfo_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
};

extern struct seq_operations zoneinfo_op;
static int zoneinfo_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &zoneinfo_op);
}

static const struct file_operations proc_zoneinfo_file_operations = {
	.open		= zoneinfo_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
};

static int version_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len;

	len = snprintf(page, PAGE_SIZE, linux_proc_banner,
		utsname()->sysname,
		utsname()->release,
		utsname()->version);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

extern struct seq_operations cpuinfo_op;
static int cpuinfo_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &cpuinfo_op);
}

static const struct file_operations proc_cpuinfo_operations = {
	.open		= cpuinfo_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
};

static int devinfo_show(struct seq_file *f, void *v)
{
	int i = *(loff_t *) v;

	if (i < CHRDEV_MAJOR_HASH_SIZE) {
		if (i == 0)
			seq_printf(f, "Character devices:\n");
		chrdev_show(f, i);
	}
#ifdef CONFIG_BLOCK
	else {
		i -= CHRDEV_MAJOR_HASH_SIZE;
		if (i == 0)
			seq_printf(f, "\nBlock devices:\n");
		blkdev_show(f, i);
	}
#endif
	return 0;
}

static void *devinfo_start(struct seq_file *f, loff_t *pos)
{
	if (*pos < (BLKDEV_MAJOR_HASH_SIZE + CHRDEV_MAJOR_HASH_SIZE))
		return pos;
	return NULL;
}

static void *devinfo_next(struct seq_file *f, void *v, loff_t *pos)
{
	(*pos)++;
	if (*pos >= (BLKDEV_MAJOR_HASH_SIZE + CHRDEV_MAJOR_HASH_SIZE))
		return NULL;
	return pos;
}

static void devinfo_stop(struct seq_file *f, void *v)
{
	/* Nothing to do */
}

static struct seq_operations devinfo_ops = {
	.start = devinfo_start,
	.next  = devinfo_next,
	.stop  = devinfo_stop,
	.show  = devinfo_show
};

static int devinfo_open(struct inode *inode, struct file *filp)
{
	return seq_open(filp, &devinfo_ops);
}

static const struct file_operations proc_devinfo_operations = {
	.open		= devinfo_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
};

extern struct seq_operations vmstat_op;
static int vmstat_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &vmstat_op);
}
static const struct file_operations proc_vmstat_file_operations = {
	.open		= vmstat_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
};

#ifdef CONFIG_PROC_HARDWARE
static int hardware_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_hardware_list(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}
#endif

#ifdef CONFIG_PROC_FREEMEM
extern struct seq_operations freemem_op;
static int freemem_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &freemem_op);
}
static struct file_operations proc_freemem_operations = {
	.open           = freemem_open,
	.read           = seq_read,
	.llseek         = seq_lseek,
	.release        = seq_release,
};
#endif

#ifdef CONFIG_STRAM_PROC
static int stram_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_stram_list(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}
#endif

#ifdef CONFIG_BLOCK
extern struct seq_operations partitions_op;
static int partitions_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &partitions_op);
}
static const struct file_operations proc_partitions_operations = {
	.open		= partitions_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
};

extern struct seq_operations diskstats_op;
static int diskstats_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &diskstats_op);
}
static const struct file_operations proc_diskstats_operations = {
	.open		= diskstats_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
};
#endif

#ifdef CONFIG_MODULES
extern struct seq_operations modules_op;
static int modules_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &modules_op);
}
static const struct file_operations proc_modules_operations = {
	.open		= modules_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
};
#endif

#ifdef CONFIG_SLAB
static int slabinfo_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &slabinfo_op);
}
static const struct file_operations proc_slabinfo_operations = {
	.open		= slabinfo_open,
	.read		= seq_read,
	.write		= slabinfo_write,
	.llseek		= seq_lseek,
	.release	= seq_release,
};

#ifdef CONFIG_DEBUG_SLAB_LEAK
extern struct seq_operations slabstats_op;
static int slabstats_open(struct inode *inode, struct file *file)
{
	unsigned long *n = kzalloc(PAGE_SIZE, GFP_KERNEL);
	int ret = -ENOMEM;
	if (n) {
		ret = seq_open(file, &slabstats_op);
		if (!ret) {
			struct seq_file *m = file->private_data;
			*n = PAGE_SIZE / (2 * sizeof(unsigned long));
			m->private = n;
			n = NULL;
		}
		kfree(n);
	}
	return ret;
}

static const struct file_operations proc_slabstats_operations = {
	.open		= slabstats_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release_private,
};
#endif
#endif

#ifdef CONFIG_DEBUG_SLAB_ACCOUNT

extern struct seq_operations slab_account_op;

static int slab_account_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &slab_account_op);
}

static struct file_operations proc_slab_account_operations = {
	.open		= slab_account_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
};

#endif

#ifdef CONFIG_SNSC_PAGES_ACCOUNTING

extern struct seq_operations pages_account_op;

static int pages_account_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &pages_account_op);
}

extern void pages_account_clear(void);

static ssize_t pages_account_write(struct file *f, const char __user *buf, size_t size, loff_t *pos)
{
	pages_account_clear();
	return size;
}

static struct file_operations proc_pages_account_operations = {
	.open		= pages_account_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
	.write		= pages_account_write,
};

#endif

static int show_stat(struct seq_file *p, void *v)
{
	int i;
	unsigned long jif;
	cputime64_t user, nice, system, idle, iowait, irq, softirq, steal;
	u64 sum = 0;

	user = nice = system = idle = iowait =
		irq = softirq = steal = cputime64_zero;
	jif = - wall_to_monotonic.tv_sec;
	if (wall_to_monotonic.tv_nsec)
		--jif;

	for_each_possible_cpu(i) {
		int j;

		user = cputime64_add(user, kstat_cpu(i).cpustat.user);
		nice = cputime64_add(nice, kstat_cpu(i).cpustat.nice);
		system = cputime64_add(system, kstat_cpu(i).cpustat.system);
		idle = cputime64_add(idle, kstat_cpu(i).cpustat.idle);
		iowait = cputime64_add(iowait, kstat_cpu(i).cpustat.iowait);
		irq = cputime64_add(irq, kstat_cpu(i).cpustat.irq);
		softirq = cputime64_add(softirq, kstat_cpu(i).cpustat.softirq);
		steal = cputime64_add(steal, kstat_cpu(i).cpustat.steal);
		for (j = 0 ; j < NR_IRQS ; j++)
			sum += kstat_cpu(i).irqs[j];
	}

	seq_printf(p, "cpu  %llu %llu %llu %llu %llu %llu %llu %llu\n",
		(unsigned long long)cputime64_to_clock_t(user),
		(unsigned long long)cputime64_to_clock_t(nice),
		(unsigned long long)cputime64_to_clock_t(system),
		(unsigned long long)cputime64_to_clock_t(idle),
		(unsigned long long)cputime64_to_clock_t(iowait),
		(unsigned long long)cputime64_to_clock_t(irq),
		(unsigned long long)cputime64_to_clock_t(softirq),
		(unsigned long long)cputime64_to_clock_t(steal));
	for_each_online_cpu(i) {

		/* Copy values here to work around gcc-2.95.3, gcc-2.96 */
		user = kstat_cpu(i).cpustat.user;
		nice = kstat_cpu(i).cpustat.nice;
		system = kstat_cpu(i).cpustat.system;
		idle = kstat_cpu(i).cpustat.idle;
		iowait = kstat_cpu(i).cpustat.iowait;
		irq = kstat_cpu(i).cpustat.irq;
		softirq = kstat_cpu(i).cpustat.softirq;
		steal = kstat_cpu(i).cpustat.steal;
		seq_printf(p, "cpu%d %llu %llu %llu %llu %llu %llu %llu %llu\n",
			i,
			(unsigned long long)cputime64_to_clock_t(user),
			(unsigned long long)cputime64_to_clock_t(nice),
			(unsigned long long)cputime64_to_clock_t(system),
			(unsigned long long)cputime64_to_clock_t(idle),
			(unsigned long long)cputime64_to_clock_t(iowait),
			(unsigned long long)cputime64_to_clock_t(irq),
			(unsigned long long)cputime64_to_clock_t(softirq),
			(unsigned long long)cputime64_to_clock_t(steal));
	}
	seq_printf(p, "intr %llu", (unsigned long long)sum);

#if !defined(CONFIG_PPC64) && !defined(CONFIG_ALPHA) && !defined(CONFIG_IA64)
	for (i = 0; i < NR_IRQS; i++)
		seq_printf(p, " %u", kstat_irqs(i));
#endif

	seq_printf(p,
		"\nctxt %llu\n"
		"btime %lu\n"
		"processes %lu\n"
		"procs_running %lu\n"
		"procs_blocked %lu\n",
		nr_context_switches(),
		(unsigned long)jif,
		total_forks,
		nr_running(),
		nr_iowait());

	return 0;
}

static int stat_open(struct inode *inode, struct file *file)
{
	unsigned size = 4096 * (1 + num_possible_cpus() / 32);
	char *buf;
	struct seq_file *m;
	int res;

	/* don't ask for more than the kmalloc() max size, currently 128 KB */
	if (size > 128 * 1024)
		size = 128 * 1024;
	buf = kmalloc(size, GFP_KERNEL);
	if (!buf)
		return -ENOMEM;

	res = single_open(file, show_stat, NULL);
	if (!res) {
		m = file->private_data;
		m->buf = buf;
		m->size = size;
	} else
		kfree(buf);
	return res;
}
static const struct file_operations proc_stat_operations = {
	.open		= stat_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= single_release,
};

/*
 * /proc/interrupts
 */
static void *int_seq_start(struct seq_file *f, loff_t *pos)
{
	return (*pos <= NR_IRQS) ? pos : NULL;
}

static void *int_seq_next(struct seq_file *f, void *v, loff_t *pos)
{
	(*pos)++;
	if (*pos > NR_IRQS)
		return NULL;
	return pos;
}

static void int_seq_stop(struct seq_file *f, void *v)
{
	/* Nothing to do */
}


extern int show_interrupts(struct seq_file *f, void *v); /* In arch code */
static struct seq_operations int_seq_ops = {
	.start = int_seq_start,
	.next  = int_seq_next,
	.stop  = int_seq_stop,
	.show  = show_interrupts
};

static int interrupts_open(struct inode *inode, struct file *filp)
{
	return seq_open(filp, &int_seq_ops);
}

static const struct file_operations proc_interrupts_operations = {
	.open		= interrupts_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
};

static int filesystems_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_filesystem_list(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

static int cmdline_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len;

	len = sprintf(page, "%s\n", saved_command_line);
	return proc_calc_metrics(page, start, off, count, eof, len);
}

#ifdef CONFIG_FILE_LOCKING
static int locks_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_locks_status(page, start, off, count);

	if (len < count)
		*eof = 1;
	return len;
}
#endif

static int execdomains_read_proc(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	int len = get_exec_domain_list(page);
	return proc_calc_metrics(page, start, off, count, eof, len);
}
#ifdef CONFIG_SNSC_USE_NODE_ORDER_DEBUG
extern int nodeorder_read_proc(char *page, char **start, off_t off,
			       int count, int *eof, void *data);
#endif

#ifdef CONFIG_KFT
extern struct seq_operations kft_data_op;
static int kft_data_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &kft_data_op);
}
static struct file_operations proc_kft_data_operations = {
	.open		= kft_data_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
};
#endif


#ifdef CONFIG_MAGIC_SYSRQ
/*
 * writing 'C' to /proc/sysrq-trigger is like sysrq-C
 */
static ssize_t write_sysrq_trigger(struct file *file, const char __user *buf,
				   size_t count, loff_t *ppos)
{
	if (count) {
		char c;

		if (get_user(c, buf))
			return -EFAULT;
		__handle_sysrq(c, NULL, 0);
	}
	return count;
}

static const struct file_operations proc_sysrq_trigger_operations = {
	.write		= write_sysrq_trigger,
};
#endif

struct proc_dir_entry *proc_root_kcore;

void create_seq_entry(char *name, mode_t mode, const struct file_operations *f)
{
	struct proc_dir_entry *entry;
	entry = create_proc_entry(name, mode, NULL);
	if (entry)
		entry->proc_fops = f;
}

void __init proc_misc_init(void)
{
	static struct {
		char *name;
		int (*read_proc)(char*,char**,off_t,int,int*,void*);
	} *p, simple_ones[] = {
		{"loadavg",     loadavg_read_proc},
		{"uptime",	uptime_read_proc},
		{"meminfo",	meminfo_read_proc},
		{"version",	version_read_proc},
#ifdef CONFIG_PROC_HARDWARE
		{"hardware",	hardware_read_proc},
#endif
#ifdef CONFIG_STRAM_PROC
		{"stram",	stram_read_proc},
#endif
		{"filesystems",	filesystems_read_proc},
		{"cmdline",	cmdline_read_proc},
#ifdef CONFIG_FILE_LOCKING
		{"locks",	locks_read_proc},
#endif
		{"execdomains",	execdomains_read_proc},
#ifdef CONFIG_SNSC_USE_NODE_ORDER_DEBUG
		{"nodeorder", nodeorder_read_proc},
#endif
		{NULL,}
	};
	for (p = simple_ones; p->name; p++)
		create_proc_read_entry(p->name, 0, NULL, p->read_proc, NULL);
#ifdef CONFIG_PROC_MEMMAP
	/*
	 * get the amount of free pages right here, as this
	 * should be the lowest amount of allocated pages
	 * the system ever sees.  After log in there
	 * should just be more allocated pages.
	 */
	lowwater = num_physpages - nr_free_pages();
#endif
	proc_symlink("mounts", NULL, "self/mounts");

	/* And now for trickier ones */
#ifdef CONFIG_PRINTK
	{
		struct proc_dir_entry *entry;
		entry = create_proc_entry("kmsg", S_IRUSR, &proc_root);
		if (entry)
			entry->proc_fops = &proc_kmsg_operations;
	}
#endif
	create_seq_entry("devices", 0, &proc_devinfo_operations);
	create_seq_entry("cpuinfo", 0, &proc_cpuinfo_operations);
#ifdef CONFIG_BLOCK
	create_seq_entry("partitions", 0, &proc_partitions_operations);
#endif
	create_seq_entry("stat", 0, &proc_stat_operations);
	create_seq_entry("interrupts", 0, &proc_interrupts_operations);
#ifdef CONFIG_SLAB
	create_seq_entry("slabinfo",S_IWUSR|S_IRUGO,&proc_slabinfo_operations);
#ifdef CONFIG_DEBUG_SLAB_LEAK
	create_seq_entry("slab_allocators", 0 ,&proc_slabstats_operations);
#endif
#endif
#ifdef CONFIG_DEBUG_SLAB_ACCOUNT
	create_seq_entry("slab_account",S_IRUGO,&proc_slab_account_operations);
#endif
#ifdef CONFIG_SNSC_PAGES_ACCOUNTING
	create_seq_entry("pages",S_IRUGO,&proc_pages_account_operations);
#endif
	create_seq_entry("buddyinfo",S_IRUGO, &fragmentation_file_operations);
	create_seq_entry("pagetypeinfo", S_IRUGO, &pagetypeinfo_file_ops);
	create_seq_entry("vmstat",S_IRUGO, &proc_vmstat_file_operations);
	create_seq_entry("zoneinfo",S_IRUGO, &proc_zoneinfo_file_operations);
#ifdef CONFIG_BLOCK
	create_seq_entry("diskstats", 0, &proc_diskstats_operations);
#endif

#ifdef CONFIG_PROC_FREEMEM
	create_seq_entry("freemem",S_IWUSR|S_IRUGO,&proc_freemem_operations);
#endif

#ifdef CONFIG_PROC_MEMMAP
	create_seq_entry("memmap", 0, &proc_memmap_file_operations);
#endif

#ifdef CONFIG_MODULES
	create_seq_entry("modules", 0, &proc_modules_operations);
#endif
#ifdef CONFIG_SCHEDSTATS
	create_seq_entry("schedstat", 0, &proc_schedstat_operations);
#endif
#ifdef CONFIG_KFT
	create_seq_entry("kft_data", 0, &proc_kft_data_operations);
#endif
#ifdef CONFIG_PROC_KCORE
	proc_root_kcore = create_proc_entry("kcore", S_IRUSR, NULL);
	if (proc_root_kcore) {
		proc_root_kcore->proc_fops = &proc_kcore_operations;
		proc_root_kcore->size =
				(size_t)high_memory - PAGE_OFFSET + PAGE_SIZE;
	}
#endif
#ifdef CONFIG_PROC_VMCORE
	proc_vmcore = create_proc_entry("vmcore", S_IRUSR, NULL);
	if (proc_vmcore)
		proc_vmcore->proc_fops = &proc_vmcore_operations;
#endif

#ifdef CONFIG_MAGIC_SYSRQ
	{
		struct proc_dir_entry *entry;
		entry = create_proc_entry("sysrq-trigger", S_IWUSR, NULL);
		if (entry)
			entry->proc_fops = &proc_sysrq_trigger_operations;
	}
#endif
#ifdef CONFIG_PROC_PPGSTAT
	 create_seq_entry("ppgstat", 0, &proc_ppgstat_file_operations);
#endif
#ifdef CONFIG_SNSC
        /* NSC Linux version */
        nlver_init();
#endif
}
