/* 2007-06-22: File changed by Sony Corporation */
/*
 * kernel/power/main.c - PM subsystem core functionality.
 *
 * Copyright (c) 2003 Patrick Mochel
 * Copyright (c) 2003 Open Source Development Lab
 *
 * Copyright 2005,2006,2007 Sony Corporation
 * 
 * This file is released under the GPLv2
 *
 */

#include <linux/module.h>
#include <linux/suspend.h>
#include <linux/kobject.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/console.h>
#include <linux/cpu.h>
#include <linux/resume-trace.h>
#include <linux/freezer.h>
#include <linux/vmstat.h>
#include <linux/device.h>
#include <linux/reboot.h>

#ifdef	CONFIG_SNSC_SAFE_SUSPEND
extern int suspend_remount(void);
extern int resume_remount(void);
#endif

#ifdef CONFIG_SNSC_DEFERRED_RESUME
#include <linux/kthread.h>
#endif

#include "power.h"

/*This is just an arbitrary number */
#define FREE_PAGE_NUMBER (100)

DEFINE_MUTEX(pm_mutex);

struct pm_ops *pm_ops;

/**
 *	pm_set_ops - Set the global power method table. 
 *	@ops:	Pointer to ops structure.
 */

void pm_set_ops(struct pm_ops * ops)
{
	mutex_lock(&pm_mutex);
	pm_ops = ops;
	mutex_unlock(&pm_mutex);
}

/**
 * pm_valid_only_mem - generic memory-only valid callback
 *
 * pm_ops drivers that implement mem suspend only and only need
 * to check for that in their .valid callback can use this instead
 * of rolling their own .valid callback.
 */
int pm_valid_only_mem(suspend_state_t state)
{
	return state == PM_SUSPEND_MEM;
}


static inline void pm_finish(suspend_state_t state)
{
	if (pm_ops->finish)
		pm_ops->finish(state);
}

/**
 *	suspend_prepare - Do prep work before entering low-power state.
 *	@state:		State we're entering.
 *
 *	This is common code that is called for each state that we're 
 *	entering. Allocate a console, stop all processes, then make sure
 *	the platform can enter the requested state.
 */

static int suspend_prepare(suspend_state_t state)
{
	int error;
	unsigned int free_pages;

	if (!pm_ops || !pm_ops->enter)
		return -EPERM;

	pm_prepare_console();

	if (freeze_processes()) {
		error = -EAGAIN;
		goto Thaw;
	}

	if ((free_pages = global_page_state(NR_FREE_PAGES))
			< FREE_PAGE_NUMBER) {
		unsigned int ret;
		pr_debug("PM: free some memory\n");
		ret = shrink_all_memory(FREE_PAGE_NUMBER - free_pages);
		free_pages += ret;
		if (free_pages < FREE_PAGE_NUMBER) {
#if 0			/* EJ: 
			 *  In case of the embedded system, this condition often makes true,
			 *  when the return value of shrink_all_memory() is 0.
			 *  We didn't know the reason of FREE_PAGE_NUMBER == 100, so we commented out like this.
			 */
			error = -ENOMEM;
			printk(KERN_ERR "PM: No enough memory\n");
			goto Thaw;
#else
			printk(KERN_WARNING "PM: free_pages = %u, shrink_all_memory() = %u\n", free_pages, ret);
			show_mem();
#endif
		}
	}

#ifdef	CONFIG_SNSC_SAFE_SUSPEND
	(void) suspend_remount();
#endif
	if (pm_ops->set_target) {
		error = pm_ops->set_target(state);
		if (error)
			goto Thaw;
	}
	suspend_console();
	error = device_suspend(PMSG_SUSPEND);
	if (error) {
		printk(KERN_ERR "Some devices failed to suspend\n");
		goto Resume_console;
	}
	if (pm_ops->prepare) {
		if ((error = pm_ops->prepare(state)))
			goto Resume_devices;
	}

	error = disable_nonboot_cpus();
	if (!error)
		return 0;

	enable_nonboot_cpus();
	pm_finish(state);
 Resume_devices:
	device_resume();
 Resume_console:
	resume_console();
 Thaw:
#ifdef	CONFIG_SNSC_SAFE_SUSPEND
	(void) resume_remount();
#endif
	thaw_processes();
	pm_restore_console();
	return error;
}

/* default implementation */
void __attribute__ ((weak)) arch_suspend_disable_irqs(void)
{
	local_irq_disable();
}

/* default implementation */
void __attribute__ ((weak)) arch_suspend_enable_irqs(void)
{
	local_irq_enable();
}

int suspend_enter(suspend_state_t state)
{
	int error = 0;

	arch_suspend_disable_irqs();
	BUG_ON(!irqs_disabled());

	if ((error = device_power_down(PMSG_SUSPEND))) {
		printk(KERN_ERR "Some devices failed to power down\n");
		goto Done;
	}
	error = pm_ops->enter(state);
	device_power_up();
 Done:
	arch_suspend_enable_irqs();
	BUG_ON(irqs_disabled());
	return error;
}

#ifdef CONFIG_SNSC_DEFERRED_RESUME

#include <linux/completion.h>

static DECLARE_COMPLETION(resume_thread_wakeup);

static struct task_struct *resume_thread_tsk = NULL;
static int enable_deferred_resume = 0;
static u32 resume_state;

#endif /* CONFIG_SNSC_DEFERRED_RESUME */

/**
 *	suspend_finish - Do final work before exiting suspend sequence.
 *	@state:		State we're coming out of.
 *
 *	Call platform code to clean up, restart processes, and free the 
 *	console that we've allocated. This is not called for suspend-to-disk.
 */

static void suspend_finish(suspend_state_t state)
{
	enable_nonboot_cpus();
	pm_finish(state);
	device_resume();
	resume_console();
#ifdef	CONFIG_SNSC_SAFE_SUSPEND
	(void) resume_remount();
#endif
	thaw_processes();
	pm_restore_console();
}


#if defined(CONFIG_FAIL_SAFE_RESUME)
#define FAIL_SAFE_DEFAULT FAIL_SAFE_RESUME
#elif defined(CONFIG_FAIL_SAFE_EXCEPTION)
#define FAIL_SAFE_DEFAULT FAIL_SAFE_EXCEPTION
#elif defined(CONFIG_FAIL_SAFE_REBOOT)
#define FAIL_SAFE_DEFAULT FAIL_SAFE_REBOOT
#else
#define FAIL_SAFE_DEFAULT FAIL_SAFE_HALT
#endif

enum {
	FAIL_SAFE_RESUME = 1,
	FAIL_SAFE_EXCEPTION,
	FAIL_SAFE_REBOOT,
	FAIL_SAFE_HALT,
	FAIL_SAFE_NUM
};
static unsigned fail_safe = FAIL_SAFE_DEFAULT;


static const char * const pm_states[PM_SUSPEND_MAX] = {
	[PM_SUSPEND_STANDBY]	= "standby",
	[PM_SUSPEND_MEM]	= "mem",
#ifdef PM_SUSPEND_DISK
	[PM_SUSPEND_DISK]       = "disk"
#endif
};

static inline int valid_state(suspend_state_t state)
{
	/* All states need lowlevel support and need to be valid
	 * to the lowlevel implementation, no valid callback
	 * implies that none are valid. */
	if (!pm_ops || !pm_ops->valid || !pm_ops->valid(state))
		return 0;
	return 1;
}

void __attribute__((weak)) pm_suspend_error(int mode)
{
	if (mode == PM_ERROR_REBOOT)
		machine_restart(NULL);
	else
		machine_halt();
}

/**
 *	enter_state - Do common work of entering low-power state.
 *	@state:		pm_state structure for state we're entering.
 *
 *	Make sure we're the only ones trying to enter a sleep state. Fail
 *	if someone has beat us to it, since we don't want anything weird to
 *	happen when we wake up.
 *	Then, do the setup for suspend, enter the state, and cleaup (after
 *	we've woken up).
 */

static int enter_state(suspend_state_t state)
{
	int error;

	if (!valid_state(state))
		return -ENODEV;
	if (!mutex_trylock(&pm_mutex))
		return -EBUSY;

	pr_debug("PM: Preparing system for %s sleep\n", pm_states[state]);
	if ((error = suspend_prepare(state)))
		goto Unlock;

	pr_debug("PM: Entering %s sleep\n", pm_states[state]);
	error = suspend_enter(state);

#ifdef CONFIG_SNSC_DEFERRED_RESUME

	if (enable_deferred_resume && resume_thread_tsk) {
		if (!error) {
			device_resume();

			/* wake up related threads */
			thaw_process_group(process_group(current));
			printk("woke up minimal threads\n");

			/* wake up resume thread */
			resume_state=state;
			smp_wmb();
			complete(&resume_thread_wakeup);

			pr_debug("PM: Return without finishing up.\n");
			printk("Return without finishing up\n");

			return 0;
		}
	}
#endif /* CONFIG_SNSC_DEFERRED_RESUME */

	pr_debug("PM: Finishing wakeup.\n");
	suspend_finish(state);
 Unlock:
	mutex_unlock(&pm_mutex);

	if(error){
		switch(fail_safe){
		case FAIL_SAFE_EXCEPTION:
			printk(KERN_ERR "Error occurred in suspending kernel.\n");
			BUG();
			break;
		case FAIL_SAFE_REBOOT:
			pm_suspend_error(PM_ERROR_REBOOT);
			break;
		case FAIL_SAFE_HALT:
			pm_suspend_error(PM_ERROR_HALT);
			break;
		default:
			break;
		}
	}

	fail_safe = FAIL_SAFE_DEFAULT;
	return error;
}


#ifdef CONFIG_SNSC_DEFERRED_RESUME
static int resume_thread(void *state_p)
{
	struct k_sigaction sa;
	sigset_t blocked;
	u32 st;

	/* resume-thread is not freezable */
	current->flags |= PF_NOFREEZE;

	/* set scheduling params. */
	set_user_nice(current, 10);

	/* Block and flush all signals */
	sigfillset(&blocked);
	sigprocmask(SIG_BLOCK, &blocked, NULL);
	flush_signals(current);

	/* SIG_IGN makes children autoreap: see do_notify_parent(). */
	sa.sa.sa_handler = SIG_IGN;
	sa.sa.sa_flags = 0;
	siginitset(&sa.sa.sa_mask, sigmask(SIGCHLD));
	do_sigaction(SIGCHLD, &sa, (struct k_sigaction *)0);

	pr_debug("init\n");

	set_current_state(TASK_INTERRUPTIBLE);
	while (!kthread_should_stop()) {

		__set_current_state(TASK_RUNNING);

		pr_debug("waiting completion\n");

		wait_for_completion (&resume_thread_wakeup); /* wait until resume */
		st = *(u32 *) state_p;

		printk("begin deferred finishing up\n");

		pr_debug("PM(deferred): Finishing up.\n");
		suspend_finish(st);

		printk("end deferred finishing up\n");

		mutex_unlock(&pm_mutex);

		set_current_state(TASK_INTERRUPTIBLE);
	}
	__set_current_state(TASK_RUNNING);

	pr_debug("end\n");

	resume_thread_tsk = NULL;
	enable_deferred_resume = 0;
	mb();
	return 0;
}

static int __init resume_thread_init(void)
{
	struct task_struct *p = NULL;
	p = kthread_create(resume_thread, &resume_state, "kresume");
	if (IS_ERR(p)) {
		resume_thread_tsk = NULL;
	}
	else {
		resume_thread_tsk = p;
		wake_up_process(resume_thread_tsk);
	}
	return 0;
}

__initcall(resume_thread_init);

#endif /* CONFIG_SNSC_DEFERRED_RESUME */

/**
 *	pm_suspend - Externally visible function for suspending system.
 *	@state:		Enumerated value of state to enter.
 *
 *	Determine whether or not value is within range, get state 
 *	structure, and enter (above).
 */

int pm_suspend(suspend_state_t state)
{
	if (state > PM_SUSPEND_ON && state <= PM_SUSPEND_MAX)
		return enter_state(state);
	return -EINVAL;
}

EXPORT_SYMBOL(pm_suspend);

decl_subsys(power,NULL,NULL);


/**
 *	state - control system power state.
 *
 *	show() returns what states are supported, which is hard-coded to
 *	'standby' (Power-On Suspend), 'mem' (Suspend-to-RAM), and
 *	'disk' (Suspend-to-Disk).
 *
 *	store() accepts one of those strings, translates it into the 
 *	proper enumerated value, and initiates a suspend transition.
 */

static ssize_t state_show(struct kset *kset, char *buf)
{
	int i;
	char * s = buf;

	for (i = 0; i < PM_SUSPEND_MAX; i++) {
		if (pm_states[i] && valid_state(i))
			s += sprintf(s,"%s ", pm_states[i]);
	}
#ifdef CONFIG_SOFTWARE_SUSPEND
	s += sprintf(s, "%s\n", "disk");
#else
	if (s != buf)
		/* convert the last space to a newline */
		*(s-1) = '\n';
#endif
	return (s - buf);
}

static ssize_t state_store(struct kset *kset, const char *buf, size_t n)
{
	suspend_state_t state = PM_SUSPEND_STANDBY;
	const char * const *s;
	char *p;
	int error;
	int len;

	p = memchr(buf, '\n', n);
	len = p ? p - buf : n;

#ifndef PM_SUSPEND_DISK
	/* First, check if we are requested to hibernate */
	if (len == 4 && !strncmp(buf, "disk", len)) {
		error = hibernate();
		return error ? error : n;
	}
#endif /* !PM_SUSPEND_DISK */

	for (s = &pm_states[state]; state < PM_SUSPEND_MAX; s++, state++) {
		if (*s && len == strlen(*s) && !strncmp(buf, *s, len))
			break;
	}
	if (state < PM_SUSPEND_MAX && *s)
		error = enter_state(state);
	else
		error = -EINVAL;
	return error ? error : n;
}

power_attr(state);

#ifdef CONFIG_SNSC_FAST_CLEAN_SHUTDOWN

#include <linux/reboot.h>
#include <linux/fs.h>

extern void tcp_shutdown_all(void);

void fast_clean_shutdown(void)
{
	(void)freeze_processes();
	(void)suspend_remount();
	tcp_shutdown_all();
	device_shutdown(); /* try to flush cache on device side */
	machine_power_off();

	panic("fast_shutdown: can't power off\n");

	/* NOTREACHED */

}

static ssize_t fast_clean_shutdown_show(struct kset * subsys, char * buf)
{
	return sprintf(buf,"enabled\n");
}

static ssize_t fast_clean_shutdown_store(struct kset * subsys, const char * buf,
				  size_t n)
{
	if (n > 2)
		return -EINVAL;
	if ((n == 2) && (buf[1] != '\n'))
		return -EINVAL;
	if (!(buf[0] == '1' || buf[0] == 'y'))
		return -EINVAL;

	fast_clean_shutdown();
	return n;
}

power_attr(fast_clean_shutdown);

#endif /* CONFIG_SNSC_FAST_CLEAN_SHUTDOWN */

#ifdef CONFIG_SNSC_DEFERRED_RESUME

static char * deferred_resume_states[] = {
	"disable",
	"enable",
	NULL
};


static ssize_t deferred_resume_show(struct kset * subsys, char * buf)
{
	char * s = buf;

	s += sprintf(s, deferred_resume_states[enable_deferred_resume]);
	s += sprintf(s, "\n");
	return (s - buf);
}

static ssize_t deferred_resume_store(struct kset * subsys, const char * buf, size_t n)
{
	char ** s;
	char *p;
	int len;
	int i;

	p = memchr(buf, '\n', n);
	len = p ? p - buf : n;

	i = 0;
	s = &deferred_resume_states[0];
	while (*s) {
		if (!strncmp(buf, *s, len))
			break;
		s++; i++;
	}

	if (!*s) {
		return -EINVAL;
	}

	if (resume_thread_tsk)
		enable_deferred_resume = i;
	return n;
}

power_attr(deferred_resume);

#endif /* CONFIG_SNSC_DEFERRED_RESUME */

static ssize_t on_failure_show(struct kset * subsys, char * buf)
{
	static char *fst[3] = {"comeback","reboot","shutdown"};
	char *src = buf;
	int i;
	
	for(i=FAIL_SAFE_RESUME;i<FAIL_SAFE_NUM;i++)
		src += sprintf(src, "%s%s\n",fail_safe==i?"*":" ", fst[i-FAIL_SAFE_RESUME]);
	
	return (src - buf);
}

static ssize_t on_failure_store(struct kset * subsys, const char * buf, size_t n)
{
	char *endp;
	unsigned value;
	
	value = simple_strtoul(buf, &endp, 0);
	if(*endp || value<FAIL_SAFE_RESUME || value>=FAIL_SAFE_NUM)
		return -EINVAL;

	fail_safe = value;

	return n;
}

power_attr(on_failure);

#ifdef CONFIG_PM_TRACE
int pm_trace_enabled;

static ssize_t pm_trace_show(struct kset *kset, char *buf)
{
	return sprintf(buf, "%d\n", pm_trace_enabled);
}

static ssize_t
pm_trace_store(struct kset *kset, const char *buf, size_t n)
{
	int val;

	if (sscanf(buf, "%d", &val) == 1) {
		pm_trace_enabled = !!val;
		return n;
	}
	return -EINVAL;
}

power_attr(pm_trace);

static struct attribute * g[] = {
	&state_attr.attr,
	&pm_trace_attr.attr,
#ifdef CONFIG_SNSC_FAST_CLEAN_SHUTDOWN
	&fast_clean_shutdown_attr.attr,
#endif
#ifdef CONFIG_SNSC_DEFERRED_RESUME
	&deferred_resume_attr.attr,
#endif /* CONFIG_SNSC_DEFERRED_RESUME */
 	&on_failure_attr.attr,
	NULL,
};
#else
static struct attribute * g[] = {
	&state_attr.attr,
#ifdef CONFIG_SNSC_FAST_CLEAN_SHUTDOWN
	&fast_clean_shutdown_attr.attr,
#endif
#ifdef CONFIG_SNSC_DEFERRED_RESUME
	&deferred_resume_attr.attr,
#endif /* CONFIG_SNSC_DEFERRED_RESUME */
 	&on_failure_attr.attr,
	NULL,
};
#endif /* CONFIG_PM_TRACE */

static struct attribute_group attr_group = {
	.attrs = g,
};


static int __init pm_init(void)
{
	int error = subsystem_register(&power_subsys);
	if (!error)
		error = sysfs_create_group(&power_subsys.kobj,&attr_group);
	return error;
}

core_initcall(pm_init);
