/*
 * Copyright (C) 2003 Bernardo Innocenti <bernie@develer.com>
 *
 * Based on former do_div() implementation from asm-parisc/div64.h:
 *	Copyright (C) 1999 Hewlett-Packard Co
 *	Copyright (C) 1999 David Mosberger-Tang <davidm@hpl.hp.com>
 *
 *
 * Generic C version of 64bit/32bit division and modulo, with
 * 64bit result and 32bit remainder.
 *
 * The fast case for (n>>32 == 0) is handled inline by do_div(). 
 *
 * Code generated for this function might be very inefficient
 * for some CPUs. __div64_32() can be overridden by linking arch-specific
 * assembly versions such as arch/ppc/lib/div64.S and arch/sh/lib/div64.S.
 */

#include <linux/types.h>
#include <linux/module.h>
#include <asm/div64.h>

/* Not needed on 64bit architectures */
#if BITS_PER_LONG == 32

uint32_t __attribute__((weak)) __div64_32(uint64_t *n, uint32_t base)
{
	uint64_t rem = *n;
	uint64_t b = base;
	uint64_t res, d = 1;
	uint32_t high = rem >> 32;

	/* Reduce the thing a bit first */
	res = 0;
	if (high >= base) {
		high /= base;
		res = (uint64_t) high << 32;
		rem -= (uint64_t) (high*base) << 32;
	}

	while ((int64_t)b > 0 && b < rem) {
		b = b+b;
		d = d+d;
	}

	do {
		if (rem >= b) {
			rem -= b;
			res += d;
		}
		b >>= 1;
		d >>= 1;
	} while (d);

	*n = res;
	return rem;
}

EXPORT_SYMBOL(__div64_32);

uint64_t udivmod64(uint64_t dividend, uint64_t divisor, uint64_t *rem)
{
	uint32_t dh,dl, nl, r, q;
	uint32_t msb, lsh;
	uint64_t nh;

	if (dividend < divisor) {
		*rem = dividend;
		return 0ULL;
	}

	dl = divisor;
	dh = divisor >> 32;
	if (dh == 0) {  /* (64bit)/(32bit) */
		*rem = do_div(dividend, dl); /* quot is stored in arg1 */
		return dividend;
	}

	/* Normalize divisor(bit63=1) to utilize do_div()
             dh:dl = NORM(divisor)	; dh:32bit, dl:32bit
             nh:nl = NORM(dividend)	; nh:64bit, nl:32bit
               2^31 <= dh < 2^32	; bit31 = 1
               dh <= nh < 2^63		; divisor >= 2^32
	   quot = nh:nl/dh:dl
	   nh/(dh+1) <= quot <= nh/dh
	     quot < 2^32		; nh < 2^63  &&  dh >= 2^31
	   nh/dh - nh/(dh+1) = (nh*(dh+1) - nh*dh)/(dh*(dh+1))
	     = nh/(dh*(dh+1)) < nh/(dh*dh) < (2^63/2^62) = 2
	   quot = nh/dh or nh/dh - 1
	*/
	msb = fls(dh);		/* 1..32 */
	lsh = 32 - msb;		/* 0..31 */
	/* dh:dl = divisor << lsh */
	dl = divisor << lsh;
	dh = divisor >> msb;	/* dh: bit31 = 1 */
	/* nh:nl = dividend << lsh */
	nl = dividend << lsh;
	nh = dividend >> msb;	/* nh: bit63 = 0 (msb >= 1) */
	/* q = nh/dh */
	r = do_div(nh, dh); /* quot is stored in arg1 */
	q = nh;
	/* if dh:dl * q > nh:nl then q--
	     nh = dh * q + r
	     (dh<<32) * q + dl * q > ((dh * q + r)<<32) + nl
	     dl * q > (r<<32) + nl
	*/
	if ((uint64_t)dl * q > ((uint64_t)r << 32) + nl)
		q--;

	*rem = dividend - divisor * q;
	return (uint64_t)q;
}

/* 64bit divisor, dividend and result. dynamic precision */
uint64_t div64_64(uint64_t dividend, uint64_t divisor)
{
	uint32_t high, d;

	high = divisor >> 32;
	if (high) {
		unsigned int shift = fls(high);

		d = divisor >> shift;
		dividend >>= shift;
	} else
		d = divisor;

	do_div(dividend, d);

	return dividend;
}
EXPORT_SYMBOL(div64_64);

#endif /* BITS_PER_LONG == 32 */
