/* 2007-10-12: File added by Sony Corporation */
/*
 *  Boot time
 *
 *  Copyright 2002-2007 Sony Corporation
 *
 *  This program is free software; you can redistribute	 it and/or modify it
 *  under  the terms of	 the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the	License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED	  ``AS	IS'' AND   ANY	EXPRESS OR IMPLIED
 *  WARRANTIES,	  INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO	EVENT  SHALL   THE AUTHOR  BE	 LIABLE FOR ANY	  DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED	  TO, PROCUREMENT OF  SUBSTITUTE GOODS	OR SERVICES; LOSS OF
 *  USE, DATA,	OR PROFITS; OR	BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN	 CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/module.h>
#include <linux/seq_file.h>
#include <linux/dma-mapping.h>
#include <linux/bootmem.h>	/* reserve_bootmem() */
#include <asm/io.h>
#include <asm/string.h>
#include <asm/uaccess.h>
#include <asm/pgalloc.h>
#include <linux/snsc_boot_time.h>

#ifdef CONFIG_SNSC_NBLARGS
#include <linux/snsc_nblargs.h>
#endif

#define BOOT_TIME_CLEARKEY    "CLEAR"

struct boot_time *pBootTime = NULL;
static struct proc_dir_entry *proc_boot_time;
static struct boot_time_md *pmd;

#define PENTRY(off)      ((struct boot_time_entry *)((unsigned long)pBootTime + off))
#define INC_OFFSET(off)  do { (off) += sizeof(struct boot_time_entry); } while(0)
#define WORD_ALIGNED(addr)      (((unsigned long)(addr) & (4 - 1)) == 0)

/* add new measurement point */
void
boot_time_add(char *comment)
{
	struct boot_time_entry *p;

	if (pBootTime == NULL) {
		return;
	}
	if (pBootTime->offNext > pBootTime->offMax) {
		return;
	}

	p = PENTRY(pBootTime->offNext);
	pmd->count_set(p);
	if (comment) {
		strncpy(p->comment, comment, MAX_COMMENT);
	} else {
		p->comment[0] = '\0';
	}
	INC_OFFSET(pBootTime->offNext);
	return ;
}

static void *
boot_time_seq_start(struct seq_file *m, loff_t *pos)
{
	u_int32_t off;
	loff_t n = *pos;

	if (pBootTime == NULL) {
		return NULL;
	}

	off = pBootTime->offHead;
	if (off == pBootTime->offNext)
		return NULL;
	while (n--) {
		INC_OFFSET(off);
		if (off == pBootTime->offNext)
			return NULL;
	}
	return (void *)off;
}

static void
boot_time_seq_stop(struct seq_file *m, void *v)
{
	/* nothing to do */
}

static void *
boot_time_seq_next(struct seq_file *m, void *v, loff_t *pos)
{
	u_int32_t off = (u_int32_t)v;
	(*pos)++;
	INC_OFFSET(off);
	return (off == pBootTime->offNext) ? NULL : (void *)off;
}

static int
boot_time_seq_show(struct seq_file *m, void *v)
{
	u_int32_t off = (u_int32_t)v;
	struct boot_time_entry *p;
	int i;
	int space = 0;

	if (off == pBootTime->offHead) {
		seq_printf(m, "boot time (%p)\n", pBootTime);
	}

	p = PENTRY(off);
	for(i =0; i < MAX_COMMENT; i++) {
		if (space || p->comment[i] == '\0' || p->comment[i] == '\n') {
			seq_putc(m, ' ');
			space = 1;
		} else {
			seq_printf(m, "%c", p->comment[i]);
		}
	}

	if (pmd->print) {
		pmd->print(m, p);
	} else {
		seq_printf(m, " : 0x%08x 0x%08x\n", p->count_hi, p->count_lo);
	}

	return 0;
}

static void
do_cache_sync(const void *addr, size_t size, int rw)
{
#ifdef CONFIG_ARM
	consistent_sync(addr, size, rw);
#elif defined(CONFIG_PPC) || defined(CONFIG_MIPS)
	dma_cache_sync(NULL, addr, size, rw);
#else
#error "Please define do_cache_sync() for your arch."
#endif
}

/* clear boot-time region */
static void
boot_time_clear(void)
{
	if (pBootTime == NULL) {
		return;
	}

	pBootTime->offNext = pBootTime->offHead;
	do_cache_sync(pBootTime,
		      (unsigned long)pBootTime->offMax
		      + sizeof(struct boot_time_entry),
		      DMA_TO_DEVICE);
}

static int
proc_boot_time_write(struct file *file, const char *buffer, size_t count, loff_t *ppos)
{
	unsigned long count0 = count;
	int len1;
	char comment[MAX_COMMENT+1];

	if (count0 > MAX_COMMENT) {
		len1 = MAX_COMMENT;
	} else {
		len1 = count0;
	}
	count0 -= len1;
	if (copy_from_user(comment, buffer, len1)) {
		return -EFAULT;
	}
	comment[len1] = '\0';

	if (strncmp(comment, BOOT_TIME_CLEARKEY, strlen(BOOT_TIME_CLEARKEY)) == 0)
		boot_time_clear();
	else
		boot_time_add(comment);
	return count;
}

static struct seq_operations proc_boot_time_seqop = {
	start:  boot_time_seq_start,
	next:   boot_time_seq_next,
	stop:   boot_time_seq_stop,
	show:   boot_time_seq_show
};

static int proc_boot_time_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &proc_boot_time_seqop);
}

static struct file_operations proc_boot_time_operations = {
	open:           proc_boot_time_open,
	read:           seq_read,
	write:          proc_boot_time_write,
	llseek:         seq_lseek,
	release:        seq_release,
};


static void
reserve_default_boottime_region(void)
{
	reserve_bootmem(BOOT_TIME_BASE, BOOT_TIME_SIZE);
	printk(KERN_INFO "reserved memory at 0x%08x size 0x%x\n",
	       BOOT_TIME_BASE, BOOT_TIME_SIZE);
}

/*
 *  Initialize boottime
 */
void
boot_time_init(struct boot_time_md *pboot_time_md)
{
#ifdef CONFIG_SNSC_BOOT_TIME_USE_NBLARGS
	struct nblargs_entry na;
#endif
	static int first = 1;
	u32 size;

	if (!first)
		return;
	first = 0;

	pmd = pboot_time_md;

#ifdef CONFIG_SNSC_BOOT_TIME_USE_NBLARGS
	if (nblargs_get_key(SNSC_BOOT_TIME_KEY, &na) < 0) {
		printk(KERN_INFO "NBLArgs key \"" SNSC_BOOT_TIME_KEY
		       "\" not found, using default(0x%08x)\n",
		       BOOT_TIME_BASE);

		reserve_default_boottime_region();
		pBootTime = (struct boot_time *)__va(BOOT_TIME_BASE);
		size = BOOT_TIME_SIZE;
	} else {
		printk(KERN_INFO "NBLArgs key \"" SNSC_BOOT_TIME_KEY
		       "\" found(0x%08x)\n", na.addr);

		/* specified boottime region is already reserved by
		   nblargs_init() */
		pBootTime = (struct boot_time *)nbl_to_va(na.addr);
		size = na.size;
	}
#else
	reserve_default_boottime_region();
	pBootTime = (struct boot_time *)__va(BOOT_TIME_BASE);
	size = BOOT_TIME_SIZE;
#endif

	if (size < sizeof(struct boot_time_entry)) {
		printk(KERN_INFO
		       "too small boot time save area(0x%x)\n", size);
		pBootTime = NULL;
		return;
	}
	pBootTime->offMax = size - sizeof(struct boot_time_entry);

	if ((pBootTime->magic != BOOT_TIME_MAGIC)
	    || ((u_int32_t)pBootTime->offHead > (u_int32_t)pBootTime->offMax)
	    || ((u_int32_t)pBootTime->offNext > (u_int32_t)pBootTime->offMax)
	    || (!WORD_ALIGNED(pBootTime->offHead))
	    || (!WORD_ALIGNED(pBootTime->offNext))) {
		/* initialize boot-time data region */
		pBootTime->magic = BOOT_TIME_MAGIC;
		pBootTime->offHead = (unsigned long)&(pBootTime->first_entry)
			- (unsigned long)pBootTime;
		pBootTime->offNext = pBootTime->offHead;

		if (pmd->init)
			pmd->init(pBootTime, 0);
	} else {
		/* boot-time region is already initialized by rom */
		if (pmd->init)
			pmd->init(pBootTime, 1);
	}
}

/* initialize /proc for boot time */
void
boot_time_proc_init(void)
{
	if ((proc_boot_time = create_proc_entry("snsc_boot_time", 0666, &proc_root))) {
		proc_boot_time->proc_fops = &proc_boot_time_operations;
	}
}

EXPORT_SYMBOL(boot_time_add);
