/* 2007-09-27: File added by Sony Corporation */
/*
 *  kinspect - kernel inspect tool
 */
/*
 * This program is free software; you can redistribute	it and/or modify it
 * under  the terms of	the GNU General	 Public License as published by the
 * Free Software Foundation;  either version 2 of the  License.
 *
 * THIS	 SOFTWARE  IS PROVIDED	 ``AS  IS'' AND	  ANY  EXPRESS OR IMPLIED
 * WARRANTIES,	 INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.	 IN
 * NO  EVENT  SHALL   THE AUTHOR  BE	LIABLE FOR ANY	 DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED	 TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 * USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN	CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write  to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/console.h>
#include <linux/tty_driver.h>
#include <linux/string.h>
#include <linux/sysrq.h>
#include <linux/smp.h>
#include <linux/netpoll.h>
#include <linux/pid.h>
#include <linux/stat.h>
#include <linux/if_ether.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,16)
#else
#include <linux/marker.h>
#endif
#include <linux/seq_file.h>
#include <linux/utsname.h>
#include <asm/uaccess.h>

#include "snsc_kinspect.h"

/* #define KI_DEBUG */

#ifdef KI_DEBUG
#define KI_DPRN(fmt, args ...) \
	    printk(KERN_DEBUG "kipct:%s(%d): " fmt, \
		   __FUNCTION__, __LINE__, ##args);
#else /* KI_DEBUG */
#define KI_DPRN(fmt, args ...)
#endif /* KI_DEBUG */

#define KI_INFO(fmt, args ...) \
	    printk(KERN_INFO "kipct:%s(%d): " fmt, \
		   __FUNCTION__, __LINE__, ##args);

#define KI_WARN(fmt, args ...) \
	    printk(KERN_WARNING "kipct:%s(%d): " fmt, \
		   __FUNCTION__, __LINE__, ##args);

#define KI_ERR(fmt, args ...) \
	    printk(KERN_ERR "kipct:%s(%d): " fmt, \
		   __FUNCTION__, __LINE__, ##args);

#ifdef CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE
#define MAX_TIME_COUNT (4096)
#define SURVEY		 0
#define OUTPUT		 1
#define MAX_MT		 3
typedef struct mt {
	unsigned long long diff[MAX_TIME_COUNT];
	unsigned int cnt;
	unsigned int max_cnt;
} mt_t ;
typedef struct mtimes {
	mt_t mt[MAX_MT];
	struct semaphore sem;
	struct proc_dir_entry *proc_file;
} mtimes_t;

static mtimes_t measure;

static void cntup(int id)
{
	mt_t *mt = &measure.mt[id];

	if (mt->max_cnt < MAX_TIME_COUNT) {
		mt->max_cnt = mt->cnt + 1;
	}
	mt->cnt++;
	if (mt->cnt >= MAX_TIME_COUNT) {
		mt->cnt = 0;
		mt->max_cnt = MAX_TIME_COUNT;
	}
}

static int
kipct_proc_read_measure(char *page, char **start, off_t off,
			int count, int *eof, void *data);
static int
kipct_proc_write_measure(struct file *file, const char *buf,
			 unsigned long count, void *data);

#endif /* CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE */

#define MAX_NETWORK_CHUNK (1280)
#define MAX_KIPCT_BUF \
	CONFIG_SNSC_KINSPECT_MAX_BUF
#define MAX_KIPCT_STBUF \
	CONFIG_SNSC_KINSPECT_MAX_STORED_BUF
#define DEFAULT_SURVEY_INTERVAL \
	CONFIG_SNSC_KINSPECT_SURVEY_INTERVAL
#define OUTPUT_INFO_COUNTER \
	CONFIG_SNSC_KINSPECT_INFOOUT_MULTIPLE

#define KIPCT_FBUF 256

#define KIPCT_CLCTR_TGT_INFO  0
#define KIPCT_CLCTR_TGT_INFO_NAME "target_info"

#define KI_SEARCH   0
#define KI_REGISTER 1

#define KIPCT_NOT_MARK 0
#define KIPCT_MARK 1

static struct netpoll np = {
	.name = "kinspect",
	.dev_name = "eth0",
	.local_port = 6667,
	.remote_port = 6668,
	.remote_mac = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff},
};


/* collector management */
typedef struct kipct_clctr {
	struct list_head list;
	int clctr_id; /* collector id */
	char *clctr_name; /* collector name */
	kipct_cb cb;
	kipct_cb infocb;
	int enable;
} kipct_clctr_t;

typedef struct kipct_stbuf {
	char *store;
	spinlock_t buf_lock;
	unsigned long offset;
	unsigned long total;
	unsigned long wtotal;
} kipct_stbuf_t;

typedef struct kipct_outinfo {
	kipct_stbuf_t stbuf;
	int network_enable;
} kipct_outinfo_t;

/* kipct  */
typedef struct kipct {
	struct semaphore sem;
	struct list_head clist;
	int survey; /* big switch */
	int clctr_num;
	char *page_buf;
	spinlock_t page_buf_lock;
	char *buf;
	unsigned int max_buf_str_len;
	spinlock_t net_out_lock;
	unsigned int info_counter;
	kipct_outinfo_t outinfo;
	kipct_write_t winfo;
	struct proc_dir_entry *proc_file;
} kipct_t;

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,16)
void
kipct_out_probe(char *buf, unsigned long len);
#else
typedef struct kipct_probe_data {
	const char *name;
	const char *format;
	marker_probe_func *probe_func;
	char* buf;
	unsigned long len;
} kipct_probe_data_t;

void kipct_out_probe(const struct __mark_marker_data *mdata,
		     const char *format, ...);

static kipct_probe_data_t kipct_probe_ary[] =
{
	{
		.name = "kinspect_output",
		.format = MARK_NOARGS,
		.probe_func = kipct_out_probe,
		.buf = NULL,
		.len = 0
	}
};
#endif

static kipct_t s_kipct;

static int configured = 0;

static int kipct_running;
static int exit_kipct = 0;
static DECLARE_WAIT_QUEUE_HEAD(kipct_task_wq);

#define KIPCT_SURVEY_STOP   0
#define KIPCT_SURVEY_START  1
static int kipct_survey = KIPCT_SURVEY_START;
static DECLARE_WAIT_QUEUE_HEAD(kipct_survey_wq);


/*
 * prototype
 */

static int option_setup(char *opt);
static int start_kipct_thread(void);
static int kipct_thread(void *unused);
static int kipct_netpoll_setup(void);
static int do_kipct(void);
static kipct_clctr_t * clctr_search(char *name, int create);
static int clctr_disable(kipct_clctr_rgst_t *rgst);
static int is_clctr_disable(void);

static int kipct_output(char *buf, unsigned int len);

/*
 * module parameter
 */

/* netpoll configulation */
static char config[256];
module_param_string(netpoll_info, config, 256, 0);
MODULE_PARM_DESC(netpoll_info, " netpoll_info=[src-port]@[src-ip]/[dev],[tgt-port]@<tgt-ip>/[tgt-macaddr]\n");

static int survey = 1;
/* survey setter */
static int survey_set(const char *val, struct kernel_param *kp)
{
	long int stat = 0;
	int retVal = 0;

	if (val) {
		stat = simple_strtol(val, NULL, 0);
		KI_DPRN("%s:set: val= <%ld>\n", kp->name, stat);
	}
	retVal = param_set_int(val, kp);
	if (retVal == 0) {
		if (stat) {
			kipct_survey = KIPCT_SURVEY_START;
			wake_up(&kipct_survey_wq);
		}
		else {
			kipct_survey = KIPCT_SURVEY_STOP;
		}
	}

	return retVal;
}

/* survey getter */
static int survey_get(char *buffer, struct kernel_param *kp)
{
	KI_DPRN("%s:set: val= <%d>\n", kp->name, *((int *)kp->arg));
	return (param_get_int(buffer, kp));
}

module_param_call(survey, survey_set,
		  survey_get, &survey, 0644);
MODULE_INFO(paramtype, "survey:int");
MODULE_PARM_DESC(survey, "start(1), stop(0)\n");





/*
 * internal function
 */

static int option_setup(char *opt)
{
	configured = !netpoll_parse_options(&np, opt);
	return 0;
}
__setup("netpoll_info=", option_setup);


/* output Information */
static int
kipct_info(void)
{
	char *buf = s_kipct.buf;
	unsigned long maxlen = MAX_KIPCT_BUF;
	unsigned int len = 0;
	int retVal = 0;
	kipct_write_t *info = NULL;
	kipct_buf_t bufinfo;
	unsigned int header_len = 0;
	unsigned int size = 0;

	bufinfo.buf = buf;
	bufinfo.max_buf_len = maxlen;
	info = &s_kipct.winfo;

	KI_DPRN("uts_name=%s\n", utsname()->nodename);

	/* kinspect info clctr */
	retVal = kipct_sprintf(buf, maxlen, &len, "%I:0:kipct:%d:%s\n",
			       KIPCT_CLCTR_TGT_INFO, KIPCT_CLCTR_TGT_INFO_NAME);
	if (retVal < 0) {
		return retVal;
	}

	(void)kipct_buf_write(buf, len);

	len = 0;
	retVal = kipct_write_header(&bufinfo, info, &len);
	if (retVal < 0) { return retVal; }
	header_len = len;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,16)
	down_read(&uts_sem);
	retVal = kipct_sprintf(buf, maxlen, &len, "hostname=%s\n",
			       system_utsname.nodename);
	up_read(&uts_sem);

#else
	retVal = kipct_sprintf(buf, maxlen, &len, "hostname=%s\n",
			       utsname()->nodename);
#endif
	if (retVal < 0) { return retVal; }
	retVal = kipct_sprintf(buf, maxlen, &len, "%\n");
	if (retVal < 0) { return retVal; }
	retVal = kipct_read_file(buf, &len, "/proc/cpuinfo",
				 maxlen);
	if (retVal < 0) { return retVal; }
	size = len - header_len;
	retVal = kipct_write_tailer(&bufinfo, info, &len, size);
	if (retVal < 0) { return retVal; }

	(void)kipct_buf_write(buf, len);
	(info->seq_id)++;

	return 0;
}

/* Initializes message buffer and kernel thread. */
static int
start_kipct_thread(void)
{
	KI_DPRN("\n");
	s_kipct.buf = kzalloc(MAX_KIPCT_BUF, GFP_KERNEL);
	if (s_kipct.buf == NULL) {
		KI_ERR("kzalloc() failed\n");
		return -ENOMEM;
	}

	KI_DPRN("\n");
	kernel_thread(kipct_thread, NULL, CLONE_FS | CLONE_FILES);
	return 0;
}

/*
 * Main loop for kinspect
 */
static int
kipct_thread(void *unused)
{
	int retVal = 0;
	struct sched_param param = { .sched_priority = MAX_RT_PRIO-1 };
	DECLARE_WAITQUEUE(wait, current);

	KI_DPRN("\n");

	daemonize("kinspect");
	sched_setscheduler(current, SCHED_RR, &param);

	kipct_running = 1;

	add_wait_queue(&kipct_task_wq, &wait);

	for (;;) {
		KI_DPRN("\n");
		if (exit_kipct)
			break;
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(msecs_to_jiffies(DEFAULT_SURVEY_INTERVAL));
		if (do_kipct())
			break;
	}
	remove_wait_queue(&kipct_task_wq, &wait);

	KI_DPRN("\n");

	netpoll_cleanup(&np);

	kipct_running = 0;

	return retVal;
}


static int
do_kipct(void)
{
	kipct_clctr_t *p = NULL;
	kipct_cb_arg_t arg;

	int retVal = 0;

#ifdef CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE
	unsigned long long start = 0;
	unsigned long long end = 0;
	mt_t *mt;
#endif /* CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE */

	KI_DPRN("\n");

	arg.bufinfo.buf = s_kipct.buf;
	arg.bufinfo.max_buf_len = MAX_KIPCT_BUF;

	retVal = wait_event_interruptible(kipct_survey_wq,
					  kipct_survey == KIPCT_SURVEY_START);
	if (retVal < 0) {
		return retVal;
	}

	if (down_interruptible(&s_kipct.sem)) {
		return -ERESTARTSYS;
	}

	if (configured && (s_kipct.outinfo.network_enable == 0)) {
		(void)kipct_netpoll_setup();
	}

	if (is_clctr_disable()) {
		kipct_survey = KIPCT_SURVEY_STOP;
		goto err;
	}

#ifdef CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE
	start = sched_clock();
#endif /* CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE */
	list_for_each_entry(p, &(s_kipct.clist), list) {
		if (s_kipct.info_counter == 0) {
			KI_DPRN("\n");
			(void)kipct_info();
			if (p->infocb && p->enable == 1) {
				p->infocb(&arg);
			}
		}
		KI_DPRN("\n");
		if (p->cb && p->enable == 1) {
			p->cb(&arg);
		}
	}
#ifdef CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE
	end = sched_clock();
	if (down_interruptible(&measure.sem)) {
		return -ERESTARTSYS;
	}
	mt = &measure.mt[SURVEY];
	mt->diff[mt->cnt] = end - start;
	cntup(SURVEY);
	up(&measure.sem);
#endif /* CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE */

	if (s_kipct.info_counter == 0) {
		s_kipct.info_counter = OUTPUT_INFO_COUNTER;
	}
	else {
		s_kipct.info_counter--;
	}

	KI_DPRN("\n");

 err:
	up(&s_kipct.sem);

	return 0;
}

static void
kipct_netpoll_send(char *buf, unsigned int len)
{
	int frag, left;

	for(left = len; left; ) {
		frag = min(left, MAX_NETWORK_CHUNK);
		netpoll_send_udp(&np, buf, frag);
		buf += frag;
		left -= frag;
	}
}

static void
output_netpoll(char *buf, unsigned int len, int mark_flag)
{
	unsigned long flags = 0;
#ifdef CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE
	unsigned long long start = 0;
	unsigned long long end = 0;
	mt_t *mt;
#endif /* CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE */

	if (!np.dev)
		return;

	KI_DPRN("buf=%p len=%u\n", buf, len);

#ifdef CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE
	start = sched_clock();
#endif /* CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE */

	if (in_interrupt())
		spin_lock_irqsave(&s_kipct.net_out_lock, flags);
	else
		spin_lock(&s_kipct.net_out_lock);

	if (mark_flag == KIPCT_MARK) {
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,16)
		kipct_out_probe(buf, len);
#else
		kipct_probe_ary[0].buf = buf;
		kipct_probe_ary[0].len = len;
		MARK(kinspect_output, MARK_NOARGS);
#endif
	}
	else {
		kipct_netpoll_send(buf, len);
	}

	if (in_interrupt())
		spin_unlock_irqrestore(&s_kipct.net_out_lock, flags);
	else
		spin_unlock(&s_kipct.net_out_lock);

#ifdef CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE
	end = sched_clock();
	mt = &measure.mt[OUTPUT];
	mt->diff[mt->cnt] = end - start;
	cntup(OUTPUT);
#endif /* CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE */

}

static void
output_stbuf(char *buf, unsigned int len)
{
	unsigned long flags = 0;
	kipct_stbuf_t *stbuf = &s_kipct.outinfo.stbuf;

	KI_DPRN("buf=%p len=%u\n", buf, len);

	if (in_interrupt())
		spin_lock_irqsave(&stbuf->buf_lock, flags);
	else
		spin_lock(&stbuf->buf_lock);

	if (stbuf->offset + len < stbuf->total) {
		memcpy(stbuf->store + stbuf->offset, buf, len);
		stbuf->offset += len;
	}
	else {
		KI_INFO("stored buffer is full\n");
		if (!configured) {
			kipct_survey = KIPCT_SURVEY_STOP;
			KI_INFO("logging will be stop\n");
		}
	}
	stbuf->wtotal += len;

	if (in_interrupt())
		spin_unlock_irqrestore(&stbuf->buf_lock, flags);
	else
		spin_unlock(&stbuf->buf_lock);

}

/*
 * initialize stored buffer
 */
static void
init_stbuf(kipct_stbuf_t *stbuf)
{
	spin_lock_init(&stbuf->buf_lock);
	stbuf->total = MAX_KIPCT_STBUF;
}

static int
init_page_buf(void)
{
	spin_lock_init(&s_kipct.page_buf_lock);
	s_kipct.page_buf = (char*)__get_free_page(GFP_KERNEL | __GFP_ZERO);
	if (s_kipct.page_buf == NULL) {
		KI_ERR("no memory\n");
		return -ENOMEM;
	}
	return 0;
}

static int
kipct_stbuf_output(kipct_stbuf_t *stbuf)
{
	KI_DPRN("\n");

	output_netpoll(stbuf->store, min(stbuf->wtotal, stbuf->total),
		       KIPCT_NOT_MARK);
	if (stbuf->store) {
		kfree(stbuf->store);
		stbuf->store = NULL;
	}
	return 0;
}

/*
 * netpoll setup
 */
static int
kipct_netpoll_setup(void)
{
	int retVal = 0;
	int i = 0;
	int is_bc = 1;

	for (i = 0; i < ETH_ALEN; i++) {
		if (np.remote_mac[i] != 0xff) {
			is_bc = 0;
			break;
		}
	}
	if (is_bc) {
		KI_ERR("kinspect: set remote mac_addr\n");
		configured = 0;
	}

	retVal = netpoll_setup(&np);
	if( retVal == 0) {
		KI_INFO("network logging started\n");
		s_kipct.outinfo.network_enable = 1;
		if (s_kipct.outinfo.stbuf.store != NULL) {
			KI_DPRN("stbuf output\n");
			kipct_stbuf_output(&s_kipct.outinfo.stbuf);
		}
	}

	return retVal;
}


static void
unregister_all(void)
{
	kipct_clctr_t *p = NULL;
	kipct_clctr_t *p_next = NULL;

	if (down_interruptible(&s_kipct.sem)) {
		return;
	}

	if (list_empty(&s_kipct.clist)) {
		up(&s_kipct.sem);
		return;
	}
	list_for_each_entry_safe(p, p_next, &(s_kipct.clist), list) {
		list_del(&p->list);
		kfree(p);
		p = NULL;
	}
	s_kipct.clctr_num = 0;
	up(&s_kipct.sem);
}

/* search clctr and create new clctr */
static kipct_clctr_t *
clctr_search(char *name, int create)
{
	kipct_clctr_t *p = NULL;

	list_for_each_entry(p, &(s_kipct.clist), list) {
		if (strncmp(p->clctr_name, name, strlen(name)) == 0) {
			return p;
		}
	}

	if (create) {
		p = kzalloc(sizeof(kipct_clctr_t), GFP_KERNEL);
		if (p == NULL) {
			KI_ERR("no memory\n");
			return NULL;
		}
		p->clctr_id = s_kipct.clctr_num;
		p->clctr_name = name;
		s_kipct.clctr_num++;
		list_add_tail(&(p->list), &(s_kipct.clist));
		return p;
	}

	return NULL;
}


/* disable clctr */
static int
clctr_disable(kipct_clctr_rgst_t *rgst)
{
	kipct_clctr_t *p = NULL;

	if (list_empty(&s_kipct.clist)) {
		return 0;
	}

	p = clctr_search(rgst->clctr_name, KI_SEARCH);
	if (p == NULL) {
		return -EINVAL;
	}

	p->enable = 0;

	return 0;
}

/* check clctr is all disabled */
static int
is_clctr_disable(void)
{
	kipct_clctr_t *p = NULL;

	if (list_empty(&s_kipct.clist)) {
		return 1;
	}

	list_for_each_entry(p, &(s_kipct.clist), list) {
		if (p->enable == 1) {
			return 0;
		}
	}

	return 1;
}

static void
kipct_calc_max_buf_len(void)
{
	char str[100]="";

	snprintf(str, 100, "%u", MAX_KIPCT_BUF);

	s_kipct.max_buf_str_len = strlen(str);

	KI_DPRN("max_buf_str_len=%u\n", s_kipct.max_buf_str_len);
}

static int
kipct_output(char *buf, unsigned int len)
{
	if (s_kipct.outinfo.network_enable == 1) {
		KI_DPRN("net\n");
		output_netpoll(buf, len, KIPCT_MARK);
	}
	else {
		KI_DPRN("memory\n");
		if (s_kipct.outinfo.stbuf.store == NULL) {
			s_kipct.outinfo.stbuf.store =
				kzalloc(s_kipct.outinfo.stbuf.total,
					GFP_ATOMIC);
			if (s_kipct.outinfo.stbuf.store == NULL) {
				KI_ERR("no memory available\n")
				return -ENOMEM;
			}
		}
		output_stbuf(buf, len);
	}

	return 0;
}



/*
 * proc filesystem
 */
static void *kipct_store_seq_start(struct seq_file *s, loff_t *pos)
{
	if (*pos >= 1) {
		return NULL;
	}
	return &s_kipct;
}

static void *kipct_store_seq_next(struct seq_file *s, void *v, loff_t *pos)
{
	(*pos)++;
	return NULL;
}

static void kipct_store_seq_stop(struct seq_file *s, void *v)
{
	; /* nothing to do */
}

static int kipct_store_seq_show(struct seq_file *s, void *v)
{
	kipct_t *ki = (kipct_t *)v;
	kipct_stbuf_t *stbuf = NULL;
	unsigned long flags = 0;

	stbuf = &(ki->outinfo.stbuf);

	spin_lock_irqsave(&stbuf->buf_lock, flags);
	if (stbuf->store != NULL) {
		seq_printf(s, "%s\n",stbuf->store);
	}
	spin_unlock_irqrestore(&stbuf->buf_lock, flags);

	return 0;
}

static struct seq_operations kipct_store_seq_ops = {
	.start = kipct_store_seq_start,
	.next  = kipct_store_seq_next,
	.stop  = kipct_store_seq_stop,
	.show  = kipct_store_seq_show
};

static int kipct_store_proc_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &kipct_store_seq_ops);
}

static struct file_operations kipct_store_proc_ops = {
	.owner	 = THIS_MODULE,
	.open	 = kipct_store_proc_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release
};

static int
kipct_create_proc(void)
{
	struct proc_dir_entry *root_dir = NULL;
	struct proc_dir_entry *store_entry = NULL;
#ifdef CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE
	struct proc_dir_entry *measure_entry = NULL;
#endif /* CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE */

	/* make "/proc/kipct" dir */
	root_dir = proc_mkdir("kinspect", 0);
	if (root_dir == NULL) {
		KI_ERR("fault proc_mkdir no memory\n");
		return -ENOMEM;
	}
	s_kipct.proc_file = root_dir;


#ifdef CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE
	/* read: measure */
	measure_entry = create_proc_read_entry("measure", 0, root_dir,
					       kipct_proc_read_measure,
					       &measure);
	if (measure_entry != NULL) {
		measure_entry->write_proc = kipct_proc_write_measure;
	}
#endif /* CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE */

	/* read: stored buffer */
	store_entry = create_proc_entry("storedbuf", 0, root_dir);
	if (store_entry != NULL) {
		store_entry->proc_fops = &kipct_store_proc_ops;
	}

	return 0;
}

static int
kipct_remove_proc(void)
{
#ifdef CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE
	remove_proc_entry("measure", s_kipct.proc_file);
#endif /* CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE */
	remove_proc_entry("storedbuf", s_kipct.proc_file);

	remove_proc_entry("kinspect", 0);

	return 0;
}

#ifdef CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE
static int
kipct_proc_read_measure(char *page, char **start, off_t off,
			int count, int *eof, void *data)
{
	mtimes_t *m;
	unsigned int len = 0;
	unsigned int invalid_num = 1;
	int mc = 0;
	mt_t *mt;
	unsigned long long avg_diff[MAX_MT];

	int i = 0;

	for (mc = 0; mc < MAX_MT; mc++) {
		avg_diff[mc] = 0;
	}

	m = (mtimes_t *)data;

	if (down_interruptible(&m->sem)) {
		return -ERESTARTSYS;
	}


	for (mc = 0; mc < MAX_MT; mc++) {
		mt = &(m->mt[mc]);
		if (mt->max_cnt == 1) {
			up(&m->sem);
			return 0;
		}
		for (i = 1; i < mt->max_cnt; i++) {
			avg_diff[mc] += mt->diff[i];
		}
		do_div(avg_diff[mc], mt->max_cnt - invalid_num);
	}

	len += sprintf(page + len, "====================================="
		       "==================\n");
	len += sprintf(page + len, "%4s %16s \n", "cnt", "survey");
	len += sprintf(page + len, "-------------------------------------"
		       "------------------\n");
	len += sprintf(page + len, "%4u %16llu \n",
		       m->mt[SURVEY].max_cnt, avg_diff[SURVEY]);
	len += sprintf(page + len, "====================================="
		       "==================\n");
	len += sprintf(page + len, "%4s %16s \n", "cnt", "output");
	len += sprintf(page + len, "-------------------------------------"
		       "------------------\n");
	len += sprintf(page + len, "%4u %16llu \n",
		       m->mt[OUTPUT].max_cnt, avg_diff[OUTPUT]);


	up(&m->sem);

	return len;
}
static int
kipct_proc_write_measure(struct file *file, const char *buf,
			 unsigned long count, void *data)
{
	mtimes_t *m;
	int i = 0;

	m = (mtimes_t *)data;

	if (down_interruptible(&m->sem)) {
		return -ERESTARTSYS;
	}

	if (count == 0) {
		return -EINVAL;
	}

	for (i = 0; i < MAX_MT; i++) {
		memset(&(m->mt[i]), 0, sizeof(mt_t));
	}
	up(&m->sem);

	return count;
}

#endif /* CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE */

/*
 * external function
 */
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,16)
void
kipct_out_probe(char *buf, unsigned long len)
{
	kipct_netpoll_send(buf, len);
}
#else
void
kipct_out_probe(const struct __mark_marker_data *mdata,
		const char *format, ...)
{
	kipct_probe_data_t *pdata = (kipct_probe_data_t *)mdata->pdata;

	KI_DPRN("pdata(%p) buf=%p len=%lX\n", pdata, pdata->buf, pdata->len);

	kipct_netpoll_send(pdata->buf, pdata->len);
}
#endif

int
kipct_clctr_register(kipct_clctr_rgst_t *rgst)
{
	kipct_clctr_t *p = NULL;
	int retVal = 0;

	if (down_interruptible(&s_kipct.sem)) {
		return -ERESTARTSYS;
	}

	p = clctr_search(rgst->clctr_name, KI_REGISTER);
	if (p == NULL) {
		retVal = -ENOMEM;
		goto err;
	}
	p->enable = 1;
	rgst->clctr_id = p->clctr_id;
	p->cb = rgst->cb;
	p->infocb = rgst->infocb;
	kipct_survey = KIPCT_SURVEY_START;
	s_kipct.info_counter = 0;
	wake_up(&kipct_survey_wq);

 err:
	up(&s_kipct.sem);

	return retVal;
}
int
kipct_clctr_unregister(kipct_clctr_rgst_t *rgst)
{
	int retVal = 0;

	if (down_interruptible(&s_kipct.sem)) {
		return -ERESTARTSYS;
	}

	retVal = clctr_disable(rgst);

	up(&s_kipct.sem);

	return retVal;
}

int
kipct_format_write(kipct_write_t *info, const char *format, ...)
{
	char buf[KIPCT_FBUF]="";
	kipct_buf_t bufinfo;
	int retVal = 0;
	unsigned int len = 0;
	unsigned int size = 0;
	unsigned int header_len = 0;
	va_list args;

	bufinfo.buf = buf;
	bufinfo.max_buf_len = KIPCT_FBUF;

	/* write header */
	retVal = kipct_write_header(&bufinfo, info, &len);
	if (retVal < 0) {
		KI_ERR("fault: write_header\n");
		return retVal;
	}
	header_len = len;

	/* write body */
	if (len > bufinfo.max_buf_len)
		return -EINVAL;

	va_start(args, format);
	len += vsnprintf(bufinfo.buf + len,
			 bufinfo.max_buf_len - len, format, args);
	va_end(args);

	/* write tailer */
	size = len - header_len;
	retVal = kipct_write_tailer(&bufinfo, info, &len, size);
	if (retVal < 0) {
		return retVal;
	}

	/* output */
	retVal = kipct_output(bufinfo.buf, len);

	return retVal;
}


int
kipct_lformat_write(kipct_write_t *info, const char *format, ...)
{
	kipct_buf_t bufinfo;
	int retVal = 0;
	unsigned int len = 0;
	va_list args;
	unsigned long flags = 0;
	unsigned int size = 0;
	unsigned int header_len = 0;

	bufinfo.buf = s_kipct.page_buf;
	bufinfo.max_buf_len = PAGE_SIZE;

	va_start(args, format);
	if (in_interrupt())
		spin_lock(&s_kipct.page_buf_lock);
	else
		spin_lock_irqsave(&s_kipct.page_buf_lock, flags);


	/* write header */
	retVal = kipct_write_header(&bufinfo, info, &len);
	if (retVal < 0) {
		KI_ERR("fault: write_header\n");
		return retVal;
	}
	header_len = len;

	/* write body */
	if (len > bufinfo.max_buf_len)
		return -EINVAL;

	va_start(args, format);
	len += vsnprintf(bufinfo.buf + len,
			 bufinfo.max_buf_len - len, format, args);
	va_end(args);

	/* write tailer */
	size = len - header_len;
	retVal = kipct_write_tailer(&bufinfo, info, &len, size);
	if (retVal < 0) {
		return retVal;
	}

	/* output */
	retVal = kipct_output(bufinfo.buf, len);

	if (in_interrupt())
		spin_unlock(&s_kipct.page_buf_lock);
	else
		spin_unlock_irqrestore(&s_kipct.page_buf_lock, flags);
	va_end(args);

	return retVal;
}

int
kipct_file_write(kipct_write_t *info, const char* file)
{
	kipct_buf_t bufinfo;
	int retVal = 0;
	unsigned int size = 0;
	unsigned int len = 0;
	unsigned int header_len = 0;

	bufinfo.buf = s_kipct.buf;
	bufinfo.max_buf_len = MAX_KIPCT_BUF;

	/* write header */
	retVal = kipct_write_header(&bufinfo, info, &len);
	if (retVal < 0) {
		return retVal;
	}
	header_len = len;

	/* write file */
	retVal = kipct_read_file(bufinfo.buf, &len, file,
				 bufinfo.max_buf_len);
	if (retVal < 0) {
		return retVal;
	}

	/* write tailer */
	size = len - header_len;
	retVal = kipct_write_tailer(&bufinfo, info, &len, size);
	if (retVal < 0) {
		return retVal;
	}

	/* output */
	retVal = kipct_output(bufinfo.buf, len);

	return retVal;
}

int
kipct_buf_write(char *buf, unsigned long len)
{
	return kipct_output(buf, len);
}


loff_t
kipct_get_file_size(const char * file)
{
	int retVal = 0;
	struct file *filp = NULL;
	loff_t size = 0;
	struct inode *inode = NULL;

	filp = filp_open(file, O_RDONLY, 0);
	if (filp == NULL || IS_ERR((void*)filp)) {
		KI_DPRN("fault: open %s\n", file);
		return -ENOENT;
	}

	inode = filp->f_dentry->d_inode;
	size = i_size_read(inode->i_mapping->host);

	retVal = filp_close(filp, current->files);
	if (retVal < 0) {
		return retVal;
	}

	return size;
}

int
kipct_sprintf(char *buf, unsigned int maxlen,
	      unsigned int *lenp, const char *format, ...)
{
	unsigned int len = *lenp;
	va_list args;

	if (len > maxlen) {
                KI_ERR("too many output. len(0x%08X) > maxlen(0x%08X), "
                       "kinspect thread buffer size is 0x%08lX\n",
                       len, maxlen, MAX_KIPCT_BUF);
		return -EINVAL;
        }

	va_start(args, format);
	len += vsnprintf(buf + len, maxlen - len, format, args);
	va_end(args);

	*lenp = len;

	return 0;
}

int
kipct_write_header(kipct_buf_t *bufinfo, kipct_write_t *info,
		   unsigned int *lenp)
{
	return kipct_sprintf(bufinfo->buf, bufinfo->max_buf_len,
			     lenp,"%s%d:%d\n%llu,%u,%0*u\n",
			     "%S", info->clctr_id, info->unit_id, sched_clock(),
			     info->seq_id, s_kipct.max_buf_str_len, 0);
}

int
kipct_write_tailer(kipct_buf_t *bufinfo, kipct_write_t *info,
		   unsigned int *lenp, unsigned int body_length)
{
	int retVal = 0;
	int header_body_pos = 0;

	header_body_pos =
		*lenp - body_length - (s_kipct.max_buf_str_len + 1);

	if (header_body_pos <= 0) {
		KI_ERR("fault: length is invalid")
		return -EINVAL;
	}
	KI_DPRN("header_body_pos=%d\n", header_body_pos);

	retVal = kipct_sprintf(bufinfo->buf, bufinfo->max_buf_len,
			       &header_body_pos,
			       "%0*u", s_kipct.max_buf_str_len, body_length);
	if (retVal < 0) {
		return retVal;
	}
	if (header_body_pos > bufinfo->max_buf_len) {
		return -EINVAL;
	}
	*(bufinfo->buf + header_body_pos) = '\n';

	return kipct_sprintf(bufinfo->buf, bufinfo->max_buf_len, lenp,
			     "%s%d:%d\n",
			     "%E", info->clctr_id, info->unit_id);
}

int
kipct_read_file(char *buf, int *lenp, const char* file,
		unsigned int max_buf_len)
{
	struct file *filp = NULL;
	int len = *lenp;
	int sz, offset;
	int retVal = 0;

	filp = filp_open(file, O_RDONLY, 0);
	if (IS_ERR(filp)) {
		KI_DPRN("filp_open(%s) failed, it will be retried.\n", file);
		retVal = -EINVAL;
		goto err1;
	}

	offset = 0;
	do {
		if (len > max_buf_len) {
			KI_ERR("buffer full len=%d max=%u\n",
			       len, max_buf_len);
			retVal = -EINVAL;
			goto err2;
		}
		sz = kernel_read(filp, offset, buf + len, max_buf_len - len);
		if (sz < 0) {
			KI_ERR("read(%s) failed\n", file);
			retVal = -EINVAL;
			goto err2;
		}
		offset += sz;
		len += sz;
	} while (sz != 0);

err2:
	filp_close(filp, current->files);
	*lenp = len;
err1:
	return retVal;
}

int
kipct_survey_start(void)
{
	kipct_survey = KIPCT_SURVEY_START;
	return 0;
}

int
kipct_survey_stop(void)
{
	kipct_survey = KIPCT_SURVEY_STOP;
	return 0;
}

/*
 * insmod/rmmod
 */

static int init_kipct(void)
{
	int retVal = 0;
	int id = 0;

#ifdef CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE
	memset(&measure, 0, sizeof(mtimes_t));
	init_MUTEX(&measure.sem);
#endif /* CONFIG_SNSC_KINSPECT_DEBUG_MEASURE_MODE */

	memset(&s_kipct, 0, sizeof(kipct_t));
	init_MUTEX(&s_kipct.sem);
	INIT_LIST_HEAD(&s_kipct.clist);
	s_kipct.clctr_num = 1;
	s_kipct.winfo.clctr_id = 0;
	s_kipct.winfo.unit_id = KIPCT_CLCTR_TGT_INFO;
	s_kipct.winfo.seq_id = 0;
	retVal = init_page_buf();
	if (retVal < 0) {
		return retVal;
	}
	init_stbuf(&s_kipct.outinfo.stbuf);
	kipct_calc_max_buf_len();

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,16)
#else
	for (id = 0; id < ARRAY_SIZE(kipct_probe_ary); id++) {
		retVal = marker_set_probe(kipct_probe_ary[id].name,
					  kipct_probe_ary[id].format,
					  kipct_probe_ary[id].probe_func,
					  &kipct_probe_ary[id]);
		if (retVal == 0) {
			KI_ERR("fault: marker_set_probe ret=%d\n", retVal);
			return retVal;
		}
	}
#endif

	if(strlen(config))
		option_setup(config);

	if(!configured) {
		KI_INFO("netpoll_info is not configured\n");
		/* if stbuf is written, output to network */
	}

	if (start_kipct_thread())
		return -EINVAL;

	retVal = kipct_create_proc();
	if (retVal < 0) {
		KI_ERR("fault create_proc()\n");
		return retVal;
	}

	printk(KERN_INFO "kinspect: insmod\n");
	return 0;
}

static void cleanup_kipct(void)
{
	int id = 0;

	exit_kipct = 1;
	while (kipct_running)
		schedule();
	unregister_all();
	kipct_remove_proc();

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,16)
#else
	for (id = 0; id < ARRAY_SIZE(kipct_probe_ary); id++) {
		marker_remove_probe(kipct_probe_ary[id].name);
	}
	synchronize_sched();	/* Wait for probes to finish */
#endif

	if (s_kipct.buf) {
		kfree(s_kipct.buf);
		s_kipct.buf = NULL;
	}
	if (s_kipct.page_buf) {
		free_page((unsigned long)s_kipct.page_buf);
		s_kipct.page_buf = NULL;
	}

	printk(KERN_INFO "kinspect: rmmod\n");
}

late_initcall(init_kipct);
module_exit(cleanup_kipct);

/*
 * module information
 */

EXPORT_SYMBOL(kipct_survey_start);
EXPORT_SYMBOL(kipct_survey_stop);
EXPORT_SYMBOL(kipct_clctr_register);
EXPORT_SYMBOL(kipct_clctr_unregister);
EXPORT_SYMBOL(kipct_format_write);
EXPORT_SYMBOL(kipct_lformat_write);
EXPORT_SYMBOL(kipct_file_write);
EXPORT_SYMBOL(kipct_buf_write);
EXPORT_SYMBOL(kipct_get_file_size);
EXPORT_SYMBOL(kipct_sprintf);
EXPORT_SYMBOL(kipct_write_header);
EXPORT_SYMBOL(kipct_write_tailer);
EXPORT_SYMBOL(kipct_read_file);


MODULE_DESCRIPTION("kernel inspect tool");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Sony Corporation");
