/* 2007-07-17: File added by Sony Corporation */
/*
 * Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 * Released under the terms of the GNU GPL v2.0.
 *
 * Introduced single menu mode (show all sub-menus in one large tree).
 * 2002-11-06 Petr Baudis <pasky@ucw.cz>
 *
 * i18n, 2005, Arnaldo Carvalho de Melo <acme@conectiva.com.br>
 */

#include <sys/ioctl.h>
#include <sys/wait.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <locale.h>

#define LKC_DIRECT_LINK
#include "lkc.h"

static void get_prompt_str(struct gstr *r, struct property *prop)
{
	int i;
	struct menu *submenu[8], *menu;

	str_printf(r, "Prompt: \"%s\"\n", prop->text);
	str_printf(r, "Defined: %s:%d\n", prop->menu->file->name,
		prop->menu->lineno);
	if (!expr_is_yes(prop->visible.expr)) {
		str_append(r, "Depends: ");
		expr_gstr_print(prop->visible.expr, r);
		str_append(r, "\n");
	}
	menu = prop->menu->parent;
	for (i = 0; menu != &rootmenu && i < 8; menu = menu->parent)
		submenu[i++] = menu;
	if (i > 0) {
		str_printf(r, "MenuLocation: ");
		while (--i >= 0) {
			menu = submenu[i];
			str_printf(r, "\"%s\"", menu_get_prompt(menu));
			if (menu->sym) {
				str_printf(r, " (%s [=%s])", menu->sym->name ?
					menu->sym->name : "<choice>",
					sym_get_string_value(menu->sym));
			}
			if (i > 0)
				str_append(r, " -> ");
		}
	}
	str_append(r, "\n");
}

static void get_symbol_str(struct gstr *r, struct symbol *sym)
{
	bool hit;
	struct property *prop;

	str_printf(r, "Symbol: %s\n", sym->name);
	str_printf(r, "Value: %s\n", sym_get_string_value(sym));
	for_all_prompts(sym, prop)
		get_prompt_str(r, prop);
	hit = false;
	for_all_properties(sym, prop, P_SELECT) {
		if (!hit) {
			str_append(r, "Selects: ");
			hit = true;
		} else
			str_printf(r, " && ");
		expr_gstr_print(prop->expr, r);
	}
	if (hit)
		str_append(r, "\n");
	if (sym->rev_dep.expr) {
		str_append(r, "SelectedBy: ");
		expr_gstr_print(sym->rev_dep.expr, r);
	}
	str_printf(r, "Help: %s\n", sym->help);
	str_append(r, "\n");
}

static struct gstr get_relations_str(struct symbol **sym_arr)
{
	struct symbol *sym;
	struct gstr res = str_new();
	int i;

	for (i = 0; sym_arr && (sym = sym_arr[i]); i++)
		get_symbol_str(&res, sym);
	if (!i)
		str_append(&res, "No matches found.");
	return res;
}

static char filename[PATH_MAX+1];
static void set_config_filename(const char *config_filename)
{
	int size;
#if 0
	struct symbol *sym;
	sym = sym_lookup("KERNELVERSION", 0);
	sym_calc_value(sym);
	printf("%s - Linux Kernel v%s Configuration\n",
	       config_filename, sym_get_string_value(sym));
#endif
	size = snprintf(filename, sizeof(filename), "%s", config_filename);
	if (size >= sizeof(filename))
		filename[sizeof(filename)-1] = '\0';
}

static void search_conf(char *re_pattern)
{
	struct symbol **sym_arr;
	struct gstr res;

	sym_arr = sym_re_search(re_pattern);
	res = get_relations_str(sym_arr);
	free(sym_arr);
	printf("%s\n", str_get(&res));
	str_free(&res);
}

int main(int ac, char **av)
{
	conf_parse(av[1]);
	conf_read(NULL);
	set_config_filename(conf_get_configname());

	if (ac > 2)
		search_conf(av[2]);

	return 0;
}
